/*!
 * \file CommandLine.h
 * \brief Declares the CommandLine class.
 * \author Joe Forte
 * \date 7/27/2008
 */

#ifndef COMMANDLINE_H
#define COMMANDLINE_H

#include <map>
#include <string>
#include <vector>
#include "Common/BaseTypes.h"

/*!
 * \class CommandLine
 * \headerfile CommandLine.h "Application/CommandLine.h"
 * \brief Class that parses the command line and allows for easy access to the commands.
 *
 * Commands are expected to be proceeded with '+' or '-' all strings after a command are assumed to be that commands arguments.   
 */
class CommandLine
{
public:
	/*!
	 * \brief Parses the command line input.
	 *
	 * \param[in] argc the length of the \a argv array.
	 * \param[in] argv the array of input arguments.
	 */
	void Parse(int argc, char *argv[]);
	
	/*!
	 * \brief Searches for a command.
	 *
	 * \param[in] cmd the name of the command to search for.
	 * \param[out] args pointer to a vector that will store any arguments to \a cmd.
	 *
	 * \return True if the command was found, false otherwise.
	 */
	bool FindCmd(const char *cmd, std::vector<std::string> *args = NULL) const;
	
	/*!
	 * \brief Gets the number of commands specified in the command line.
	 *
	 * \return The number of commands.
	 */
	uint GetCmdCount() const;
	
	/*!
	 * \brief Searches for an argument in the command line.
	 *
	 * \param[in] arg the name of the argument to search for.
	 *
	 * \return The index to the argument or -1 if the argument does not exist.
	 */
	int Find(const char *arg) const;
	
	/*!
	 * \brief Gets the \a i th argument from the command line.
	 *
	 * \param[in] i the index.
	 *
	 * \return The \a i th command line argument.
	 */
	const char *Get(uint i) const;
	
	/*!
	 * \brief Gets the number of arguments in the command line.
	 *
	 * \return The number of arguments in the command line.
	 */
	uint GetArgCount() const;
	
private:
	typedef std::vector<std::string> argType;
	argType args;
	
	// Maps commads to a list of its arguments.
	typedef std::map<std::string, argType> cmdMapType;
	cmdMapType cmds;
};

#endif
