/*!
 * \file ConVar.cpp
 * \brief Implements the ConVar class.
 * \author Joe Forte
 * \date 6/9/2008
 */

#include <sstream>
#include "Application/Console.h"
#include "Application/ConVar.h"

//--------------------------------------------------------------------
ConVar::ConVar(const char *n, const char *defaultVal, const char *helpStr, fnConCommad c, void *user, Type t, Console *con)
:
name(n),
value(defaultVal),
defaultValue(defaultVal),
help(helpStr),
cmd(c),
userData(user),
type(t),
owner(con)
{
	Convert(defaultVal);
}

//--------------------------------------------------------------------
int ConVar::GetInt() const
{
	return intValue;
}

//--------------------------------------------------------------------
uint ConVar::GetUInt() const
{
	return uintValue;
}

//--------------------------------------------------------------------
Vector2f ConVar::GetVector() const
{
	return Vector2f(xValue, yValue);
}

//--------------------------------------------------------------------
Point2f ConVar::GetPoint() const
{
	return Point2f(xValue, yValue);
}

//--------------------------------------------------------------------
Color4f ConVar::GetColor() const
{
	return Color4f(rValue, gValue, bValue, aValue);
}

//--------------------------------------------------------------------
float ConVar::GetFloat() const
{
	return floatValue;
}

//--------------------------------------------------------------------
const char *ConVar::GetString() const
{
	return value.c_str();
}

//--------------------------------------------------------------------
fnConCommad ConVar::GetCommand() const
{
	return cmd;
}

//--------------------------------------------------------------------
void *ConVar::GetUserData() const
{
	return userData;
}

//--------------------------------------------------------------------
void ConVar::SetInt(int v)
{
	std::stringstream s;
	
	s << v;
	value = s.str();
	
	intValue = v;
}

//--------------------------------------------------------------------
void ConVar::SetUInt(uint v)
{
	std::stringstream s;
	
	s << v;
	value = s.str();
	
	uintValue = v;
}

//--------------------------------------------------------------------
void ConVar::SetVector(const Vector2f &v)
{
	std::stringstream s;
	
	s << v.x << " " << v.y;
	value = s.str();
	
	xValue = v.x;
	yValue = v.y;
}

//--------------------------------------------------------------------
void ConVar::SetPoint(const Point2f &v)
{
	std::stringstream s;
	
	s << v.x << " " << v.y;
	value = s.str();
	
	xValue = v.x;
	yValue = v.y;
}

//--------------------------------------------------------------------
void ConVar::SetColor(const Color4f &v)
{
	std::stringstream s;
	
	s << v.r << " " << v.g << " " << v.b << " " << v.a;
	value = s.str();
	
	rValue = v.r;
	gValue = v.g;
	bValue = v.b;
	aValue = v.a;
}

//--------------------------------------------------------------------
void ConVar::SetFloat(float v)
{
	std::stringstream s;
	
	s << v;
	value = s.str();
	
	floatValue = v;
}

//--------------------------------------------------------------------
void ConVar::SetString(const char *v)
{
	if( Convert(v) )
	{
		value = v;
	}
}

//--------------------------------------------------------------------
void ConVar::SetString(const std::string &v)
{
	if( Convert(v.c_str()) )
	{
		value = v;
	}
}

//--------------------------------------------------------------------
void ConVar::Revert()
{
	value = defaultValue;
	Convert(defaultValue.c_str());
}

//--------------------------------------------------------------------
void ConVar::AddListener(ConVarListener *lsn)
{
	listeners.push_back(lsn);
}

//--------------------------------------------------------------------
void ConVar::RemoveListener(ConVarListener *lsn)
{
	for(std::vector<ConVarListener*>::iterator i = listeners.begin(); i != listeners.end(); i++)
	{
		if(*i == lsn)
		{
			listeners.erase(i);
			return;
		}
	}
}

//--------------------------------------------------------------------
void ConVar::ImformListenersOfChange(const char *cmdLine, const std::vector<std::string> &args)
{
	for(std::vector<ConVarListener*>::iterator i = listeners.begin(); i != listeners.end(); i++)
	{
		assert(*i);
		(*i)->OnConVarChange(this, owner, cmdLine, args);
	}
}

//--------------------------------------------------------------------
bool ConVar::Convert(const char *str)
{
	std::istringstream iss(str);
	
	switch(type)
	{
		case TYPE_INT:
		{
			int i;
			
			// Checks if the value is in hex format (proceeded by 0x)
			if( *str && *str == '0' && 
			   *(str + 1) && ( *(str + 1) == 'x' || *(str + 1) == 'X' ) )
			{
				iss >> std::hex;
			}
			
			if( (iss >> i).fail() )
			{
				owner->PrintError() << "The given value '" << str << "' is not an integer value.\n";
				return false;
			}
			
			intValue = i;
		}
			break;
			
		case TYPE_UNSIGNED_INT:
		{
			uint i;
			
			// Checks if the value is in hex format (proceeded by 0x)
			if( *str && *str == '0' && 
			   *(str + 1) && ( *(str + 1) == 'x' || *(str + 1) == 'X' ) )
			{
				iss >> std::hex;
			}
			
			if( (iss >> i).fail() )
			{
				owner->PrintError() << "The given value '" << str << "' is not an integer value.\n";
				return false;
			}
			
			uintValue = i;
		}
			break;
			
		case TYPE_FLOAT:
		{
			float fl;
			if( (iss >> fl).fail() )
			{
				owner->PrintError() << "The given value '" << str << "' is not a floating point value.\n";
				return false;
			}
			
			floatValue = fl;
		}
			break;
			
		case TYPE_VECTOR:
		case TYPE_POINT:
		{
			float x, y;
			if( (iss >> x >> y).fail() )
			{
				owner->PrintError() << "The given value '" << str << "' does not contain 2 floating point values.\n";
				return false;
			}
			
			xValue = x;
			yValue = y;
		}
			break;
			
			
		case TYPE_COLOR:
		{
			float r, g, b, a;
			if( (iss >> r >> g >> b >> a).fail() )
			{
				owner->PrintError() << "The given value '" << str << "' does not contain 4 floating point values.\n";
				return false;
			}
			
			rValue = r;
			gValue = g;
			bValue = b;
			aValue = a;
		}
			break;
	};
	
	return true;
}
