/*!
 * \file ConVar.h
 * \brief Declares the ConVar class used for interacting with the application through the console.
 * \author Joe Forte
 * \date 6/9/2008
 */

#ifndef CONVAR_H
#define CONVAR_H

#include <string>
#include "Math/MathCommon.h"
#include "Graphics/Color.h"

class ConVar;
class Console;

/*!
 * \brief The console command function prototype.
 *
 * \param[in] con pointer to the console calling the command.
 * \param[in] cmd the full command string that initiated the comand.
 * \param[in] userData user data associated with the command.
 * \param[in] args the command line split into seperate strings.
 */
typedef void (*fnConCommad)(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);

/*!
 * \class ConVarListener
 * \headerfile ConVar.h "Application/ConVar.h"
 * \brief Interface for listening to console variable changes.
 */
class ConVarListener
{
public:
	
	/*!
	 * \brief Called when a console variable changes.
	 *
	 * \param[in] v the console variable that changed.
	 * \param[in] con the console.
	 * \param[in] cmdLine the full command line that caused the change.
	 * \param[in] args the command line split into seperate strings.
	 */
	virtual void OnConVarChange(ConVar *v, Console *con, const char *cmdLine, const std::vector<std::string> &args) = 0;
};

/*!
 * \class ConVar
 * \headerfile ConVar.h "Application/ConVar.h"
 * \brief Represents a single console variable or command.
 */
class ConVar
{
public:
	
	/*!
	 * \enum Type
	 * \brief The data type of the console variable.
	 */
	enum Type
	{
		TYPE_INT = 0,		/*!< A 32-bit signed integer. */
		TYPE_UNSIGNED_INT,	/*!< A 32-bit unsigned integer. */
		TYPE_VECTOR,		/*!< A 2 component floating point vector. */
		TYPE_POINT,			/*!< A 2 component floating point point. */
		TYPE_COLOR,			/*!< A 4 component floating point color. */
		TYPE_FLOAT,			/*!< A floating point value. */
		TYPE_STRING,		/*!< An arbitrarily sized string. */
		TYPE_COMMAND,		/*!< A console command. */
		
		TYPE_COUNT
	};
	
	/*!
	 * \brief Returns the ConVar value as a 32-bit signed int.
	 */
	int GetInt() const;
	
	/*!
	 * \brief Returns the ConVar value as a 32-bit unsigned int.
	 */
	uint GetUInt() const;
	
	/*!
	 * \brief Returns the ConVar value as a float.
	 */
	float GetFloat() const;
	
	/*!
	 * \brief Returns the ConVar value as a 2 component floating point vector.
	 */
	Vector2f GetVector() const;
	
	/*!
	 * \brief Returns the ConVar value as a 2 component floating point point.
	 */
	Point2f GetPoint() const;
	
	/*!
	 * \brief Returns the ConVar value as a 4 component floating point color.
	 */
	Color4f GetColor() const;
	
	/*!
	 * \brief Returns the ConVar value as a string.
	 */
	const char *GetString() const;
	
	/*!
	 * \brief Returns the console command function.
	 */
	fnConCommad GetCommand() const;
	
	/*!
	 * \brief Returns the user data associated with the variable.
	 */
	void *GetUserData() const;
	
	/*!
	 * \brief Sets the int value of the ConVar.
	 */
	void SetInt(int v);
	
	/*!
	 * \brief Sets the unsigned int value of the ConVar.
	 */
	void SetUInt(uint v);
	
	/*!
	 * \brief Sets the float value of the ConVar.
	 */
	void SetFloat(float v);
	
	/*!
	 * \brief Sets the vector value of the ConVar.
	 */
	void SetVector(const Vector2f &v);
	
	/*!
	 * \brief Sets the point value of the ConVar.
	 */
	void SetPoint(const Point2f &v);
	
	/*!
	 * \brief Sets the color value of the ConVar.
	 */
	void SetColor(const Color4f &v);
	
	/*!
	 * \brief Sets the string value of the ConVar.
	 */
	void SetString(const char *v);
	void SetString(const std::string &v);
	
	/*!
	 * \brief Reverts the variable to its default variable.
	 */
	void Revert();
	
	/*!
	 * \brief Returns the name of the console variable.
	 */
	const std::string &GetNameString() const { return name; }
	
	/*!
	 * \brief Returns the name of the console variable.
	 */
	const char *GetName() const { return name.c_str(); }
	
	/*!
	 * \brief Returns the defualt console variable.
	 */
	const char *GetDefaultValue() const { return defaultValue.c_str(); }
	
	/*!
	 * \brief Returns the help string for the console variable.
	 */
	const char *GetHelpStr() const { return help.c_str(); }
	
	/*!
	 * \brief Returns the type of the console variable.
	 */
	ConVar::Type GetType() const { return type; }
	
	/*!
	 * \brief Adds a listner to the console variable.
	 */
	void AddListener(ConVarListener *lsn);
	
	/*!
	 * \brief Removes a listner from the console variable.
	 */
	void RemoveListener(ConVarListener *lsn);
	
private:
	
	// Only the console class is allowed to create ConVars
	ConVar(const char *n, const char *defaultVal, const char *helpStr, fnConCommad c, void *user, Type t, Console *con);
	
	// Converts the current string value to the proper type for this ConVar.  Returns true if the given string
	// is valid.  Returns false if the string does not represent the proper type.
	bool Convert(const char *str);
	
	std::string name;
	std::string defaultValue;
	std::string value;
	std::string help;
	
	union
	{
		float floatValue;
		int intValue;
		uint uintValue;
		
		// Point and Vector data
		struct
		{
			float xValue, yValue;
		};
		
		// Color data
		struct
		{
			float rValue, gValue, bValue, aValue;
		};
	};
	
	fnConCommad cmd;
	void *userData;
	
	Type type;
	
	void ImformListenersOfChange(const char *cmdLine, const std::vector<std::string> &args);
	
	std::vector<ConVarListener*> listeners;
	
	Console *owner;
	
	friend class Console;
};

#endif
