/*!
 * \file ConfigFile.cpp
 * \brief Implements the ConfigFile class.
 * \author Joe Forte
 * \date 2/2/2009
 */

#include <fstream>
#include <sstream>
#include "Utility/StringUtility.h"
#include "Application/ConfigFile.h"

const char COMMENT = ';';
const char ASSIGNMENT = '=';

//--------------------------------------------------------------------
ConfigFile::ConfigFile(const char *file)
{
	Load(file);
}

//--------------------------------------------------------------------
bool ConfigFile::Load(const char *file)
{
	std::fstream cfg(file);
	
	if(!cfg.is_open())
		return false;
	
	std::string line;
	while( getline(cfg, line) )
	{
		// Removes any comments from the line.
		size_t commentPos = line.find_first_of(COMMENT);
		if(commentPos != std::string::npos)
			line = line.substr(0, commentPos);
		
		if(line.size() == 0)
			continue;
		
		// Extracts both sides of the assignment.
		size_t assignPos = line.find_first_of(ASSIGNMENT);
		if(assignPos == std::string::npos)
			continue;
		
		std::string key = Trim( line.substr(0, assignPos) );
		std::string value = Trim( line.substr(assignPos + 1) );
		
		SetValue(key, value);
	}
	
	return true;
}

//--------------------------------------------------------------------
bool ConfigFile::Save(const char *name)
{
	return false;
}

//--------------------------------------------------------------------
const ConfigFile::ConfigValue *ConfigFile::FindValueByName(const char *name) const
{
	std::map<std::string, ConfigValue>::const_iterator val = values.find(name);
	if(val != values.end())
		return &val->second;

	return NULL;
}

//--------------------------------------------------------------------
const ConfigFile::ConfigValue &ConfigFile::operator[](const char *name) const
{
	return *FindValueByName(name);
}

//--------------------------------------------------------------------
bool ConfigFile::SetValue(const char *name, const char *value)
{	
	return SetValue(std::string(name), std::string(value));
}

//--------------------------------------------------------------------
bool ConfigFile::SetValue(const std::string &name, const std::string &value)
{
	std::map<std::string, ConfigValue>::iterator old = values.find(name);
	if(old != values.end())
	{
		old->second.SetValue(value);
	}
	else
	{
		ConfigValue val(name);
		val.SetValue(value);
		values[name] = val;
	}
	
	return true;
}

//--------------------------------------------------------------------
void ConfigFile::ConfigValue::SetValue(const std::string &val)
{
	value = val;
	type = DetermineAndConvertType(val);
}

//--------------------------------------------------------------------
ConfigFile::ValueType ConfigFile::ConfigValue::DetermineAndConvertType(const std::string &val)
{
	int i;
	if(IsWholeInteger(val, &i))
	{
		intVal = i;
		return ConfigFile::TYPE_INTEGER;
	}
	
	float f;
	if(IsFloat(val, &f))
	{
		fltVal = f;
		return ConfigFile::TYPE_FLOAT;
	}
	
	bool b;
	if(IsBoolean(val, &b))
	{
		intVal = (int)b;
		return ConfigFile::TYPE_BOOL;
	}
	
	Vector2f v;
	if(IsVector2(val, v))
	{
		vec2[0] = v.x;
		vec2[1] = v.y;
		return ConfigFile::TYPE_VECTOR2;
	}
	
	Color4f c;
	if(IsColor4(val, c))
	{
		col4[0] = c.r;
		col4[1] = c.g;
		col4[2] = c.b;
		col4[3] = c.a;
		return ConfigFile::TYPE_COLOR4;
	}
	
	return ConfigFile::TYPE_STRING;
}

//--------------------------------------------------------------------
bool ConfigFile::ConfigValue::GetOperatorAndParams(const std::string &val, std::string &op, std::vector<std::string> &args)
{
	size_t paramStart = val.find_first_of("(");
	size_t paramEnd = val.find_last_of(")");
	
	if(paramStart == std::string::npos || paramEnd == std::string::npos || paramEnd < paramStart)
		return false;
	
	op = Trim( val.substr(0, paramStart) );
	
	size_t comma = paramStart + 1;
	while(comma < paramEnd)
	{
		size_t nextComma = val.find_first_of(",", comma);
		if(nextComma == std::string::npos)
			nextComma = paramEnd;
		
		args.push_back( Trim(val.substr(comma, nextComma - comma)) );
		
		comma = nextComma + 1;
	}
	
	return true;
}

//--------------------------------------------------------------------
bool ConfigFile::ConfigValue::IsColor4(const std::string &val, Color4f &col)
{
	std::string op;
	std::vector<std::string> args;
	if(!GetOperatorAndParams(val, op, args))
		return false;
	
	ToLower(op);
	if(op != "color")
		return false;
	
	if(args.size() == 4 && 
	   IsFloat(args[0], &col.r) && 
	   IsFloat(args[1], &col.g) && 
	   IsFloat(args[2], &col.b) && 
	   IsFloat(args[3], &col.a) )
	{
		return true;
	}
	
	return false;
}

//--------------------------------------------------------------------
bool ConfigFile::ConfigValue::IsVector2(const std::string &val, Vector2f &vec)
{
	std::string op;
	std::vector<std::string> args;
	if(!GetOperatorAndParams(val, op, args))
		return false;
	
	ToLower(op);
	if(op != "vector")
		return false;
	
	if(args.size() == 2 && 
	   IsFloat(args[0], &vec.x) && 
	   IsFloat(args[1], &vec.y) )
	{
		return true;
	}
	
	return false;
}
