/*!
 * \file ConfigFile.h
 * \brief Declares the ConfigFile class that loads a configuration file.
 * \author Joe Forte
 * \date 2/2/2009
 */

#ifndef CONFIGFILE_H
#define CONFIGFILE_H

#include <map>
#include <string>
#include <vector>
#include "Math/Point.h"
#include "Math/Vector.h"
#include "Graphics/Color.h"
#include "Common/BaseTypes.h"

/*!
 * \class ConfigFile
 * \headerfile ConfigFile.h "Application/ConfigFile.h"
 * \brief Loads a configuration file.
 *
 * Each line of the config defines a key value pair in the following format:
 *
 *    key = value
 *
 * Where key is a case-sensitve single string of charachters without spaces or newlines.
 * Value is the data value associated with the key. To define a color use the color operator 
 * Color(r, g, b, a) where r, g, b, and a are in the range [0,1]. To define a vector use the Vector
 * operator Vector(x, y). All other data types will be auto detected.  If the type cannot be deduced,
 * it is assumed to be a string.
 *
 * Line comments are started with the ';' symbole.  All characters in a line following a semi-colon
 * are ignored.
 *
 * Duplicate occurances of a single key will overwrite the previous value.  The file is read from the 
 * top down.
 */
class ConfigFile
{
public:
	
	/*!
	 * \brief Constructs an empty config file.
	 */
	ConfigFile() {}
	
	/*!
	 * \brief Constructs and loads a config file.
	 *
	 * \param[in] file the name of the config file to load.
	 */
	ConfigFile(const char *file);
	
	/*!
	 * \brief Deconstructor.
	 */
	~ConfigFile() {}
	
	/*!
	 * \enum ValueType
	 * \headerfile ConfigFile.h "Application/ConfigFile.h"
	 * \brief Defins the diffeant types a value can have.
	 */
	enum ValueType
	{
		TYPE_UNKNOWN = -1, /*!< Unknown type. */
		TYPE_FLOAT = 0,    /*!< 32-bit floating point number. */
		TYPE_INTEGER,      /*!< 32-bit signed integer. */
		TYPE_STRING,       /*!< A null terminated string. */
		TYPE_BOOL,         /*!< A boolean value. */
		TYPE_VECTOR2,      /*!< A 2D floating point vector. */
		TYPE_COLOR4,       /*!< A 4 component floating point color. */
		
		/* Add new types */
		
		TYPE_COUNT         /*!< The number of types. */
	};
	
	/*!
	 * \class ConfigValue
	 * \headerfile ConfigFile.h "Application/ConfigFile.h"
	 * \brief Holds a single config value.
	 */
	class ConfigValue
	{
	public:
		
		/*!
		 * \brief Constructor.
		 */
		ConfigValue() {}
		
		/*!
		 * \brief Gets the name associated with the value.
		 *
		 * \return the name.
		 */
		const char *Name() const { return name.c_str(); }
		
		/*!
		 * \brief Gets the data type of the value.
		 *
		 * \return the data type.
		 *
		 */
		ValueType Type() const { return type; }
		
		/*!
		 * \brief Gets the value as a string.
		 *
		 * \return the value as a null terminated c-string.
		 */
		const char *StringValue() const { return value.c_str(); }
		
		/*!
		 * \brief Gets the value as a float.
		 *
		 * \return the value as a float.
		 */
		float FloatValue() const { return fltVal; }
		
		/*!
		 * \brief Gets the value as an integer.
		 *
		 * \return the value as an integer.
		 */
		int IntegerValue() const { return intVal; }
		
		/*!
		 * \brief Gets the value as a 2D Vector.
		 *
		 * \return the value as a 2D Vector.
		 */
		Vector2f VectorValue() const { return Vector2f(vec2[0], vec2[1]); }
		
		/*!
		 * \brief Gets the value as a 2D Point.
		 *
		 * \return the value as a 2D Point.
		 */
		Point2f PointValue() const { return Point2f(vec2[0], vec2[1]); }
		
		/*!
		 * \brief Gets the value as Color.
		 *
		 * \return the value as a Color.
		 */
		Color4f ColorValue() const { return Color4f(col4[0], col4[1], col4[2], col4[3]); }
		
	private:
		
		ConfigValue(const std::string &n) : name(n) {}
		
		void SetValue(const std::string &val);
		
		ValueType DetermineAndConvertType(const std::string &val);
		
		/*!
		 * \brief Converts a operator type into its component parts.
		 *
		 * \param[in] val the string with the operator type of the form: Operator(arg1, arg2, ... , argN)
		 * \param[in,out] op the name of the operator.
		 * \param[in,out] args the arguments to the operator.
		 *
		 * \return true if the operator is valid, false otherwise.
		 */
		bool GetOperatorAndParams(const std::string &val, std::string &op, std::vector<std::string> &args);
		
		bool IsColor4(const std::string &val, Color4f &col);
		bool IsVector2(const std::string &val, Vector2f &vec);
		
		std::string name;
		std::string value;
		
		ValueType type;
		
		union
		{
			int intVal;
			float fltVal;
			float vec2[2];
			float col4[4];
		};
		
		friend class ConfigFile;
	};
	
	/*!
	 * \brief Loads a config file.
	 *
	 * \param[in] file the name of the config file to load.
	 *
	 * \return true if the config file was loaded, false otherwise.
	 */
	bool Load(const char *file);
	
	/*!
	 * \brief Writes a config file.
	 *
	 * \param[in] name the name of the config file to write to.
	 *
	 * \return true if the config file was saved, false otherwise.
	 */
	bool Save(const char *name);
	
	/*!
	 * \brief Searches for the existance of a value by name.
	 *
	 * \param[in] name the name of the value to search for.
	 *
	 * \return The config value or NULL if the value does not exist.
	 */
	const ConfigValue *FindValueByName(const char *name) const;
	
	/*!
	 * \brief Retrieves a value.
	 *
	 * \param[in] name the name of the value.
	 *
	 * \return the requested value.
	 */
	const ConfigValue &operator[](const char *name) const;
	
	/*!
	 * \brief Creates or sets a value.
	 *
	 * \param[in] name the name of the value.
	 * \param[in] value the value to associate with the value.
	 *
	 * \return true if the value was set, false otherwise.
	 */
	bool SetValue(const char *name, const char *value);
	
	/*!
	 * \brief Creates or sets a value.
	 *
	 * \param[in] name the name of the value.
	 * \param[in] value the value to associate with the value.
	 *
	 * \return true if the value was set, false otherwise.
	 */
	bool SetValue(const std::string &name, const std::string &value);
	
protected:
	std::map<std::string, ConfigValue> values;
};

#endif
