/*!
 * \file Console.cpp
 * \brief Implements the Console class.
 * \author Joe Forte
 * \date 6/9/2008
 */

#include <ctime>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <algorithm>
#include "Common/Globals.h"
#include "Math/MathCommon.h"
#include "Input/Keyboard.h"
#include "Input/Mouse.h"
#include "Graphics/MathDraw.h"
#include "Application/DataPaths.h"
#include "Application/SDLApplication.h"
#include "Application/Console.h"
#include "Utility/StringUtility.h"

const char TAB_STRING[] = "    "; // The sting that replaces tab ('\t') characters.
const uint TAB_STRING_LEN = sizeof(TAB_STRING) / sizeof(char); 

const char *INPUT_INDICATOR = ">";
const char CONSOLE_ESCAPE = '|'; // The escape character used.  We don't use '\' because it can result in problems 
								 // when printing paths to the console.

// Keys that will be filtered from keyboard input.
const Key FILTERED_CONSOLE_KEYS[] = 
{
	Keys::KEY_RETURN,
	Keys::KEY_KP_ENTER,
	Keys::KEY_TAB,
};

const uint FILTERED_CONSOLE_KEYS_SIZE = sizeof(FILTERED_CONSOLE_KEYS) / sizeof(Key);

//--------------------------------------------------------------------
bool PredictiveStringMatch(const std::string &txt, const std::string &str)
{
	// Tests if the characters of txt matche the first characters of str.
	uint txtLen = (uint)txt.length();
	uint strLen = (uint)str.length();
	
	if(txtLen > strLen)
		return false;
	
	for(uint i = 0; i < txtLen; i++)
	{
		if(txt[i] != str[i])
			return false;
	}
	
	return true;
}

//--------------------------------------------------------------------
std::string ExtractCommonChars(uint offset, const std::vector<ConVar*> &vars)
{
	std::string common;
	
	if(vars.size() == 0)
		return common;

	// loops until we can no longer find any common characters between all of the vars.
	while(true)
	{
		if(offset >= vars[0]->GetNameString().length())
			return common;
		
		char commonChar = vars[0]->GetNameString()[offset];
		
		std::vector<ConVar*>::const_iterator i;
		for(i = vars.begin(); i != vars.end(); i++)
		{
			const std::string &str = (*i)->GetNameString();
			
			if(offset >= str.length() || str[offset] != commonChar)
				return common;
		}
		   
		offset++;
		common += commonChar;
	}
}

//--------------------------------------------------------------------
Console::Console()
:
isOpen(false),
lines(0),
startText(0),
height(0.0f),
width(std::numeric_limits<float>::max()),
characterWidth(0.0f),
maxCharactersPerLine(0),
cursorDraw(true),
cursorTimer(0.0f),
bufferedInput(FILTERED_CONSOLE_KEYS, FILTERED_CONSOLE_KEYS_SIZE)
{
	// Registers built in commands
	RegisterCmd("help", "Displays help information for the given command.", Console::ConCmdHelp);
	RegisterCmd("clear", "Clears all text in the console window.", Console::ConCmdClear);
	RegisterCmd("clearHistory", "Clears the console command history.", Console::ConCmdClearHistory);
	RegisterCmd("close", "Closes the console window.", Console::ConCmdClose);
	RegisterCmd("time", "Prints the current time.", Console::ConCmdTime);
	RegisterCmd("top", "Scrolls to the top of the console window.", Console::ConCmdTop);
	RegisterCmd("bottom", "Scrolls to the bottom of the console window.", Console::ConCmdBottom);
	RegisterCmd("revert", "Reverts the given console variable to its default value.", Console::ConCmdRevert);
	RegisterCmd("print", "Prints text to the console window.", Console::ConCmdPrint);
	RegisterCmd("find", "Searches the registered console commands and variables.", Console::ConCmdFind);
	RegisterCmd("version", "Prints the application version information.", Console::ConCmdVersion);
	RegisterCmd("dump", "Dumps all text and relevent information about the console to a file.", Console::ConCmdDump);

	// Only used for testing the console
	RegisterCmd("fib", "Calculates the nth fibonacci.", Console::ConCmdFib);
	
	con_bgColor = RegisterVar("con_bgColor", "0.0 0.0 0.0 0.9", "The background color of the console window.", ConVar::TYPE_COLOR);
	con_maxLines = RegisterVar("con_maxLines", "1000", "The maximum number of lines the console will hold.", ConVar::TYPE_UNSIGNED_INT);
    con_cmdHistory = RegisterVar("con_cmdHistory", "10", "The number of commands stored for easy access latter.", ConVar::TYPE_UNSIGNED_INT);
}

//--------------------------------------------------------------------
Console::~Console()
{
	std::map<std::string, ConVar*>::iterator i;
	for(i = conVars.begin(); i != conVars.end(); i++)
	{
		if(i->second)
			delete i->second;
	}
}

//--------------------------------------------------------------------
ConVar *Console::RegisterVar(const char *name, const char *defaultValue, const char *helpStr, ConVar::Type type)
{
	ConVar *var = new ConVar(name, defaultValue, helpStr, NULL, NULL, type, this);
	
	std::string cmdName = name;
	ToLower(cmdName);
	conVars[cmdName] = var;
	return var;
}

//--------------------------------------------------------------------
void Console::RegisterCmd(const char *name, const char *helpStr, fnConCommad cmd, void *userData)
{
	ConVar *var = new ConVar(name, "", helpStr, cmd, userData, ConVar::TYPE_COMMAND, this);
	
	std::string cmdName = name;
	ToLower(cmdName);
	conVars[cmdName] = var;
}

//--------------------------------------------------------------------
void Console::Open()
{
	isOpen = true;
}

//--------------------------------------------------------------------
void Console::Close()
{	
	isOpen = false;
}

//--------------------------------------------------------------------
bool Console::IsOpened() const
{
	return isOpen;
}

//--------------------------------------------------------------------
void Console::Update(float dt, const IMouse &m, const IKeyboard &kb)
{
	if(!IsOpened())
		return;
	
	if( kb.KeyPressed(Keys::KEY_RETURN) || kb.KeyPressed(Keys::KEY_KP_ENTER) )
	{
		if(bufferedInput.GetInput().size() > 0)
		{
			AddCmd(bufferedInput.GetInput().c_str());
			cmdHistoryIndex = -1;

			ExecuteCommand( bufferedInput.GetInput().c_str() );
			
			bufferedInput.ClearInput();
			
			// Its possible for a command to close the console
			if(!IsOpened())
				return;
		}
	}
	else if( kb.KeyPressedRepeat(Keys::KEY_UP) && (kb.KeyDown(Keys::KEY_LSHIFT) || kb.KeyDown(Keys::KEY_RSHIFT)) )
	{
		ScrollUp();
	}
	else if( kb.KeyPressedRepeat(Keys::KEY_UP) )
	{
		NextCmd();
	}
	else if( kb.KeyPressedRepeat(Keys::KEY_DOWN) && (kb.KeyDown(Keys::KEY_LSHIFT) || kb.KeyDown(Keys::KEY_RSHIFT)) )
	{
		ScrollDown();
	}
	else if( kb.KeyPressedRepeat(Keys::KEY_DOWN) )
	{
		PrevCmd();
	}
	else if( kb.KeyPressedRepeat(Keys::KEY_PAGEUP) )
	{
		PageUp();
	}
	else if( kb.KeyPressedRepeat(Keys::KEY_PAGEDOWN) )
	{
		PageDown();
	}
	
	bufferedInput.Update(kb);

	if(m.WheelUp())
	{
		ScrollUp();
	}
	else if(m.WheelDown())
	{
		ScrollDown();
	}
	
	std::string input = bufferedInput.GetInput();
	uint cursorPos = bufferedInput.GetCursorPos();
	
	// Handels predictive text
	if(kb.KeyPressed(Keys::KEY_TAB) && input.length() > 0)
	{
		std::string preQuery = "";
		std::string query = "";
		
		size_t start = input.find_last_of(" ");
		if(start == std::string::npos)
		{
			query = input;
		}
		else
		{
			preQuery = input.substr(0, start + 1);
			query = input.substr(start + 1);
		}
		
		if(query.size() > 0)
		{
			ToLower(query);
			
			std::vector<ConVar*> vars;
			Predict(query, vars);
			
			if(vars.size() == 1)
			{
				input = preQuery + vars[0]->GetNameString() + " ";
			}
			else if(vars.size() > 1)
			{
				ConStream out = Print(Color::Yellow);
				
				out << "Possible matches (" << Console::SetColor(Color::White) << (uint)vars.size() << Console::RevertColor() << "):\n";
				
				std::vector<ConVar*>::reverse_iterator i;
				for(i = vars.rbegin(); i != vars.rend(); i++)
				{
					out << "  " << (*i)->GetName() << "\n";
				}
				
				std::string common = ExtractCommonChars((uint)query.length(), vars);
				if(common.size() > 0)
				{
					input = preQuery + query + common;
				}
			}
			
			bufferedInput.SetInput(input.c_str());
			bufferedInput.SetCursorPos((uint)input.length());
		}
	}
				 
	cursorTimer += dt;
	if(cursorTimer >= CURSOR_BLINK_TIME)
	{
		cursorDraw = !cursorDraw;
		cursorTimer = 0.0f;
	}
	
	EnforceMaxLineCount();
}

//--------------------------------------------------------------------
void Console::Draw()
{
	// Offset from the size of the screen when drawing text
	const float xOffset = 2.0f;
	const float yOffset = 5.0f;
	
	// The offset from the input indicator
	const float inputOffset = 3.0f;
	
	if(!IsOpened() || !font.IsValid())
		return;
	
	// Draws the background
	Rectf r = GetScreen()->GetRect();
	r.h = height;
	
	DrawRect(r, true, con_bgColor->GetColor());
	
	Line line;
	line.a = Point2f(r.x, r.y - r.h);
	line.b = Point2f(r.x + r.w, r.y - r.h);
	DrawLine(line, Color::Gray, 1.0f);
	
	// Draws the input line
	float lineHeight = font->LineHeight();
	
	float drawX = r.x + xOffset;
	float drawY = r.y - r.h + lineHeight + yOffset;
	Point2f drawPos(drawX, drawY);
	
	font->DrawTextStr(INPUT_INDICATOR, Color::Green, drawPos);
	drawPos.x += font->StringWidth(INPUT_INDICATOR) + inputOffset;
	
	std::string input = bufferedInput.GetInput();
	font->DrawTextStr(input.c_str(), Color::Green, drawPos);
	
	// Draws the input cursor
	if(cursorDraw)
	{
		float cw = font->StringWidth(input.substr(0, bufferedInput.GetCursorPos()).c_str());
		float x = drawPos.x + cw;
		float y = drawPos.y;
		
		Line cursor;
		cursor.a = Point2f(x, y - yOffset);
		cursor.b = Point2f(x, y - lineHeight);
		
		DrawLine(cursor, Color::White, 1.0f);
	}
	
	// Draws the text
	uint linesToDraw = GetNumberOfVisibleLines();
	ConTextType::reverse_iterator i = text.rbegin();
	
	// moves to the correct text line
	{
		int count = 0;
		while(i != text.rend() && count < startText)
		{
			i++;
			count++;
		}
	}
	
	// Skips the first line if its empty
	if(i != text.rend() && startText == 0 && i->size() == 0)
		i++;
	
	while(i != text.rend() && linesToDraw > 0)
	{
		drawPos.y += lineHeight;
		drawPos.x = drawX;
		
		for(ConLine::iterator j = i->begin(); j != i->end(); j++)
		{
			font->DrawTextStr(j->str.c_str(), j->c, drawPos);
			drawPos.x += font->StringWidth(j->str.c_str());
		}
		
		linesToDraw--;
		i++;
	}
}

//--------------------------------------------------------------------
uint Console::GetNumberOfVisibleLines() const
{
	return std::max( (int)std::ceil( (height / font->LineHeight()) - 1.0f ), 0 );
}

//--------------------------------------------------------------------
void Console::SetFont(FontRef f)
{
	font = f;
	
	characterWidth = f->StringWidth("M");
	maxCharactersPerLine = (uint)std::floor(width / characterWidth);
}

//--------------------------------------------------------------------
void Console::SetWidth(float w)
{
	width = w;
	maxCharactersPerLine = (uint)std::floor(width / characterWidth);
}

//--------------------------------------------------------------------
void Console::PushPrependText(const char *str, const Color4f &c)
{
	PushPrependText(std::string(str), c);
}

//--------------------------------------------------------------------
void Console::PushPrependText(const std::string &str, const Color4f &c)
{
	ConString conStr;
	conStr.str = str;
	conStr.c = c;
	
	Replace(conStr.str, "\n\r", ' ');
	Replace(conStr.str, "\t", TAB_STRING);
	
	prependText.push_back(conStr);
}

//--------------------------------------------------------------------
void Console::PopPrependText()
{
	if(!prependText.empty())
		prependText.pop_back();
}

//--------------------------------------------------------------------
void Console::Print(const char *str, const Color4f &c)
{
	OutputToConsole(str, c, NULL);
}

//--------------------------------------------------------------------
void Console::PrintLine(const char *str, const Color4f &c)
{
	std::string s = str;
	s += "\n";
	OutputToConsole(s.c_str(), c, NULL);
}

//--------------------------------------------------------------------
void Console::OutputToConsole(const char *str, const Color4f &c, ConStream *stream)
{
	std::string line = str;

	Color4f currColor = c;
	
	uint i = 0;
	uint start = 0;
	uint len = 0;
	while(line[i])
	{
		if(line[i] == '\n')
		{
			AddLine(line.substr(start, len), currColor);
			NewLine();
			start = i + 1;
			len = 0;
			i++;
		}
		else if(line[i] == '\t')
		{
			line.replace(i, 1, TAB_STRING);
			len = len + TAB_STRING_LEN;
			i = i + TAB_STRING_LEN;
		}
		else
		{
			len++;
			i++;
		}
	}
	
	if(len > 0)
		AddLine(line.substr(start, len), currColor);
}

//--------------------------------------------------------------------
void Console::NewLine()
{
	ConLine newLine;
	text.push_back(newLine);
	lines++;
}

//--------------------------------------------------------------------
void Console::AddLine(const char *str, const Color4f &c)
{
	AddLine(std::string(str), c);
}

//--------------------------------------------------------------------
void Console::AddLine(const std::string &str, const Color4f &c)
{
	std::vector<ConLine> wrappedLine;
	
	if(text.size() > 0)
	{
		ConLine &last = text.back();
		if(last.size() > 0)
		{
			if(last.back().c == c)
			{
				last.back().str += str;
			}
			else
			{
				ConString conStr;
				conStr.str = str;
				conStr.c = c;
				last.push_back(conStr);
			}
		}
		else
		{
			last = prependText;
			
			ConString s;
			s.str = str;
			s.c = c;
			last.push_back(s);
		}
		
		ConLineWrap(last, wrappedLine);
		text.pop_back();
		lines--;
	}
	else
	{
		ConString s;
		s.str = str;
		s.c = c;
		
		ConLine line = prependText;
		line.push_back(s);
		
		ConLineWrap(line, wrappedLine);
	}
	
	size_t count = wrappedLine.size();
	for(size_t i = 0; i < count; i++)
	{
		text.push_back(wrappedLine[i]);
		lines++;
	}
}

//--------------------------------------------------------------------
void Console::SplitString(uint charsLeft, const std::string &str, std::string &left, std::string &right)
{
	uint len = (uint)str.size();
	uint over = len - charsLeft; // the number of characters that are over the max limit
	uint end = charsLeft; // the index to the character to split on
	
	// Searches for a suitable split point. Stops before we get to the start of the string so we dont get stuck in an infinite loop.
	while(end > 0 && !Match(str[end], " ")) 
	{
		end--;
	}
	
	// A suitable split character could not be found so we use the original end location
	if(end == 0) 
		end = charsLeft;
	else if(end < charsLeft) // Makes sure we split the string after the deliminator
		end++;

	left = str.substr(0, end);
	right = str.substr(end, len);
}

//--------------------------------------------------------------------
uint Console::SplitConString(uint charsLeft, const ConString &conStr, ConLine &currLine, std::vector<ConLine> &out)
{
	ConString left;
	ConString right;
	left.c = conStr.c;
	right.c = conStr.c;
	
	std::string toSplit = conStr.str;
	
	uint len = (uint)toSplit.size();
	while(len > maxCharactersPerLine)
	{
		std::string tmp = right.str;
		SplitString(charsLeft, toSplit, left.str, right.str);

		len = (uint)right.str.size();
		charsLeft = len;

		if(len > maxCharactersPerLine) // Kinda ugly but it avoids a needless string copy
			toSplit = right.str;

		currLine.push_back(left);
		out.push_back(currLine);
		currLine.clear();
	}

	if(len > 0)
		currLine.push_back(right);

	assert(maxCharactersPerLine >= len);

	return maxCharactersPerLine - len;
}

//--------------------------------------------------------------------
void Console::ConLineWrap(const ConLine &line, std::vector<ConLine> &out)
{
	if(maxCharactersPerLine == 0)
	{
		out.push_back(line);
		return;
	}
	
	ConLine currLine;
	uint charsLeft = maxCharactersPerLine;
	for(ConLine::const_iterator i = line.begin(); i != line.end(); i++)
	{
		uint len = (uint)i->str.size();
		if(len > charsLeft)
		{
			charsLeft = SplitConString(charsLeft, *i, currLine, out);
		}
		else
		{
			currLine.push_back(*i);
			charsLeft -= len;
			if(charsLeft == 0)
			{
				out.push_back(currLine);
				currLine.clear();
			}
		}
	}
	
	if(currLine.size() > 0)
		out.push_back(currLine);
}

//--------------------------------------------------------------------
void Console::EnforceMaxLineCount()
{
	int maxLines = con_maxLines->GetInt();
	if(lines > maxLines)
	{
		int linesToRemove = lines - con_maxLines->GetInt();
		while(linesToRemove > 0)
		{
			text.pop_front();
			linesToRemove--;
			lines--;
		}
	}
}

//--------------------------------------------------------------------
void Console::ExecuteCommand(const char *cmd)
{
	std::vector<std::string> args;
	ParseCommand(cmd, args);
	
	if(args.size() == 0)
		return;

	ConVar *var = FindVar(args[0].c_str());
	
	if(var)
	{		
		if(var->GetType() == ConVar::TYPE_COMMAND)
		{
			fnConCommad func = var->GetCommand();
			
			if(func)
			{
				func(this, cmd, var->GetUserData(), args);
			}
			else
			{
				PrintWarning() << "The console command '" << args[0].c_str() << "' is not bound to a valid function.\n";
			}
		}
		else
		{
			if(var)
			{
				if(args.size() > 1)
				{
					const char *in = cmd;
					while(*in && *in == ' ') in++; // Skips any leading white spaces
					while(*in && *in != ' ') in++; // advances to the end of the var name

					// Skips the deliminating space
					if(*in) in++;

					var->SetString(in);

					var->ImformListenersOfChange(in, args);
				}

				Print(Color::White) << var->GetName() << " = " << var->GetString() << "\n";
			}
			else
			{
				PrintWarning() << "The console variable '" << args[0].c_str() << "' was not properly constructed.\n";
			}
		}
	}
	else
	{
		PrintWarning() << "Unknown console command or variable '" << Console::SetColor(Color::Green) << args[0].c_str() << Console::RevertColor() << "'\n";
	}
}

//--------------------------------------------------------------------
void Console::ParseCommand(const char *cmd, std::vector<std::string> &args) const
{
	const char *tokenStart = NULL;
	uint tokenLen = 0;
	
	// skips over any leading spaces
	while(*cmd && *cmd == ' ')
		cmd++;
	
	tokenStart = cmd;
	
	// Parses the string adding each token found to the args vector (tokens are deliminated by whitespace)
	while(*cmd)
	{
		if(*cmd == ' ')
		{
			args.push_back( std::string(tokenStart, tokenLen) );
			
			tokenLen = 0;
			
			// Skips any white space between tokens
			while(*cmd && *cmd == ' ')
				cmd++;
			
			tokenStart = cmd;
		}
		else
		{
			cmd++;
			tokenLen++;
		}
	}
	
	if(tokenLen > 0)
		args.push_back( std::string(tokenStart, tokenLen) );
}

//--------------------------------------------------------------------
ConVar *Console::FindVar(const char *name)
{
	std::string cmd = name;
	ToLower(cmd);
	
	std::map<std::string, ConVar*>::iterator i = conVars.find(cmd);
	if(i != conVars.end())
		return i->second;
	else
		return NULL;
}

//--------------------------------------------------------------------
void Console::NextCmd()
{
	if(cmdHistory.size() == 0)
		return;
	
	if(cmdHistoryIndex == -1)
	{
		currCmdInput = bufferedInput.GetInput();
		currCmdInputCursorPos = bufferedInput.GetCursorPos();
	}
	
	cmdHistoryIndex = std::min(cmdHistoryIndex + 1, (int)cmdHistory.size() - 1);
	
	assert(cmdHistoryIndex >= 0);
	
	std::list<std::string>::const_iterator it = cmdHistory.begin();
	for(int i = 0; i < cmdHistoryIndex; i++) it++;
	
	bufferedInput.SetInput(*it);
	bufferedInput.MoveCursorToEnd();
}

//--------------------------------------------------------------------
void Console::PrevCmd()
{
	if(cmdHistory.size() == 0 || cmdHistoryIndex == -1)
		return;
	
	assert(cmdHistoryIndex >= 0);
	
	cmdHistoryIndex--;
	
	// We store the current input state when changing to an input in the history list for the first time
	if(cmdHistoryIndex == -1)
	{		
		bufferedInput.SetInput(currCmdInput);
		bufferedInput.SetCursorPos(currCmdInputCursorPos);
	}
	else
	{
		std::list<std::string>::const_iterator it = cmdHistory.begin();
		for(int i = 0; i < cmdHistoryIndex; i++) it++;
		
		bufferedInput.SetInput(*it);
		bufferedInput.MoveCursorToEnd();
	}
}

//--------------------------------------------------------------------
void Console::AddCmd(const char *cmd)
{
	// Skips over the command if it is the same as the last command entered
	if(cmdHistory.size() > 0 && cmdHistory.front() == cmd)
		return;

	// We have to account for the fact that con_cmdHistory can change at any time
	if(cmdHistory.size() >= con_cmdHistory->GetUInt())
	{
		uint toRemove = (uint)cmdHistory.size() - con_cmdHistory->GetUInt();
		while(toRemove > 0)
		{
			cmdHistory.pop_back();
			toRemove--;
		}
	}

	cmdHistory.push_front( std::string(cmd) );
}

//--------------------------------------------------------------------
void Console::ScrollUp()
{
	startText = std::min(startText + 1, lines - 1);
}

//--------------------------------------------------------------------
void Console::ScrollDown()
{
	startText = std::max(startText - 1, 0);
}

//--------------------------------------------------------------------
void Console::ScrollTop()
{
	int page = GetNumberOfVisibleLines();
	startText = std::max(lines - page + 1, 0);
}

//--------------------------------------------------------------------
void Console::ScrollBottom()
{
	startText = 0;
}

//--------------------------------------------------------------------
void Console::PageUp()
{
	int page = GetNumberOfVisibleLines();
	startText = std::min(startText + page, lines - 1);
}

//--------------------------------------------------------------------
void Console::PageDown()
{
	int page = GetNumberOfVisibleLines();
	startText = std::max(startText - page, 0);
}

//--------------------------------------------------------------------
void Console::DumpToFile(const char *name)
{
	char buff[64];
	time_t sec = time(NULL);
	
#ifdef _WIN32
	tm local;
	localtime_s(&local, &sec);
	strftime(buff, 64, "%m_%d_%Y_%H%M%S", &local); // month_day_year_HoursMinutesSeconds
#else
	tm *local = localtime(&sec);
	strftime(buff, 64, "%m_%d_%Y_%H%M%S", local); // month_day_year_HoursMinutesSeconds
#endif

	std::string fileName;
	if(!name)
	{
		fileName = "condump_"; 
		fileName += buff;
		fileName += ".txt";
	}
	else
	{
		fileName = name;
	}

	fileName = ConcatPath(DataPaths::GetPath("Dump"), fileName);

	std::ofstream file(fileName.c_str());
	if(!file.is_open())
	{
		PrintError() << "Console could not open the file '" << fileName << "' for writting.\n";
		return;
	}

	// Outputs general console related data
	file << "Console Dump Performed on: ";
#ifdef _WIN32
	asctime_s(buff, 64, &local);
	file << buff;
#else // Are there equivalent safe versions for gcc?	
	file << asctime(local);
#endif
	file << "\n\n";

	// Counts the number of Console variables making note of their specific type
	uint varTypes[ConVar::TYPE_COUNT];
	memset((void*)varTypes, 0, sizeof(uint)*ConVar::TYPE_COUNT);
	{
		std::map<std::string, ConVar*>::const_iterator i = conVars.begin();
		while(i != conVars.end())
		{
			ConVar *cv = i->second;
			if(cv)
			{
				varTypes[cv->GetType()]++;
			}
			i++;
		}
	}

	uint total = 0;
	uint totalVars = 0;
	for(uint t = 0; t < ConVar::TYPE_COUNT; t++)
	{
		if(t != ConVar::TYPE_COMMAND)
			totalVars += varTypes[t];
		total += varTypes[t];
	}

	file << "Total number of console variables and commands registered:\n";
	file << "     Total console variables : " << totalVars << "\n";
	file << "                         int : " << varTypes[ConVar::TYPE_INT] << "\n";
	file << "                unsigned int : " << varTypes[ConVar::TYPE_UNSIGNED_INT] << "\n";
	file << "                       float : " << varTypes[ConVar::TYPE_FLOAT] << "\n";
	file << "                      string : " << varTypes[ConVar::TYPE_STRING] << "\n";
	file << "                      vector : " << varTypes[ConVar::TYPE_VECTOR] << "\n";
	file << "                       point : " << varTypes[ConVar::TYPE_POINT] << "\n";
	file << "                       color : " << varTypes[ConVar::TYPE_COLOR] << "\n";
	file << "      Total console commands : " << varTypes[ConVar::TYPE_COMMAND] << "\n";
	file << "                       TOTAL : " << total << "\n\n";

	file << "         Max number of lines : " << con_maxLines->GetUInt() << "\n";
	file << "     Current number of lines : " << lines << "\n\n";

	file << "---BEGIN TEXT DUMP----------------------------------------------\n\n\n";

	// Outputs the actual text data
	{
		// determins the number of characters for the largest line number 
		std::stringstream maxLineNumber;
		maxLineNumber << lines;
		uint width = (uint)maxLineNumber.str().length();

		uint count = 0;
		std::list<ConLine>::const_iterator i = text.begin();
		while(i != text.end())
		{
			file << std::setw(width) << count << "| ";
			for(ConLine::const_iterator j = i->begin(); j != i->end(); j++)
			{
				file << j->str;
			}
			
			file << "\n";
			
			count++;
			i++;
		}
	}

	file << "\n---END TEXT DUMP------------------------------------------------\n\n";
}

//--------------------------------------------------------------------
void Console::ClearHistory()
{
	cmdHistory.clear();
}

//--------------------------------------------------------------------
void Console::Clear()
{
	text.clear();
	lines = 0;
}

//--------------------------------------------------------------------
bool ConVarSort(const ConVar *v1, const ConVar *v2)
{
	return v1->GetNameString() < v2->GetNameString();
}

//--------------------------------------------------------------------
void Console::Search(const std::string &txt, std::vector<ConVar*> &vars)
{
	std::string cmd = txt;
	ToLower(cmd);
	
	std::map<std::string, ConVar*>::const_iterator i;
	for(i = conVars.begin(); i != conVars.end(); i++)
	{
		if( i->first.find(cmd) != std::string::npos)
		{
			vars.push_back(i->second);
		}
	}
	
	std::sort(vars.begin(), vars.end(), ConVarSort);
}

//--------------------------------------------------------------------
void Console::Predict(const std::string &txt, std::vector<ConVar*> &vars)
{
	std::map<std::string, ConVar*>::const_iterator i;
	for(i = conVars.begin(); i != conVars.end(); i++)
	{
		if(PredictiveStringMatch(txt, i->first))
		{
			vars.push_back(i->second);
		}
	}
	
	std::sort(vars.begin(), vars.end(), ConVarSort);
}

//--------------------------------------------------------------------
void Console::ConCmdHelp(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	if(args.size() > 1)
	{
		ConVar *cv = con->FindVar(args[1].c_str());
		if(!cv)
		{
			con->Print(Color::Yellow) << "The console variable '" << SetColor(Color::Green) << args[1] << RevertColor() << "' could not found.\n"; 
		}
		else
		{
			ConVar::Type type = cv->GetType();
			
			con->Print(Color::Green) << args[1] << "\n";
			con->Print(Color::White) << "    Help         : " << cv->GetHelpStr() << "\n";
			con->Print(Color::White) << "    Type         : ";
			switch(type)
			{
				case ConVar::TYPE_INT:
					con->Print(Color::Blue) << "int\n";
					break;

				case ConVar::TYPE_UNSIGNED_INT:
					con->Print(Color::Blue) << "unsigned int\n";
					break;

				case ConVar::TYPE_VECTOR:
					con->Print(Color::Blue) << "vector\n";
					break;

				case ConVar::TYPE_POINT:
					con->Print(Color::Blue) << "point\n";
					break;

				case ConVar::TYPE_COLOR:
					con->Print(Color::Blue) << "color RGBA\n";
					break;

				case ConVar::TYPE_FLOAT:
					con->Print(Color::Blue) << "float\n";
					break;
					
				case ConVar::TYPE_STRING:
					con->Print(Color::Blue) << "string\n";
					break;
					
				case ConVar::TYPE_COMMAND:
					con->Print(Color::Blue) << "command\n";
					break;
			};
			
			if(type != ConVar::TYPE_COMMAND)
			{
				con->Print(Color::White) << "    Current Value: " << cv->GetString() << "\n";
				con->Print(Color::White) << "    Default Value: " << cv->GetDefaultValue() << "\n";
			}
		}
	}
	else
	{
		con->Print(Color::White) << SetColor(Color::Yellow) << "Usage: " << SetColor(Color::Green) << "help " 
								 << RevertColor() << "[" << SetColor(0xFFD600FF) << "command" << RevertColor() << "]\n";
		con->Print(Color::White) << "       Displays the help information for the given " << SetColor(Color::Yellow) << "command" 
								 << RevertColor() << ".\n";
	}
}

//--------------------------------------------------------------------
void Console::ConCmdClear(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	con->Clear();
}

//--------------------------------------------------------------------
void Console::ConCmdClearHistory(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	con->Print(Color::White) << "Console histoy cleared\n";
	con->ClearHistory();
}

//--------------------------------------------------------------------
void Console::ConCmdClose(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	con->Close();
}

//--------------------------------------------------------------------
void Console::ConCmdTime(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	time_t sec = time(NULL);

#ifdef _WIN32
	tm local;
	localtime_s(&local, &sec);

	char buff[64];
	asctime_s(buff, 64, &local);

	con->Print(Color::White) << buff;
#else // Are there equivalent safe versions for gcc?
	tm *local = localtime(&sec);
	
	con->Print(Color::White) << asctime(local);
#endif
}

//--------------------------------------------------------------------
void Console::ConCmdTop(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	con->ScrollTop();
}

//--------------------------------------------------------------------
void Console::ConCmdBottom(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	con->ScrollBottom();
}

//--------------------------------------------------------------------
void Console::ConCmdRevert(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	if(args.size() > 1)
	{
		ConVar *cv = con->FindVar(args[1].c_str());
		if(!cv)
		{
			con->Print(Color::Yellow) << "The console variable '" << args[1] << "' could not found.\n"; 
		}
		else if(cv->GetType() == ConVar::TYPE_COMMAND)
		{
			con->PrintWarning() << "'" << args[1] << "' is not a variable and cannot be reverted.\n";
		}
		else
		{
			con->Print(Color::Green) << args[1] << "\n";
			con->Print(Color::White) << "    Old value: " << cv->GetString() << "\n";
			cv->Revert();
			con->Print(Color::White) << "    New value: " << cv->GetString() << "\n";
		}
	}
	else
	{		
		con->Print(Color::White) << SetColor(Color::Yellow) << "Usage: " << SetColor(Color::Green) << "revert " 
								 << RevertColor() << "[" << SetColor(0xFFD600FF) << "console variable" << RevertColor() << "]\n";
		con->Print(Color::White) << SetColor(Color::Yellow) << "       console variable" << RevertColor() << " - Reverts the given to its default state.\n";
	}	
}

//--------------------------------------------------------------------
void Console::ConCmdPrint(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	Console::ConStream stream = con->Print(Color::White);
	
	uint len = (uint)args.size();
	for(uint i = 1; i < len; i++)
	{
		stream << args[i] << " ";
	}
	
	stream << "\n";
}

//--------------------------------------------------------------------
void Console::ConCmdFind(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	if(args.size() <= 1)
	{
		con->Print(Color::White) << SetColor(Color::Yellow) << "Usage: " << SetColor(Color::Green) << "find " 
								 << RevertColor() << "[" << SetColor(0xFFD600FF) << "text" << RevertColor() << "]\n";
		con->Print(Color::White) << SetColor(Color::Yellow) << "       text" << RevertColor() << " - The string to search for in the console commands\\vars.\n";
	}
	else
	{
		std::vector<ConVar*> vars;
		con->Search(args[1], vars);
		
		if(vars.size() == 0)
		{
			con->Print(Color::Yellow) << "No matches found.\n";
		}
		else
		{
			ConStream out = con->Print(Color::Yellow);
			ConStream outCmd = con->Print(Color::Green);
			ConStream outHelp = con->Print(Color::White);
			
			out << "Possible matches found (" << SetColor(Color::White) << (uint)vars.size() << RevertColor() << "):\n";
			for(std::vector<ConVar*>::iterator i = vars.begin(); i != vars.end(); i++)
			{
				outCmd << "  " << (*i)->GetName() << "\n";
				outHelp << "    " << (*i)->GetHelpStr() << "\n";
			}
		}
	}
}

//--------------------------------------------------------------------
void Console::ConCmdVersion(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	SDLApplication *app = GetEnv()->GetApp();
	con->Print() << app->GetAppName() << " " << app->GetVersionString() << "\n";
}

//--------------------------------------------------------------------
void Console::ConCmdDump(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	if(args.size() > 1)
		con->DumpToFile(args[1].c_str());
	else
		con->DumpToFile();

	con->Print(Color::White) << "Console dump complete.\n";
}

//--------------------------------------------------------------------
void Console::ConCmdFib(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	if(args.size() > 1)
	{
		std::istringstream iss(args[1]);
		uint n;
		if( (iss >> n).fail() )
		{
			con->Print(Color::Yellow) << "'" << args[1] << "' is not a positive integer.\n";
		}

		uint a = 1;
		uint b = 1;

		con->Print(Color::White) << 0 << "\n";
		while(n > 0)
		{
			con->Print(Color::White) << a << "\n";
			uint tmp = a;
			a = b;
			b = tmp + b;
			n--;
		}
	}
	else
	{		
		con->Print(Color::White) << SetColor(Color::Yellow) << "Usage: " << SetColor(Color::Green) << "fib " 
								 << RevertColor() << "[" << SetColor(0xFFD600FF) << "n" << RevertColor() << "]\n";
		con->Print(Color::White) << "       Prints out the first " << SetColor(0xFFD600FF) << "n" << RevertColor() << " Fibonacci numbers.\n";
		con->Print(Color::White) << SetColor(Color::Yellow) << "       n" << RevertColor() << " - must be a positive integer.\n";
	}
}

//--------------------------------------------------------------------
Console::ConStream &Console::ConStream::operator<<(const char *str)
{
	con->OutputToConsole(str, currColor, this);
	return *this;
}

//--------------------------------------------------------------------
Console::ConStream &Console::ConStream::operator<<(const std::string &str)
{
	con->OutputToConsole(str.c_str(), currColor, this);
	return *this;
}

//--------------------------------------------------------------------
Console::ConStream &Console::ConStream::operator<<(float v)
{
	std::stringstream str;
	str << v;
	con->OutputToConsole(str.str().c_str(), currColor, this);
	return *this;
}

//--------------------------------------------------------------------
Console::ConStream &Console::ConStream::operator<<(int i)
{
	std::stringstream str;
	str << i;
	con->OutputToConsole(str.str().c_str(), currColor, this);
	return *this;
}

//--------------------------------------------------------------------
Console::ConStream &Console::ConStream::operator<<(uint ui)
{
	std::stringstream str;
	str << ui;
	con->OutputToConsole(str.str().c_str(), currColor, this);
	return *this;
}

//--------------------------------------------------------------------
Console::ConStream &Console::ConStream::operator<<(byte b)
{
	std::stringstream str;
	str << (int)b;
	con->OutputToConsole(str.str().c_str(), currColor, this);
	return *this;
}
