/*!
 * \file Console.h
 * \brief Declares the Console class used for outputting text and accepting user input in-game.
 * \author Joe Forte
 * \date 6/9/2008
 */

#ifndef CONSOLE_H
#define CONSOLE_H

#include <map>
#include <list>
#include <string>
#include <vector>
#include "Math/MathCommon.h"
#include "Graphics/Color.h"
#include "Graphics/Font.h"
#include "Application/ConVar.h"
#include "Input/BufferedKeyboardInput.h"

class IKeyboard;
class IMouse;

/*!
 * \brief The amount of time between blinks of the input cursor.
 */
const float CURSOR_BLINK_TIME = 0.5f;

/*!
 * \class Console
 * \headerfile Console.h "Application/Console.h
 * \brief Class that allows for in-game text output and user input. 
 */
class Console
{
public:
	/*!
	 * \brief Default constructor.
	 */
	Console();
	
	/*!
	 * \brief De-Constructor.
	 */
	~Console();
	
	/*!
	 * \brief Sets the toggle key that sould be ignored duing input buffering.
	 */
	void IgnoreToggleKey(Key k) { bufferedInput.AddFilteredKey(k); }
	
	/*!
	 * \brief Registers a Console Variable with the Console.
	 *
	 * \param[in] name the name used to refer to the console variable. Not case sensitive.
	 * \param[in] defaultValue the initial value of the console variable.  Assumed to be formatted correctly for the choosen type.
	 * \param[in] helpStr a string describing the console variable.
	 * \param[in] type the data type the console variable represents.
	 *
	 * \return A pointer to the newly registered Console variable.  DO NOT deallocate the returned pointer. 
	 */
	ConVar *RegisterVar(const char *name, const char *defaultValue, const char *helpStr, ConVar::Type type);
	
	/*!
	 * \brief Registers a Console Command with the Console.
	 *
	 * \param[in] name the name used to refer to the console command. Not case sensitive.
	 * \param[in] helpStr a string describing the console variable.
	 * \param[in] cmd pointer to the function that will be called when the console command is evoked.
	 * \param[in] userData a user defined pointer that will be passed to 'cmd' when it is called.  Defaults to NULL.
	 */
	void RegisterCmd(const char *name, const char *helpStr, fnConCommad cmd, void *userData = NULL);
	
	/*!
	 * \brief Searches the console for matching command or variable.
	 *
	 * \param[in] name the name of a console variable or command to find.  Not case sensitive.
	 *
	 * \return A pointer to the console variable or NULL if no variable or command is found.
	 */
	ConVar *FindVar(const char *name);
	
	/*!
	 * \brief Opens the console window.
	 */
	void Open();
	
	/*!
	 * \brief Closes the console window.
	 */
	void Close();
	
	/*!
	 * \brief Determines if the console windows is open or not.
	 *
	 * \return True if the console window is open, false otherwise.
	 */
	bool IsOpened() const;
	
	/*!
	 * \brief Updates the console, should be called once every frame.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 * \param[in] m the mouse to accept input from.
	 * \param[in] kb the keyboard to accept input from.
	 */
	void Update(float dt, const IMouse &m, const IKeyboard &kb);
	
	/*!
	 * \brief Draws the console window.
	 *
	 * NOTE: The console window is only drawn if the console window is actually open.
	 */
	void Draw();
	
	/*!
	 * \brief Sets the font used to draw the console text.
	 *
	 * \param[in] f the font.
	 *
	 * NOTE: The font is assumed to be a mono space font.  Its not required but it works best with the type of word wrapping used.
	 */
	void SetFont(FontRef f);
	
	/*!
	 * \brief Sets the height of the console window.
	 *
	 * \param[in] h the height of the console window.
	 */
	void SetHeight(float h) { height = h; }
	
	/*!
	 * \brief Sets the width of the console window.
	 *
	 * \param[in] w the width of the console window.
	 */
	void SetWidth(float w);
	
	/*!
	 * \brief Executes a console command.
	 *
	 * \param[in] cmd the command to execute.
	 */
	void ExecuteCommand(const char *cmd);
	
	/*!
	 * \brief Determines the maximum number of lines that can be visible in the console window.
	 *
	 * \return The number of lines.
	 */
	uint GetNumberOfVisibleLines() const;
	
	/*!
	 * \brief Scrolls the text in the console window up by one line.
	 */
	void ScrollUp();
	
	/*!
	 * \brief Scrolls the text in the console window down by one line.
	 */
	void ScrollDown();
	
	/*!
	 * \brief Scrolls the text in the console window to the top line.
	 */
	void ScrollTop();
	
	/*!
	 * \brief Scrolls the text in the console window to the bottom line.
	 */
	void ScrollBottom();
	
	/*!
	 * \brief Scrolls the text in the console window up by one page.
	 */
	void PageUp();
	
	/*!
	 * \brief Scrolls the text in the console window down by one page.
	 */
	void PageDown();

	/*!
	 * \brief Dumps console text and statistics to a file.
	 *
	 * \param[in] name optional name of the file to dump to.  Defaults to NULL.
	 */
	void DumpToFile(const char *name = NULL);
	
	/*!
	 * \brief Clears the console command history list.
	 */
	void ClearHistory();
	
	/*!
	 * \brief Clears the console output.
	 */
	void Clear();
	
	/*!
	 * \brief Searches the console variables for matches.
	 *
	 * \param[in] txt the command to search for.  NOTE: Should be a single non-space delimited string.
	 * \param[out] vars vector to hold all found console variables sorted by command name length from shortest to longest.
	 */
	void Search(const std::string &txt, std::vector<ConVar*> &vars);
	
	/*!
	 * \brief Predicts possible console commands given a partial command.
	 *
	 * \param[in] txt the partial command to predict.  NOTE: Should be a single non-space delimited string.
	 * \param[out] vars vector to hold all potential console variables sorted by command name length from shortest to longest.
	 */
	void Predict(const std::string &txt, std::vector<ConVar*> &vars);
	
	/*!
	 * \class ConStream
	 * \headerfile Console.h "Application/Console.h"
	 * \brief An input stream to the console.
	 * 
	 * The following types can be passed to the ConStream using the overloaded '<<' operator:
	 *	-# const char * (NULL terminated cstring)
	 *  -# std::string
	 *  -# float
	 *  -# int
	 *  -# unsigned int
	 *  -# byte
	 */
	class ConStream
	{
	public:
		ConStream(Console *c, const Color4f &col) : con(c), color(col), currColor(col) {}
		
		ConStream &operator<<(const char *str);
		ConStream &operator<<(const std::string &str);
		ConStream &operator<<(float v);
		ConStream &operator<<(int i);
		ConStream &operator<<(uint ui);
		ConStream &operator<<(byte b);
		
		class SetStreamColorClass
		{
		public:
			SetStreamColorClass(const Color4f &c) : color(c) {}
			Color4f color;
		};
		
		class RevertStreamColorClass
		{
		public:
			RevertStreamColorClass() {}
		};
		
		static SetStreamColorClass SetColor(const Color4f &c) { return SetStreamColorClass(c); }
		static RevertStreamColorClass RevertColor() { return RevertStreamColorClass(); }

		ConStream &operator<<(const SetStreamColorClass &c)
		{
			SetCurrentColor(c.color);
			return *this;
		}
		
		ConStream &operator<<(const RevertStreamColorClass &c)
		{
			RevertStreamColor();
			return *this;
		}

		Color4f Color() { return currColor; }
		void RevertStreamColor() { currColor = color; }
		void SetCurrentColor(const Color4f &c) { currColor = c; }
		
	private:
		Console *con;
		Color4f color;
		Color4f currColor;
	};

	/*!
	 * \brief A useful shortcut for changing console stream colors.
	 *
	 * \param[in] c the color.
	 *
	 * \return A Console stream object that changes the color of the console stream.
	 */
	static ConStream::SetStreamColorClass SetColor(const Color4f &c) { return ConStream::SetColor(c); }
	static ConStream::SetStreamColorClass SetColor(uint c) { return ConStream::SetColor(Color4f(c)); }

	/*!
	 * \brief A useful shortcut for reverting a console streams color.
	 *
	 * \return A Console stream object that changes reverts a console streams color.
	 */
	static ConStream::RevertStreamColorClass RevertColor() { return ConStream::RevertColor(); }
	
	/*!
	 * \brief Adds a string of text that will be printed at the start of each newline.
	 *
	 * \param[in] str the text string. NOTE: newline characters are ignored.
	 * \param[in] c the color of the text. Defaults to white.
	 */
	void PushPrependText(const char *str, const Color4f &c = Color::White);
	void PushPrependText(const std::string &str, const Color4f &c = Color::White);
	
	/*!
	 * \brief Pops an element from the newline text stack.
	 */
	void PopPrependText();
	
	/*!
	 * \brief Creates a console stream that can be used to print to the console output in a std::iostream like fashion.
	 *
	 * \param[in] c the color the draw the text. Defaults to white.
	 *
	 * \return A Console stream that can be used to further print to the console.
	 */
	ConStream Print(const Color4f &c = Color::White) { return ConStream(this, c); }
	
	/*!
	 * \brief Creates a console stream that suitable for displaying a warning.
	 *
	 * \return A Console stream that can be used to further print to the console.
	 */
	ConStream PrintWarning() { Print("WARNING: ", Color::Yellow); return Print(Color::White); }
	
	/*!
	 * \brief Creates a console stream that suitable for displaying an error.
	 *
	 * \return A Console stream that can be used to further print to the console.
	 */
	ConStream PrintError() { Print("ERROR: ", Color::Red); return Print(Color::White); }
	
	/*!
	 * \brief Appends a line of text to the console output.
	 *
	 * \param[in] str the string to append.
	 * \param[in] c the color to draw 'str'. Defaults to white.
	 */
	void Print(const char *str, const Color4f &c = Color::White);
	
	/*!
	 * \brief Adds a line of text to the console output.
	 *
	 * \param[in] str the string to add to the console output.
	 * \param[in] c the color to draw 'str'. Defaults to white.
	 */ 
	void PrintLine(const char *str, const Color4f &c = Color::White);
	
	/*!
	 * \brief Creates a newline in the console output.
	 */
	void NewLine();
	
private:
	
	// Built in console commands
	static void ConCmdHelp(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdClear(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdClearHistory(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdClose(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdTime(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdTop(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdBottom(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdPrint(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdRevert(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdFind(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdVersion(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdDump(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);

	// Only used for testing
	static void ConCmdFib(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);

	void ParseCommand(const char *cmd, std::vector<std::string> &args) const;

	// internal print method called by the ConStream class and Print/PrintLine.
	void OutputToConsole(const char *str, const Color4f &c, ConStream *stream);
	
	void AddLine(const char *str, const Color4f &c);
	void AddLine(const std::string &str, const Color4f &c);
	
	void EnforceMaxLineCount();
	
	void NextCmd();
	void PrevCmd();
	void AddCmd(const char *cmd);
	
	// built in console variables
	ConVar *con_bgColor;
	ConVar *con_maxLines;
	ConVar *con_cmdHistory;
	
	struct ConString
	{
		std::string str;
		Color4f c;
	};
	
	typedef std::vector<ConString> ConLine;
	typedef std::list<ConLine> ConTextType;
	ConTextType text;
	int lines;
	int startText; // index to the first line to draw
	
	ConLine prependText; // Text added to each newline
	
	void SplitString(uint charsLeft, const std::string &str, std::string &left, std::string &right);
	uint SplitConString(uint charsLeft, const ConString &conStr, ConLine &currLine, std::vector<ConLine> &out);
	void ConLineWrap(const ConLine &line, std::vector<ConLine> &out);
	
	FontRef font;
	
	float height; // The height of the console window
	float width;
	
	// Used to facilitate word wrapping
	float characterWidth;
	uint maxCharactersPerLine;
	
	bool isOpen;
	
	// NOTE ConVar names are all stored in lower case
	std::map<std::string, ConVar*> conVars;
	
	// User input data
	KeyFilteredBufferedKeyboardInput bufferedInput;
	
	float cursorTimer;
	bool cursorDraw;
	
	// Previous commands
	std::list<std::string> cmdHistory;
	int cmdHistoryIndex;
	
	// Stores the current input before cycling through the command history
	std::string currCmdInput;
	uint currCmdInputCursorPos;
};

#endif
