/*!
 * \file SDLApplication.cpp
 * \brief Implements the SDLApplication class.
 * \author Joe Forte
 * \date 3/8/2008
 */

#include <assert.h>
#include <fstream>
#include <sstream>
#include <iostream>
#include <SDL.h>
#include <gl/glew.h>
#include "Common/Globals.h"
#include "Common/FileSystem.h"
#include "Input/SDLMouse.h"
#include "Input/SDLKeyboard.h"
#include "Graphics/SDLScreen.h"
#include "Graphics/Font.h"
#include "Graphics/MathDraw.h"
#include "Math/MathCommon.h"
#include "Application/DataPaths.h"
#include "Application/SDLApplication.h"

IEnvironment *gEnv = NULL;

const float MAX_FPS = 60.0f;

class Environment : public IEnvironment
{
public:
	Environment() 
	: 
	currTime(0), 
	frameTime(0.0f), 
	fps(0.0f), 
	screen(NULL), 
	mouse(NULL),
	keyboard(NULL),
	imageMgr(NULL), 
	fontMgr(NULL),
	soundSampleMgr(NULL),
	soundStreamMgr(NULL),
	app(NULL),
	console(NULL)
	{}
		
	virtual uint ScreenWidth() const					{ return screen->Width(); }
	virtual uint ScreenHeight() const					{ return screen->Height(); }
	virtual void ScreenDim(uint &w, uint &h) const		{ w = ScreenWidth(); h = ScreenHeight(); }
	virtual uint CurrentTimeMS() const					{ return currTime; }
	virtual float FrameTime() const						{ return frameTime; }
	virtual float FPS() const							{ return fps; }
	virtual IScreen *GetScreen() const					{ return screen; }
	virtual IMouse *GetMouse() const                    { return mouse; }
	virtual IKeyboard *GetKeyboard() const              { return keyboard; }
	virtual ImageMgr *ImageManager() const				{ return imageMgr; }
	virtual FontMgr *FontManager() const				{ return fontMgr; }
	virtual SoundSampleMgr *SoundSampleManager() const	{ return soundSampleMgr; }
	virtual SoundStreamMgr *SoundStreamManager() const	{ return soundStreamMgr; }
	virtual Console *GetConsole() const					{ return console; }
	virtual SDLApplication *GetApp() const				{ return app; }
	virtual std::string AppName() const					{ return appName; }
		
	uint currTime;
	float frameTime;
	float fps;
	std::string appName;
		
	IScreen *screen;
	IMouse *mouse;
	IKeyboard *keyboard;
	
	ImageMgr *imageMgr;
	FontMgr *fontMgr;
	SoundSampleMgr *soundSampleMgr;
	SoundStreamMgr *soundStreamMgr;
		
	SDLApplication *app;
	Console *console;
};

//--------------------------------------------------------------------
const IEnvironment *GetEnv() { return gEnv; }

//--------------------------------------------------------------------
SDLApplication::SDLApplication(int w, int h, bool useFullscreen) 
: 
width(w),
height(h),
fullscreen(useFullscreen),
mouse(NULL),
screen(NULL),
keyboard(NULL),
localEnv(NULL),
hasFocus(true),
isRunning(true)
{}

//--------------------------------------------------------------------
SDLApplication::~SDLApplication()
{
	if(keyboard)
		delete keyboard;
	if(mouse)
		delete mouse;
	if(screen)
		delete screen;
	
	if(localEnv == gEnv)
		gEnv = NULL;
	
	if(localEnv)
		delete localEnv;
	
	UnInitSDL();
}

//--------------------------------------------------------------------
bool SDLApplication::Init(int argc, char *argv[])
{
	if( width < 0 || height < 0 )
		return false;

	commandLine.Parse(argc, argv);
	
	console.ExecuteCommand("time");
	console.NewLine();
	console.Print() << "--------------------------------------------------------------------\n";
	console.Print() << "    " << GetAppName() << " " << GetVersionString() << " initialization started\n";
	console.Print() << "--------------------------------------------------------------------\n";
	console.PushPrependText("\t");
	console.Print() << "Setting proper working directory\n";
	console.PushPrependText("\t-> ");
	
	std::string newCwd = GetPathPart( CreatePath(commandLine.Get(0)) );
	if(!ChangeWorkingDirectory(newCwd.c_str()))
		return false;
	
	console.Print(Color::Green) << "Done\n";
	console.PopPrependText();
	console.Print() << "Initializing SDL\n";
	console.PushPrependText("\t-> ");

	if(!InitSDL())
		return false;

	console.Print(Color::Green) << "Done\n";
	console.PopPrependText();
	console.Print() << "Creating application window\n";
	console.PushPrependText("\t-> ");

#ifdef _WIN32
	if(GetIconName())
	{
		SDL_WM_SetIcon(SDL_LoadBMP(GetIconName()), NULL);
	}
#endif
	
	if(screen) delete screen;
	screen = new SDLScreen;
	
	if(!screen->Set(width, height, Color::Black, GetAppName()))
	{
		console.Print(Color::Red) << "Failed\n";
		console.DumpToFile();
		return false;
	}

	screen->CenterScreen();
	
	console.Print(Color::Green) << "Done\n";
	console.PopPrependText();
	console.Print() << "Initializing OpenGL extensions\n";
	console.PushPrependText("\t-> ");
	
	GLenum err = glewInit();
	if(err != GLEW_OK)
	{
		console.Print(Color::Red) << "Failed\n";
		console.DumpToFile();
		return false;
	}
	
	if(!GLEW_VERSION_2_0)
	{
		console.Print(Color::Red) << "Failed\n";
		console.DumpToFile();
		return false;
	}
	
	console.Print(Color::Green) << "Done\n";
	console.PopPrependText();
	console.Print() << "Creating global environment\n";
	console.PushPrependText("\t-> ");
	
	if(InitEnv())
	{
		console.Print(Color::Green) << "Done\n";
	}
	else
	{
		console.Print(Color::Red) << "Failed\n";
		console.DumpToFile();
		return false;
	}

	console.PopPrependText();
	console.Print() << "Initializing data paths\n";
	console.PushPrependText("\t-> ");
	
	DataPaths::Load("DataPaths.xml");
	
	try
	{
		imageMgr.SetResourcePath( DataPaths::GetPath("Textures").c_str() );
		fontMgr.SetResourcePath( DataPaths::GetPath("Fonts").c_str() );
		soundSampleMgr.SetResourcePath( DataPaths::GetPath("Sounds").c_str() );
		soundStreamMgr.SetResourcePath( DataPaths::GetPath("Music").c_str() );
		
		console.Print(Color::Green) << "Done\n";
	}
	catch(InvalidArgumentError e)
	{
		console.Print(Color::Red) << "Failed\n";
		console.DumpToFile();
		return false;
	}
	
	console.PopPrependText();
	console.Print() << "Initializing resource managers\n";
	console.PushPrependText("\t-> ");
	
	imageMgr.SetDefaultResource(Image::ERROR_IMAGE_PATH.c_str());
	fontMgr.SetDefaultResource(Font::DEFAULT_FONT_PATH.c_str(), &Font::DEFAULT_FONT_SIZE);
	
	console.Print(Color::Green) << "Done\n";
	console.PopPrependText();
	console.Print() << "Initializing sound system\n";
	console.PushPrependText("\t-> ");
	
	if(InitSoundSubSystem())
	{
		console.Print(Color::Green) << "Done\n";
	}
	else
	{
		console.Print(Color::Red) << "Failed\n";
		console.DumpToFile();
		return false;
	}
	
	console.PopPrependText();
	console.Print() << "Creating system fonts\n";
	console.PushPrependText("\t-> ");
	
	systemFontSmall = LoadFontResource("VeraMono.ttf", 14);
	systemFont = LoadFontResource("VeraMono.ttf", 16);
	systemFontLarge = LoadFontResource("VeraMono.ttf", 36);
	
	console.Print(Color::Green) << "Done\n";
	console.PopPrependText();
	console.Print() << "Initializing keyboard and mouse input\n";
	console.PushPrependText("\t-> ");
	
	if(keyboard) delete keyboard;
	keyboard = new SDLKeyboard;
	
	if(mouse) delete mouse;
	mouse = new SDLMouse;

	console.Print(Color::Green) << "Done\n";
	console.PopPrependText();
	console.Print() << "Initializing console resources\n";
	console.PushPrependText("\t-> ");

	console.SetWidth((float)screen->Width());
	
	consoleFont = systemFont;
	console.SetFont(consoleFont);
	console.SetHeight(screen->Height() / 2.0f);
	
	// Registers console commands and variables
	console.RegisterCmd("quit", "Exits the game.", ConCmdQuit, this);
	console.RegisterCmd("exit", "Exits the game.", ConCmdQuit, this);
	console.RegisterCmd("pause", "Pauses and unpauses the game.", ConCmdPause, this);
	console.RegisterCmd("dataUsage", "Prints current estimate of data usage.", ConCmdDataUsage, this);
	console.RegisterCmd("dumpFontMgr", "Dumps information about the Font manager to the console.", ConCmdDumpFontMgr, this);
	console.RegisterCmd("dumpImageMgr", "Dumps information about the Image manager to the console.", ConCmdDumpImageMgr, this);
	console.RegisterCmd("dumpSoundSampleMgr", "Dumps information about the SoundSample manager to the console.", ConCmdDumpSoundSampleMgr, this);
	console.RegisterCmd("dumpSoundStreamMgr", "Dumps information about the SoundStream manager to the console.", ConCmdDumpSoundStreamMgr, this);
	console.RegisterCmd("centerScreen", "Centers the game window.", ConCmdCenterScreenMgr, this);

	g_showFps = console.RegisterVar("g_showFps", "0", "Set to 1 to display the FPS, set to 0 to disable the FPS display", ConVar::TYPE_INT);
	
	console.Print(Color::Green) << "Done\n";
	console.PopPrependText();
	console.Print() << "Initializing application SubSystems\n";
	console.PushPrependText("\t-> ");
	
	srand(TimeSeed());
	
	if(InitSubSystems())
	{
		console.Print(Color::Green) << "Done\n";
	}
	else
	{
		console.Print(Color::Red) << "Failed\n";
		console.DumpToFile();
		return false;
	}
	
	console.PopPrependText();
	console.Print() << "Initializing application\n";
	console.PushPrependText("\t-> ");
	
	if(ApplicationInit())
	{
		console.Print(Color::Green) << "Done\n";
	}
	else
	{
		console.Print(Color::White) << "[" << Console::SetColor(Color::Red) << "failed" << Console::RevertColor() << "]\n";
		console.DumpToFile();
		return false;
	}
	
	console.PopPrependText();
	console.PopPrependText();
	console.Print() << "--------------------------------------------------------------------\n";
	console.Print() << "    " << GetAppName() << " " << GetVersionString() << " initialization finished\n";
	console.Print() << "--------------------------------------------------------------------\n";
	
	return true;
}

//--------------------------------------------------------------------
bool SDLApplication::InitSDL()
{
	if( SDL_Init(SDL_INIT_TIMER | SDL_INIT_VIDEO) )
	{
		console.PrintError() <<  "ERROR: SDL could not be initialized: " << SDL_GetError() << "\n";
		return false;
	}
	
	return true;
}

//--------------------------------------------------------------------
bool SDLApplication::InitEnv()
{
	Environment *env    = new Environment; 
	env->currTime		= 0;
	env->frameTime		= 0.0f;
	env->fps			= 0.0f;
	env->appName		= GetAppName();
	env->screen		    = screen;
	env->mouse          = mouse;
	env->keyboard       = keyboard;
	env->imageMgr		= &imageMgr;
	env->fontMgr		= &fontMgr;
	env->soundSampleMgr = &soundSampleMgr;
	env->soundStreamMgr = &soundStreamMgr;
	env->console		= &console;
	env->app			= this;
	
	localEnv = env;
	gEnv = env;

	return true;
}

//--------------------------------------------------------------------
void SDLApplication::UnInitSDL()
{
	SDL_Quit();
}

//--------------------------------------------------------------------
bool SDLApplication::HaveFocus() const
{
	return hasFocus;
}

//--------------------------------------------------------------------
uint SDLApplication::GetTickCount() const
{
	return (uint)SDL_GetTicks();
}

//--------------------------------------------------------------------
void SDLApplication::Exit()
{
	isRunning = !OnApplicationExitRequest();
}

//--------------------------------------------------------------------
int SDLApplication::Run()
{
	const float MAX_SEC_PER_FRAME = 1.0f / MAX_FPS;
	const float MAX_MS_PER_FRAME = 1000.0f * MAX_SEC_PER_FRAME;
	uint lastTick = GetTickCount();
	
	SDL_Event msg;
	while(isRunning)
	{
		// Gets time elapsed and updates frame rate and the local environment
		uint curTick = GetTickCount();
		float timeDelta = (float)(curTick - lastTick);
		float dt = timeDelta / 1000.0f;

		// If it takes more then MAX_FRAME_TIME to render a frame then there was most likely an
		// event that pause execution of the program (user moving the game window etc.)
		if(dt > MAX_SEC_PER_FRAME)
			dt = MAX_SEC_PER_FRAME;

		lastTick = curTick;
		fps.Update(dt);
		
		Environment *env = static_cast<Environment*>(localEnv);
		env->currTime = curTick;
		env->frameTime = dt;
		env->fps = fps.Get();

		PreFrame();
		
		// Update the console before the keyboard is updated so we don't capture the input 
		// key (usually the tilda ~) that closes the console.
		console.Update(dt, *mouse, *keyboard);

		mouse->Update();
		keyboard->Update(dt);
	
		while(SDL_PollEvent(&msg))
		{
			switch( msg.type )
			{
				case SDL_QUIT:
					Exit();
					break;
				
				case SDL_ACTIVEEVENT:
				
					// informs the mouse that it has lost or gained focus
					if( msg.active.state & SDL_APPMOUSEFOCUS )
					{
						if( msg.active.gain == 1 )
						{
							mouse->GainedFocus();
							OnGainedMouseFocus();
						}
						else
						{
							mouse->LostFocus();
							OnLostMouseFocus();
						}
					}
				
					// pauses the game if we lose focus
					if( msg.active.state & SDL_APPACTIVE || msg.active.state & SDL_APPINPUTFOCUS)
					{
						if( msg.active.gain == 1 )
						{
							hasFocus = true;
							OnGainedFocus();
						}
						else
						{
							hasFocus = false;
							OnLostFocus();
						}
					}
					break;

				case SDL_KEYDOWN:
					/* Handles system specific quit commands. */
					#ifdef _WIN32
						if((msg.key.keysym.mod & KMOD_ALT) &&  (msg.key.keysym.sym == SDLK_F4))
						{
							Exit();
							break;
						}
					#elif defined(__APPLE__)
						if((msg.key.keysym.mod & KMOD_ALT) &&  (msg.key.keysym.sym == SDLK_q))
						{
							Exit();
							break;
						}
					#endif
				
				case SDL_KEYUP:
					keyboard->HandleEvent( msg.key );
					break;
				
				case SDL_MOUSEBUTTONUP:
				case SDL_MOUSEBUTTONDOWN:
					mouse->HandleEvent( msg.button );
					break;
			};
		}

		// Application update
		isRunning = Update(dt);

		screen->BeginFrame();
		
		Draw(dt);

		screen->DrawQueuedText();

		if(g_showFps->GetInt())
			DrawFPS();

		screen->EndFrame();

		PostFrame();
		
		// Delays loop to fix the frame rate
		int delay = (int)MAX_MS_PER_FRAME - (int)(GetTickCount() - curTick);
		if(delay > 0)
		{
			SDL_Delay((uint)delay);
		}
	}

	return 0;
}

//--------------------------------------------------------------------
void SDLApplication::DrawFPS()
{
	std::stringstream str;
	str << "FPS: " << fps.Get();
	
	screen->QueueText(systemFont, str.str().c_str(), Color::Red, Font::ALIGN_TOP_LEFT, Point2f(0.01f, -0.01f));
}

//--------------------------------------------------------------------
template<class ResourceMgrType>
void DumpResourceData(Console &con, ResourceMgrType &mgr)
{
	const std::vector<typename ResourceMgrType::resource_type*> &resources = mgr.GetResources();
	
	Console::ConStream out = con.Print();
	
	float kilobytes = mgr.GetDataUsageInBytes() / 1024.0f;
	out << "Number of resources loaded (" << mgr.GetNumberOfResourcesLoaded() << ") totaling " << (kilobytes/1024.0f) << "M (" << kilobytes << "K)\n";
	
	uint count = 1;
	size_t len = resources.size();
	for(size_t i = 0; i < len; i++)
	{
		if(resources[i])
		{
			float size = resources[i]->GetDataUsageInBytes() / 1024.0f;
			out << "  " << count << ". " << resources[i]->GetFileName() << " " << size / 1024.0f << "M (" << size << "K) ";
			out << "Resource ID: " << resources[i]->GetID() << " Ref Count: " << resources[i]->GetRefCount() << "\n";
			count++;
		}
	}
}

//--------------------------------------------------------------------
void SDLApplication::ConCmdDumpFontMgr(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	FontMgr *mgr = GetFontManager();
	DumpResourceData(*con, *mgr);
}

//--------------------------------------------------------------------
void SDLApplication::ConCmdDumpSoundSampleMgr(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	SoundSampleMgr *mgr = GetSoundSampleManager();
	DumpResourceData(*con, *mgr);
}

//--------------------------------------------------------------------
void SDLApplication::ConCmdDumpSoundStreamMgr(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	SoundStreamMgr *mgr = GetSoundStreamManager();
	DumpResourceData(*con, *mgr);
}

//--------------------------------------------------------------------
void SDLApplication::ConCmdDumpImageMgr(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	ImageMgr *mgr = GetImageManager();
	DumpResourceData(*con, *mgr);
}

//--------------------------------------------------------------------
void SDLApplication::ConCmdQuit(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	con->Print(Color::White) << "Exiting game...\n";
	SDLApplication *g = static_cast<SDLApplication*>(userData);
	g->Exit();
}

//--------------------------------------------------------------------
void SDLApplication::ConCmdPause(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	SDLApplication *g = static_cast<SDLApplication*>(userData);	
	g->TogglePause();
}

//--------------------------------------------------------------------
void SDLApplication::ConCmdDataUsage(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	float imageUsage = GetImageManager()->GetDataUsageInBytes() / 1024.0f;
	uint imageCount = GetImageManager()->GetNumberOfResourcesLoaded();
	
	float fontUsage = GetFontManager()->GetDataUsageInBytes() / 1024.0f;
	uint fontCount = GetFontManager()->GetNumberOfResourcesLoaded();
	
	float sampleUsage = GetSoundSampleManager()->GetDataUsageInBytes() / 1024.0f;
	uint sampleCount = GetSoundSampleManager()->GetNumberOfResourcesLoaded();
	
	float streamUsage = GetSoundStreamManager()->GetDataUsageInBytes() / 1024.0f;
	uint streamCount = GetSoundStreamManager()->GetNumberOfResourcesLoaded();
	
	float totalUsage = imageUsage + fontUsage + sampleUsage + streamUsage;
	
	con->Print(Color::White) << "  Total memory used: " << totalUsage / 1024.0f << "M (" << totalUsage << "K)\n";
	con->Print(Color::White) << "      Images loaded: " << imageCount << " totaling " << imageUsage / 1024.0f << "M (" << imageUsage << "K)\n";
	con->Print(Color::White) << "       Fonts loaded: " << fontCount << " totaling " << fontUsage / 1024.0f << "M (" << fontUsage << "K)\n";
	con->Print(Color::White) << "SoundSamples loaded: " << sampleCount << " totaling " << sampleUsage / 1024.0f << "M (" << sampleUsage << "K)\n";
	con->Print(Color::White) << "SoundStreams loaded: " << streamCount << " totaling " << streamUsage / 1024.0f << "M (" << streamUsage << "K)\n";
}

//--------------------------------------------------------------------
void SDLApplication::ConCmdCenterScreenMgr(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
#ifdef __APPLE__

	ConWarning() << "Centering the game window is unsupported on Mac OSX.\n";

#endif

	GetScreen()->CenterScreen();
}
