/*!
 * \file SDLApplication.h
 * \brief Declares the SDLApplication class that represents a realtime app using SDL.
 * \author Joe Forte
 * \date 3/8/2008
 */

#ifndef SDLAPPLICATION_H
#define SDLAPPLICATION_H

#include <string>
#include <vector>
#include "Common/Error.h"
#include "Common/Version.h"
#include "Application/CommandLine.h"
#include "Application/FramesPerSecond.h"
#include "Application/Console.h"
#include "Graphics/Image.h"
#include "Graphics/Font.h"
#include "Sound/Sound.h"

class IEnvironment;
class SDLScreen;
class SDLMouse;
class SDLKeyboard;

/*!
 * \class SDLApplication
 * \headerfile SDLApplication.h "Application/SDLApplication.h"
 * \brief Represents a realtime app using SDL.
 */
class SDLApplication
{
public:

	/*!
	 * \brief Constructor.
	 *
	 * \param[in] w the initial width of the application screen.
	 * \param[in] h the initial height of the application screen.
	 * \param[in] useFullscreen set to true to start the application fullscreen.
	 */
	SDLApplication(int w, int h, bool useFullscreen);

	/*!
	 * \brief De-constructor.
	 */
	virtual ~SDLApplication();
	
	/*!
	 * \brief Returns the major version number of the game.
	 */
	virtual uint GetMajorVersion() const { return MAJOR_VERSION_INT; }
	
	/*!
	 * \brief Returns the minor version number of the game.
	 */
	virtual uint GetMinorVersion() const { return MINOR_VERSION_INT; }
	
	/*!
	 * \brief Returns the build number number of the game.
	 */
	virtual uint GetBuildNumber() const { return BUILD_NUMBER_INT; }
	
	/*!
	 * \brief Returns an optional versioning identifier (alpha, beta, RC, etc.)
	 */
	virtual const char *GetVersionType() const { return VERSION_TYPE_CSTR; }
	
	/*!
	 * \brief Returns the complete version string of the game.
	 */
	virtual const char *GetVersionString() const { return VERSION_CSTR; }

	/*!
	 * \brief Initializes the application.
	 *
	 * \param[in] argc the number of argument strings in \a argv.
	 * \param[in] argv an array of null-terminated c strings holding the arguments.
	 *
	 * \return True if the initialization was successful, false otherwise.
	 *
	 * NOTE: This function must be called (and return true) before calling Run.
	 */
	bool Init(int argc, char *argv[]);

	/*!
	 * \brief Enters the applications main event loop.
	 *
	 * NOTE: Assumes that Init() has been called and returned true.
	 */
	int Run();

	/*!
	 * \brief Returns the image name of the application icon.
	 *
	 * \return The file name of the icon or NULL if there is no icon.
	 *
	 * NOTE: Only used on Windows.
	 */
	virtual const char *GetIconName() { return NULL; }

	/*!
	 * \brief Pauses the application if it is unpaused or unpauses the application if it is paused.
	 */
	virtual void TogglePause() {}

	/*!
	 * \brief Pauses the application.
	 */
	virtual void Pause() {}

	/*!
	 * \brief Unpauses the application.
	 */
	virtual void UnPause() {}

	/*!
	 * \brief Returns if the application is paused or not.
	 *
	 * \return True if the application is paused or false if the application is unpaused.
	 */
	virtual bool IsPaused() const  { return false; }

	/*!
	 * \brief Determines if the application has input focus.
	 *
	 * \return True if the application has input focus.
	 */
	virtual bool HaveFocus() const;

	/*!
	 * \brief Returns the current tick count.
	 */
	virtual uint GetTickCount() const;

	/*!
	 * \brief Informs the application to exit at the end of the current frame.
	 */
	void Exit();

	/*!
	 * \brief Gets the applications name.
	 *
	 * \return The name of the application.
	 */
	virtual std::string GetAppName() { return std::string("SDLApplication"); }

protected:

	/*!
	 * \brief Initializes any subsystems needed by the application.
	 *
	 * \return True if the initialization was successful, false otherwise.
	 */
	virtual bool InitSubSystems() { return true; }

	/*!
	 * \brief Initializes the application.  All subsystems (including the global environment) is valid
	 * when this is called.
	 *
	 * \return True if the initialization was successful, false otherwise.
	 */
	virtual bool ApplicationInit() { return true; }

	/*!
	 * \brief Called before the start of the frame.
	 *
	 * NOTE: The environment is valid for the current frame.
	 */
	virtual void PreFrame() {}

	/*!
	 * \brief Called after the frame is complete.
	 *
	 * NOTE: The environment is valid for the current frame but not the next.
	 */
	virtual void PostFrame() {}

	/*!
	 * \brief Update method called once each frame.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 *
	 * \return True if the application should continue.  Returns false if the application should exit.
	 */
	virtual bool Update(float dt) = 0;

	/*!
	 * \brief Draw method called once each frame.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 */
	virtual void Draw(float dt) = 0;

	/*!
	 * \brief Called when the application looses focus.
	 */
	virtual void OnLostFocus() {}

	/*!
	 * \brief Called when the application gains focus.
	 */
	virtual void OnGainedFocus() {}

	/*!
	 * \brief Called when the application mouse focus.
	 */
	virtual void OnLostMouseFocus() {}

	/*!
	 * \brief Called when the application gains mouse focus.
	 */
	virtual void OnGainedMouseFocus() {}
	
	/*!
	 * \brief Called when the application receives a message to quit from the OS or a call to Exit().
	 *
	 * \return True if the application should exit at the end of the frame.  Returns false if the application should ignore the quit message.
	 */
	virtual bool OnApplicationExitRequest() { return true; }

protected:

	CommandLine commandLine;       /*!< The command line information for the application. */

	int width;			           /*!< the init width of the application screen. */
	int height;                    /*!< the init height of the application screen. */
	bool fullscreen;	           /*!< the init value indicating fullscreen use. */

	SDLScreen *screen;             /*!< the screen the application draws to. */
	SDLMouse *mouse;               /*!< pointer to the mouse. */
	SDLKeyboard *keyboard;         /*!< pointer to the keyboard. */

	Console console;               /*!< application console. */

	FramesPerSecond fps;           /*!< calculates the frames per second. */

	ImageMgr imageMgr;		       /*!< the image manager. */
	FontMgr fontMgr;			   /*!< the font manager. */
	SoundSampleMgr soundSampleMgr; /*!< the sound sampler manager. */
	SoundStreamMgr soundStreamMgr; /*!< the sound stream manager. */

	FontRef systemFontSmall;       /*!< the small font used by the application. */
	FontRef systemFont;            /*!< the normal font used by the application. */
	FontRef systemFontLarge;       /*!< the large font used by the application. */

	FontRef consoleFont;           /*!< the font used by the console. */

	ConVar *g_showFps;             /*!< console variable to control the fps display. */

	IEnvironment *localEnv;        /*!< the Environment associated with this application. */

private:

	// console commands
	static void ConCmdQuit(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdPause(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdDataUsage(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdDumpFontMgr(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdDumpImageMgr(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdDumpSoundSampleMgr(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdDumpSoundStreamMgr(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdCenterScreenMgr(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);

	bool InitSDL();
	bool InitEnv();

	void UnInitSDL();

	void DrawFPS();

	bool hasFocus;
	bool isRunning;
};

#endif
