/*!
 * \file Globals.h
 * \brief Defines the global data set maintained by the application.
 * \author Joe Forte
 * \date 5/19/2008
 */

#ifndef GLOBALS_H
#define GLOBALS_H

#include <assert.h>
#include <string>
#include "Application/Console.h"
#include "Graphics/Screen.h"
#include "Graphics/Image.h"
#include "Graphics/Font.h"
#include "Sound/Sound.h"
#include "Common/BaseTypes.h"

class IMouse;
class IKeyboard;
class SDLApplication;

/*!
 * \class IEnvironment
 * \headerfile Globals.h "Common/Globals.h"
 * \brief Interface class to the global environment of the currently running application.
 */
class IEnvironment
{
public:
	
	/*!
	 * \brief Gets the width of the drawable area of the currently active Screen.
	 *
	 * \return The width of the drawable area of the currently active Screen.
	 */
	virtual uint ScreenWidth() const = 0;
	
	/*!
	 * \brief Gets the height of the drawable area of the currently active Screen.
	 *
	 * \return The height of the drawable area of the currently active Screen.
	 */
	virtual uint ScreenHeight() const = 0;
	
	/*!
	 * \brief Gets the width and height of the drawable area of the currently active Screen.
	 *
	 * \param[out] w the width of the drawable area of the currently active Screen.
	 * \param[out] h the height of the drawable area of the currently active Screen.
	 */
	virtual void ScreenDim(uint &w, uint &h) const = 0;
	
	/*!
	 * \brief Gets the number of milliseconds that have passed since the application started.
	 *
	 * \return The time since the application started in milliseconds.
	 */
	virtual uint CurrentTimeMS() const = 0;
	
	/*!
	 * \brief Gets the number of seconds between this and the last frame
	 *
	 * \return The number of seconds between this and the last frame
	 */
	virtual float FrameTime() const = 0;
	
	/*!
	 * \brief Gets the frames per second of the application.
	 *
	 * \return The frames per second of the application.
	 */
	virtual float FPS() const = 0;

	/*!
	 * \brief Gets the interface to the currently active Screen,
	 *
	 * \return The interface to the active Screen.
	 */
	virtual IScreen *GetScreen() const = 0;
	
	/*!
	 * \brief Gets the interface for the active mouse.
	 *
	 * \return The mouse interface.
	 */
	virtual IMouse *GetMouse() const = 0;
	
	/*!
	 * \brief Gets the interface for the active keyboard.
	 *
	 * \return The keyboard interface.
	 */
	virtual IKeyboard *GetKeyboard() const = 0;
	
	/*!
	 * \brief Gets the applications ImageMgr.
	 *
	 * \return A pointer to the applications ImageMgr.
	 */
	virtual ImageMgr *ImageManager() const = 0;
	
	/*!
	 * \brief Gets the applications FontMgr.
	 *
	 * \return A pointer to the applications FontMgr.
	 */
	virtual FontMgr *FontManager() const = 0;
	
	/*!
	 * \brief Gets the applications SoundSampleMgr.
	 *
	 * \return A pointer to the applications SoundSampleMgr.
	 */
	virtual SoundSampleMgr *SoundSampleManager() const = 0;

	/*!
	 * \brief Gets the applications SoundStreamMgr.
	 *
	 * \return A pointer to the applications SoundStreamMgr.
	 */
	virtual SoundStreamMgr *SoundStreamManager() const = 0;
	
	/*!
	 * \brief Gets the applications Console.
	 *
	 * \return A pointer to the applications Console.
	 */
	virtual Console *GetConsole() const = 0;
	
	/*!
	 * \brief Gets the application itself.
	 *
	 * \return A pointer to the application.
	 */
	virtual SDLApplication *GetApp() const = 0;
	
	/*!
	 * \brief Gets the name of the active application.
	 *
	 * \return the name of the active application.
	 */
	virtual std::string AppName() const = 0;
};

/*!
 * \brief Gets the version string for the active application.
 *
 * \return the version string for the active application.
 */
const char *GetVersionString();

/*!
 * \brief Gets the global enviroment for the active application.
 *
 * \return A pointer to an IEnviroment object describing the current application
 */
const IEnvironment *GetEnv();

/*!
 * \brief Gets the active application.
 *
 * \return A pointer to the active application.
 */
inline SDLApplication *GetApp()
{
	const IEnvironment *env = GetEnv();
	assert(env);
	return env->GetApp();
}

/*!
 * \brief Gets the Screen for the active application.
 *
 * \return A pointer to the active Screen for current application
 */
inline IScreen *GetScreen()
{
	const IEnvironment *env = GetEnv();
	assert(env);
	return env->GetScreen();
}


/*!
 * \brief Gets the ImageMgr for the active application.
 *
 * \return A pointer to the ImageMgr for current application
 */
inline ImageMgr *GetImageManager()
{
	const IEnvironment *env = GetEnv();
	assert(env);
	return env->ImageManager();
}

/*!
 * \brief Gets the FontMgr for the active application.
 *
 * \return A pointer to the FontMgr for current application
 */
inline FontMgr *GetFontManager()
{
	const IEnvironment *env = GetEnv();
	assert(env);
	return env->FontManager();
}

/*!
 * \brief Gets the SoundSampleMgr for the active application.
 *
 * \return A pointer to the SoundSampleMgr for current application
 */
inline SoundSampleMgr *GetSoundSampleManager()
{
	const IEnvironment *env = GetEnv();
	assert(env);
	return env->SoundSampleManager();
}

/*!
 * \brief Gets the SoundStreamMgr for the active application.
 *
 * \return A pointer to the SoundStreamMgr for current application
 */
inline SoundStreamMgr *GetSoundStreamManager()
{
	const IEnvironment *env = GetEnv();
	assert(env);
	return env->SoundStreamManager();
}

/*!
 * \brief Gets the Console for the active application.
 *
 * \return A pointer to the Console for current application
 */
inline Console *GetConsole()
{
	const IEnvironment *env = GetEnv();
	assert(env);
	return env->GetConsole();
}

/*!
 * \brief Reports a warning message to the console.
 *
 * \return A ConStream that can be used similiarlt to the std iostreams. 
 */
Console::ConStream ConWarning();

/*!
 * \brief Reports an error message to the console.
 *
 * \return A ConStream that can be used similiarlt to the std iostreams. 
 */
Console::ConStream ConError();

/*!
 * \brief Reports a message to the console.
 *
 * \return A ConStream that can be used similiarlt to the std iostreams. 
 */
Console::ConStream ConMsg();

#endif
