/*!
 * \file GUIPanel.cpp
 * \brief Implements the GUIPanel class, the base class for all GUI elements
 * \author Joe Forte
 * \date 6/2/2008
 */

#include <assert.h>
#include <algorithm>
#include "GUI/GUIPanel.h"

//--------------------------------------------------------------------
bool GUIPanelSortFunc(const GUIPanel &p1, const GUIPanel &p2)
{
	return p1.GetDepth() > p2.GetDepth();
}

//--------------------------------------------------------------------
class GUIPanelRefitVisitor : public GUIPanelVisitor
{
public:
	virtual bool Visit(GUIPanel *p, uint index) const
	{
		p->Refit();
		return true;
	}
};

//--------------------------------------------------------------------
class GUIPanelDrawVisitor : public GUIPanelVisitor
{
public:
	virtual bool Visit(GUIPanel *p, uint index) const
	{
		if(p->IsVisible())
			p->Draw();
		return true;
	}
};

//--------------------------------------------------------------------
class GUIPanelMouseHoverVisitor : public GUIPanelVisitor
{
public:
	GUIPanelMouseHoverVisitor(const IMouse *m, const Point2f &p) : mouse(m), mousePos(p) {} 
	
	virtual bool Visit(GUIPanel *p, uint index) const
	{
		if(p->IsActive() && p->InActiveArea(mousePos))
		{
			p->OnMouseHover(*mouse);
			return false;
		}
		return true;
	}
	
	const IMouse *mouse;
	Point2f mousePos;
};

//--------------------------------------------------------------------
class GUIPanelMouseClickVisitor : public GUIPanelVisitor
{
public:
	GUIPanelMouseClickVisitor(const IMouse *m, const Point2f &p) : mouse(m), mousePos(p) {} 
	
	virtual bool Visit(GUIPanel *p, uint index) const
	{
		if(p->IsActive() && p->InActiveArea(mousePos))
		{
			p->OnMouseClick(*mouse);
			return false;
		}
		return true;
	}
	
	const IMouse *mouse;
	Point2f mousePos;
};

//--------------------------------------------------------------------
class GUIPanelKeyboardUpdateVisitor : public GUIPanelVisitor
{
public:
	GUIPanelKeyboardUpdateVisitor(const IKeyboard *k) : kb(k) {} 
		
	virtual bool Visit(GUIPanel *p, uint index) const
	{
		if(p->IsActive())
			p->KeyboardUpdate(*kb);
		
		return true;
	}
		
	const IKeyboard *kb;
};

//--------------------------------------------------------------------
class GUIPanelUpdateVisitor : public GUIPanelVisitor
{
public:
	GUIPanelUpdateVisitor(float t) : dt(t) {} 
		
	virtual bool Visit(GUIPanel *p, uint index) const
	{
		if(p->IsActive())
			p->Update(dt);
		
		return true;
	}
		
	float dt;
};

//--------------------------------------------------------------------
class GUIPanelFocusGainedVisitor : public GUIPanelVisitor
{
public:
	GUIPanelFocusGainedVisitor(const IKeyboard *k) : kb(k) {} 
		
	virtual bool Visit(GUIPanel *p, uint index) const
	{
		if(p->IsActive())
		{
			p->OnFocusGained(*kb);
			return false;
		}
		return true;
	}
		
	const IKeyboard *kb;
};

//--------------------------------------------------------------------
class GUIPanelFocusLostVisitor : public GUIPanelVisitor
{
public:
	GUIPanelFocusLostVisitor(const IKeyboard *k) : kb(k) {} 
		
	virtual bool Visit(GUIPanel *p, uint index) const
	{
		if(p->IsActive())
		{
			p->OnFocusLost(*kb);
			return false;
		}
		return true;
	}
		
	const IKeyboard *kb;
};

//--------------------------------------------------------------------
class GUIPanelDeleteVisitor : public GUIPanelVisitor
{
public:
	virtual bool Visit(GUIPanel *p, uint index) const
	{
		delete p;
		return true;
	}
};

//--------------------------------------------------------------------
GUIPanel::GUIPanel() : parent(NULL), width(0), height(0), active(true), visible(true) {}

//--------------------------------------------------------------------
GUIPanel::GUIPanel(GUIPanel *p) : parent(p), width(0), height(0), active(true), visible(true) {}

//--------------------------------------------------------------------
GUIPanel::~GUIPanel() 
{
	VisitChildren( GUIPanelDeleteVisitor() );
}

//--------------------------------------------------------------------
void GUIPanel::OnMouseOver(const IMouse &m)
{
}

//--------------------------------------------------------------------
void GUIPanel::OnMouseLeave(const IMouse &m)
{
}

//--------------------------------------------------------------------
void GUIPanel::OnMouseHover(const IMouse &m)
{
	VisitChildrenFrontToBack( GUIPanelMouseHoverVisitor(&m, m.GetPosNorm()) );
}

//--------------------------------------------------------------------
void GUIPanel::OnMouseClick(const IMouse &m)
{
	VisitChildrenFrontToBack( GUIPanelMouseClickVisitor(&m, m.GetPosNorm()) );
}

//--------------------------------------------------------------------
void GUIPanel::OnMouseRelease(const IMouse &m)
{
}

//--------------------------------------------------------------------
void GUIPanel::OnFocusGained(const IKeyboard &kb)
{
	VisitChildrenFrontToBack( GUIPanelFocusGainedVisitor(&kb) );
}

//--------------------------------------------------------------------
void GUIPanel::OnFocusLost(const IKeyboard &kb)
{
	VisitChildrenFrontToBack( GUIPanelFocusLostVisitor(&kb) );
}

//--------------------------------------------------------------------
void GUIPanel::KeyboardUpdate(const IKeyboard &kb)
{
	VisitChildrenFrontToBack( GUIPanelKeyboardUpdateVisitor(&kb) );
}

//--------------------------------------------------------------------
void GUIPanel::Update(float dt)
{
	if(IsActive())
	{
		OnUpdate(dt);
		VisitChildren( GUIPanelUpdateVisitor(dt) );
	}
}

//--------------------------------------------------------------------
bool GUIPanel::InActiveArea(const Point2f &p) const
{
	AABB bounds = GetBounds();
	return bounds.Inside(p);
}

//--------------------------------------------------------------------
void GUIPanel::Draw()
{
	if(IsVisible())
	{
		Paint();
		VisitChildren( GUIPanelDrawVisitor() );
	}
}

//--------------------------------------------------------------------
void GUIPanel::SetPosition(const Point2f &p)
{	
	Rectf area = GetArea();
	
	area.x = p.x;
	area.y = p.y;
	
	SetArea(area);
}

//--------------------------------------------------------------------
void GUIPanel::SetArea(const Rectf &r)
{
	position = Point2f(r.x, r.y);
	width = r.w;
	height = r.h;
}

//--------------------------------------------------------------------
void GUIPanel::SetArea(const AABB &a)
{
	SetArea( AABBToRect<float>(a) );
}

//--------------------------------------------------------------------
Point2f GUIPanel::GetPosition() const 
{ 
	return position; 
}

//--------------------------------------------------------------------
AABB GUIPanel::GetBounds() const 
{ 
	return RectToAABB<float>( GetArea() ); 
}

//--------------------------------------------------------------------
Rectf GUIPanel::GetArea() const 
{
	Rectf r;
	r.x = position.x;
	r.y = position.y;
	r.w = width;
	r.h = height;
	return r;
}

//--------------------------------------------------------------------
void GUIPanel::VisitChildren(const GUIPanelVisitor &visitor)
{
	uint index = 0;
	
	std::vector<GUIPanel*>::iterator i = children.begin();
	while(i != children.end() && visitor.Visit(*i, index))
	{
		i++;
		index++;
	}
}

//--------------------------------------------------------------------
void GUIPanel::VisitChildrenFrontToBack(const GUIPanelVisitor &visitor)
{
	uint index = 0;
	
	std::vector<GUIPanel*>::reverse_iterator i = children.rbegin();
	while(i != children.rend() && visitor.Visit(*i, index))
	{
		i++;
		index++;
	}
}

//--------------------------------------------------------------------
GUIPanel *GUIPanel::Query(const Point2f &p) const
{
	std::vector<GUIPanel*>::const_reverse_iterator i;
	for(i = children.rbegin(); i != children.rend(); i++)
	{
		if((*i)->IsActive() && (*i)->InActiveArea(p))
		{
			GUIPanel *panel = (*i)->Query(p);
			if(panel)
				return panel;
			return *i;
		}
	}

	return NULL;
}

//--------------------------------------------------------------------
void GUIPanel::CenterInArea(const Rectf &r, const Point2f offset /* = Point2f(0.0f, 0.0f) */)
{
	Rectf area = GetArea();
	
	Point2f center(r.x + r.w / 2.0f, r.y - r.h / 2.0f);
	Point2f pos = center + Point2f(-area.w / 2.0f, -area.h / 2.0f) + offset;
	SetPosition(pos);
}

//--------------------------------------------------------------------
void GUIPanel::Refit()
{
	VisitChildren( GUIPanelRefitVisitor() );
	RecomputeLayout();
}

//--------------------------------------------------------------------
void GUIPanel::MoveToFront()
{
	depth = 0;
	
	if(parent)
	{
		uint index;
		bool res = parent->GetIndex(this, index);
		
		assert(res);
		
		parent->MoveChildToFront(index);
	}
}

//--------------------------------------------------------------------
void GUIPanel::SetDepth(uint z)
{
	depth = z;

	if(parent)
		parent->DepthChanged();
}

//--------------------------------------------------------------------
bool GUIPanel::GetIndex(GUIPanel *p, uint &index) const
{
	index = 0;
	for(std::vector<GUIPanel*>::const_iterator i = children.begin(); i != children.end(); i++)
	{
		if(*i == p)
			return true;
		index++;
	}
	return false;
}

//--------------------------------------------------------------------
void GUIPanel::MoveChildToFront(uint i)
{
	assert(i < children.size());
	
	GUIPanel *p = children[i];
	children.erase(children.begin() + i);
	children.push_back(p);
}

//--------------------------------------------------------------------
void GUIPanel::DepthChanged()
{
	std::sort(children.begin(), children.end(), GUIPanelSortFunc);
}

//--------------------------------------------------------------------
void GUIPanel::AddChild(GUIPanel *p)
{
	assert(p);
	p->parent = this;
	children.push_back(p);
}

//--------------------------------------------------------------------
void GUIPanel::RemoveChild(GUIPanel *p)
{
	std::vector<GUIPanel*>::iterator i;
	for(i = children.begin(); i != children.end(); i++)
	{
		if(*i == p)
		{
			delete *i;
			children.erase(i);
			return;
		}
	}
}

//--------------------------------------------------------------------
void GUIPanel::RemoveChild(uint i)
{
	assert(i < NumberOfChildren());
	delete children[i];
	children.erase( children.begin() + i );
}

//--------------------------------------------------------------------
uint GUIPanel::NumberOfChildren() const
{
	return (uint)children.size();
}

//--------------------------------------------------------------------
GUIPanel *GUIPanel::GetParent() const
{
	return parent;
}

//--------------------------------------------------------------------
GUIPanel *GUIPanel::GetChild(uint i) const
{
	assert(i < NumberOfChildren());
	return children[i];
}
