/*!
 * \file GUIPanel.h
 * \brief Declares the GUIPanel class, the base class for all GUI elements
 * \author Joe Forte
 * \date 6/2/2008
 */

#ifndef GUIPANEL_H
#define GUIPANEL_H

#include <vector>
#include "Math/MathCommon.h"
#include "Input/Mouse.h"
#include "Input/Keyboard.h"

class GUIPanel;

class GUIPanelVisitor
{
public:
	virtual bool Visit(GUIPanel *p, uint index) const = 0;
};

/*!
 * \class GUIPanel
 * \headerfile GUIPanel.h "GUI/GUIPanel.h"
 * \brief The base class for all GUI elements.
 *
 * NOTE: The GUIPanel class takes complete ownership of all of its children and assumes they
 * were dynamically allocated with a call to new.
 */
class GUIPanel
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	GUIPanel();
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] p the panels parent.
	 */
	GUIPanel(GUIPanel *p);
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~GUIPanel();
	
	/*!
	 * \brief Method called when the mouse moves over the panel.
	 */
	virtual void OnMouseOver(const IMouse &m);
	
	/*!
	 * \brief Method called when the mouse moves out of the panel.
	 */
	virtual void OnMouseLeave(const IMouse &m);
	
	/*!
	 * \brief Method called when the mouse is hovering over the panel.
	 */
	virtual void OnMouseHover(const IMouse &m);
	
	/*!
	 * \brief Method called when a mouse clicks a button on the panel.
	 */
	virtual void OnMouseClick(const IMouse &m);
	
	/*!
	 * \brief Method called when a mouse releases a button over the panel.
	 */
	virtual void OnMouseRelease(const IMouse &m);
	
	/*!
	 * \brief Returns true if the point 'p' is inside this panel.
	 */
	virtual bool InActiveArea(const Point2f &p) const;
	
	/*!
	 * \brief Called when the panel gains keyboard focus.
	 */
	virtual void OnFocusGained(const IKeyboard &kb);
	
	/*!
	 * \brief Called when the panel loses keyboard focus.
	 */
	virtual void OnFocusLost(const IKeyboard &kb);
	
	/*!
	 * \brief Called when the panel is activated.
	 */
	virtual void OnActivation() {}
	
	/*!
	 * \brief Called when the panel is deactivated.
	 */
	virtual void OnDeactivation() {}
	
	/*!
	 * \brief Called each frame when the panel has focus.
	 */
	virtual void KeyboardUpdate(const IKeyboard &kb);
	
	/*!
	 * \brief Gets the string type identifier for the GUI element.
	 *
	 * \return The string type identifier.
	 */
	virtual const char *GetPanelTypeName() const { return "GUI::Panel"; }
	
	/*!
	 * \brief Updates the panel and all of its active children.
	 *
	 * NOTE: This controls the propagation of the OnUpdate calls down the tree.
	 */
	virtual void Update(float dt);
	
	/*!
	 * \brief Draws the GUI element and its children.
	 *
	 * NOTE: Do not override.
	 */
	void Draw();
	
	/*!
	 * \brief Requests keyborad focus.
	 */
	void RequestFocus() { InternalRequestFocus(this); }
	
	/*!
	 * \brief Visits each child util the visistors Vist method returns false or all children are visited. 
	 *
	 * NOTE: Visits children in back to front order.
	 */
	void VisitChildren(const GUIPanelVisitor &visitor);
	
	/*!
	 * \brief Visits each child util the visistors Vist method returns false or all children are visited. 
	 *
	 * NOTE: Visits chidren front to back.
	 */
	void VisitChildrenFrontToBack(const GUIPanelVisitor &visitor);
	
	/*!
	 * \brief Adds a child panel.
	 *
	 * \param[in] p the panel.
	 */
	void AddChild(GUIPanel *p);
	
	/*!
	 * \brief Removes a child panel.
	 *
	 * \param[in] p the panel to remove.
	 */
	void RemoveChild(GUIPanel *p);
	
	/*!
	 * \brief Removes a child panel.
	 *
	 * \param[in] i the index of the child to remove.
	 */
	void RemoveChild(uint i);
	
	/*!
	 * \brief Returns the number of child panels.
	 */
	uint NumberOfChildren() const;

	/*!
	 * \brief Finds the first panel under the given point.
	 *
	 * \param[in] p the query point.
	 *
	 * \return the first panel under the given point.
	 */
	GUIPanel *Query(const Point2f &p) const;
	
	/*!
	 * \brief Returns the panels parent.
	 */
	GUIPanel *GetParent() const;
	
	/*!
	 * \brief Returns the ith of child panel.
	 */
	GUIPanel *GetChild(uint i) const;
	
	/*!
	 * \brief Sets the position of the top left corner of the GUI panel.
	 *
	 * \param[in] p the position.
	 */
	void SetPosition(const Point2f &p);
	
	/*!
	 * \brief Sets the width of the panels bounding area.
	 *
	 * \param[in] w the width.
	 */
	void SetWidth(float w) { width = w; }
	
	/*!
	 * \brief Sets the height of the panels bounding area.
	 *
	 * \param[in] h the height.
	 */
	void SetHeight(float h) { height = h; }
	
	/*!
	 * \brief Sets the panels area.
	 */
	void SetArea(const Rectf &r);
	void SetArea(const AABB &a);
	
	/*!
	 * \brief Positions the panel into the center of the given area.
	 */
	void CenterInArea(const Rectf &r, const Point2f offset = Point2f(0.0f, 0.0f));
	
	/*!
	 * \brief Returns the positioin of the top-left corner of the panel.
	 */
	Point2f GetPosition() const;
	
	/*!
	 * \brief Returns the bounds of the panel as an AABB.
	 */
	AABB GetBounds() const;
	
	/*!
	 * \brief Returns the area of the panel as an Rect.
	 */
	Rectf GetArea() const;
	
	/*!
	 * \brief Gets the width of the bounding area for the panel.
	 *
	 * \return the width.
	 */
	float GetWidth() const { return width; }
	
	/*!
	 * \brief Gets the height of the bounding area for the panel.
	 *
	 * \return the height.
	 */
	float GetHeight() const { return height; }
	
	/*
	 * \brief Computes the bounds for the panel and its children.
	 *
	 * NOTE: Do not override.
	 */
	void Refit();
	
	/*!
	 * \brief Moves the panel to the front of the z-order.
	 */
	void MoveToFront();
	
	/*!
	 * \brief Sets the z-order of the panel.
	 *
	 * NOTE: Depth moves into the screen.  0 is the front of the screen.
	 */
	void SetDepth(uint z);
	
	/*!
	 * \brief Gets the z-order of the panel.
	 *
	 * NOTE: Depth moves into the screen.  0 is the front of the screen.
	 */
	uint GetDepth() const { return depth; }
	
	/*!
	 * \brief Returns true if the panel is active.
	 */
	bool IsActive() const { return active; }
	
	/*!
	 * \brief Activates the panel.
	 */
	void Activate() { active = true; OnActivation(); }
	
	/*!
	 * \brief Deactivates the panel.
	 */
	void Deactivate() { active = false; OnDeactivation(); }
	
	/*!
	 * \brief Returns true if the panel is visable.
	 */
	bool IsVisible() const { return visible; }
	
	/*!
	 * \brief Sets if the panal is visable or not.
	 */
	void SetVisibility(bool vis) { visible = vis; }
	
protected:
	
	/*!
	 * \brief Does the actual rendering of the GUI element.
	 *
	 * NOTE: Defaults to a no-op.  Override in child class were applicable.
	 */
	virtual void Paint() {}
	
	/*!
	 * \brief Recomputes the bounds and layout for the panel. Called by the Refit() method.
	 *
	 * NOTE: Defaults to a no-op.  Override in child class were applicable.
	 */
	virtual void RecomputeLayout() {}
	
	/*!
	 * \brief Call once a frame when the panel is active.
	 *
	 * NOTE: Defaults to a no-op.  Override in child class were applicable.
	 */
	virtual void OnUpdate(float dt) {}
	
protected:
	
	Point2f position; // The top left corner
	float width;
	float height;
	uint depth;
	
	bool active;
	bool visible;
	
protected:
	
	/*!
	 * \brief Moves up the tree until a panel that can grant focus is reached.
	 *
	 * NOTE: Do not override this method.
	 */
	virtual void InternalRequestFocus(GUIPanel *p) { if(GetParent()) GetParent()->InternalRequestFocus(p); }

private:
	
	// returns false if 'p' is not a child of this panel
	bool GetIndex(GUIPanel *p, uint &index) const;
	
	// Moves the child at index i to the front of the stack of child panels
	void MoveChildToFront(uint i);
	
	// Called when depth is changed by a child panel. 
	void DepthChanged();
	
	GUIPanel *parent;
	std::vector<GUIPanel*> children;
};

#endif
