/*!
 * \file GUIProgressBar.h
 * \brief Defines the progress bar gui element.
 * \author Joe Forte
 * \date 1/18/2009
 */

#ifndef GUIPROGRESSBAR_H
#define GUIPROGRESSBAR_H

#include <string>
#include "GUI/GUIPanel.h"
#include "Graphics/Color.h"

/*!
 * \class GUIProgressBar
 * \headerfile GUIProgressBar.h "GUI/GUIProgressBar.h"
 * \brief Class that represents a progress bar.
 */
class GUIProgressBar : public GUIPanel
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	GUIProgressBar() : progress(0.0f), borderColor(Color::White), barColor(Color::White), bgColor(Color::Black) {}
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~GUIProgressBar() {}
	
	/*!
	 * \brief Sets the background color of the progress bar.
	 *
	 * \param[in] c the color.
	 */
	void SetBackgroundColor(const Color4f &c) { bgColor = c; }
	
	/*!
	 * \brief Sets the color of the progress bar border.
	 *
	 * \param[in] c the color.
	 */
	void SetBorderColor(const Color4f &c) { borderColor = c; }
	
	/*!
	 * \brief Sets the color of the progress bar.
	 *
	 * \param[in] c the color.
	 */
	void SetBarColor(const Color4f &c) { barColor = c; }
	
	/*!
	 * \brief Increments the progress bar.
	 *
	 * \param[in] amt the amount to increment the progress bar.  Clamped to the range [0,1].
	 */
	void IncrementProgress(float amt);
	
	/*!
	 * \brief Sets the progress.
	 *
	 * \param[in] val the new progress clamped to the range [0, 1]. 
	 */
	void SetProgress(float val);
	
	/*!
	 * \brief Gets the current progress of the progress bar in the range [0, 1].
	 *
	 * \return The progress in the range [0, 1].
	 */
	float GetProgress() const { return progress; }
	
	/*!
	 * \brief Resets the progress of the progress bar.
	 */
	void ResetProgress() { progress = 0.0f; }
	
	/*!
	 * \brief Sets the inner text to draw.
	 */
	void SetInnerText(const char *str) { innerText = str; }
	
	/*!
	 * \brief Sets the inner text font.
	 */
	void SetInnerTextFont(FontRef font) { innerTextFont = font; }
	
protected:
	
	/*!
	 * \brief Draws the progress bar.
	 */
	virtual void Paint();
	
protected:
	
	float progress;
	
	std::string innerText;
	FontRef innerTextFont;
	
	Color4f borderColor;
	Color4f barColor;
	Color4f bgColor;
};

#endif
