/*!
 * \file GUITextEntry.h
 * \brief Defines text entry gui element.
 * \author Joe Forte
 * \date 1/18/2009
 */

#ifndef GUITEXTENTRY_H
#define GUITEXTENTRY_H

#include "GUI/GUIPanel.h"
#include "Graphics/Font.h"
#include "Graphics/Color.h"
#include "Input/BufferedKeyboardInput.h"

/*!
 * \class GUITextEntry
 * \headerfile GUIText.h "GUI/GUITextEntry.h"
 * \brief Text entry gui element.
 */
class GUITextEntry : public GUIPanel
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	GUITextEntry();
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] p the panels parent.
	 */
	GUITextEntry(GUIPanel *p);
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] len the max length of the input.
	 */
	GUITextEntry(uint len);
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] len the max length of the input.
	 * \param[in] keys an array of keys to filter.
	 * \param[in] n the length of the keys array.
	 */
	GUITextEntry(uint len, const Key *keys, uint n);
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~GUITextEntry() {}
	
	/*!
	 * \brief Sets the font to use when drawing text.
	 *
	 * \param[in] f the referance to the font.
	 */
	virtual void SetFont(FontRef f) { font = f; }
	
	/*!
	 * \brief Returns the font used for drawing text.
	 */
	FontRef GetFont() const { return font; }
	
	/*!
	 * \brief Sets the color to use when drawing text.
	 *
	 * \param[in] c the color.
	 */
	void SetColor(const Color4f &c) { color = c; }
	
	/*!
	 * \brief Returns the input text.
	 */
	const std::string &GetText() const;
	
	/*!
	 * \brief Clears the input text.
	 */
	void ClearText();
	
	/*!
	 * \brief Returns the cursor position for the input text.
	 */
	uint GetCursorPosition() const;
	
	/*!
	 * \brief Returns the max length for the input text.
	 */
	uint GetMaxInputLength() const;
	
	/*!
	 * \brief Method called when a mouse clicks a button on the panel.
	 */
	virtual void OnMouseClick(const IMouse &m);
	
	/*!
	 * \brief Called each frame when the panel has focus.
	 */
	virtual void KeyboardUpdate(const IKeyboard &kb);
	
	/*!
	 * \brief Gets the string type identifier for the GUI element.
	 *
	 * \return The string type identifier.
	 */
	virtual const char *GetPanelTypeName() const { return "GUI::TextEntry"; }
	
protected:
	
	/*!
	 * \brief Does the actual rendering of the GUI element.
	 *
	 * NOTE: Defaults to a no-op.  Override in child class were applicable.
	 */
	virtual void Paint();
	
	/*!
	 * \brief Recomputes the bounds and layout for the panel. Called by the Refit() method.
	 */
	virtual void RecomputeLayout();
	
private:
	
	FontRef font;
	Color4f color;
	
	LengthLimitedBufferedKeyboardInput bufferedInput;
	
};

#endif
