/*!
 * \file Ball.h
 * \brief Declares the ball class.
 * \author Joe Forte
 * \date 3/18/2008
 */

#ifndef BALL_H
#define BALL_H

#include "Math/MathCommon.h"
#include "Graphics/Sprite.h"
#include "Graphics/ParticleSystem.h"
#include "Sound/Sound.h"
#include "Physics/PhysicsWorld.h"
#include "Physics/PhysicsObject.h"

/*!
 * \class BallTrailPolicy
 * \headerfile Ball.h "Game/Ball.h"
 * \brief Policy that the creates a Ball trail ParticleSystem.
 */
class BallTrailPolicy
{
public:
	
	typedef ContinuousEmitPolicy<> Emit;

	class Init
	{
	public:
		float radius;
		Color4f color;
		
		Init() : radius(0) {}
		
		inline void PreEmit(const Point2f &pos, uint amount) { position = pos; }
		
		inline void InitParticle(Particle &p)
		{
			p.color = color;
			p.life = RandomFloat(0.5f, 1.0f);
			p.age = 0.0f;
			p.angle = 0.0f;
			p.angularVel = 0.0f;
			
			float size = RandomFloat(10.0f, 25.0f);
			p.width = size;
			p.height = size;
			
			float angle = RandomFloat(0.0f, _2PI);
			Vector2f vel(std::cos(angle), std::sin(angle));
			p.pos = position + radius*vel;
			
			p.vel = RandomFloat(1.0f, 5.0f) * vel;
		}
		
	private:
		Point2f position;
	};

	class Update
	{
	public:

		inline void PreUpdate(float dt) {}

		inline bool ApplyUpdate(Particle &p, float dt)
		{
			p.pos = p.pos + p.vel*dt;
			
			p.age += dt;
			
			float delta = p.age / p.life;
			p.color.a = 1.0f - delta;
			
			return p.age < p.life;
		}
	};
};


/*!
 * \class BallHitSmokePolicy
 * \headerfile Ball.h "Game/Ball.h"
 * \brief Policy that the creates the Ball hit smoke ParticleSystem.
 */
class BallHitSmokePolicy
{
public:

	class Init
	{
	public:
		Vector2f hitDir;

		inline void PreEmit(const Point2f &pos, uint amount) { position = pos; }

		inline void InitParticle(Particle &p)
		{
			p.pos = position;
			p.color = Color::White;
			p.color.a = 1.0f;
			p.life = RandomFloat(0.2f, 0.5f);
			p.age = 0.0f;
			p.angle = 0.0f;
			p.angularVel = 0.0f;
			
			float size = RandomFloat(10.0f, 20.0f);
			p.width = size;
			p.height = size;
			
			Vector2f d = hitDir.Perp();
			
			p.vel = RandomFloat(-100.0f, 100.0f) * d + RandomFloat(0.0f, 50.0f) * hitDir;
		}
		
	private:
		Point2f position;
	};

	class Update
	{
	public:
		inline void PreUpdate(float dt) {}
		
		inline bool ApplyUpdate(Particle &p, float dt)
		{
			p.pos = p.pos + p.vel*dt;
			
			p.age += dt;
			
			p.color.a = 1.0f - (p.age/p.life);
			
			return p.age < p.life;
		}
	};
};

/*!
 * \class Ball
 * \headerfile Ball.h "Game/Ball.h"
 * \brief Class that represents the brickout ball.
 */
class Ball : public PhysicsObject
{
public:
	
	Ball();
	
	virtual ~Ball() {}
	
	virtual bool Touch(PhysicsObject *obj, const PhysicsCollisionInfo &hitInfo);

	virtual PhysicsObjectType GetObjectType() const { return POBJ_BALL; }

	void SetImage(const char *file);
	
	void SetRadius(float r);
	
	void Update(float dt);
	
	void Draw();
	
	void SetAttachment(bool val) { attached = val; }
	
	Circle GetCircle() const;
	
	float GetRadius() const { return radius; }
	
	AABB GetBBox() { return ball.GetBBox(); }
	
private:
	Sprite ball;
	
	float radius;
	
	bool attached; // set to true when the ball is attached to a paddle.
	
	ParticleSystem<100, BallTrailPolicy::Init, BallTrailPolicy::Update, BallTrailPolicy::Emit> trail1;
	ParticleSystem<100, BallTrailPolicy::Init, BallTrailPolicy::Update, BallTrailPolicy::Emit> trail2;
	
	ParticleSystem<200, BallHitSmokePolicy::Init, BallHitSmokePolicy::Update> smokeHit;

	Sound hitSound;
};

#endif
