/*!
 * \file BrickGroup.cpp
 * \brief Implements the BrickGroup class.
 * \author Joe Forte
 * \date 3/27/2008
 */

#include <assert.h>
#include "Game/Ball.h"
#include "Game/BrickGroup.h"
#include "Game/ConsoleVars.h"

//--------------------------------------------------------------------
BrickGroup::BrickGroup(CollisionGroup belongsTo, CollisionGroup collidesWith) 
: 
listener(NULL),
belongsToGroup(belongsTo),
collidsWithGroup(collidesWith)
{
	brickCount = 0;
}

//--------------------------------------------------------------------
BrickGroup::~BrickGroup()
{
	std::vector<Brick*>::iterator i;
	for(i = bricks.begin(); i != bricks.end(); i++)
	{
		(*i)->Destroy();
		delete *i;
	}
}

//--------------------------------------------------------------------
void BrickGroup::SetAttributes(const BrickGroup::Attributes &attributes)
{
	attribs = attributes;
}

//--------------------------------------------------------------------
void BrickGroup::Init(PhysicsWorld *w)
{
	world = w;

	uint len = (uint)attribs.imageFiles.size();
	for(uint i = 0; i < len; i++)
	{
		Sprite s;
		s.Load(attribs.imageFiles[i].c_str());
		s.SetSize((float)attribs.brickWidth, (float)attribs.brickHeight);
		images.push_back(s);
	}

	CreateBrickDistrobution(attribs.points);
}

//--------------------------------------------------------------------
void BrickGroup::CreateBrickDistrobution(const std::vector<Point2f> &pos)
{
	std::vector<Point2f>::const_iterator i;
	for(i = pos.begin(); i != pos.end(); i++)
	{
		CreateBrick( *i );
	}
}

//--------------------------------------------------------------------
void BrickGroup::Update(float dt, Ball &b)
{
	if(!IsInvincible())
	{
		// Destroys all dead bricks
		std::vector<Brick*>::iterator i = bricks.begin();
		while(i != bricks.end())
		{
			if((*i)->life == 0)
			{
				(*i)->Destroy();
				delete *i;
				i = bricks.erase(i);
				brickCount--;
			}
			else
			{
				i++;
			}
		}
	}
}

//--------------------------------------------------------------------
void BrickGroup::CreateBrick(const Point2f &p)
{
	Brick *b = new Brick(this);
	b->MakeStatic();
	b->AttachToWorld(world);
	b->SetShape( CreatePhysicsShapeBox((float)attribs.brickWidth, (float)attribs.brickHeight) );
	b->SetPosition(p);
	b->SetCollisionGroups(belongsToGroup, collidsWithGroup);
	b->life = attribs.brickHP;

	bricks.push_back(b);
	brickCount++;
}

//--------------------------------------------------------------------
void BrickGroup::HitBrick(Brick *b, const PhysicsCollisionInfo &hitInfo)
{
	assert(b);
	if(b->life == 0)
		return;

	if(!IsInvincible())
		b->life--;
	
	int points = 0;
	if(b->life > 0)
		points = attribs.pointsPerHit;
	else
		points = attribs.pointsPerKill + attribs.pointsPerHit;

	if(listener)
		listener->BrickHit(*this, *b, hitInfo, points);
}

//--------------------------------------------------------------------
AABB BrickGroup::GetBrickBounds(uint i) const
{
	if((int)i >= brickCount)
		return AABB(0.0f, 0.0f, 0.0f, 0.0f);
	return AABB(bricks[i]->GetPosition(), (float)attribs.brickWidth, (float)attribs.brickHeight);
}

//--------------------------------------------------------------------
void BrickGroup::Draw()
{
	bool drawShapes = !!phy_drawCollisionShapes->GetInt();
	
	std::vector<Brick*>::iterator i;
	for(i = bricks.begin(); i != bricks.end(); i++)
	{
		if((*i)->life > 0)
		{
			int index = std::min<uint>(attribs.brickHP - (*i)->life, (uint)images.size() - 1);
		
			images[index].SetPosition( (*i)->GetPosition() );
			images[index].SetAngle( (*i)->GetAngle() ); 
			images[index].Draw();
			
			if(drawShapes)
			   (*i)->GetShapes().DrawOutline();
		}
	}
}

//--------------------------------------------------------------------
bool Brick::Touch(PhysicsObject *obj, const PhysicsCollisionInfo &hitInfo)
{
	assert(owner);
	
	if(obj->GetObjectType() == POBJ_BALL)
		owner->HitBrick(this, hitInfo);
	
	return true;
}
