/*!
 * \file BrickGroup.h
 * \brief Declares the BrickGroup class.
 * \author Joe Forte
 * \date 3/27/2008
 */

#ifndef BRICKGROUP_H
#define BRICKGROUP_H

#include <string>
#include <vector>
#include "Math/MathCommon.h"
#include "Graphics/Sprite.h"
#include "Physics/PhysicsObject.h"

class PhysicsWorld;
class Ball;

class BrickGroup;

class Brick : public PhysicsObject
{
public:
	Brick(BrickGroup *bg) : owner(bg) {}

	virtual bool Touch(PhysicsObject *obj, const PhysicsCollisionInfo &hitInfo);

	virtual PhysicsObjectType GetObjectType() const { return POBJ_BRICK; }

	int life;

private:
	BrickGroup *owner;
};

class BrickGroupListener
{
public:
	
	/*!
	 * \brief Called each time a brick is hit.
	 *
	 * \param[in] group referance to the appropriate BrickGroup.
	 * \param[in] brick referance to the Brick that was hit.
	 * \param[in] info the collision information for the brick hit.
	 * \param[in] points the number of points awarded for the hit.
	 */
	virtual void BrickHit(BrickGroup &group, Brick &brick, const PhysicsCollisionInfo &hitInfo, int points) = 0;
};

class BrickGroup
{
public:
	BrickGroup(CollisionGroup belongsTo = CG_GROUP1, CollisionGroup collidesWith = CG_GROUP1);
	~BrickGroup();

	struct Attributes
	{
		std::string name;
		
		std::vector<std::string> imageFiles;
		int brickWidth;
		int brickHeight;
		int brickHP;
		int pointsPerHit;
		int pointsPerKill;  // number of points for clearing a brick
		bool invincible;
		
		std::vector<Point2f> points;
	};
	
	void SetAttributes(const BrickGroup::Attributes &attributes);

	void Init(PhysicsWorld *w);

	void Update(float dt, Ball &b);
	
	int GetBrickCount() const { return brickCount; }
	
	void Draw();

	void SetListener(BrickGroupListener *lsn) { listener = lsn; }
	
	/*!
	 * \brief Determines the maximum number of points that can be obtained from a single brick.
	 *
	 * \return The number of points.
	 */
	int GetMaxPointsPerBrick() const { return attribs.pointsPerHit * attribs.brickHP + attribs.pointsPerKill; }
	
	/*!
	 * \brief Determines if the bricks have negitive point values.
	 *
	 * \return True if negitive points can be given, false otherwise.
	 */
	bool HasNegativePoints() const { return attribs.pointsPerHit < 0 || attribs.pointsPerKill < 0; }

	/*!
	 * \brief determines if the bricks can be destroyed.
	 *
	 * \return true if the bricks are indestructable, false otherwise.
	 */
	bool IsInvincible() const { return attribs.invincible; }

protected:

	AABB GetBrickBounds(uint i) const;
	
	void CreateBrick(const Point2f &p);

	void HitBrick(Brick *b, const PhysicsCollisionInfo &hitInfo);
	
	void SetImageScale(int i);
	
	void SetImage(int i, const char *str);
	
	void CreateBrickDistrobution(const std::vector<Point2f> &pos);

private:
	
	CollisionGroup belongsToGroup;
	CollisionGroup collidsWithGroup;
	
	Attributes attribs;
	
	std::vector<Brick*> bricks;
	int brickCount;
	
	std::vector<Sprite> images;
	
	PhysicsWorld *world;

	BrickGroupListener *listener;

	friend class Brick;
};


#endif
