/*!
 * \file BrickOutGame.cpp
 * \brief Implements the BrickOutGame class.
 * \author Joe Forte
 * \date 3/18/2008
 */

#include <assert.h>
#include <fstream>

#include "Input/SDLMouse.h"
#include "Input/SDLKeyboard.h"

#include "Graphics/SDLScreen.h"
#include "Graphics/MathDraw.h"

#include "Math/MathCommon.h"

#include "Utility/StringUtility.h"

#include "Common/FileSystem.h"

#include "Game/GameMouse.h"
#include "Game/GUIMainMenu.h"
#include "Game/GUIScoreBoard.h"
#include "Game/HighScores.h"
#include "Game/GameSplashScreen.h"
#include "Game/LevelParser.h"
#include "Game/BrickOutGame.h"
#include "Game/ConsoleVars.h"
#include "Game/StarFieldEffect.h"
#include "Game/GUINameEntryScreen.h"
#include "Game/MusicVisualizerEffect.h"

#include "Application/DataPaths.h"

const char *ICON_NAME_WINDOWS = "BrickOutReDuXIcon32.bmp";

const char *DEFAULT_HIGHSCORE_NAMES[] = 
{
	"Unknown Warrior",
	"An Impatient Guy",
	"Anonymous Coward",
	"Some Paranoide Guy",
	"Smitty Werbenjegarmanjensen",
	"Anonymous Coward",
	"A Paranoide Guy",
	"Some Dude",
	"Some Impatient Guy",
	"The Dude",
};

const uint DEFAULT_HIGHSCORE_NAMES_COUNT = 10;

const Key TOGGLE_CONSOLE_KEY = Keys::KEY_BACKQUOTE;

//--------------------------------------------------------------------
const char *ChooseRandomHighScoreName()
{
	return DEFAULT_HIGHSCORE_NAMES[ rand() % DEFAULT_HIGHSCORE_NAMES_COUNT ];
}

//--------------------------------------------------------------------
BrickOutGame::BrickOutGame(int w, int h, bool useFullscreen)
:
SDLApplication(w, h, useFullscreen),
mainMenu(NULL),
splashScreen(NULL),
mouseEffect(NULL),
bgEffect(NULL),
scoreBoard(NULL),
level(NULL)
{}

//--------------------------------------------------------------------
BrickOutGame::~BrickOutGame()
{
	if(level)
		delete level;
	
	if(splashScreen)
		delete splashScreen;

	if(mouseEffect)
		delete mouseEffect;
	
	if(scoreBoard)
		delete scoreBoard;

	if(highScores)
		delete highScores;

	if(bgEffect)
		delete bgEffect;
}

//--------------------------------------------------------------------
const char *BrickOutGame::GetIconName()
{
#ifdef _WIN32
	return ICON_NAME_WINDOWS;
#endif
	return NULL;
}

//--------------------------------------------------------------------
bool BrickOutGame::InitSubSystems()
{
	console.RegisterCmd("level", "Loads a spacific level.", ConCmdLevel, this);
	console.RegisterCmd("clearHighScores", "Permanently erases all high scores.", ConCmdClearHighScores, this);

	console.RegisterCmd("mp_next", "Moves the music player to the next song.", ConCmdMpNext, this);
	console.RegisterCmd("mp_previous", "Moves the music player to the previous song.", ConCmdMpPrevious, this);
	console.RegisterCmd("mp_pause", "Pauses the music player.", ConCmdMpPause, this);
	console.RegisterCmd("mp_play", "Sets the music player to play.", ConCmdMpPlay, this);
	console.RegisterCmd("mp_stop", "Stops the music player.", ConCmdMpStop, this);
	console.RegisterCmd("mp_volume", "Gets/Sets the music players volume.", ConCmdMpVolume, this);
	console.RegisterCmd("mp_info", "Prints out the music players info.", ConCmdMpInfo, this);
	console.RegisterCmd("mp_repeat", "Controls the music players repeat function.", ConCmdMpRepeat, this);
	console.RegisterCmd("mp_loadPlaylist", "Loads a new playlist.", ConCmdMpLoadPlaylist, this);
	console.RegisterCmd("mp_printPlaylist", "Prints the current playlist.", ConCmdMpPrintPlaylist, this);
	
	// Registers global console vars
	RegisterGlobalConsoleVars(console);
	
	console.IgnoreToggleKey(TOGGLE_CONSOLE_KEY);

	return true;
}

//--------------------------------------------------------------------
void BrickOutGame::TogglePause()
{
	paused = !paused;
}

//--------------------------------------------------------------------
void BrickOutGame::Pause()
{
	paused = true;
}

//--------------------------------------------------------------------
void BrickOutGame::UnPause()
{
	paused = false;
}

//--------------------------------------------------------------------
bool BrickOutGame::IsPaused() const
{
	return paused;
}

//--------------------------------------------------------------------
bool BrickOutGame::BallLost(Level &level)
{
	scoreBoard->DecrementBallCount();
	return scoreBoard->GetBallCount() >= 0;
}

//--------------------------------------------------------------------
void BrickOutGame::ExtraBall(Level &level)
{
	scoreBoard->IncrementBallCount();
}

//--------------------------------------------------------------------
void BrickOutGame::ScoreIncrease(Level &level, int amt)
{
	scoreBoard->IncrementScore(amt);
}

//--------------------------------------------------------------------
void BrickOutGame::ClearHighScores()
{
	if(highScores)
	{
		highScores->ClearHighScores();
		ConMsg() << "High scores cleared.\n";
	}
}

//--------------------------------------------------------------------
bool BrickOutGame::ApplicationInit()
{
	if( !gameSettings.Load(ConcatPath(DataPaths::GetPath("Config"), GAME_CONFIG_FILE).c_str()) )
	{
		return false;
	}
	
	extraLargeFont = LoadFontResource("VeraMono.ttf", 200);
	
	// Sets up the main menu
	Rectf aboutArea;
	aboutArea.w = width / 2.0f;
	aboutArea.h = height / 2.0f;
	aboutArea.x = -width / 4.0f;
	aboutArea.y = height / 4.0f;
	
	mainMenu = new GUIMainMenu(&state, 
							   systemFont, 
							   LoadFontResource("VeraMono.ttf", 16), 
							   aboutArea, 
							   ABOUT_TEXT,
							   LoadFontResource("VeraMono.ttf", 10));
	
	gui.SetInput(mouse, keyboard);
	gui.AddChild(mainMenu);
	
	nameEntryScreen = new GUINameEntryScreen(LoadFontResource("VeraMono.ttf", 24), 
											 LoadFontResource("VeraMono.ttf", 24),
											 systemFont);
	
	nameEntryScreen->Deactivate();
	nameEntryScreen->SetVisibility(false);
	gui.AddChild(nameEntryScreen);
	
	LoadLevelList(gameSettings.DefaultLevelListFile(), levelFiles);
	currLevel = 0;
	paused = false;
	
	scoreBoard = new GUIScoreBoard(gameSettings);
	
	completeScreen.Init(systemFontLarge, systemFont, 1.0f);

	// Make sure we get told when we lose a ball or score some points.
	level = new Level;
	level->SetListener(this);
	
	state.SetListener(this);

	mouse->SetCursor("cursor.png");
	
	// Sets up the transition effect
	transEffect.SetListener(this);

	transEffect.PushTransition( new AlphaFadeInTransition(0.5f) );
	//transEffect.PushTransition( new SlantedSlideInTransition(1.0f) );

	transEffect.PushTransition( new SolidColorDelayTransition(0.2f) );

	transEffect.PushTransition( new AlphaFadeOutTransition(0.5f) );
	//transEffect.PushTransition( new SlantedSlideOutTransition(1.0f) );
	
	transEffect.SetArea( GetScreen()->GetRect() );
	transEffect.ForceCompletion();

	std::vector<std::string> args;
	if(commandLine.FindCmd("-level", &args) && args.size() > 0)
	{
		LoadLevel( ConcatPath(DataPaths::GetPath("Levels"), CreatePath(args[0])).c_str() );
	}
	else if(!commandLine.FindCmd("-nosplash"))
	{
		splashScreen = new GameSplashScreen;
		state.RequestStateChange(GameState::GS_SPLASH_SCREEN);
	}

	mouseEffect = new GameMouseEffect();

	bgEffect = new MusicVisualizerEffect( GetScreen()->GetRect() );
	//bgEffect = new StarFieldEffect( GetScreen()->GetRect() );
	if(!bgEffect->Init())
	{
		ConWarning() << "Default background effect could not be initialized.";
		return false;
	}
	
	// Setup the game credits
	credits.SetArea( GetScreen()->GetRect() );
	
	if( !credits.Load(ConcatPath( DataPaths::GetPath("Config"), GAME_CREDITS_FILE ).c_str()) )
	{
		ConWarning() << "Could not load game credits.";
		return false;
	}
	
	scoreBoard->SetScore(0);
	scoreBoard->SetBallCount(3);

	highScores = new HighScores(gameSettings);
	highScores->LoadScoresFromFile( ConcatPath(DataPaths::GetPath("Config"), "scores.dat").c_str() );

	//console.ExecuteCommand("mp_loadPlaylist -m Playlist1.xml");
	
	return true;
}

//--------------------------------------------------------------------
bool BrickOutGame::OnApplicationExitRequest()
{
	state.RequestStateChange(GameState::GS_EXIT);	
	return false;
}

//--------------------------------------------------------------------
void BrickOutGame::OnLostFocus()
{
	Pause();
}

//--------------------------------------------------------------------
void BrickOutGame::OnGainedFocus()
{
	UnPause();
}

//--------------------------------------------------------------------
void BrickOutGame::OnLostMouseFocus()
{
	mouseEffect->LostFocus();
}

//--------------------------------------------------------------------
void BrickOutGame::OnGainedMouseFocus()
{
	mouseEffect->GainedFocus();
}

//--------------------------------------------------------------------
void BrickOutGame::PreFrame()
{
	// Determines if we should be in fixed mouse mode or not
	if(state.Get() == GameState::GS_LEVEL && !console.IsOpened() && HaveFocus())
	{
		if(!mouse->InFixedPositionMode())
			mouse->StartFixedPositionMode();
	}
	else
	{
		if(mouse->InFixedPositionMode())
			mouse->EndFixedPositionMode();
	}
}

//--------------------------------------------------------------------
void BrickOutGame::PostFrame()
{}

//--------------------------------------------------------------------
bool BrickOutGame::GameComplete() const
{
	return level->IsComplete() && !loadLevelRequested && currLevel == levelFiles.size() - 1;
}

//--------------------------------------------------------------------
bool BrickOutGame::ShouldNotDoTransitionEffect(GameState::State current, GameState::State next)
{
	return next == GameState::GS_START_GAME || 
			(current == GameState::GS_MAIN_MENU_OVERLAY && next != GameState::GS_EXIT);
}

//--------------------------------------------------------------------
void BrickOutGame::SetupTransitionEffect(TransitionEffect &effect, GameState::State current, GameState::State next)
{
	if(next == GameState::GS_SPLASH_SCREEN)
	{
		Transition::TransitionType order[] =
		{
			Transition::TRANSITION_DELAY,
			Transition::TRANSITION_IN
		};

		effect.ResetOrder(order, sizeof(order) / sizeof(Transition::TransitionType));
	}
	else if(next == GameState::GS_EXIT)
	{
		Transition::TransitionType order[] = { Transition::TRANSITION_OUT };

		effect.ResetOrder(order, sizeof(order) / sizeof(Transition::TransitionType));
	}
	else
	{
		Transition::TransitionType order[] = 
		{ 
			Transition::TRANSITION_OUT,
			Transition::TRANSITION_DELAY,
			Transition::TRANSITION_IN,
		};

		effect.ResetOrder(order, sizeof(order) / sizeof(Transition::TransitionType));
	}
}

//--------------------------------------------------------------------
void BrickOutGame::OnTransitionOutCompleted(TransitionEffect &effect, Transition &trans)
{
	if( state.IsStateChangeRequestPending() )
	{
		state.Transition();
	}
}

//--------------------------------------------------------------------
void BrickOutGame::OnTransitionInCompleted(TransitionEffect &effect, Transition &trans)
{}

//--------------------------------------------------------------------
void BrickOutGame::OnTransitionDelayCompleted(TransitionEffect &effect, Transition &trans)
{
	if( state.Get() == GameState::GS_STARTUP && state.GetRequestedState() == GameState::GS_SPLASH_SCREEN )
	{
		state.Transition();
	}
}

//--------------------------------------------------------------------
void BrickOutGame::OnTransitionEffectCompleted(TransitionEffect &effect)
{}

//--------------------------------------------------------------------
bool BrickOutGame::Update(float dt)
{	
	musicPlayer.Update(dt);
	
	// Handles the various meanings of the Escape key
	if( keyboard->KeyPressed(Keys::KEY_ESCAPE) )
	{
		if( transEffect.IsComplete() && state.Get() != GameState::GS_LOAD_LEVEL )
		{
			if(console.IsOpened())
			{
				console.Close();
			}
			else if(state.Get() == GameState::GS_MAIN_MENU_OVERLAY)
			{
				if(mainMenu->IsAboutMenuOpen())
					mainMenu->CloseAboutMenu();
				else
					state.ForceStateReversion();
			}
			else if(state.Get() == GameState::GS_SPLASH_SCREEN || 
					state.Get() == GameState::GS_HIGHSCORES || 
					state.Get() == GameState::GS_CREDITS)
			{
				state.RequestStateChange(GameState::GS_MAIN_MENU);
			}
			else if(state.Get() == GameState::GS_ENTER_NAME)
			{
				nameEntryScreen->SetNameEntryComplete();
			}
			else if(state.Get() != GameState::GS_MAIN_MENU)
			{
				state.ForceStateChange(GameState::GS_MAIN_MENU_OVERLAY);
			}
		}
		else // Transition effect in progress
		{
			if(console.IsOpened())
			{
				console.Close();
			}
		}
	}
	
	if( keyboard->KeyPressed(TOGGLE_CONSOLE_KEY) )
	{
		if(console.IsOpened())
			console.Close();
		else
			console.Open();
	}
	
	if( keyboard->KeyPressed(Keys::KEY_F5) )
	{
		static int lastSceenShot = 0;
		std::stringstream name;
		name << "screenshot" << lastSceenShot;
		std::string str = name.str();
		if(screen->SaveScreenshot(str))
		{
			console.Print() << "Screenshot saved.\n";
			lastSceenShot++;
		}
		else
		{
			console.PrintWarning() << "Screenshot could not be saved.\n";
		}
	}
	
	/// TEMP
#ifdef _DEBUG

	if(keyboard->KeyPressed(Keys::KEY_RIGHTBRACKET))
		scoreBoard->DecrementBallCount();
	else if(keyboard->KeyPressed(Keys::KEY_LEFTBRACKET))
		scoreBoard->IncrementBallCount();

#endif

	// Handles state transitions and transition effects
	if( transEffect.IsComplete() )
	{
		if( state.IsStateChangeRequestPending() )
		{
			GameState::State curr = state.Get();
			GameState::State next = state.GetRequestedState();
			
			if( ShouldNotDoTransitionEffect(curr, next) )
			{
				state.Transition();
			}
			else
			{
				SetupTransitionEffect(transEffect, curr, next);
			}
		}
		else
		{
			UpdateGameState(dt);
		}
	}
	else
	{
		BackgroundUpdateGameState(dt);
		transEffect.Update(dt);
	}
	
	mouseEffect->Update(dt, *mouse);

	if(bgEffect)
		bgEffect->Update(dt);

	return state.Get() != GameState::GS_EXIT;
}

//--------------------------------------------------------------------
void BrickOutGame::UpdateGameState(float dt)
{	
	switch(state.Get())
	{
		case GameState::GS_MAIN_MENU:
			{
				static float timer = 0.0f;
				timer += dt;
				if(mainMenu->IsAboutMenuOpen())
				{
					timer = 0.0f;
				}
				else if(timer > 15.0f)
				{
					timer = 0.0f;
					state.RequestStateChange(GameState::GS_HIGHSCORES);
				}
			}
		case GameState::GS_MAIN_MENU_OVERLAY:
			gui.Update(dt);
			break;
			
		case GameState::GS_LEVEL:
			if( keyboard->KeyPressed(Keys::KEY_p) && !console.IsOpened() )
					TogglePause();
			
			if(!IsPaused() && !console.IsOpened())
			{
				scoreBoard->Update(dt);
				level->Update(dt, *keyboard, *mouse);
			}
			
			if(scoreBoard->GetBallCount() < 0)
			{
				std::stringstream scoreStream;
				scoreStream << "Final Score: " << scoreBoard->GetScore();
				
				completeScreen.Reset("Game Over", scoreStream.str());
				completeScreen.SetIsNewHighScore( highScores->IsHighScore(scoreBoard->GetScore()) );
				
				state.ForceStateChange(GameState::GS_GAME_OVER);
			}
			else
			{
				if(level->IsComplete())
				{
					if(GameComplete())
					{
						std::stringstream scoreStream;
						scoreStream << "Final Score: " << scoreBoard->GetScore();
						
						completeScreen.Reset("Game Complete!", scoreStream.str());
						completeScreen.SetIsNewHighScore( highScores->IsHighScore(scoreBoard->GetScore()) );
						
						state.ForceStateChange(GameState::GS_GAME_COMPLETE);
					}
					else
					{
						std::stringstream scoreStream;
						scoreStream << "Score: " << scoreBoard->GetScore();
						
						completeScreen.Reset("Level Complete!", scoreStream.str());
						completeScreen.SetIsNewHighScore(false);
						
						state.ForceStateChange(GameState::GS_LEVEL_COMPLETE);
					}
				}
			}
			break;

        case GameState::GS_GAME_OVER:
		case GameState::GS_GAME_COMPLETE:	
		case GameState::GS_LEVEL_COMPLETE:
			completeScreen.Update(dt);
			if(!completeScreen.FadeInComplete())
			{
				level->Update(dt, *keyboard, *mouse);
			}
			else
			{
				static float timer = 0.0f;
				timer += dt;
				if(timer > 4.0f || mouse->ButtonPressed(Keys::MOUSE_BUTTON_LEFT))
				{
					timer = 0.0f;
					
					GameState::State next;
					GameState::State s = state.Get();
					if(loadLevelRequested)
					{
						next = GameState::GS_MAIN_MENU;
					}
					else if(s == GameState::GS_GAME_COMPLETE)
					{
						next = GameState::GS_CREDITS;
					}
					else if(s == GameState::GS_GAME_OVER)
					{						
						next = GameState::GS_HIGHSCORES;
					}
					else
					{
						currLevel++;
						next = GameState::GS_LOAD_LEVEL;
					}
					
					if(s == GameState::GS_GAME_COMPLETE || s == GameState::GS_GAME_OVER)
					{
						if(highScores->IsHighScore( scoreBoard->GetScore() ))
						{
							next = GameState::GS_ENTER_NAME;
						}
					}
					
					state.RequestStateChange(next);
				}
			}
			break;

		case GameState::GS_SPLASH_SCREEN:
			{
				static float timer = 0.0f;
				timer += dt;
				if(timer > 3.0f || keyboard->KeyPressed(Keys::KEY_SPACE) || mouse->ButtonPressed(Keys::MOUSE_BUTTON_LEFT))
				{
					timer = 0.0f;
					state.RequestStateChange(GameState::GS_MAIN_MENU);
				}
			}
			break;
			
		case GameState::GS_HIGHSCORES:
			{
				static float timer = 0.0f;
				timer += dt;
				if(timer > 10.0f || keyboard->KeyPressed(Keys::KEY_SPACE) || mouse->ButtonPressed(Keys::MOUSE_BUTTON_LEFT))
				{
					timer = 0.0f;
					state.RequestStateChange(GameState::GS_MAIN_MENU);
				}
			}
			break;
			
		case GameState::GS_ENTER_NAME:
			gui.Update(dt);
			if(nameEntryScreen->IsNameEntryComplete())
			{
				std::string name = nameEntryScreen->GetName();
				if(name.empty())
					name = ChooseRandomHighScoreName();
				
				highScores->AddHighScore(name.c_str(), scoreBoard->GetScore());
				if(GameComplete())
					state.RequestStateChange(GameState::GS_CREDITS);
				else
					state.RequestStateChange(GameState::GS_HIGHSCORES);
				
				nameEntryScreen->Deactivate();
			}
			break;
			
		case GameState::GS_START_GAME:			
		case GameState::GS_START_NEW_GAME:
			state.RequestStateChange(GameState::GS_LOAD_LEVEL);
			break;
			
		case GameState::GS_CREDITS:
			credits.Update(dt);
			if(credits.IsComplete())
				state.RequestStateChange(GameState::GS_HIGHSCORES);
			break;
			
		case GameState::GS_LOAD_LEVEL:
			{
				const char *levelName;
				if(loadLevelRequested)
					levelName = requestedLevelName.c_str();
				else
					levelName = levelFiles[currLevel].c_str();
				
				uint startTime = GetTickCount();
				if( parser.LoadLevel(levelName, *level) )
				{
					level->Init(&gameSettings);
					
					float loadTime = (GetTickCount() - startTime) / 1000.0f;
					console.Print() << levelName << " \"" << level->GetName() << "\" loaded in " << loadTime << " seconds.\n";
				
					state.RequestStateChange(GameState::GS_LEVEL);
				}
				else
				{
					state.RequestStateChange(GameState::GS_MAIN_MENU);
				}
			}
			break;
	};
}

//--------------------------------------------------------------------
void BrickOutGame::BackgroundUpdateGameState(float dt)
{
	switch(state.Get())
	{
		case GameState::GS_MAIN_MENU:
		case GameState::GS_MAIN_MENU_OVERLAY:
			break;
			
		case GameState::GS_LEVEL:
			break;
			
        case GameState::GS_GAME_OVER:
		case GameState::GS_GAME_COMPLETE:	
		case GameState::GS_LEVEL_COMPLETE:
			completeScreen.Update(dt);
			break;
			
		case GameState::GS_SPLASH_SCREEN:
			break;
			
		case GameState::GS_HIGHSCORES:
			break;
			
		case GameState::GS_START_GAME:
			break;
			
		case GameState::GS_CREDITS:
			break;
			
		case GameState::GS_LOAD_LEVEL:
			break;
	};	
}

//--------------------------------------------------------------------
void BrickOutGame::Draw(float dt)
{
	DrawFrame();
	
	screen->DrawQueuedText();
	
	GameState::State curr = state.Get();

	if( !console.IsOpened() &&
		curr != GameState::GS_EXIT &&
		curr != GameState::GS_STARTUP &&
		curr != GameState::GS_SPLASH_SCREEN &&
	    curr != GameState::GS_LOAD_LEVEL &&
		curr != GameState::GS_LEVEL_COMPLETE &&
	    curr != GameState::GS_CREDITS &&
	    curr != GameState::GS_HIGHSCORES &&
		curr != GameState::GS_GAME_OVER &&
		curr != GameState::GS_GAME_COMPLETE &&
		curr != GameState::GS_LEVEL )
	{
		mouseEffect->Draw();
		mouse->Draw();
	}

	if(!transEffect.IsComplete())
		transEffect.Draw();

	if(console.IsOpened())
	{
		console.Draw();
		mouseEffect->Draw();
		mouse->Draw();
	}
}

//--------------------------------------------------------------------
void BrickOutGame::DrawFrame()
{
	switch(state.Get())
	{
		case GameState::GS_STARTUP:
			DrawFullscreenQuad( Color::Black );
			break;

		case GameState::GS_START_NEW_GAME:
		case GameState::GS_MAIN_MENU_OVERLAY:
			if(bgEffect) 
				bgEffect->Draw();
			DrawLevel();
			screen->DrawQueuedText();
			DrawFullscreenQuad( Color4f(0.0f, 0.0f, 0.0f, 0.8f) );
			gui.Draw();
			break;

		case GameState::GS_START_GAME:
		case GameState::GS_MAIN_MENU:
			if(bgEffect) 
				bgEffect->Draw();
			gui.Draw();
			break;

		case GameState::GS_SPLASH_SCREEN:
			splashScreen->Draw();
			break;

		case GameState::GS_GAME_OVER:
		case GameState::GS_GAME_COMPLETE:
		case GameState::GS_LEVEL_COMPLETE:
			if(bgEffect) 
				bgEffect->Draw();
			DrawLevel();
			screen->DrawQueuedText();
			completeScreen.Draw();
			break;
			
		case GameState::GS_LEVEL:
			if(bgEffect) 
				bgEffect->Draw();
			DrawLevel();
			break;
			
		case GameState::GS_CREDITS:
			credits.Draw();
			break;
			
		case GameState::GS_HIGHSCORES:
			if(bgEffect) 
				bgEffect->Draw();
			highScores->Draw();
			break;
			
		case GameState::GS_ENTER_NAME:
			if(bgEffect) 
				bgEffect->Draw();
			gui.Draw();
			break;
			
		case GameState::GS_LOAD_LEVEL:
			screen->QueueText(systemFont, "Loading...", Color::White, Font::ALIGN_BOTTOM_RIGHT, Point2f(-0.05f, 0.05f));
			break;
	};
}

//--------------------------------------------------------------------
void BrickOutGame::LoadLevel(const char *levelName)
{
	requestedLevelName = levelName;
	loadLevelRequested = true;
	state.RequestStateChange(GameState::GS_LOAD_LEVEL);
}

//--------------------------------------------------------------------
void BrickOutGame::LoadLevelList(const char *file, std::vector<std::string> &levelFiles)
{
	std::string filePath = ConcatPath(DataPaths::GetPath("Levels"), CreatePath(file));
	std::string path = GetPathPart(filePath);
	
	std::fstream f( filePath.c_str() );
	if( f.is_open() )
	{
		std::string line;
		while(getline(f, line))
		{
			std::string lvl = ConcatPath(path, Trim(line));
			if(FileExists(lvl.c_str()))
				levelFiles.push_back(lvl.c_str());
			else
				ConWarning() << "Could not find level file '" << lvl << "'\n";
		}
	}
	else
	{
		ConError() << "Cannot open level list file '" << path << "'\n";
	}
}

//--------------------------------------------------------------------
void BrickOutGame::DrawLevel()
{
	level->Draw();
	
	scoreBoard->Draw();
	
	level->DrawGui();
	
	if(IsPaused())
	{
		if(state.Get() != GameState::GS_MAIN_MENU_OVERLAY && state.Get() != GameState::GS_START_NEW_GAME)
			DrawFullscreenQuad( Color4f(0.0f, 0.0f, 0.0f, 0.8f) );
		screen->QueueText(extraLargeFont, "PAUSED", Color::Red, Font::ALIGN_CENTER, Point2f(-0.01f, -0.01f));
	}
}

//--------------------------------------------------------------------
void BrickOutGame::OnStateChangeRequest(const GameState &gs, GameState::State requested)
{
	if(requested == GameState::GS_EXIT)
		musicPlayer.FadeOut(1.0f);
}

//--------------------------------------------------------------------
void BrickOutGame::OnForcedStateChange(const GameState &gs, GameState::State state)
{}

//--------------------------------------------------------------------
void BrickOutGame::OnPreStateTransition(const GameState &gs, GameState::State current, GameState::State next)
{
	if(current == GameState::GS_GAME_OVER && next == GameState::GS_MAIN_MENU)
		level->Destroy();
	
	if(next == GameState::GS_CREDITS)
		credits.Reset();
	
	if(next == GameState::GS_MAIN_MENU_OVERLAY)
		mainMenu->EnableInGameMode();
	
	if(next == GameState::GS_MAIN_MENU)
		mainMenu->DisableInGameMode();
	
	if(next == GameState::GS_ENTER_NAME)
	{
		nameEntryScreen->Activate();
		nameEntryScreen->SetVisibility(true);
		nameEntryScreen->Reset();
		
		mainMenu->Deactivate();
		mainMenu->SetVisibility(false);
	}
}

//--------------------------------------------------------------------
void BrickOutGame::OnPostStateTransition(const GameState &gs, GameState::State current, GameState::State last)
{
	if(last == GameState::GS_START_NEW_GAME || last == GameState::GS_START_GAME)
	{
		level->Destroy();
		UnPause();
		scoreBoard->SetScore(0);
		scoreBoard->SetBallCount(3);
		currLevel = 0;
		loadLevelRequested = false;
	}
	
	if(last == GameState::GS_HIGHSCORES)
		highScores->ResetHighlighting();
	
	if(last == GameState::GS_ENTER_NAME)
	{
		nameEntryScreen->Deactivate();
		nameEntryScreen->SetVisibility(false);
		
		mainMenu->Activate();
		mainMenu->SetVisibility(true);
	}
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdLevel(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	if(args.size() < 2)
	{
		con->Print() << "Usage : level [levelname]\n";
		con->Print() << "        levelname - the name of the level to load.\n";
	}
	else
	{
		BrickOutGame *g = static_cast<BrickOutGame*>(userData);	
		g->LoadLevel( ConcatPath(DataPaths::GetPath("Levels"), CreatePath(args[1])).c_str() );
	}
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdClearHighScores(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	BrickOutGame *g = static_cast<BrickOutGame*>(userData);	
	g->ClearHighScores();
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdMpNext(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	assert(userData);

	BrickOutGame *app = (BrickOutGame*)userData;
	app->GetMusicPlayer().Next();
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdMpPrevious(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	assert(userData);

	BrickOutGame *app = (BrickOutGame*)userData;
	app->GetMusicPlayer().Previous();
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdMpPause(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	assert(userData);

	BrickOutGame *app = (BrickOutGame*)userData;
	app->GetMusicPlayer().Pause();
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdMpPlay(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	assert(userData);

	BrickOutGame *app = (BrickOutGame*)userData;
	app->GetMusicPlayer().Play();
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdMpStop(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	assert(userData);

	BrickOutGame *app = (BrickOutGame*)userData;
	app->GetMusicPlayer().Stop();
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdMpVolume(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	assert(userData);
	
	BrickOutGame *app = (BrickOutGame*)userData;
	MusicPlayer &mp = app->GetMusicPlayer();
	
	if(args.size() <= 1)
	{
		con->Print(Color::White) << Console::SetColor(Color::Yellow) << "Usage: " << Console::SetColor(Color::Green) << "mp_volume " 
								 << Console::RevertColor() << "[" << Console::SetColor(0xFFD600FF) << "volume" << Console::RevertColor() << "]\n";
		con->Print(Color::White) << Console::SetColor(Color::Yellow) << "       volume" << Console::RevertColor() << " - The volume in the range [0,255].\n";
	}
	else
	{
		int n;
		if(IsInteger(args[1], &n))
		{
			n = Clamp(n, (int)std::numeric_limits<byte>::min(), (int)std::numeric_limits<byte>::max());
			byte b = (byte)n;
			mp.SetVolume(b);
		}
	}
	
	con->Print(Color::White) << "Volume = " << Console::SetColor(Color::Yellow) << mp.GetVolume() << "\n";
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdMpInfo(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	assert(userData);
	
	BrickOutGame *app = (BrickOutGame*)userData;
	MusicPlayer &mp = app->GetMusicPlayer();
	
	Playlist::ConstIterator curr = mp.GetCurrentSongIterator();
	
	if( curr.IsValid() )
	{
		con->Print(Color::White) << "Current Song : " << Console::SetColor(Color::Yellow) << curr->name << "\n";
		con->Print(Color::White) << "      Artist : " << Console::SetColor(Color::Yellow) << curr->artist << "\n";
	}
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdMpRepeat(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	assert(userData);
	
	BrickOutGame *app = (BrickOutGame*)userData;
	MusicPlayer &mp = app->GetMusicPlayer();
	
	if(args.size() <= 1)
	{
		con->Print(Color::White) << Console::SetColor(Color::Yellow) << "Usage: " << Console::SetColor(Color::Green) << "mp_repeat " 
								 << Console::RevertColor() << "[" << Console::SetColor(0xFFD600FF) << "on" << Console::RevertColor() 
								 << "|" << Console::SetColor(0xFFD600FF) << "off" << Console::RevertColor() << "|" << Console::SetColor(0xFFD600FF) 
								 << "single" << Console::RevertColor() << "]\n";
		con->Print(Color::White) << Console::SetColor(Color::Yellow) << "       on" << Console::RevertColor() << "     - Turns repeat mode on.\n";
		con->Print(Color::White) << Console::SetColor(Color::Yellow) << "       off" << Console::RevertColor() << "    - Turns repeat mode off.\n";
		con->Print(Color::White) << Console::SetColor(Color::Yellow) << "       single" << Console::RevertColor() << " - Turns on single songe repeat mode.\n";
	}
	else
	{
		std::string param = args[1];
		ToLower(param);
		
		if(param == "on")
			mp.SetRepeatMode(MusicPlayer::REPEAT_ON);
		else if(param == "off")
			mp.SetRepeatMode(MusicPlayer::REPEAT_OFF);
		else if(param == "single")
			mp.SetRepeatMode(MusicPlayer::REPEAT_SINGLE);
	}
	
	con->Print(Color::White) << "Repeat mode = ";
	MusicPlayer::RepeatMode mode = mp.GetRepeatMode();
	if(mode == MusicPlayer::REPEAT_OFF)
		con->Print(Color::Yellow) << "off\n";
	else if(mode == MusicPlayer::REPEAT_ON)
		con->Print(Color::Yellow) << "on\n";
	else
		con->Print(Color::Yellow) << "single\n";
}

//--------------------------------------------------------------------
void PrintConCmdMpLoadPlaylistUsage(Console *con)
{
	con->Print(Color::White) << Console::SetColor(Color::Yellow) << "Usage: " << Console::SetColor(Color::Green) << "mp_loadPlaylist " 
							 << Console::RevertColor() << "-[" << Console::SetColor(0xFFD600FF) << "r" << Console::RevertColor() 
							 << "|" << Console::SetColor(0xFFD600FF) << "a" << Console::RevertColor() << "|" << Console::SetColor(0xFFD600FF) 
							 << "m" << Console::RevertColor() << "]" << Console::SetColor(0xFFD600FF) << " path\n";
	con->Print(Color::White) << Console::SetColor(Color::Yellow) << "       r" << Console::RevertColor() << "    - The path is relative to the games root directory.\n";
	con->Print(Color::White) << Console::SetColor(Color::Yellow) << "       a" << Console::RevertColor() << "    - The path is an absolute path to the playlist.\n";
	con->Print(Color::White) << Console::SetColor(Color::Yellow) << "       m" << Console::RevertColor() << "    - The path is relative to the Music folder (default behavior).\n";
	con->Print(Color::White) << Console::SetColor(Color::Yellow) << "       path" << Console::RevertColor() << " - The path to the playlist file.\n";
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdMpLoadPlaylist(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	assert(userData);
	
	if(args.size() == 1)
	{
		PrintConCmdMpLoadPlaylistUsage(con);
		return;
	}
	
	// Build the proper path to the playlist and loads it.
	std::string path;
	std::string playlistFilePath;
	char pathType = 'm';

	if(args.size() > 1 && args[1].size() == 2 && args[1].at(0) == '-')
	{
		pathType = args[1].at(1);
		if(!Match(pathType, "ram"))
		{
			con->Print(Color::White) << "Unknown option '" << Console::SetColor(Color::Yellow) << args[1] << Console::RevertColor() << "'\n";
			PrintConCmdMpLoadPlaylistUsage(con);
			return;
		}

		// If we have a path type then we must also have a path for everything to work.
		if(args.size() < 2)
		{
			PrintConCmdMpLoadPlaylistUsage(con);
			return;
		}

		playlistFilePath = CreatePath(args[2]);
	}
	else
	{
		playlistFilePath = CreatePath(args[1]);
	}

	if(pathType == 'm')
	{
		path = ConcatPath(DataPaths::GetPath("Music"), playlistFilePath);
	}
	else if(pathType == 'r')
	{
		path = ConcatPath(DataPaths::GetPath("Root"), playlistFilePath);
	}
	else // pathType == 'a'
	{
		path = playlistFilePath;
	}

	Playlist playlist;
	if( !LoadPlaylistFromXMLFile(path, playlist) )
	{
		con->Print(Color::White) << "Error loading playlist '" << Console::SetColor(Color::Yellow) << path << Console::RevertColor() << "'\n";
		return;
	}

	BrickOutGame *app = (BrickOutGame*)userData;
	MusicPlayer &mp = app->GetMusicPlayer();

	mp.SetPlaylist(playlist);
}

//--------------------------------------------------------------------
void BrickOutGame::ConCmdMpPrintPlaylist(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args)
{
	assert(userData);
	
	BrickOutGame *app = (BrickOutGame*)userData;
	MusicPlayer &mp = app->GetMusicPlayer();
	
	Playlist list = mp.GetPlaylist();
	Playlist::ConstIterator it = list.Begin();
	
	for(uint count = 1; it != list.End(); it++, count++)
	{
		con->Print(Color::White) << count << ". " << Console::SetColor(Color::Yellow) << it->name << Console::RevertColor() 
			<< " by " << Console::SetColor(Color::Yellow) << it->artist << "\n";
	}
}
