/*!
 * \file BrickOutGame.h
 * \brief Declares the BrickOutGame class is the starting point for BrickOut |ReDuX| game.
 * \author Joe Forte
 * \date 3/18/2008
 */

#ifndef BRICKOUTGAME_H
#define BRICKOUTGAME_H

#include <map>
#include <string>
#include "Application/SDLApplication.h"
#include "GUI/GUIViewport.h"
#include "Sound/MusicPlayer.h"
#include "Game/GameSettings.h"
#include "Game/GameCompleteScreen.h"
#include "Game/TransitionEffect.h"
#include "Game/LevelParser.h"
#include "Game/Level.h"
#include "Game/Credits.h"
#include "Game/GameState.h"

class GameMouseEffect;
class GUIMainMenu;
class GUIScoreBoard;
class HighScores;
class GameSplashScreen;
class BackgroundEffect;
class GUINameEntryScreen;

const std::string GAME_NAME = "BrickOut |ReDuX|";
const int DEFAULT_WIDTH = 800;
const int DEFAULT_HEIGHT = 600;
const bool DEFAULT_FULLSCREEN = false;

const std::string ABOUT_TEXT = "BrickOut |ReDuX|\n\n"
                               "Art & Programming: Joe Forte\n"
							   "           E-Mail: Joe@JoeForte.net\n"
                               "             Date: June 23, 2008\n\n"
							   "    BrickOut |ReDuX| is a remake of the\n"
                               "classic BrickOut game of yore. Based on\n" 
                               "the simple concept of breaking bricks\n"
                               "with a ball and a paddle. BrickOut\n"
                               "|ReDuX| offers 7 levels of classic brick\n"
							   "breaking action.\n\n"
							   "Enjoy!!";

const std::string GAME_CONFIG_FILE = "Game.cfg";
const std::string GAME_CREDITS_FILE = "Credits.xml";

/*!
 * \class BrickOutGame
 * \headerfile BrickOutGame.h "Game/BrickOutGame.h"
 * \brief Represents the BrickOut |ReDuX| Game.
 */
class BrickOutGame : public SDLApplication, LevelListener, protected TransitionEffectListener, GameStateListener
{
public:

	/*!
	 * \brief Constructor.
	 *
	 * \param[in] w the initial width of the application screen. Defaults to DEFAULT_WIDTH.
	 * \param[in] h the initial height of the application screen. Defaults to DEFAULT_HEIGHT.
	 * \param[in] useFullscreen set to true to start the application fullscreen. Defaults to false.
	 */
	BrickOutGame(int w = DEFAULT_WIDTH, int h = DEFAULT_HEIGHT, bool useFullscreen = false);
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~BrickOutGame();

	/*!
	 * \brief Returns the image name of the application icon.
	 *
	 * \return The file name of the icon or NULL if there is no icon.
	 */
	virtual const char *GetIconName();
	
	/*!
	 * \brief Pauses the application if it is unpaused or unpauses the application if it is paused. 
	 */
	virtual void TogglePause();
	
	/*!
	 * \brief Pauses the application. 
	 */
	virtual void Pause();
	
	/*!
	 * \brief Unpauses the application. 
	 */
	virtual void UnPause();
	
	/*!
	 * \brief Returns if the application is paused or not.
	 *
	 * \return True if the application is paused or false if the application is unpaused.
	 */
	virtual bool IsPaused() const;
	
	/*!
	 * \brief Loads the given level.
	 *
	 * \param[in] levelName the name of the level to load.
	 */
	void LoadLevel(const char *levelName);
	
	/*!
	 * \brief Clears all hich scores.
	 */
	void ClearHighScores();

	/*!
	 * \brief Called when a ball is lost.
	 *
	 * \param[in] level referance to the appropriate level.
	 *
	 * \return True if you wish to have another ball spawned.  Return false to not have another ball spawned.
	 */
	virtual bool BallLost(Level &level);
	
	/*!
	 * \brief Called when an extra ball is aquired.
	 *
	 * \param[in] level referance to the level that called the method.
	 */
	virtual void ExtraBall(Level &level);

	/*!
	 * \brief Called when the user earns some points.
	 *
	 * \param[in] level referance to the appropriate level.
	 * \param[in] amt the score that the user aquired.
	 */
	virtual void ScoreIncrease(Level &level, int amt);

	/*!
	 * \brief Gets the applications name.
	 *
	 * \return The name of the application.
	 */
	virtual std::string GetAppName() { return GAME_NAME; }

	/*!
	 * \brief Returns the games music player.
	 */
	MusicPlayer &GetMusicPlayer() { return musicPlayer; }
	
protected:
	
	/*!
	 * \brief Called when a state change request is made.
	 */
	virtual void OnStateChangeRequest(const GameState &gs, GameState::State requested);
	
	/*!
	 * \brief Called when a state change is forcefully made.
	 */
	virtual void OnForcedStateChange(const GameState &gs, GameState::State state);
	
	/*!
	 * \brief Called before a state transition is made.
	 */
	virtual void OnPreStateTransition(const GameState &gs, GameState::State current, GameState::State next);
	
	/*!
	 * \brief Called after a state transition is made.
	 */
	virtual void OnPostStateTransition(const GameState &gs, GameState::State current, GameState::State last);

protected:

	/*!
	 * \brief Called when a transition out is completed.
	 *
	 * \param[in] effect the transition effect.
	 * \param[in] trans the transition that completed.
	 */
	virtual void OnTransitionOutCompleted(TransitionEffect &effect, Transition &trans);

	/*!
	 * \brief Called when a transition in is completed.
	 *
	 * \param[in] effect the transition effect.
	 * \param[in] trans the transition that completed.
	 */
	virtual void OnTransitionInCompleted(TransitionEffect &effect, Transition &trans);

	/*!
	 * \brief Called when a transition delay is completed.
	 *
	 * \param[in] effect the transition effect.
	 * \param[in] trans the transition that completed.
	 */
	virtual void OnTransitionDelayCompleted(TransitionEffect &effect, Transition &trans);

	/*!
	 * \brief Called when the transition effect is completed.
	 *
	 * \param[in] effect the transition effect.
	 */
	virtual void OnTransitionEffectCompleted(TransitionEffect &effect);

private:

	/*!
	 * \brief Initializes any subsystems needed by the application.
	 * 
	 * \return True if the initialization was successful, false otherwise.
	 */
	virtual bool InitSubSystems();

	/*!
	 * \brief Initializes the application.  All subsystems (including the global environment) is valid
	 * when this is called.
	 * 
	 * \return True if the initialization was successful, false otherwise.
	 */
	virtual bool ApplicationInit();

	/*!
	 * \brief Called before the start of the frame.
	 *
	 * NOTE: The environment is valid for the current frame.
	 */
	virtual void PreFrame();

	/*!
	 * \brief Called aftern the frame is complete.
	 *
	 * NOTE: The environment is valid for the current frame but not the next.
	 */
	virtual void PostFrame();

	/*!
	 * \brief Update method called once each frame.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 *
	 * \return True if the application should continue.  Returns false if the application should exit.
	 */
	virtual bool Update(float dt);
	
	/*!
	 * \brief Draw method called once each frame.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 */
	virtual void Draw(float dt);

	/*!
	 * \brief Called when the application receives a message to quit from the OS or a call to Exit().
	 *
	 * \return True if the application should exit at the end of the frame.  Returns false if the application should ignore the quit message.
	 */
	virtual bool OnApplicationExitRequest();

	/*!
	 * \brief Called when the application looses focus.
	 */
	virtual void OnLostFocus();

	/*!
	 * \brief Called when the application gains focus.
	 */
	virtual void OnGainedFocus();

	/*!
	 * \brief Called when the application mouse focus.
	 */
	virtual void OnLostMouseFocus();

	/*!
	 * \brief Called when the application gains mouse focus.
	 */
	virtual void OnGainedMouseFocus();

	/*!
	 * \brief Return true if we should NOT do a transition effect between current and next.
	 */
	bool ShouldNotDoTransitionEffect(GameState::State current, GameState::State next);

	/*!
	 * \brief Setup the transition effect for the given state transition.
	 */
	void SetupTransitionEffect(TransitionEffect &effect, GameState::State current, GameState::State next);

	void UpdateGameState(float dt);
	
	bool GameComplete() const;
	
	// Called when the screen is fadding in/out
	void BackgroundUpdateGameState(float dt);

	void DrawFrame();
	void DrawLevel();
	
	void LoadLevelList(const char *file, std::vector<std::string> &levelFiles);
	
	// Console commands
	static void ConCmdLevel(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdClearHighScores(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);

	// Music player commands
	static void ConCmdMpNext(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdMpPrevious(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdMpPause(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdMpPlay(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdMpStop(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdMpVolume(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdMpInfo(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdMpRepeat(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdMpLoadPlaylist(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);
	static void ConCmdMpPrintPlaylist(Console *con, const char *cmd, void *userData, const std::vector<std::string> &args);

private:
	
	GUIViewport gui;
	GUIMainMenu *mainMenu;
	GUINameEntryScreen *nameEntryScreen;
	GameSplashScreen *splashScreen;
	
	GameCompleteScreen completeScreen;
	
	GUIScoreBoard *scoreBoard;
	HighScores *highScores;

	MusicPlayer musicPlayer;
	
	// Game Data
	bool paused;
	Level *level;
	GameState state;

	TransitionEffect transEffect;
	
	GameSettings gameSettings;

	GameMouseEffect *mouseEffect;
	
	Credits credits;

	LevelParser parser;	
	std::vector<std::string> levelFiles;
	uint currLevel;

	BackgroundEffect *bgEffect;
	
	bool loadLevelRequested; // set to true when a level is loaded with a call to LoadLevel.
	std::string requestedLevelName;
	
	FontRef extraLargeFont;
};

#endif
