/*!
 * \file Credits.h
 * \brief Implements the Credits class.
 * \author Joe Forte
 * \date 2/9/2009
 */

#include "Common/Globals.h"
#include "Utility/FormattedText.h"
#include "Game/Credits.h"

#include "tinyxml.h"

// Xml tag names
const std::string CREDITS_TAG_STR = "Credits";
const std::string SECTION_TAG_STR = "Section";
const std::string ELEMENT_TAG_STR = "Element";
const std::string IMAGEELEMENT_TAG_STR = "ImageElement";
const std::string ENDIMAGE_TAG_STR = "EndImage";

//--------------------------------------------------------------------
CreditSection::CreditSection(const char *name)
{
	header.SetText(name);
}

//--------------------------------------------------------------------
void CreditSection::AddElement(const char *text, FontRef f, const Color4f &c)
{
	AddElement(text, f, c, 0.0f, 0.0f);
}

//--------------------------------------------------------------------
void CreditSection::AddElement(const char *text, FontRef f, const Color4f &c, float t, float b)
{
	elements.push_back( Element() );
	Element &elm = elements.back();

	elm.SetText(text);
	elm.SetFont(f);
	elm.SetColor(c);
	elm.SetPadding(t, b);
}

//--------------------------------------------------------------------
void CreditSection::AddImageElement(const Sprite &image, 
									const char *text, 
									FontRef f, 
									const Color4f &c, 
									float t, 
									float b)
{
	AddImageElement(image, text, f, c, t, b, 0.0f);
}

//--------------------------------------------------------------------
void CreditSection::AddImageElement(const Sprite &image, 
									const char *text, 
									FontRef f, 
									const Color4f &c, 
									float t, 
									float b,
									float spacing)
{
	elements.push_back( Element() );
	Element &elm = elements.back();

	elm.SetText(text);
	elm.SetFont(f);
	elm.SetColor(c);
	elm.SetImageTextPadding(spacing);
	elm.SetPadding(t, b);
	elm.SetImage(image);
}

//--------------------------------------------------------------------
void CreditSection::GetDimensions(float &w, float &h) const
{	
	w = header.GetWidth();
	h = header.GetHeight();
	
	for(std::vector<Element>::const_iterator i = elements.begin(); i != elements.end(); ++i)
	{
		w = std::max(w, i->GetWidth());
		h = h + i->GetHeight();
	}
}

//--------------------------------------------------------------------
void CreditSection::Draw(const Point2f &p) const
{
	float w, h;
	GetDimensions(w, h);
	
	float offset = w / 2.0f;
	
	Point2f pos = p;
	if(!header.IsEmpty())
	{
		header.Draw(pos + Point2f(offset - (header.GetWidth() / 2.0f), 0.0f));
		pos.y -= header.GetHeight();
	}
	
	for(std::vector<Element>::const_iterator i = elements.begin(); i != elements.end(); ++i)
	{
		if(!i->IsEmpty())
		{
			i->Draw(pos + Point2f(offset - (i->GetWidth() / 2.0f), 0.0f));
			pos.y -= i->GetHeight();
		}
	}
}

//--------------------------------------------------------------------
CreditSection::Element::Element()
:
color(Color::White),
imageTextPadding(0.0f),
width(0.0f),
height(0.0f),
paddingTop(0.0f),
paddingBottom(0.0f)
{}

//--------------------------------------------------------------------
void CreditSection::Element::SetText(const char *str)
{
	text.clear();	
	FormatText(str, font, text);

	ComputeDimensions();
}

//--------------------------------------------------------------------
void CreditSection::Element::SetColor(const Color4f &c)
{
	color = c;
}

//--------------------------------------------------------------------
void CreditSection::Element::SetFont(FontRef f)
{
	font = f;
	ComputeDimensions();
}

//--------------------------------------------------------------------
void CreditSection::Element::SetImage(const Sprite &im)
{
	image = im;
	ComputeDimensions();
}

//--------------------------------------------------------------------
void CreditSection::Element::SetImageTextPadding(float p) 
{ 
	imageTextPadding = p;
	ComputeDimensions();
}

//--------------------------------------------------------------------
void CreditSection::Element::SetPadding(float top, float bottom) 
{ 
	paddingTop = top;
	paddingBottom = bottom;
	ComputeDimensions();
}

//--------------------------------------------------------------------
void CreditSection::Element::Draw(const Point2f &p) const
{
	if(font.IsValid())
	{
		Point2f pos = p - Point2f(0.0f, paddingTop);
		if(image.CanDraw())
		{
			// Centers the image vertically next to the text
			Point2f drawPos = pos - Point2f(0.0f, (height - image.GetHeight()) / 2.0f);
			image.Draw(drawPos, Image::TOPLEFT);
			pos.x += image.GetWidth() + imageTextPadding;
		}

		for(std::vector<std::string>::const_iterator i = text.begin(); i != text.end(); ++i)
		{
			font->DrawTextStr(i->c_str(), color, pos);
			pos.y -= font->LineHeight();
		}
	}
}


//--------------------------------------------------------------------
void CreditSection::Element::GetDimensions(float &w, float &h) const
{
	w = width;
	h = height;
}

//--------------------------------------------------------------------
void CreditSection::Element::ComputeDimensions()
{
	if(font.IsValid())
	{
		width = 0.0f;
		height = 0.0f;
		
		for(std::vector<std::string>::const_iterator i = text.begin(); i != text.end(); ++i)
		{
			float w = font->StringWidth(i->c_str());

			width = std::max(width, w);
			height = height + font->LineHeight();
		}

		if(image.CanDraw())
		{
			width += image.GetWidth() + imageTextPadding;
			height = std::max(height, image.GetHeight());
		}

		height += paddingTop + paddingBottom;
	}
}

//--------------------------------------------------------------------
Credits::Credits() : scrollSpeed(50.0f), delay(3.0f), delayTimer(0.0f), complete(false) {}

//--------------------------------------------------------------------
void Credits::ResetSection(Section &sec)
{
	sec.section.GetDimensions(sec.w, sec.h);
	sec.active = true;
}

//--------------------------------------------------------------------
void Credits::Reset()
{
	complete = false;
	delayTimer = 0.0f;

	ResetSection(title);
	title.position = drawArea.CenterPosition() + Point2f(-title.w / 2.0f, title.h / 2.0f);
		
	Point2f pos = drawArea.CenterPosition() - Point2f(0.0f, drawArea.Height() / 2.0f);
	for(std::vector<Section>::iterator i = sections.begin(); i != sections.end(); ++i)
	{
		ResetSection(*i);
		i->position = pos - Point2f(i->w / 2.0f, 0.0f);

		pos.y -= i->h;
	}

	if(endImage.CanDraw())
	{
		endImage.SetPosition(pos - Point2f(0.0f, (endImage.GetHeight() + drawArea.Height()) / 2.0f));
	}
}

//--------------------------------------------------------------------
void Credits::SetArea(const Rectf &area)
{
	drawArea = RectToAABB(area);
}

//--------------------------------------------------------------------
bool Credits::IsComplete() const
{
	return complete;
}

//--------------------------------------------------------------------
void Credits::Update(float dt)
{
	Point2f amount(0.0f, scrollSpeed * dt);

	if(title.active)
	{
		title.position += amount;
		if((title.position.y - title.h) > drawArea.max.y)
			title.active = false;
	}

	for(std::vector<Section>::iterator i = sections.begin(); i != sections.end(); ++i)
	{
		if(i->active)
		{
			i->position += amount;
			if((i->position.y - i->h) > drawArea.max.y)
				i->active = false;
		}
	}

	if(endImage.CanDraw() && endImage.Position().y < 0.0f)
	{
		endImage.Move(amount);
		if(endImage.Position().y > 0.0f)
			endImage.SetPosition( Point2f(endImage.Position().x, 0.0f) );
	}

	// Updates the delay timer and checks if the credits are done.
	if(endImage.CanDraw())
	{
		if(endImage.Position().y >= 0.0f)
			delayTimer = std::min(delayTimer + dt, delay);
	}
	else // No end image is being used
	{
		Section &end = title;
		if(sections.size() > 0)
			end = sections.back();

		if(!end.active)
			delayTimer = std::min(delayTimer + dt, delay);
	}

	if(delayTimer >= delay)
		complete = true;
}

//--------------------------------------------------------------------
void Credits::Draw() const
{
	if(title.active && drawArea.Overlap(title.GetBounds()))
	{
		title.section.Draw(title.position);
	}

	for(std::vector<Section>::const_iterator i = sections.begin(); i != sections.end(); ++i)
	{
		if(i->active && drawArea.Overlap(i->GetBounds()))
		{
			i->section.Draw(i->position);
		}
	}

	if(endImage.CanDraw() && drawArea.Overlap(endImage.GetBBox()))
		endImage.Draw();
}

//--------------------------------------------------------------------
bool Credits::Load(const char *fileName)
{
	return LoadCreditsXMLFile(fileName);
}

//--------------------------------------------------------------------
bool Credits::LoadCreditsXMLFile(const char *file)
{
	TiXmlDocument doc(file);
	if( !doc.LoadFile() )
	{
		ConError() << "Credits file '" << file << "' could not be parsed.\n";
		return false;
	}
	
	TiXmlHandle hDoc(&doc);
	
	TiXmlNode *Root = hDoc.FirstChildElement().Node();
	if( !Root || 
		Root->ValueStr() != CREDITS_TAG_STR || 
		!ParseXMLCreditsTag(Root->ToElement()) )
	{
		return false;
	}

	TiXmlNode *curr = Root->FirstChild();
	while(curr)
	{
		if(curr->Type() == TiXmlNode::ELEMENT)
		{
			TiXmlElement *elm = curr->ToElement();
			
			if(elm->ValueStr() == SECTION_TAG_STR)
			{
				if( !ParseXMLSectionTag(elm) )
					return false;
			}
			else if(elm->Value() == ENDIMAGE_TAG_STR)
			{
				if( !ParseXMLEndImageTag(elm) )
					return false;
			}
			else if(elm->ValueStr() == CREDITS_TAG_STR)
			{
				ConWarning() << "The tag" << CREDITS_TAG_STR << "' is incorrectly placed in the credits file '" << file << "'.\n";
			}
			else if(elm->ValueStr() == ELEMENT_TAG_STR)
			{
				ConWarning() << "The tag" << ELEMENT_TAG_STR << "' is incorrectly placed in the credits file '" << file << "'.\n";
			}
			else
			{
				ConWarning() << "Unknown XML tag '" << elm->ValueStr() << "' in the credits file '" << file << "'.\n";
			}
		}
		
		curr = Root->IterateChildren(curr);
	}
	
	return true;
}

//--------------------------------------------------------------------
bool Credits::ParseXMLCreditsTag(TiXmlElement *elm)
{
	TagAttributes attribs;
	if( !ParseXMLTagAttributes(elm, attribs) )
		return false;
	
	ApplySectionTagAttributes(title.section, attribs);

	float speed;
	if( elm->QueryFloatAttribute("scrollSpeed", &speed) != TIXML_NO_ATTRIBUTE )
		scrollSpeed = speed;

	float endDelay;
	if( elm->QueryFloatAttribute("endDelay", &endDelay) != TIXML_NO_ATTRIBUTE )
		delay = endDelay;

	return true;
}

//--------------------------------------------------------------------
bool Credits::ParseXMLSectionTag(TiXmlElement *elm)
{
	sections.push_back( Section() );
	Section &sec = sections.back();
	
	TagAttributes attribs;
	if( !ParseXMLTagAttributes(elm, attribs) )
		return false;

	ApplySectionTagAttributes(sec.section, attribs);
	
	TiXmlNode *curr = elm->FirstChild();
	while( curr )
	{
		if( curr->Type() == TiXmlNode::ELEMENT )
		{
			TiXmlElement *e = curr->ToElement();
			if(e->Value() == ELEMENT_TAG_STR)
			{
				if( !ParseXMLElementTag(e, sec.section) )
					return false;
			}
			else if(e->Value() == IMAGEELEMENT_TAG_STR)
			{
				if( !ParseXMLImageElementTag(e, sec.section) )
					return false;
			}
			else if(e->Value() == CREDITS_TAG_STR)
			{
				ConWarning() << "The tag" << CREDITS_TAG_STR << "' is incorrectly placed under '" << SECTION_TAG_STR << "'.\n";
			}
			else if(e->Value() == SECTION_TAG_STR)
			{
				ConWarning() << "The tag" << SECTION_TAG_STR << "' is incorrectly placed under '" << SECTION_TAG_STR << "'.\n";
			}
			else if(e->Value() == ENDIMAGE_TAG_STR)
			{
				ConWarning() << "The tag" << ENDIMAGE_TAG_STR << "' is incorrectly placed under '" << SECTION_TAG_STR << "'.\n";
			}
			else
			{
				ConWarning() << "Unknown XML tag '" << elm->ValueStr() << "' under '" << SECTION_TAG_STR << "'.\n";
			}
		}
		
		curr = elm->IterateChildren(curr);
	}
	
	return true;
}

//--------------------------------------------------------------------
bool Credits::ParseXMLEndImageTag(TiXmlElement *elm)
{
	const char *str = elm->Attribute("fileName");
	if(str)
		endImage.Load(str);

	if(endImage.CanDraw())
	{
		float scale;
		if( elm->QueryFloatAttribute("scale", &scale) != TIXML_NO_ATTRIBUTE )
			endImage.SetScale(scale);
	}

	return true;
}

//--------------------------------------------------------------------
bool Credits::ParseXMLElementTag(TiXmlElement *elm, CreditSection &sec)
{
	TagAttributes attribs;
	if( !ParseXMLTagAttributes(elm, attribs) )
		return false;

	std::string text;
	if(!ParseXMLElementText(elm, text))
		return false;

	sec.AddElement(text.c_str(), 
				   LoadFontResource(attribs.font.c_str(), attribs.fontSize), 
				   attribs.color,
				   attribs.paddingTop,
				   attribs.paddingBottom); 

	return true;
}

//--------------------------------------------------------------------
bool Credits::ParseXMLImageElementTag(TiXmlElement *elm, CreditSection &sec)
{
	TagAttributes attribs;
	if( !ParseXMLTagAttributes(elm, attribs) )
		return false;

	Sprite im;

	const char *str = elm->Attribute("image");
	if(str)
		im.Load(str);

	if(im.CanDraw())
	{
		float scale;
		if( elm->QueryFloatAttribute("scale", &scale) != TIXML_NO_ATTRIBUTE )
			im.SetScale(scale);
	}

	float spacing;
	if( elm->QueryFloatAttribute("spacing", &spacing) == TIXML_NO_ATTRIBUTE )
		spacing = 0.0f;

	std::string text;
	if(!ParseXMLElementText(elm, text))
		return false;

	sec.AddImageElement(im,
						text.c_str(), 
					    LoadFontResource(attribs.font.c_str(), attribs.fontSize), 
						attribs.color,
						attribs.paddingTop,
						attribs.paddingBottom,
						spacing);

	return true;
}

//--------------------------------------------------------------------
bool Credits::ParseXMLElementText(TiXmlElement *elm, std::string &out)
{
	TiXmlNode *curr = elm->FirstChild();
	while( curr )
	{
		if( curr->Type() == TiXmlNode::ELEMENT )
		{
			TiXmlElement *e = curr->ToElement();
			if(e->ValueStr() == "Line")
			{
				out += e->GetText();
				out += "\n";
			}
		}
		
		curr = elm->IterateChildren(curr);
	}
	
	return true;
}

//--------------------------------------------------------------------
bool Credits::ParseXMLTagAttributes(TiXmlElement *elm, TagAttributes &attribs)
{
	const char *str = elm->Attribute("title");
	if(str)
		attribs.title = str;

	const char *font = str = elm->Attribute("font");
	if(font)
		attribs.font = str;
	else
		attribs.font = Font::DEFAULT_FONT_PATH;
	
	if( elm->QueryIntAttribute("fontSize", &attribs.fontSize) == TIXML_NO_ATTRIBUTE )
		attribs.fontSize = Font::DEFAULT_FONT_SIZE;

	if( elm->QueryFloatAttribute("paddingTop", &attribs.paddingTop) == TIXML_NO_ATTRIBUTE )
		attribs.paddingTop = 0.0f;

	if( elm->QueryFloatAttribute("paddingBottom", &attribs.paddingBottom) == TIXML_NO_ATTRIBUTE )
		attribs.paddingBottom = 0.0f;

	const char *col = elm->Attribute("color");
	if(col)
	{
		std::stringstream stream(col);
		float r, g, b;

		if((stream >> r >> g >> b).fail())
		{
			attribs.color = Color::White;
			ConWarning() << "Improperly formatted color attribute.\n";
		}
		else
		{
			attribs.color = Color4f(r, g, b);
		}

	}
	else
	{
		attribs.color = Color::White;
	}

	return true;
}

//--------------------------------------------------------------------
void Credits::ApplySectionTagAttributes(CreditSection &sec, TagAttributes &attribs)
{
	sec.SetTitle(attribs.title.c_str());
	sec.SetTitleFont( LoadFontResource(attribs.font.c_str(), attribs.fontSize) );
	sec.SetTitleColor(attribs.color);
	sec.SetTitlePadding(attribs.paddingTop, attribs.paddingBottom);
}
