/*!
 * \file Credits.h
 * \brief Defines the Credits class for displaying end game credits.
 * \author Joe Forte
 * \date 2/9/2009
 */

#ifndef CREDITS_H
#define CREDITS_H

#include <string>
#include <vector>
#include "Graphics/Font.h"
#include "Graphics/Sprite.h"
#include "Graphics/Color.h"
#include "Math/MathCommon.h"

class TiXmlElement; // TinyXml class used for parsing credits xml files

/*!
 * \class CreditSection
 * \headerfile Credits.h "Game/Credits.h"
 * \brief Holds the data for one section of the credits.
 */
class CreditSection
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	CreditSection() {}
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] name the name of the section.
	 */
	CreditSection(const char *name);
	
	/*!
	 * \brief Sets the title of the section.
	 *
	 * \param[in] title the section title.
	 */
	void SetTitle(const char *title) { header.SetText(title); }
	
	/*!
	 * \brief Sets the fonts to use for rendering the title.
	 *
	 * \param[in] font the font to use for the header text.
	 */
	void SetTitleFont(FontRef font) { header.SetFont(font); }

	/*!
	 * \brief Sets the color to use for rendering the title.
	 *
	 * \param[in] color the color to use for the header text.
	 */
	void SetTitleColor(const Color4f &color) { header.SetColor(color); }

	/*!
	 * \brief Sets the padding to use for rendering the title.
	 *
	 * \param[in] top the top padding.
	 * \param[in] bottom the bottom padding.
	 */
	void SetTitlePadding(float top, float bottom) { header.SetPadding(top, bottom); }
	
	/*!
	 * \brief Adds the given text to the section.
	 *
	 * \param[in] text the text to add.
	 * \param[in] f the font.
	 * \param[in] c the color.
	 */
	void AddElement(const char *text, FontRef f, const Color4f &c);

	/*!
	 * \brief Adds the given text to the section.
	 *
	 * \param[in] text the text to add.
	 * \param[in] f the font.
	 * \param[in] c the color.
	 * \param[in] t top padding.
	 * \param[in] b bottom padding.
	 */
	void AddElement(const char *text, FontRef f, const Color4f &c, float t, float b);

	/*!
	 * \brief Adds text with an associated image.
	 *
	 * \param[in] image the image to display with the text.
	 * \param[in] text the text to add.
	 * \param[in] f the font.
	 * \param[in] c the color.
	 * \param[in] t top padding.
	 * \param[in] b bottom padding.
	 */
	void AddImageElement(const Sprite &image, 
						 const char *text,
						 FontRef f, 
						 const Color4f &c, 
						 float t, 
						 float b);

	/*!
	 * \brief Adds text with an associated image.
	 *
	 * \param[in] image the image to display with the text.
	 * \param[in] text the text to add.
	 * \param[in] f the font.
	 * \param[in] c the color.
	 * \param[in] t top padding.
	 * \param[in] b bottom padding.
	 * \param[in] spacing the spacing between the image and the text.
	 */
	void AddImageElement(const Sprite &image, 
						 const char *text,
						 FontRef f, 
						 const Color4f &c, 
						 float t, 
						 float b,
						 float spacing);

	/*!
	 * \brief Gets the size of the section.
	 *
	 * \param[out] w the width.
	 * \param[out] h the height.
	 */
	void GetDimensions(float &w, float &h) const;
	
	/*!
	 * \brief Draws the section at the given location.
	 *
	 * \param[in] p the top left position of the area to draw to.
	 */
	void Draw(const Point2f &p) const;
		
protected:
	
	/*!
	 * \class Element
	 * \headerfile Credits.h "Game/Credits.h"
	 * \brief Represents a single element in a CreditSection.
	 */
	class Element
	{
	public:
		
		/*!
		 * \brief Constructor.
		 */
		Element();
		
		/*!
		 * \brief Sets the text of the element.
		 *
		 * \param[in] str the text.
		 */
		void SetText(const char *str);
		
		/*!
		 * \brief Sets the color of the text.
		 *
		 * \param[in] c the color.
		 */
		void SetColor(const Color4f &c);
		
		/*!
		 * \brief Sets the fonts to use for rendering.
		 *
		 * \param[in] f the font.
		 */
		void SetFont(FontRef f);

		/*!
		 * \brief Sets the image to draw with the text.
		 *
		 * \param[in] image the image.
		 */
		void SetImage(const Sprite &im);

		/*!
		 * \brief Sets the top and bottom padding for the section.
		 *
		 * \param[in] top the top padding.
		 * \param[in] bottom the bottom padding.
		 */
		void SetPadding(float top, float bottom);

		/*!
		 * \brief Sets padding between the image and the text.
		 *
		 * \param[in] p the padding.
		 */
		void SetImageTextPadding(float p);

		/*!
		 * \brief Gets the size of the element.
		 *
		 * \param[out] w the width.
		 * \param[out] h the height.
		 */
		void GetDimensions(float &w, float &h) const;
		
		/*!
		 * \brief Gets the width of the element.
		 *
		 * \return the width.
		 */
		float GetWidth() const { return width; }
		
		/*!
		 * \brief Gets the height of the element.
		 *
		 * \return the height.
		 */
		float GetHeight() const { return height; }
		
		/*!
		 * \brief Draws the section at the given location.
		 *
		 * \param[in] p the top left position of the area to draw to.
		 */
		void Draw(const Point2f &p) const;
		
		/*!
		 * \brief Determines if the element contains any data that can be drawn.
		 *
		 * \return true if the element contains no drawable data, false otherwise.
		 */
		bool IsEmpty() const { return text.size() == 0; }
		
	private:
		
		void ComputeDimensions();
		
		std::vector<std::string> text;
		
		FontRef font;
		Color4f color;
		
		Sprite image;
		float imageTextPadding;
		
		float width;
		float height;

		float paddingTop;
		float paddingBottom;
	};
	
	Element header;

	std::vector<Element> elements; 
};

/*!
 * \class Credits
 * \headerfile Credits.h "Game/Credits.h"
 * \brief Displays scrolling game credits.
 */
class Credits
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	Credits();
	
	/*!
	 * \brief Resets the credits to the beginning.
	 */
	void Reset();
	
	/*!
	 * \brief Sets the draw area for the credits.
	 *
	 * \param[in] area the area the credits are displayed in.
	 */
	void SetArea(const Rectf &area);
	
	/*!
	 * \brief Loads credits from a xml file.
	 *
	 * \param[in] fileName the name of the xml file containing the credits.
	 *
	 * \return true of the credits were loaded, false otherwise.
	 */
	bool Load(const char *fileName);
	
	/*!
	 * \brief Determins if the credits are finished.
	 *
	 * \return true if the credits have been displayed, false otherwise.
	 */
	bool IsComplete() const;
	
	/*!
	 * \brief Updates the credits.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 */
	void Update(float dt);
	
	/*!
	 * \brief Draws the credits.
	 */
	void Draw() const;
	
protected:

	// All tags in the Credits XML file contain, at a minimum, these attributes
	struct TagAttributes
	{
		std::string title;
		std::string font;
		int fontSize;
		Color4f color;

		float paddingTop;
		float paddingBottom;
	};
	
	/* Functions for loading an xml formatted credits file. */
	bool LoadCreditsXMLFile(const char *file);
	
	bool ParseXMLCreditsTag(TiXmlElement *elm);
	
	bool ParseXMLSectionTag(TiXmlElement *elm);

	bool ParseXMLEndImageTag(TiXmlElement *elm);
	
	bool ParseXMLElementTag(TiXmlElement *elm, CreditSection &sec);

	bool ParseXMLImageElementTag(TiXmlElement *elm, CreditSection &sec);

	bool ParseXMLElementText(TiXmlElement *elm, std::string &out);

	bool ParseXMLTagAttributes(TiXmlElement *elm, TagAttributes &attribs);

	void ApplySectionTagAttributes(CreditSection &sec, TagAttributes &attribs);

	struct Section
	{
		CreditSection section;

		Point2f position;
		float w;
		float h;

		AABB GetBounds() const 
		{ 
			Rectf r;
			r.x = position.x;
			r.y = position.y;
			r.w = w;
			r.h = h;
			return RectToAABB(r);
		}

		bool active;
	};

	void ResetSection(Section &sec);
	
	std::vector<Section> sections;
	
	Section title;

	Sprite endImage;
	
	AABB drawArea;
	
	float scrollSpeed;

	float delay;
	float delayTimer;

	bool complete;
};

#endif
