/*!
 * \file GUIMainMenu.cpp
 * \brief Implements the GUIMenuButton class.
 * \author Joe Forte
 * \date 6/2/2008
 */

#include "Common/Globals.h"
#include "GUI/GUIText.h"
#include "GUI/GUIImage.h"
#include "Game/GUIMainMenu.h"
#include "Game/GUIMenuButton.h"

//--------------------------------------------------------------------
GUIMainMenu::GUIMainMenu(GameState *state, FontRef f, FontRef aboutFont, const Rectf &aboutArea, const std::string &aboutTxt, FontRef versionFont)
:
inGameMode(false)
{
	// Sets up the main menu
	banner = new GUIImage();
	banner->SetImage("banner.png", 0.5f);
	banner->SetCenterPos( Point2f(0.0f, 0.5f) );
	AddChild(banner);
	
	std::string versionStr = GetVersionString();
	versionText = new GUIText();
	versionText->SetText(versionStr);
	versionText->SetFont(versionFont);
	versionText->SetColor(Color::Gray);
	
	float w, h;
	versionFont->StringDimensions(versionStr.c_str(), w, h);
	
	Rectf screen = GetScreen()->GetRect();
	
	float px = (screen.w / 2.0f) - w - 10.0f;
	float py = versionFont->LineHeight() + 10.0f - (screen.h / 2.0f);
	
	versionText->SetPosition( Point2f(px, py) );
	
	AddChild(versionText);
	
	about = new GUIAboutMenu();
	about->SetAboutText(f, aboutFont, aboutArea, Color::White, aboutTxt);
	about->Deactivate();
	about->SetVisibility(false);
	AddChild(about);
	
	startButton = new GUIStateChangeButton(GameState::GS_START_GAME, state);
	startButton->SetText( std::string("Start Game"), f, Color::White );
	startButton->SetTextBounds(0.01f, 0.01f, Point2f(0.0f, 0.0f));
	AddChild(startButton);
	
	startNewGameButton = new GUIStateChangeButton(GameState::GS_START_NEW_GAME, state);
	startNewGameButton->SetText( std::string("Start New Game"), f, Color::White );
	startNewGameButton->SetTextBounds(0.01f, 0.01f, Point2f(0.0f, -0.1f));
	startNewGameButton->Deactivate();
	startNewGameButton->SetVisibility(false);
	AddChild(startNewGameButton);
	
	resumeButton = new GUIStateChangeButton(GameState::GS_LAST_GAMESTATE, state);
	resumeButton->SetText( std::string("Resume"), f, Color::White );
	resumeButton->SetTextBounds(0.01f, 0.01f, Point2f(0.0f, 0.0f));
	resumeButton->Deactivate();
	resumeButton->SetVisibility(false);
	AddChild(resumeButton);
	
	highScoresButton = new GUIStateChangeButton(GameState::GS_HIGHSCORES, state);
	highScoresButton->SetText(std::string("High Scores"), f, Color::White);
	highScoresButton->SetTextBounds(0.01f, 0.01f, Point2f(0.0f, -0.1f));
	AddChild(highScoresButton);
	
	aboutButton = new GUIAboutButton(about);
	aboutButton->SetText( std::string("About"), f, Color::White );
	aboutButton->SetTextBounds(0.01f, 0.01f, Point2f(0.0f, -0.2f));
	AddChild(aboutButton);
	
	creditsButton = new GUIStateChangeButton(GameState::GS_CREDITS, state);
	creditsButton->SetText( std::string("Credits"), f, Color::White );
	creditsButton->SetTextBounds(0.01f, 0.01f, Point2f(0.0f, -0.3f));
	AddChild(creditsButton);
	
	quitButton = new GUIStateChangeButton(GameState::GS_EXIT, state);
	quitButton->SetText( std::string("Quit"), f, Color::White );
	quitButton->SetTextBounds(0.01f, 0.01f, Point2f(0.0f, -0.5f));
	AddChild(quitButton);
}

//--------------------------------------------------------------------
void GUIMainMenu::EnableInGameMode()
{
	if(inGameMode)
		return;
	
	resumeButton->Activate();
	resumeButton->SetVisibility(true);
	
	startNewGameButton->Activate();
	startNewGameButton->SetVisibility(true);
	
	startButton->Deactivate();
	startButton->SetVisibility(false);

	highScoresButton->Deactivate();
	highScoresButton->SetVisibility(false);
	
	creditsButton->Deactivate();
	creditsButton->SetVisibility(false);
	
	versionText->Deactivate();
	versionText->SetVisibility(false);

	quitButton->SetTextBounds(0.01f, 0.01f, Point2f(0.0f, -0.4f));
	
	inGameMode = true;
}

//--------------------------------------------------------------------
void GUIMainMenu::DisableInGameMode()
{
	if(!inGameMode)
		return;
	
	resumeButton->Deactivate();
	resumeButton->SetVisibility(false);
	
	startNewGameButton->Deactivate();
	startNewGameButton->SetVisibility(false);
	
	startButton->Activate();
	startButton->SetVisibility(true);

	highScoresButton->Activate();
	highScoresButton->SetVisibility(true);
	
	creditsButton->Activate();
	creditsButton->SetVisibility(true);
	
	versionText->Activate();
	versionText->SetVisibility(true);
	
	quitButton->SetTextBounds(0.01f, 0.01f, Point2f(0.0f, -0.5f));
	
	inGameMode = false;
}

//--------------------------------------------------------------------
bool GUIMainMenu::IsAboutMenuOpen() const 
{ 
	return about->IsVisible(); 
}

//--------------------------------------------------------------------
void GUIMainMenu::CloseAboutMenu()
{
	about->Deactivate();
	about->SetVisibility(false);
}
