/*!
 * \file GUIMenuButton.cpp
 * \brief Implements the GUIMenuButton class.
 * \author Joe Forte
 * \date 6/2/2008
 */

#include <assert.h>
#include "Common/Globals.h"
#include "Input/Keys.h"
#include "Graphics/MathDraw.h"
#include "Game/GUIMenuButton.h"
#include "Utility/FormattedText.h"

//--------------------------------------------------------------------
GUIMenuButton::GUIMenuButton() 
: 
enabled(true)
{
	color = Color::White;
	mouseClick.Load("mouse_click.wav");
	mouseRelease.Load("mouse_release.wav");
}

//--------------------------------------------------------------------
void GUIMenuButton::OnMouseOver(const IMouse &m)
{
	color = Color::Gray;
}

//--------------------------------------------------------------------
void GUIMenuButton::OnMouseLeave(const IMouse &m)
{
	color = Color::White;
}

//--------------------------------------------------------------------
void GUIMenuButton::OnMouseClick(const IMouse &m)
{
	color = Color::Red;
	mouseClick.Play();
}

//--------------------------------------------------------------------
void GUIMenuButton::OnMouseRelease(const IMouse &m)
{
	color = Color::White;
	mouseRelease.Play();
	
	DoClickAction();
}

//--------------------------------------------------------------------
void GUIMenuButton::Disable()
{
	enabled = false;
}

//--------------------------------------------------------------------
void GUIMenuButton::Enable()
{
	enabled = true;
}

//--------------------------------------------------------------------
void GUIMenuButton::Draw()
{
	GUIButton::Draw();
}

//--------------------------------------------------------------------
GUIStateChangeButton::GUIStateChangeButton(GameState::State stateTransition, GameState *stateToChange) 
: 
transition(stateTransition), 
state(stateToChange)
{}

//--------------------------------------------------------------------
void GUIStateChangeButton::OnMouseRelease(const IMouse &m)
{
	assert(state);
	
	if(enabled)
	{
		GUIMenuButton::OnMouseRelease(m);
		state->RequestStateChange(transition);
	}
}

//--------------------------------------------------------------------
GUIAboutMenu::GUIAboutMenu()
{
	back = new GUIBackButton(this);
	AddChild(back);
}

//--------------------------------------------------------------------
void GUIAboutMenu::KeyboardUpdate(const IKeyboard &kb)
{
	if(kb.KeyPressed(Keys::KEY_ESCAPE))
	{
		Deactivate();
		SetVisibility(false);
	}
}

//--------------------------------------------------------------------
void GUIAboutMenu::SetAboutText(FontRef f, FontRef aboutFont, const Rectf &r, const Color4f &c, const std::string &txt)
{
	assert(f.IsValid() && aboutFont.IsValid());
	
	font = aboutFont;
	area = r;
	color = c;
	
	back->SetText(std::string("Back"), f, Color::White);
	
	float w, h;
	f->StringDimensions("Back", w, h);
	Point2f p = ScreenToNormCoords(f->GetTextPosition("Back", area, Font::ALIGN_BOTTOM_RIGHT) + Point2f(w/2.0f, h/2.0f));
	back->SetTextBounds(0.01f, 0.0f, p);
	
	text.clear();
	
	FormatText(txt.c_str(), aboutFont, r, text);
}

//--------------------------------------------------------------------
void GUIAboutMenu::Paint()
{	
	Rectf window = ExpandRect(area, 15.0f, 15.0f);
	Rectf border = ExpandRect(area, 10.0f, 10.0f);
	
	DrawRect(window, true, Color4f(0.0f, 0.0f, 0.0f, 0.9f));
	DrawRect(border, false, Color::White, 1.0f);
	
	Point2f drawPos(area.x, area.y);
	float h = font->LineHeight();
	
	for(std::vector<std::string>::iterator i = text.begin(); i != text.end(); i++)
	{
		font->DrawTextStr(i->c_str(), color, drawPos);
		drawPos.y -= h; 
	}
	
	RequestFocus();
}

//--------------------------------------------------------------------
GUIAboutButton::GUIAboutButton(GUIAboutMenu *a)
{
	assert(a);
	about = a;
}

//--------------------------------------------------------------------
void GUIAboutButton::OnMouseRelease(const IMouse &m)
{
	GUIMenuButton::OnMouseRelease(m);
	about->MoveToFront();
	about->Activate();
	about->SetVisibility(true);
}
