/*!
 * \file NameEntryScreen.cpp
 * \brief Implements the NameEntryScreen class.
 * \author Joe Forte
 * \date 2/11/2009
 */

#include "Math/MathCommon.h"
#include "Common/Globals.h"
#include "Input/Keyboard.h"
#include "Game/GUINameEntryScreen.h"
#include "Graphics/MathDraw.h"

const uint MAX_NAME_LEN = 27;

// Keys that will be filtered from keyboard input.
const Key FILTERED_KEYS[] = 
{
	Keys::KEY_RETURN,
	Keys::KEY_KP_ENTER,
	Keys::KEY_TAB,
	Keys::KEY_ESCAPE,
};

const uint FILTERED_KEYS_SIZE = sizeof(FILTERED_KEYS) / sizeof(Key);

//--------------------------------------------------------------------
GUINameTextEntry::GUINameTextEntry()
:
blinkOn(false),
blinkTime(0.0f),
blinkTimer(0.4f)
{}

//--------------------------------------------------------------------
GUINameTextEntry::GUINameTextEntry(GUIPanel *p)
:
GUITextEntry(p),
blinkOn(false),
blinkTime(0.0f),
blinkTimer(0.4f)
{}

//--------------------------------------------------------------------
GUINameTextEntry::GUINameTextEntry(uint len)
:
GUITextEntry(len),
blinkOn(false),
blinkTime(0.0f),
blinkTimer(0.4f)
{}

//--------------------------------------------------------------------
GUINameTextEntry::GUINameTextEntry(uint len, const Key *keys, uint n)
:
GUITextEntry(len, keys, n),
blinkOn(false),
blinkTime(0.0f),
blinkTimer(0.4f),
charSpace(0.0f),
charPadding(0.0f)
{}

//--------------------------------------------------------------------
void GUINameTextEntry::SetFont(FontRef f)
{
	GUITextEntry::SetFont(f);
	
	charSpace = f->StringWidth("M");
	charPadding = charSpace * 0.3f;
}

//--------------------------------------------------------------------
void GUINameTextEntry::OnUpdate(float dt)
{
	blinkTime += dt;
	if(blinkTime >= blinkTimer)
	{
		blinkTime = 0.0f;
		blinkOn = !blinkOn;
	}
}

//--------------------------------------------------------------------
void GUINameTextEntry::RecomputeLayout()
{
	FontRef font = GetFont();
	if(!font.IsValid())
		return;
	
	const uint maxChars = Minimum<uint>(64, GetMaxInputLength());
	
	float nameInputWidth = (maxChars * charSpace) + (maxChars - 1) * charPadding;
	float height = font->LineHeight();
	
	SetWidth(nameInputWidth);
	SetHeight(height);
}

//--------------------------------------------------------------------
void GUINameTextEntry::Paint()
{
	FontRef font = GetFont();
	
	if(!font.IsValid())
		return;
	
	const uint cursor = GetCursorPosition();
	const uint maxChars = Minimum<uint>(64, GetMaxInputLength());
	
	Point2f drawPos = GetPosition();
	drawPos.y -= font->LineHeight() * 2.5f;
	
	for(uint i = 0; i < maxChars; i++)
	{
		Line line;
		line.a = Point2f(drawPos.x + (charSpace + charPadding) * i, drawPos.y);
		line.b = Point2f(line.a.x + charSpace, line.a.y);
		
		if((int)i == cursor)
			DrawLine(line, Color::Blue, 1.0f);
		else
			DrawLine(line, Color::White, 1.0f);
	}
	
	drawPos.y += font->LineHeight() * 1.1f;
	
	const std::string &name = GetText();
	uint len = name.length();
	for(uint i = 0; i < len; i++)
	{
		char str[2] = { name[i], '\0' };
		
		Point2f pos(drawPos.x + (charSpace + charPadding) * i, drawPos.y);
		
		if((int)i == cursor)
		{
			float t = blinkTime / blinkTimer;
			
			Color4f c;
			if(!blinkOn)
				c = lerp(Color4f(1.0f, 1.0f, 1.0f), Color4f(0.5f, 0.5f, 0.5f), t);
			else
				c = lerp(Color4f(0.5f, 0.5f, 0.5f), Color4f(1.0f, 1.0f, 1.0f), t);
			
			font->DrawTextStr(str, c, pos);
		}
		else
		{
			font->DrawTextStr(str, Color::White, pos);
		}
	}	
}

//--------------------------------------------------------------------
GUINameEntryScreen::GUINameEntryScreen(FontRef headerFont, FontRef inputFont, FontRef buttonFont)
{
	enterNameText = new GUIText;
	
	enterNameText->SetFont(headerFont);
	enterNameText->SetColor(Color::Blue);
	enterNameText->SetText("Enter Your Name:");
	
	AddChild(enterNameText);
	
	textEntry = new GUINameTextEntry(MAX_NAME_LEN, FILTERED_KEYS, FILTERED_KEYS_SIZE);
	
	textEntry->SetFont(inputFont);
	textEntry->SetColor(Color::White);
	textEntry->SetPosition(Point2f(0.0f, -25.0f));
	
	AddChild(textEntry);
	
	doneButton = new GUIMenuButton;
	
	doneButton->SetAction(this);
	doneButton->SetText( std::string("Done"), buttonFont, Color::White );
	doneButton->SetTextBounds(0.01f, 0.01f, Point2f(0.0f, -0.3f));
	
	AddChild(doneButton);
	
	// Computes the bounds for all of the children.
	Refit();
	
	Rectf screen = GetScreen()->GetRect();
	
	float yOffset = 3.0f * enterNameText->GetHeight();
	enterNameText->CenterInArea(screen, Point2f(0.0f, yOffset));
	
	yOffset = 2.0f * enterNameText->GetHeight();
	textEntry->CenterInArea(screen, Point2f(0.0f, yOffset));
}

//--------------------------------------------------------------------
void GUINameEntryScreen::OnActivation()
{
	RequestFocus();
}

//--------------------------------------------------------------------
void GUINameEntryScreen::KeyboardUpdate(const IKeyboard &kb)
{
	textEntry->KeyboardUpdate(kb);
	
	if( kb.KeyPressed(Keys::KEY_RETURN) || kb.KeyPressed(Keys::KEY_KP_ENTER) || kb.KeyPressed(Keys::KEY_ESCAPE) )
		SetNameEntryComplete();
}

//--------------------------------------------------------------------
bool GUINameEntryScreen::IsNameEntryComplete() const
{
	return nameEntryComplete;
}

//--------------------------------------------------------------------
void GUINameEntryScreen::SetNameEntryComplete()
{
	nameEntryComplete = true;
	textEntry->Deactivate();
}

//--------------------------------------------------------------------
void GUINameEntryScreen::Reset()
{
	nameEntryComplete = false;
	textEntry->Activate();
	ClearName();
}

//--------------------------------------------------------------------
void GUINameEntryScreen::ClearName()
{
	textEntry->ClearText();
}

//--------------------------------------------------------------------
std::string GUINameEntryScreen::GetName() const
{
	return textEntry->GetText();
}

//--------------------------------------------------------------------
void GUINameEntryScreen::OnClick(GUIButton *b)
{
	if(b == doneButton)
		SetNameEntryComplete();
}

//--------------------------------------------------------------------
void GUINameEntryScreen::Paint()
{
}
