/*!
 * \file GUINameEntryScreen.h
 * \brief Declares the GUINameEntryScreen class used for entering names for high scores.
 * \author Joe Forte
 * \date 2/11/2009
 */

#ifndef GUINAMEENTRYSCREEN_H
#define GUINAMEENTRYSCREEN_H

#include "GUI/GUITextEntry.h"
#include "GUI/GUIText.h"
#include "Graphics/Font.h"
#include "Input/BufferedKeyboardInput.h"
#include "Game/GUIMenuButton.h"

class IKeyboard;

/*!
 * \class GUINameTextEntry
 * \headerfile NameEntryScreen.h "Game/NameEntryScreen.h"
 * \brief The name text entry gui element. 
 */
class GUINameTextEntry : public GUITextEntry
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	GUINameTextEntry();
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] p the panels parent.
	 */
	GUINameTextEntry(GUIPanel *p);
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] len the max length of the input.
	 */
	GUINameTextEntry(uint len);
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] len the max length of the input.
	 * \param[in] keys an array of keys to filter.
	 * \param[in] n the length of the keys array.
	 */
	GUINameTextEntry(uint len, const Key *keys, uint n);
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~GUINameTextEntry() {}
	
	/*!
	 * \brief Sets the font to use when drawing text.
	 *
	 * \param[in] f the referance to the font.
	 */
	virtual void SetFont(FontRef f);
	
	/*!
	 * \brief Gets the string type identifier for the GUI element.
	 *
	 * \return The string type identifier.
	 */
	virtual const char *GetPanelTypeName() const { return "GUI::NameTextEntry"; }
	
protected:
	
	/*!
	 * \brief Does the actual rendering of the GUI element.
	 */
	virtual void Paint();
	
	/*!
	 * \brief Call once a frame when the panel is active.
	 */
	virtual void OnUpdate(float dt);
	
	/*!
	 * \brief Recomputes the bounds and layout for the panel. Called by the Refit() method.
	 */
	virtual void RecomputeLayout();
	
private:
	
	bool blinkOn;
	float blinkTime;
	float blinkTimer;
	
	float charSpace; // The space reserverd for each character in the name.
	float charPadding; // The space between each character.
	
};

/*!
 * \class GUINameEntryScreen
 * \headerfile NameEntryScreen.h "Game/NameEntryScreen.h"
 * \brief The NameEntryScreen class used for entering names for high scores.
 */
class GUINameEntryScreen : public GUIPanel, GUIButtonAction
{
public:
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] headerFont the font to draw header text with.
	 * \param[in] inputFont the font to draw input text with.
	 * \param[in] buttonFont the font to draw button text with.
	 */
	GUINameEntryScreen(FontRef headerFont, FontRef inputFont, FontRef buttonFont);
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~GUINameEntryScreen() {}

	/*!
	 * \brief Called when the panel is activated.
	 */
	virtual void OnActivation();
	
	/*!
	 * \brief Called each frame when the panel has focus.
	 */
	virtual void KeyboardUpdate(const IKeyboard &kb);
	
	/*!
	 * \brief Returns true if the given query point is in the panels bounds.
	 */
	virtual bool InActiveArea(const Point2f &p) const { return true; }
	
	/*!
	 * \brief Returns true when the user has completed their name entry.
	 */
	virtual bool IsNameEntryComplete() const;
	
	/*!
	 * \brief Sets the name entry to be completed.
	 */
	void SetNameEntryComplete();
	
	/*!
	 * \brief Resets for a new name entry.
	 */
	void Reset();
	
	/*!
	 * \brief Clears the name text.
	 */
	void ClearName();
	
	/*!
	 * \brief Gets the currently entered name.
	 */
	std::string GetName() const;
	
	/*!
	 * \brief Called when a button is clicked.
	 *
	 * \param[in] b the button that was clicked.
	 */
	virtual void OnClick(GUIButton *b);
	
protected:
	
	/*!
	 * \brief Does the actual rendering of the GUI element.
	 */
	virtual void Paint();
	
private:
		
	bool nameEntryComplete;
	
	GUIText *enterNameText;
	GUINameTextEntry *textEntry;
	GUIMenuButton *doneButton;
	
};

#endif
