/*!
 * \file GUIScoreBoard.cpp
 * \brief Declares the GUIScoreBoard class.
 * \author Joe Forte
 * \date 3/2/2009
 */

#include <sstream>
#include "Math/MathCommon.h"
#include "Common/Globals.h"
#include "Graphics/MathDraw.h"
#include "Game/GameSettings.h"
#include "Game/GUIScoreBoard.h"

//--------------------------------------------------------------------
GUIScoreBoard::GUIScoreBoard(const GameSettings &settings) 
:
ballCount(0),
currentScore(0),
realScore(0)
{
	font = LoadFontResource(settings.LevelFontFile(), settings.LevelFontSize());
		
	maxBallSize = font->LineHeight();
	ballTimer = 0.2f;
	
	ballImage.Load(settings.BallImage());
	ballImage.SetSize(maxBallSize, maxBallSize);
	
	Rectf screen = GetScreen()->GetRect();
	
	float h = font->LineHeight() * 1.5f;
	
	area.x = -(screen.w / 2.0f);
	area.y = -(screen.h / 2.0f) + h;
	area.w = screen.w;
	area.h = h;
}

//--------------------------------------------------------------------
void GUIScoreBoard::Update(float dt)
{
	if(currentScore < realScore)
		currentScore += std::max((int)((realScore - currentScore) * dt), 1);
	
	if(ballCount <= MAX_BALLS_DISPLAYED)
	{
		for(int i = 0; i < ballCount; i++)
		{
			ballInfo[i].timer += dt;
			ClampHigh(ballInfo[i].timer, ballTimer);
			ballInfo[i].size = maxBallSize * (ballInfo[i].timer / ballTimer);
		}
		
		for(int i = ballCount; i < MAX_BALLS_DISPLAYED; i++)
		{
			ballInfo[i].timer -= dt;
			ClampLow(ballInfo[i].timer, 0.0f);
			ballInfo[i].size = maxBallSize * (ballInfo[i].timer / ballTimer);
		}
	}
}

//--------------------------------------------------------------------
void GUIScoreBoard::IncrementScore(int amount)
{
	realScore += amount;
}

//--------------------------------------------------------------------
void GUIScoreBoard::SetScore(int score)
{
	realScore = score;
	currentScore = score;
}

//--------------------------------------------------------------------
void GUIScoreBoard::IncrementBallCount()
{
	if(ballCount < MAX_BALLS_DISPLAYED)
	{
		ballInfo[ballCount].timer = 0.0f;
		ballInfo[ballCount].size = 0.0f;
	}
	
	ballCount++;
}

//--------------------------------------------------------------------
void GUIScoreBoard::DecrementBallCount()
{	
	if(ballCount == MAX_BALLS_DISPLAYED + 1)
	{
		for(uint i = 0; i < MAX_BALLS_DISPLAYED; i++)
		{
			ballInfo[i].timer = 0.0f;
			ballInfo[i].size = 0.0f;
		}
	}
	
	ballCount--;
}

//--------------------------------------------------------------------
void GUIScoreBoard::SetBallCount(uint count)
{
	ballCount = (int)count;
	
	uint num = std::min(MAX_BALLS_DISPLAYED, count);
	for(uint i = 0; i < num; i++)
	{
		ballInfo[i].timer = ballTimer;
		ballInfo[i].size = maxBallSize;
	}
	
	for(uint i = num; i < MAX_BALLS_DISPLAYED; i++)
	{
		ballInfo[i].timer = 0.0f;
		ballInfo[i].size = 0.0f;
	}
}

//--------------------------------------------------------------------
void GUIScoreBoard::Paint()
{	
	DrawRect(area, true, Color4f(0.0f, 0.0f, 0.0f, 0.5f));	
	
	Line l;
	l.a = Point2f(area.x, area.y);
	l.b = Point2f(area.x + area.w, area.y);
	
	DrawLine(l, Color::White, 1.0f);
	
	std::stringstream scoreTxt;
	scoreTxt << "Score:" << currentScore;
	
	Point2f drawPos(area.x + 10.0f, area.y);
	font->DrawTextStr(scoreTxt.str().c_str(), Color::White, drawPos);
	
	drawPos.x = area.x + area.w - 10.0f;
	
	if(ballCount >= 0)
	{
		if(ballCount <= MAX_BALLS_DISPLAYED)
		{
			drawPos.y -= (maxBallSize/ 2.0f) + font->LineHeight() * 0.2f;
			drawPos.x -= maxBallSize / 2.0f;
		
			for(int i = 0; i < MAX_BALLS_DISPLAYED; i++)
			{
				if(ballInfo[i].size > 0.0f)
				{
					ballImage.SetSize(ballInfo[i].size, ballInfo[i].size);
			
					ballImage.SetPosition(drawPos);
					drawPos.x -= maxBallSize * 1.1f;
				
					ballImage.Draw();
				}
			}
		}
		else
		{
			std::stringstream ballTxt;
			ballTxt << ballCount;
			
			std::string num = ballTxt.str();
			float numWidth = font->StringWidth(num.c_str());
			
			float xWidth, xHeight;
			font->StringDimensions("x", xWidth, xHeight);
			font->DrawTextStr("x", Color::White, drawPos - Point2f(numWidth + xWidth + 1.0f, 0.0f));
			font->DrawTextStr(num.c_str(), Color::White, drawPos - Point2f(numWidth, 0.0f));
			
			ballImage.SetSize(maxBallSize, maxBallSize);
			
			Point2f ballPos = drawPos - Point2f((maxBallSize / 2.0f) + numWidth + xWidth + 2.0f, (maxBallSize / 2.0f) + font->LineHeight() * 0.2f);
			ballImage.SetPosition(ballPos);
			drawPos.x += ballImage.GetWidth() * 1.1f;
		
			ballImage.Draw();
		}
	}
}
