/*!
 * \file GameCompleteScreen.h
 * \brief Defines GameCompleteScreen class used to display the game over and level complete screens.
 * \author Joe Forte
 * \date 7/22/2008
 */

#ifndef GAMECOMPLETESCREEN_H
#define GAMECOMPLETESCREEN_H

#include "Common/Globals.h"
#include "Math/MathCommon.h"
#include "Graphics/MathDraw.h"
#include "Graphics/Font.h"

/*!
 * \class GameCompleteScreen
 * \headerfile GameCompleteScreen.h "Game/GameCompleteScreen.h"
 * \brief Class used to display the game over and level complete screens.
 */
class GameCompleteScreen
{
public:
	
	void Init(FontRef largeFont, FontRef smallFont, float fadeTime)
	{
		mainTextFont = largeFont;
		scoreTextFont = smallFont;
		
		fadeInTime = fadeTime;
		
		blinkTime = 0.3f;
		blinkTimer = 0.0f;
		blink = false;
	}
	
	void Reset(const std::string &text, const std::string &score)
	{
		mainText = text;
		scoreText = score;
		
		fadeInTimer = fadeInTime;
	}
	
	void SetIsNewHighScore(bool val) { newHighScore = val; }
	
	void Update(float dt)
	{
		fadeInTimer -= dt;
		ClampLow(fadeInTimer, 0.0f);
		
		if(blink)
		{
			blinkTimer += dt;
			if(blinkTimer > blinkTime)
				blink = !blink;
		}
		else
		{
			blinkTimer -= dt;
			if(blinkTimer <= 0.0f)
				blink = !blink;
		}
	}
	
	bool FadeInComplete() { return fadeInTimer <= 0.0f; }
	
	void Draw()
	{
		float alpha = 1.0f - (fadeInTimer / fadeInTime);
		
		DrawFullscreenQuad( Color4f(0.0f, 0.0f, 0.0f, alpha*0.5f) );
		
		Rectf r = GetScreen()->GetRect();
		mainTextFont->DrawTextStr(mainText.c_str(), Color4f(0.0f, 0.0f, 1.0f, alpha), r, Font::ALIGN_CENTER);
		
		Point2f off(0.0f, -mainTextFont->LineHeight());
		scoreTextFont->DrawTextStr(scoreText.c_str(), Color4f(1.0f, 1.0f, 1.0f, alpha), r, Font::ALIGN_CENTER, off);
		
		if(newHighScore)
		{
			float balpha = blinkTimer / blinkTime;
			
			off.y -= mainTextFont->LineHeight() * 2.0f;
			mainTextFont->DrawTextStr("New High Score!!!", Color4f(1.0f, 1.0f, 1.0f, alpha * balpha), r, Font::ALIGN_CENTER, off);
		}
	}
	
private:
	FontRef mainTextFont;
	FontRef scoreTextFont;
	
	std::string mainText;
	std::string scoreText;
	
	float fadeInTime;
	float fadeInTimer;
	
	float blinkTime;
	float blinkTimer;
	bool blink;
	
	bool newHighScore;
};

#endif
