/*!
 * \file GameMouse.h
 * \brief Defines the mouse spacific to the BrickOut game.
 * \author Joe Forte
 * \date 3/27/2008
 */

#ifndef GAMEMOUSE_H
#define GAMEMOUSE_H

#include "Input/SDLMouse.h"
#include "Graphics/ParticleSystem.h"
#include "Common/Globals.h"

/*!
 * \class SparklePolicy
 * \headerfile GameMouse.h "Game/GameMouse.h"
 * \brief ParticleSystem policy that creates a simple sprkling fountain.
 */
class SparklePolicy
{
public:
	typedef ContinuousEmitPolicy<> Emit;
		
	class Init
	{
	public:
		/*!
		 * \brief Called once before new particles are emitted.
		 *
		 * \param[in] pos the requested position to emit the particles.
		 * \param[in] amount the number of particles that will be emitted.
		 */
		inline void PreEmit(const Point2f &pos, uint amount) { position = pos; }
	
		/*!
		 * \brief Called once on each new particle that is emitted.
		 *
		 * \param[in] p the particle to initialize.
		*/
		inline void InitParticle(Particle &p)
		{
			p.pos = position;
			p.color = Color::Gold;
			p.color.a = 0.0f;
			p.life = RandomFloat(1.0f, 2.0f);
			p.age = 0.0f;
			p.angle = 0.0f;
			p.angularVel = 0.0f;
		
			float size = RandomFloat(15.0f, 25.0f);
			p.width = size;
			p.height = size;
			p.vel = Vector2f(RandomFloat(-25.0, 25.0f), -50.0f);
		
			// adds a little random jitter
			p.pos.x += RandomFloat(-10.0, 10.0f);
		}

	private:
		// updated after each call to PreEmit
		Point2f position;
	};

	class Update
	{
	public:
		// update data
		float fadeOutAge;
		Vector2f gravity;
	
		Update() : gravity(0.0f, -98.0f), fadeOutAge(0.5f) {}
	
		/*!
		 * \brief Called once before each update.
		 *
		 * \param[in] dt the time in seconds since the last frame.
		 */
		inline void PreUpdate(float dt)
		{
			at = gravity*dt;
			att = at*dt;
		}
	
		/*!
		 * \brief Called once before each update.
		 *
		 * \param[in,out] p the particle to apply the Policy to.
		 * \param[in] dt the time in seconds since the last frame.
		 *
		 * \return False if you want the particle to be destroyed, return true otherwise.
		 */
		inline bool ApplyUpdate(Particle &p, float dt)
		{
			p.pos = p.pos + p.vel*dt + att;
			p.vel = p.vel + at;
		
			p.age += dt;
			if(p.age >= fadeOutAge)
				p.color.a = 1.0f - ( (p.age - fadeOutAge) / (p.life - fadeOutAge) );
			else
				p.color.a = p.age / fadeOutAge;
		
			return p.age < p.life;
		}
	
	private:
		// updated after each call to update
		Vector2f at, att; // Acceleration times time, Acceleration times time squared
	};
};

/*!
 * \class GameMouse
 * \headerfile GameMouse.h "Game/GameMouse.h"
 * \brief A game spacific mouse. 
 */
class GameMouseEffect
{
public:
	
	GameMouseEffect()
	{
		sparkler.emitPolicy.Start();
		sparkler.emitPolicy.SetFlowRate(25.0f);
		sparkler.emitPolicy.position = Point2f(0.0f, 0.0f);
		sparkler.drawPolicy.image = LoadImageResource("sparkle.png");
		sparkler.drawPolicy.blendMode = BLENDMODE_ADDITIVE_WITH_ALPHA;
	}
	
	void Update(float dt, SDLMouse &m)
	{
		sparkler.Update(dt);
		
		float w, h;
		m.GetCursor().GetDimensions(w, h);
		
		Point2f p;
		ConvertPoint2(m.GetPos(), p);
		
		p.x += w / 2.0f;
		p.y -= h / 2.0f;
		sparkler.emitPolicy.position = p;
	}
	
	/*!
	 * \brief Informs the mouse that it has has gained screen focus.
	 */
	void GainedFocus()
	{
		sparkler.emitPolicy.Start();
	}
	
	/*!
	 * \brief Informs the mouse that it has has lost screen focus.
	 */
	void LostFocus()
	{
		sparkler.emitPolicy.Stop();
	}
	
	void Draw()
	{
		sparkler.Draw();
	}
	
private:
	ParticleSystem<100, SparklePolicy::Init, SparklePolicy::Update, SparklePolicy::Emit> sparkler;
};

#endif
