/*!
 * \file GameSettings.cpp
 * \brief Implements the class that holds games settings loaded from a config file.
 * \author Joe Forte
 * \date 2/6/2009
 */

#include "Common/Path.h"
#include "Application/DataPaths.h"
#include "Game/GameSettings.h"

#define FIND_AND_SET(obj, name, value_func, config_file) \
{ \
	const ConfigFile::ConfigValue *val = config_file.FindValueByName(name); \
	if(val) \
	{ \
		obj = val->value_func(); \
	} \
} \

//--------------------------------------------------------------------
GameSettings::GameSettings()
{
	/* Sets the default values */

	//---- Game settings ---------------------------------------------------------------

	defualtLevelListFile = "";
	
	//---- Level settings --------------------------------------------------------------
	
	levelFontFile       = "";
	launchArrowImage    = "";
	levelFontSize       = 0;
	multiplierFontSize  = 0;
	launchArrowHeight   = 0.0f;
	launchArrowFadeTime = 0.0f;
	ballFadeInTime      = 0.0f;
	ballLaunchSpeed     = 0.0f;
	ballMinSpeed        = 0.0f;
	ballMaxSpeed        = 0.0f;
	powerUpProbability  = 0.0f;
	multiplierTime      = 0.0f;
	stuckBallFixTime    = 0.0f;
	paddleStartPosition = Point2f(0.0f, 0.0f);
	
	//---- Default Paddle settings -----------------------------------------------------
	
	defaultPaddleBaseImage       = "";
	defaultPaddleBaseStickyImage = "";
	defaultPaddleThrusterImage   = "";
	defaultPaddleShape           = "";
	defaultPaddleWidth           = 0.0f;
	defaultPaddleMinWidth        = 0.0f;
	defaultPaddleMaxWidth        = 0.0f;
	defaultPaddleGrowIncrements  = 0;
	defaultPaddleThrusterWidth   = 0.0f;
	defaultPaddleThrusterHeight  = 0.0f;
	
	//---- Ball settings ---------------------------------------------------------------
	
	ballImage  = "";
	ballRadius = 0;
	
	//---- High Scores settings --------------------------------------------------------

	highScoresFontFile			= "";
	highScoresFontSize			= 0;
	highScoresHeaderFontSize	= 0;
	highScoresCount				= 0;
}

//--------------------------------------------------------------------
bool GameSettings::Load(const char *configFile)
{
	ConfigFile settings;
	if(!settings.Load(configFile))
		return false;

	//---- Game settings ---------------------------------------------------------------
	
	FIND_AND_SET(defualtLevelListFile, "DefaultLevelList", StringValue, settings)
	
	//---- Level settings --------------------------------------------------------------
	
	FIND_AND_SET(levelFontFile, "LevelFontFile", StringValue, settings)
	FIND_AND_SET(launchArrowImage, "LaunchArrowImage", StringValue, settings)
	FIND_AND_SET(multiplierFontFile, "MultiplierFontFile", StringValue, settings)
	FIND_AND_SET(levelFontSize, "LevelFontSize", IntegerValue, settings)
	FIND_AND_SET(multiplierFontSize, "MultiplierFontSize", IntegerValue, settings)
	FIND_AND_SET(launchArrowHeight, "LaunchArrowHeight", FloatValue, settings)
	FIND_AND_SET(launchArrowFadeTime, "LaunchArrowFadeTime", FloatValue, settings)
	FIND_AND_SET(ballFadeInTime, "BallFadeInTime", FloatValue, settings)
	FIND_AND_SET(ballLaunchSpeed, "BallLaunchSpeed", FloatValue, settings)
	FIND_AND_SET(ballMinSpeed, "BallMinSpeed", FloatValue, settings)
	FIND_AND_SET(ballMaxSpeed, "BallMaxSpeed", FloatValue, settings)
	FIND_AND_SET(powerUpProbability, "PowerUpProbability", FloatValue, settings)
	FIND_AND_SET(multiplierTime, "MultiplierTime", FloatValue, settings)
	FIND_AND_SET(stuckBallFixTime, "StuckBallFixTime", FloatValue, settings)
	FIND_AND_SET(paddleStartPosition, "PaddleStartPosition", PointValue, settings)
	
	//---- Default Paddle settings -----------------------------------------------------
	
	FIND_AND_SET(defaultPaddleBaseImage, "DefaultPaddleBaseImage", StringValue, settings)
	FIND_AND_SET(defaultPaddleBaseStickyImage, "DefaultPaddleBaseStickyImage", StringValue, settings)
	FIND_AND_SET(defaultPaddleThrusterImage, "DefaultPaddleThrusterImage", StringValue, settings)
	FIND_AND_SET(defaultPaddleShape, "DefaultPaddleShape", StringValue, settings)
	FIND_AND_SET(defaultPaddleWidth, "DefaultPaddleWidth", FloatValue, settings)
	FIND_AND_SET(defaultPaddleMinWidth, "DefaultPaddleMinWidth", FloatValue, settings)
	FIND_AND_SET(defaultPaddleMaxWidth, "DefaultPaddleMaxWidth", FloatValue, settings)
	FIND_AND_SET(defaultPaddleGrowIncrements, "DefaultPaddleGrowIncrements", IntegerValue, settings)
	FIND_AND_SET(defaultPaddleThrusterWidth, "DefaultPaddleThrusterWidth", FloatValue, settings)
	FIND_AND_SET(defaultPaddleThrusterHeight, "DefaultPaddleThrusterHeight", FloatValue, settings)
	
	//---- Ball settings ---------------------------------------------------------------
	
	FIND_AND_SET(ballImage, "BallImage", StringValue, settings)
	FIND_AND_SET(ballRadius, "BallRadius", FloatValue, settings)
	
	//---- High Scores settings --------------------------------------------------------

	FIND_AND_SET(highScoresFontFile, "HighScoresFontFile", StringValue, settings)
	FIND_AND_SET(highScoresFontSize, "HighScoresFontSize", IntegerValue, settings)
	FIND_AND_SET(highScoresHeaderFontSize, "HighScoresHeaderFontSize", IntegerValue, settings)
	FIND_AND_SET(highScoresCount, "HighScoresCount", IntegerValue, settings)

	return true;
}
