/*!
 * \file GameSettings.h
 * \brief Declares the class that holds games settings loaded from a config file.
 * \author Joe Forte
 * \date 2/6/2009
 */

#ifndef GAMESETTINGS_H
#define GAMESETTINGS_H

#include <string>
#include "Math/Point.h"
#include "Application/ConfigFile.h"

/*!
 * \class GameSettings
 * \headerfile GameSettings.h "Game/GameSettings.h"
 * \brief Holds settings for the brickout redux game.
 */
class GameSettings
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	GameSettings();
	
	/*!
	 * \brief Loads a game config file.
	 *
	 * \param[in] configFile the name of the config file to load settings from.
	 *
	 * \return true if the config file was loaded, false otherwise.
	 *
	 * NOTE: Multiples file can be loaded consecutively.
	 */
	bool Load(const char *configFile);

	//---- Game settings ---------------------------------------------------------------

	/*! The default level list file. */
	const char *DefaultLevelListFile() const { return defualtLevelListFile.c_str(); }
	
	//---- Level settings --------------------------------------------------------------
	
	/*! The font file the level uses. */
	const char *LevelFontFile() const { return levelFontFile.c_str(); }

	/*! The font file the multiplier timer uses. */
	const char *MultiplierFontFile() const { return multiplierFontFile.c_str(); }
	
	/*! The size of the font the level uses. */
	int LevelFontSize() const { return levelFontSize; }
	
	/*! The size of the font the multiplier GUI element uses. */
	int MultiplierFontSize() const { return multiplierFontSize; }
	
	/*! The image used when the ball is attached to the paddle. */
	const char *LaunchArrowImage() const { return launchArrowImage.c_str(); }
	
	/*! The height of the launch arrow. */
	float LaunchArrowHeight() const { return launchArrowHeight; }
	
	/*! The time it takes to fade in/out the launch arrow. */
	float LaunchArrowFadeTime() const { return launchArrowFadeTime; }
	
	/*! The time it takes to fade in the ball. */
	float BallFadeInTime() const { return ballFadeInTime; }
	
	/*! The speed to launch the ball. */
	float BallLaunchSpeed() const { return ballLaunchSpeed; }
	
	/*! The minimum speed the ball can have when active. */
	float BallMinSpeed() const { return ballMinSpeed; }
	
	/*! The maximum speed the ball can have when active. */
	float BallMaxSpeed() const { return ballMaxSpeed; }
	
	/*! The probability that a powerup will be created. */
	float PowerUpProbability() const { return powerUpProbability; }
	
	/*! The time a multiplier is active. */
	float MultiplierTime() const { return multiplierTime; }
	
	/*! The time until the ball is considered stuck. */
	float StuckBallFixTime() const { return stuckBallFixTime; }
	
	/*! The position place the paddle at the start of the level. */
	const Point2f &PaddleStartPosition() const { return paddleStartPosition; }
	
	//---- Default Paddle settings -----------------------------------------------------
	
	/*! The base image for the default paddle. */
	const char *DefaultPaddleBaseImage() const { return defaultPaddleBaseImage.c_str(); }
	
	/*! The base image for the default paddle when its sticky. */
	const char *DefaultPaddleBaseStickyImage() const { return defaultPaddleBaseStickyImage.c_str(); }
	
	/*! The thruster image for the default paddle. */
	const char *DefaultPaddleThrusterImage() const { return defaultPaddleThrusterImage.c_str(); }
	
	/*! The shape file for the default paddle. */
	const char *DefaultPaddleShape() const { return defaultPaddleShape.c_str(); }
	
	/*! The width for the default paddle. */
	float DefaultPaddleWidth() const { return defaultPaddleWidth; }
	
	/*! The minimum width for the default paddle. */
	float DefaultPaddleMinWidth() const { return defaultPaddleMinWidth; }
	
	/*! The maximum width for the default paddle. */
	float DefaultPaddleMaxWidth() const { return defaultPaddleMaxWidth; }
	
	/*! The number of times the paddle can grow/shrink untill reaching the paddles max/min widths. */
	int DefaultPaddleGrowIncrements() const { return defaultPaddleGrowIncrements; }
	
	/*! The width for the default paddle thruster. */
	float DefaultPaddleThrusterWidth() const { return defaultPaddleThrusterWidth; }
	
	/*! The height for the default paddle thruster. */
	float DefaultPaddleThrusterHeight() const { return defaultPaddleThrusterHeight; }
	
	//---- Ball settings ---------------------------------------------------------------
	
	/*! The ball image. */
	const char *BallImage() const { return ballImage.c_str(); }
	
	/*! The default radius of the ball. */
	float BallRadius() const { return ballRadius; }

	//---- High Scores settings --------------------------------------------------------

	/*! The high scores font. */
	const char *HighScoresFontFile() const { return highScoresFontFile.c_str(); }

	/*! The high scores font size. */
	int HighScoresFontSize() const { return highScoresFontSize; }

	/*! The high scores font size for header text. */
	int HighScoresHeaderFontSize() const { return highScoresHeaderFontSize; }

	/*! The number of high scores to display. */
	int HighScoresCount() const { return highScoresCount; }
	
protected:

	//---- Game settings ---------------------------------------------------------------

	std::string defualtLevelListFile;
	
	//---- Level settings --------------------------------------------------------------
	
	std::string levelFontFile;
	std::string launchArrowImage;
	std::string multiplierFontFile;
	int levelFontSize;
	int multiplierFontSize;
	float launchArrowHeight;
	float launchArrowFadeTime;
	float ballFadeInTime;
	float ballLaunchSpeed;
	float ballMinSpeed;
	float ballMaxSpeed;
	float powerUpProbability;
	float multiplierTime;
	float stuckBallFixTime;
	Point2f paddleStartPosition;
	
	//---- Default Paddle settings -----------------------------------------------------
	
	std::string defaultPaddleBaseImage;
	std::string defaultPaddleBaseStickyImage;
	std::string defaultPaddleThrusterImage;
	std::string defaultPaddleShape;
	float defaultPaddleWidth;
	float defaultPaddleMinWidth;
	float defaultPaddleMaxWidth;
	int defaultPaddleGrowIncrements;
	float defaultPaddleThrusterWidth;
	float defaultPaddleThrusterHeight;
	
	//---- Ball settings ---------------------------------------------------------------
	
	std::string ballImage;
	float ballRadius;
	
	//---- High Scores settings --------------------------------------------------------

	std::string highScoresFontFile;
	int highScoresFontSize;
	int highScoresHeaderFontSize;
	int highScoresCount;
};

#endif
