/*!
 * \file GameState.h
 * \brief Defines GameState class.
 * \author Joe Forte
 * \date 9/13/2008
 */

#ifndef GAMESTATE_H
#define GAMESTATE_H

class GameStateListener;

/*!
 * \class GameState
 * \headerfile GameState.h "Game/GameState.h"
 * \brief Defines a class that holds and maintains game states.
 */
class GameState
{
public:
	
	enum State
	{
		GS_STARTUP = 0,
		GS_SPLASH_SCREEN,
		GS_START_GAME,
		GS_START_NEW_GAME,
		GS_MAIN_MENU,
		GS_MAIN_MENU_OVERLAY,
		GS_HIGHSCORES,
		GS_ENTER_NAME,
		GS_LOAD_LEVEL,
		GS_LEVEL,
		GS_LEVEL_COMPLETE,
		GS_GAME_COMPLETE,
		GS_GAME_OVER,
		GS_CREDITS,
		GS_EXIT,
	
		GS_COUNT,
		
		GS_LAST_GAMESTATE // Special state that always refers to the last state the game was in.
	};

	GameState();
	
	void SetListener(GameStateListener *l) { listener = l; }

	void RequestStateChange(State s);

	void RequestStateReversion() { RequestStateChange(lastState); }

	void ForceStateChange(State s);

	void ForceStateReversion() { ForceStateChange(lastState); }

	bool IsStateChangeRequestPending() const { return currState != nextState; }

	GameState::State Get() const { return currState; }

	GameState::State GetLastState() const { return lastState; }

	GameState::State GetRequestedState() const { return nextState; }

	void Transition();
	
private:
	State currState;
	State lastState;
	
	State nextState; // the state that has been requested
	
	GameStateListener *listener;
};

/*!
 * \class GameStateListener
 * \headerfile GameState.h "Game/GameState.h"
 * \brief Defines an interface for listnening to game state events.
 */
class GameStateListener
{
public:
	
	/*!
	 * \brief Called when a state change request is made.
	 */
	virtual void OnStateChangeRequest(const GameState &gs, GameState::State requested) {}
	
	/*!
	 * \brief Called when a state change is forcefully made.
	 */
	virtual void OnForcedStateChange(const GameState &gs, GameState::State state) {}
	
	/*!
	 * \brief Called before a state transition is made.
	 */
	virtual void OnPreStateTransition(const GameState &gs, GameState::State current, GameState::State next) {}
	
	/*!
	 * \brief Called after a state transition is made.
	 */
	virtual void OnPostStateTransition(const GameState &gs, GameState::State current, GameState::State last) {}
	
};

#endif
