/*!
 * \file HighScores.cpp
 * \brief Implements the HighScores class.
 * \author Joe Forte
 * \date 9/13/2008
 */

#include <fstream>
#include <sstream>
#include <iomanip>
#include "Common/Globals.h"
#include "Game/GameSettings.h"
#include "Game/HighScores.h"

const uint HIGHSCORES_FILE_HEADER_SIZE = 4;
const char HIGHSCORES_FILE_HEADER[HIGHSCORES_FILE_HEADER_SIZE] = { 'B', 'R', 'H', 'S' };

const char *HIGHSCORES_HEADER_TEXT = "High Scores";

//--------------------------------------------------------------------
HighScores::HighScores(const GameSettings &settings)
{
	newlyAddedHighScore = -1;
	
	headerFont = LoadFontResource(settings.HighScoresFontFile(), settings.HighScoresHeaderFontSize());
	scoreFont = LoadFontResource(settings.HighScoresFontFile(), settings.HighScoresFontSize());

	scoresToDisplay = settings.HighScoresCount();

	HighScore score;
	score.name = "Joe Forte";
	score.score = 42;
	for(uint i = 0; i < scoresToDisplay; i++)
	{
		highScores.push_back(score);
	}
}

//--------------------------------------------------------------------
HighScores::~HighScores() {}
	
//--------------------------------------------------------------------
void HighScores::LoadScoresFromFile(const char *file)
{
	// NOTE: Not big endien safe
	assert(sizeof(uint) == 4);
	assert(sizeof(char) == 1);
	
	fileName = file;
	
	std::fstream stream(file, std::ios_base::in | std::ios_base::binary);

	if(!stream.is_open())
		return;
	
	highScores.clear();

	char header[HIGHSCORES_FILE_HEADER_SIZE] = { '\0' };
	stream.read(header, HIGHSCORES_FILE_HEADER_SIZE);
	for(uint i = 0; i < HIGHSCORES_FILE_HEADER_SIZE; i++)
	{
		if(header[i] != HIGHSCORES_FILE_HEADER[i])
			return;
	}

	uint count;
	stream.read((char*)&count, sizeof(uint));

	for(uint i = 0; i < count; i++)
	{
		uint len = 0;
		stream.read((char*)&len, sizeof(uint));
		
		char *name = new char[len + 1];
		stream.read(name, sizeof(char) * len);
		name[len] = '\0';

		uint score = 0;
		stream.read((char*)&score, sizeof(uint));

		HighScore hs;
		hs.name = name;
		hs.score = score;

		highScores.push_back(hs);

		delete name;
	}
}

//--------------------------------------------------------------------
void HighScores::SaveScoresToFile(const char *file) const
{
	// NOTE: Not big endien safe
	assert(sizeof(uint) == 4);
	assert(sizeof(char) == 1);

	std::fstream stream(file, std::ios_base::out | std::ios_base::binary | std::ios_base::trunc);

	if(!stream.is_open())
		return;

	stream.write(HIGHSCORES_FILE_HEADER, HIGHSCORES_FILE_HEADER_SIZE);

	uint len = highScores.size();
	stream.write((const char*)&len, sizeof(uint));

	for(uint i = 0; i < len; i++)
	{
		uint nameLen = highScores[i].name.size();
		stream.write((const char*)&nameLen, sizeof(uint));
		stream.write(highScores[i].name.c_str(), nameLen);
		stream.write((const char*)&highScores[i].score, sizeof(uint));
	}
}

//--------------------------------------------------------------------
void HighScores::ClearHighScores()
{
	highScores.clear();
	SaveScoresToFile(fileName.c_str());
}

//--------------------------------------------------------------------
bool HighScores::IsHighScore(uint score) const
{
	return highScores.empty() || highScores.size() < scoresToDisplay || score >= highScores.back().score;
}

//--------------------------------------------------------------------
void HighScores::AddHighScore(const char *name, uint score)
{	
	HighScore hs;
	hs.name = name;
	hs.score = score;
	
	if(highScores.empty())
	{
		highScores.push_back(hs);
		newlyAddedHighScore = highScores.size() - 1;
	}
	else
	{
		int count = (int)highScores.size();
		int index;
		for(index = count - 1; index >= 0; index--)
		{
			if(highScores[index].score > score)
				break;
		}
		
		if(highScores.front().score <= score)
		{
			highScores.insert(highScores.begin(), hs);
			newlyAddedHighScore = 0;
		}
		else if(index == (count - 1))
		{
			highScores.push_back(hs);
			newlyAddedHighScore = highScores.size() - 1;
		}
		else
		{
			highScores.insert(highScores.begin() + index + 1, hs);
			newlyAddedHighScore = index + 1;

		}
	}
	
	if(highScores.size() > scoresToDisplay)
		highScores.pop_back();
	
	SaveScoresToFile(fileName.c_str());
}

//--------------------------------------------------------------------
void HighScores::Draw()
{
	Rectf area = GetScreen()->GetRect();
	
	float headerWidth = headerFont->StringWidth(HIGHSCORES_HEADER_TEXT);
	
	Point2f drawPos(area.x + (area.w - headerWidth) / 2.0f, area.y - headerFont->LineHeight());
	headerFont->DrawTextStr(HIGHSCORES_HEADER_TEXT, Color::Blue, drawPos);
	
	float numWidth = scoreFont->StringWidth("00. ");
	float nameWidth = scoreFont->StringWidth("MMMMMMMMMMMMMMMMMMMMMMMMMMM");
	float scoreWidth = scoreFont->StringWidth("0000000000");
	float innerWidth = scoreFont->StringWidth("   ");
	
	float scoresX = area.x + (area.w - numWidth - nameWidth - scoreWidth - innerWidth) / 2.0f;
	float scoresY = area.y - headerFont->LineHeight() * 3.0f;
	
	const uint count = std::min((uint)highScores.size(), scoresToDisplay);
	for(uint i = 0; i < count; i++)
	{
		Color4f color = Color::White;
		if(i == newlyAddedHighScore)
			color = Color::Red;
		
		std::stringstream numberStream;
		numberStream << std::setfill(' ');
		numberStream << std::setw(2) << (i + 1) << ". ";
		std::string number = numberStream.str();
		
		Point2f pos(scoresX, scoresY);
		scoreFont->DrawTextStr(number.c_str(), color, pos);
		
		std::stringstream nameStream;
		nameStream << std::setfill(' ') << std::setiosflags(std::ios::left);
		nameStream << std::setw(16) << highScores[i].name;
		std::string name = nameStream.str();
		
		pos.x += numWidth;
		scoreFont->DrawTextStr(name.c_str(), color, pos);
		
		std::stringstream scoreStream;
		scoreStream << std::setfill(' ');
		scoreStream << std::setw(10) << highScores[i].score;
		std::string scoreStr = scoreStream.str();
		
		pos.x += nameWidth + innerWidth;
		scoreFont->DrawTextStr(scoreStr.c_str(), color, pos);
		
		scoresY -= scoreFont->LineHeight() * 1.5f;
	}
}
