/*!
 * \file Level.cpp
 * \brief Implement the Level class.
 * \author Joe Forte
 * \date 3/27/2008
 */

#include <sstream>
#include "Graphics/Screen.h"
#include "Graphics/Font.h"
#include "Graphics/MathDraw.h"
#include "Game/Level.h"
#include "Input/Mouse.h"
#include "Input/Keyboard.h"
#include "Common/Globals.h"
#include "Game/GameSettings.h"

//--------------------------------------------------------------------
Level::Level() 
: 
settings(NULL), 
ball(NULL), 
paddle(NULL), 
world(NULL), 
listener(NULL)
{	
	multiplierTimer = new GUIMultiplierTimer;
	viewport.AddChild(multiplierTimer);
}

//--------------------------------------------------------------------
Level::~Level()
{
	Destroy();
}

//--------------------------------------------------------------------
void Level::SetAttributes(const Level::Attributes &attributes)
{
	attribs = attributes;
}

//--------------------------------------------------------------------
void Level::Destroy()
{
	for(int i = 0; i < 3; i++)
		levelBounds[i].Destroy();
	
	if(ball)
	{
		delete ball;
		ball = NULL;
	}
	
	if(paddle)
	{
		delete paddle;
		paddle = NULL;
	}

	for(std::vector<BrickGroup*>::iterator i = brickGroups.begin(); i != brickGroups.end(); i++)
	{
		if(*i)
			delete *i;
	}
	
	brickGroups.clear();
	
	for(std::vector<PowerUp*>::iterator i = powerUps.begin(); i != powerUps.end(); i++)
	{
		if(*i)
			delete *i;
	}
	
	powerUps.clear();
	
	if(world)
	{
		delete world;
		world = NULL;
	}
	
	brickExplosionChunks.Clear();
	brickExplosionSparks.Clear();
	powerUpGrabbed.Clear();
	textParticles.Clear();
}

//--------------------------------------------------------------------
void Level::Init(const GameSettings *gameSettings)
{	
	Destroy();
	
	assert(gameSettings);
	settings = gameSettings;

	bounds = RectToAABB(GetScreen()->GetRect());
	
	textParticles.SetUpdateFunc(TextParticleUpdateFunc);
	textParticles.SetAttributes(LoadFontResource(settings->LevelFontFile(), settings->LevelFontSize()), 0.5f);
	
	AABB worldBounds = bounds;
	worldBounds.min -= Point2f(100.0f, 100.0f);
	worldBounds.max += Point2f(100.0f, 100.0f);
	world = new PhysicsWorld(worldBounds);

	// Sets the level bounds
	for(int i = 0; i < 3; i++)
	{
		levelBounds[i].MakeStatic();
		levelBounds[i].AttachToWorld(world);
	}
	
	const float w = bounds.Width();
	const float h = bounds.Height();
	const float hw = w / 2.0f;
	const float hh = h / 2.0f;
	
	const float d = 0.1f;
	const float hd = d / 2.0f;
	
	// Left wall
	levelBounds[0].SetPosition( Point2f(-hw - hd, 0.0f) );
	levelBounds[0].SetShape( CreatePhysicsShapeBox(d, h) );
	levelBounds[0].SetCollisionGroups(BOUNDS_COLLISION_GROUP, CG_ALL);
	
	// Right wall
	levelBounds[1].SetPosition( Point2f(hw + hd, 0.0f) );
	levelBounds[1].SetShape( CreatePhysicsShapeBox(d, h) );
	levelBounds[1].SetCollisionGroups(BOUNDS_COLLISION_GROUP, CG_ALL);
	
	// Ceiling
	levelBounds[2].SetPosition( Point2f(0.0f, hh + hd) );
	levelBounds[2].SetShape( CreatePhysicsShapeBox(w, d) );
	levelBounds[2].SetCollisionGroups(BOUNDS_COLLISION_GROUP, CG_ALL);
	
	// Sets up the paddle
	paddle = new DefaultPaddle(*gameSettings);
	paddle->AttachToWorld(world);

	paddle->SetBounds(AABBToRect<float>(bounds));
	paddle->SetPosition(settings->PaddleStartPosition());
		
	paddle->SetCollisionGroups(PADDLE_COLLISION_GROUP, BALL_COLLISION_GROUP | POWERUP_COLLISION_GROUP | BOUNDS_COLLISION_GROUP);
	
	// Sets up the ball
	ball = new Ball;
	ball->AttachToWorld(world);
	ball->SetShape( CreatePhysicsShapeCircle(settings->BallRadius()) );
	
	ball->SetImage(settings->BallImage());
	ball->SetRadius(settings->BallRadius());
	
	ball->SetPosition( Point2f(0.0f, worldBounds.min.y) );
	
	ball->SetCollisionGroups(BALL_COLLISION_GROUP, BRICK_COLLISION_GROUP | PADDLE_COLLISION_GROUP | BOUNDS_COLLISION_GROUP);
	
	// Sets up brick groups
	std::vector<BrickGroup::Attributes>::const_iterator i;
	for(i = attribs.brickGroupAttribs.begin(); i != attribs.brickGroupAttribs.end(); i++)
	{
		BrickGroup *bg = new BrickGroup(BRICK_COLLISION_GROUP, BALL_COLLISION_GROUP);
		bg->SetAttributes(*i);
		bg->Init(world);
		bg->SetListener(this);
		brickGroups.push_back(bg);
	}
	
	// Initializes the level data and objects
	brickExplosionChunks.drawPolicy.image = LoadImageResource("fire2.png");
	brickExplosionChunks.drawPolicy.blendMode = BLENDMODE_ADDITIVE_WITH_ALPHA;

	brickExplosionSparks.drawPolicy.image = LoadImageResource("smoke.png");
	brickExplosionSparks.drawPolicy.blendMode = BLENDMODE_ALPHA;
	
	powerUpGrabbed.drawPolicy.image = LoadImageResource("sparkle.png");
	powerUpGrabbed.drawPolicy.blendMode = BLENDMODE_ADDITIVE_WITH_ALPHA;
	
	arrow.Load(settings->LaunchArrowImage());
	
	font = LoadFontResource(settings->LevelFontFile(), settings->LevelFontSize());
	multiplierTimer->SetTimerLength(settings->MultiplierTime());
	multiplierTimer->SetFont( LoadFontResource(settings->MultiplierFontFile(), settings->MultiplierFontSize()) );
	multiplierTimer->ForceHide();
	
	fixTimer = 0.0f;
	fixTime = settings->StuckBallFixTime();
	
	pointsMultiplier = 1.0f;
	powerUpProb = settings->PowerUpProbability();
	
	launchVel = settings->BallLaunchSpeed();
	launchAngle = PI_2;
	launchAngleInc = -0.01f;
	launchBall = true;
	
	minBallVel = settings->BallMinSpeed();
	maxBallVel = settings->BallMaxSpeed();
	
	ball->SetAttachment(true);
	
	arrowFadeTimer = 0.0f;
	arrowFadeTime = settings->LaunchArrowFadeTime();
	
	ballFadeInTimer = settings->BallFadeInTime();
	ballFadeInTime = settings->BallFadeInTime();
	
	// Makes sure the arrow and ball are not drawn when the level is initially displayed
	arrow.SetTransparency(0.0f);
	ball->SetRadius(0.0f);
}

//--------------------------------------------------------------------
void Level::Update(float dt, const IKeyboard &kb, const IMouse &mouse)
{
	multiplierTimer->UpdateTimer(dt);
	if(multiplierTimer->TimerComplete())
	{
		pointsMultiplier = std::max(pointsMultiplier - 1.0f, 1.0f);
		if(pointsMultiplier > 1.0f)
		{
			multiplierTimer->SetMultiplier(pointsMultiplier);
			multiplierTimer->ResetTimer();
		}
		else
		{
			multiplierTimer->Hide();
		}
	}
	
	world->Update(dt);

	paddle->Update(dt, kb, mouse);
	ball->Update(dt);
	
	textParticles.Update(dt);

	brickExplosionChunks.Update(dt);
	brickExplosionSparks.Update(dt);
	powerUpGrabbed.Update(dt);
	
	if(launchBall)
	{
		ball->SetPosition( paddle->GetPosition() + Point2f(0.0f, (paddle->GetHeight() / 2.0f) + ball->GetRadius()) );
		
		// Fades in the ball and launch arrow
		if(ballFadeInTimer >= 0.0f)
		{
			ballFadeInTimer -= dt;
			
			float amt = 1.0f - (ballFadeInTimer / ballFadeInTime);
			arrow.SetTransparency(amt);
			arrow.SetHeightUniform(std::max(settings->LaunchArrowHeight() * amt, 0.0001f));
				
			float r = amt * settings->BallRadius();
			ball->SetRadius(r);
		}
		else
		{
			if(mouse.ButtonPressed(Keys::MOUSE_BUTTON_LEFT) || kb.KeyPressed(Keys::KEY_SPACE))
			{
				launchBall = false;
				float x = std::cos(launchAngle);
				float y = std::sin(launchAngle);
				ball->ApplyImpulse( launchVel * Vector2f(x, y) );
				
				arrowFadeTimer = arrowFadeTime;
				ballFadeInTimer = ballFadeInTime;
				
				ball->SetAttachment(false);
			}
			else
			{
				launchAngle += launchAngleInc;
				if( launchAngle < PI_4 || launchAngle > (3.0f*PI_4) )
					launchAngleInc = -launchAngleInc;
			}
			
			arrow.SetTransparency(1.0f);
		}
		
		Vector2f dir(std::cos(launchAngle), std::sin(launchAngle));
		arrow.SetAngle(launchAngle - PI_2);
		arrow.SetPosition(ball->GetPosition() + dir * (ball->GetRadius() + arrow.GetHeight() / 2.0f));
	}
	else // The ball is active
	{
		// Makes sure the ball doesn't get stuck
		fixTimer = std::min(fixTimer + dt, fixTime);
		if(fixTimer >= fixTime && (mouse.ButtonPressed(Keys::MOUSE_BUTTON_LEFT) || kb.KeyPressed(Keys::KEY_SPACE)))
		{
			// Make the fix small so as not to arouse suspicion and make sure its positive so we 
			// don't accedently screw the player and kill the ball
			ball->ApplyImpulse(15.0f*Vector2f(INV_SQRT2, INV_SQRT2));
			fixTimer = 0.0f;
		}
		
		if(paddle->IsBallAttached(ball))
			fixTimer = 0.0f;
		
		// Fades out the arrow after the ball is launched
		if(arrowFadeTimer > 0.0f)
		{
			arrowFadeTimer -= dt;
			
			float delta = arrowFadeTimer / arrowFadeTime;
			arrow.SetTransparency(delta);
			arrow.SetHeightUniform(std::max(20.0f * delta, 0.0001f));
		}
		
		if(BricksCleared())
		{
			// Slows the ball to a stop when the level is completed.
			Vector2f force = -2.0f * ball->GetVelocity();
			ball->ApplyForce(force);
		}
		else
		{
			// Enforces the minimum and maximum velocity
			Vector2f vel = ball->GetVelocity();
			float speed = Length(vel);
			
			if(!Close(speed, 0.0f))
			{
				if(speed < minBallVel)
					ball->SetVelocity( minBallVel * Normalize(vel) );
				else if(speed > maxBallVel)
					ball->SetVelocity( maxBallVel * Normalize(vel) );
			}
		}
		
		// Checks if the ball was lost
		if( ball->GetPosition().y < (bounds.min.y - (ball->GetRadius() * 2.0f)) )
		{
			// We don't consider the ball lost once all of the bricks are cleared.
			if(BricksCleared() || (listener && !listener->BallLost(*this)))
			{
				ball->SetPosition(2.0f*bounds.max);
			}
			else
			{
				launchBall = true;
				launchAngle = PI_2;
				arrowFadeTimer = 0.0f;
			
				ball->SetAttachment(true);
			}
			
			pointsMultiplier = 1.0f;
			multiplierTimer->Hide();
			
			paddle->Reset();
			
			// Since we reuse the ball after it dies it might have some velocity still 
			// associated with it, so we clear it out.
			ball->SetVelocity(Vector2f(0.0f, 0.0f));
		}
	}
	
	// Clean up unused data
	std::vector<BrickGroup*>::iterator bg = brickGroups.begin();
	while(bg != brickGroups.end())
	{
		(*bg)->Update(dt, *ball);
		
		if( (*bg)->GetBrickCount() <= 0 )
		{
			delete *bg;
			bg = brickGroups.erase(bg);
		}
		else
		{
			bg++;
		}
	}
	
	std::vector<PowerUp*>::iterator pu = powerUps.begin();
	while(pu != powerUps.end())
	{
		(*pu)->Draw();
		PowerUp *pill = *pu;
		
		if(!pill->IsActive() || pill->GetPosition().y < (bounds.min.y - 5.0f))
		{
			delete pill;
			pu = powerUps.erase(pu);
		}
		else
		{
			pu++;
		}
	}
}

//--------------------------------------------------------------------
bool Level::IsComplete() const
{
	return BricksCleared() && 
	       brickExplosionChunks.GetParticleCount() == 0 && 
	       brickExplosionSparks.GetParticleCount() == 0;
}

//--------------------------------------------------------------------
bool Level::BricksCleared() const
{
	bool brickGroupsDone = true;
	for(std::vector<BrickGroup*>::const_iterator i = brickGroups.begin(); i != brickGroups.end(); i++)
	{
		const BrickGroup &b = **i;
		if(!(b.HasNegativePoints() || b.IsInvincible()) && b.GetBrickCount() > 0)
		{
			brickGroupsDone = false;
			break;
		}
	}
	
	return brickGroupsDone;
}

//--------------------------------------------------------------------
void Level::PowerUpGrabbed(PowerUp &p, const PhysicsCollisionInfo &hitInfo)
{
	if(!p.IsActive())
		return;
	
	ApplyPowerUpAbility(p.GetAbility(), hitInfo.contacts[0].position);
	
	powerUpGrabbed.initPolicy.hitNorm = hitInfo.contacts[0].normal;
	powerUpGrabbed.Emit(hitInfo.contacts[0].position, 50);
	
	p.Deactivate();
}

//--------------------------------------------------------------------
void Level::ApplyPowerUpAbility(PowerUp::Ability a, const Point2f &p)
{
	switch(a) 
	{
		case PowerUp::ABILITY_SHRINK:
			textParticles.Emit("Shrink", p, Color::White);
			paddle->Shrink();
			break;
			
		case PowerUp::ABILITY_GROW:
			textParticles.Emit("Expand", p, Color::White);
			paddle->Grow();
			break;
			
		case PowerUp::ABILITY_STICKY:
			textParticles.Emit("Sticky", p, Color::White);
			paddle->SetSticky(true);
			break;
			
		case PowerUp::ABILITY_EXTRALIFE:
			textParticles.Emit("+1 Life", p, Color::White);
			if(listener)
				listener->ExtraBall(*this);
			break;
			
		case PowerUp::ABILITY_POINTS:
			{
				int points = (int)(pointsMultiplier * 50.0f);
			
				std::stringstream ss;
				ss << "+" << points;
				
				textParticles.Emit(ss.str().c_str(), p, Color::White);
				
				if(listener)
					listener->ScoreIncrease(*this, points);
			}
			break;
			
		case PowerUp::ABILITY_MULTIPLIER:
			textParticles.Emit("Multiplier", p, Color::White);
			pointsMultiplier += 1.0f;
			
			multiplierTimer->SetMultiplier(pointsMultiplier);
			multiplierTimer->ResetTimer();
			multiplierTimer->Display();
			break;
			
		case PowerUp::ABILITY_RANDOM:
			
			// The powerups GetAbility() method should never return the random ability
			assert(true);
			break;
			
		case PowerUp::ABILITY_NONE:
		default:
			break;
	}
}

//--------------------------------------------------------------------
void Level::BrickHit(BrickGroup &group, Brick &brick, const PhysicsCollisionInfo &hitInfo, int points)
{
	// We won't worry about fixing a stuck ball if its hitting a non-invincable brick
	if(!group.IsInvincible())
		fixTimer = 0.0f;
		
	points = (int)(points * pointsMultiplier);
	
	if(points != 0 && listener)
		listener->ScoreIncrease(*this, points);

	if(brick.life == 0)
	{
		brickExplosionChunks.Emit(brick.GetPosition(), 50);
		brickExplosionSparks.Emit(brick.GetPosition(), 50);

		if(points > 0)
		{
			float r = RandomFloat(0.0f, 1.0f);
			if(r <= powerUpProb)
			{			
				PowerUp *pill = new PowerUp(this, world, brick.GetPosition());
				powerUps.push_back(pill);
			}
		}
	}
	
	// Creates the score particle
	if(points != 0)
	{
		Color4f textColor = Color::White;
		std::stringstream ss;
	
		if(points > 0)
		{
			ss << "+";
		}
		else
		{
			ss << "-";
			textColor = Color::Red;
		}
		
		ss << points;
		std::string str = ss.str();
		
		FontRef font = textParticles.GetFont();
		
		float w, h;
		font->StringDimensions(str.c_str(), w, h);
		
		Point2f pos = hitInfo.contacts[0].position;
		pos.x = pos.x - (w / 2.0f);
		pos.y = pos.y + (h / 2.0f);
		
		textParticles.Emit(str.c_str(), pos, textColor);
	}
}

//--------------------------------------------------------------------
void Level::Draw()
{
	DrawAABB(bounds, attribs.bgColor, true);
	
	std::vector<BrickGroup*>::iterator bg;
	for(bg = brickGroups.begin(); bg != brickGroups.end(); bg++)
	{
		(*bg)->Draw();
	}
	
	std::vector<PowerUp*>::iterator pu;
	for(pu = powerUps.begin(); pu != powerUps.end(); pu++)
	{
		(*pu)->Draw();
	}
	
	if(launchBall || arrowFadeTimer >= 0.0f)
	{
		arrow.Draw();
	}
	
	paddle->Draw();
	ball->Draw();

	brickExplosionChunks.Draw();
	brickExplosionSparks.Draw();
	powerUpGrabbed.Draw();
	
	textParticles.Draw();
}

//--------------------------------------------------------------------
void Level::DrawGui()
{
	viewport.Draw();
}
