/*!
 * \file Level.h
 * \brief Declares the Level class.
 * \author Joe Forte
 * \date 3/27/2008
 */

#ifndef LEVEL_H
#define LEVEL_H

#include <vector>
#include <string>
#include "Math/MathCommon.h"
#include "Graphics/Sprite.h"
#include "Graphics/Font.h"
#include "Graphics/ParticleSystem.h"
#include "Physics/PhysicsWorld.h"
#include "GUI/GUIViewport.h"
#include "Game/GUIMultiplierTimer.h"
#include "Game/Ball.h"
#include "Game/Paddle.h"
#include "Game/PowerUp.h"
#include "Game/BrickGroup.h"
#include "Game/TextParticleSystem.h"

class Screen;
class IKeyboard;
class IMouse;
class GameSettings;

class Level;

const CollisionGroup BOUNDS_COLLISION_GROUP  = CG_GROUP2; /*!< The collision group the level bounds belongs to. */
const CollisionGroup PADDLE_COLLISION_GROUP  = CG_GROUP2; /*!< The collision group the paddle belongs to. */
const CollisionGroup BALL_COLLISION_GROUP    = CG_GROUP3; /*!< The collision group the ball belongs to. */
const CollisionGroup BRICK_COLLISION_GROUP   = CG_GROUP4; /*!< The collision group the bricks belong to. */
const CollisionGroup POWERUP_COLLISION_GROUP = CG_GROUP5; /*!< The collision group the powerups belong to. */

/*!
 * \class LevelListener
 * \headerfile Level.h "Game/Level.h"
 * \brief base class for listening to level events.
 */
class LevelListener
{
public:

	/*!
	 * \brief Called when a ball is lost.
	 *
	 * \param[in] level referance to the level that called the method.
	 *
	 * \return True if you wish to have another ball spawned.  Return false to not have another ball spawned.
	 */
	virtual bool BallLost(Level &level) = 0;
	
	/*!
	 * \brief Called when an extra ball is aquired.
	 *
	 * \param[in] level referance to the level that called the method.
	 */
	virtual void ExtraBall(Level &level) = 0;

	/*!
	 * \brief Called when the user earns some points.
	 *
	 * \param[in] level referance to the level that called the method.
	 * \param[in] score the score that the user aquired.
	 */
	virtual void ScoreIncrease(Level &level, int score) = 0;
};

/*!
 * \class BrickExplodePolicy
 * \headerfile Paddle.h "Game/Paddle.h"
 * \brief Policy that creates the Brick explosion ParticleSystem.
 */
class BrickExplodePolicy
{
public:
	class SparksInit
	{
	public:
		inline void PreEmit(const Point2f &pos, uint amount) { position = pos; }
		
		inline void InitParticle(Particle &p)
		{
			p.pos = position;
			p.color = Color::White * RandomFloat(0.1f, 1.0f);
			p.color.a = 1.0f;
			p.life = RandomFloat(0.5f, 1.0f);
			p.age = 0.0f;
			p.angularVel = 0.0f;
			
			float size = RandomFloat(5.0f, 50.0f);
			p.width = size;
			p.height = size;
			
			float a = RandomFloat(0.0f, _2PI);
			p.angle = a - PI_2;
			
			Vector2f vel(std::cos(a), std::sin(a));
			
			p.vel = RandomFloat(50.0f, 150.0f) * vel;
		}
		
	private:
		Point2f position;
	};
	
	class ChunksInit
	{
	public:
		inline void PreEmit(const Point2f &pos, uint amount) { position = pos; }

		inline void InitParticle(Particle &p)
		{
			p.pos = position;
			p.color = Color::White * RandomFloat(0.2f, 0.5f);
			p.color.a = 1.0f;
			p.life = RandomFloat(0.2f, 0.4f);
			p.age = 0.0f;
			p.angle = 0.0f;
			p.angularVel = 0.0f;
			
			float size = RandomFloat(5.0f, 50.0f);
			p.width = size;
			p.height = size;
			
			float angle = RandomFloat(0.0f, _2PI);
			Vector2f vel(std::cos(angle), std::sin(angle));
			
			p.vel = RandomFloat(20.0f, 100.0f) * vel;
		}
		
	private:
		Point2f position;
	};

	class Update
	{
	public:
		inline void PreUpdate(float dt) {}

		inline bool ApplyUpdate(Particle &p, float dt)
		{
			p.pos = p.pos + p.vel*dt;
			p.angle = p.angle + p.angularVel*dt;
			
			p.age += dt;
			
			p.color.a = 1.0f - (p.age / p.life);
			
			return p.age < p.life;
		}
	};
};

/*!
 * \class PowerUpGrabbedPolicy
 * \headerfile Paddle.h "Game/Paddle.h"
 * \brief Policy that creates the particle effect when power ups are grabbed by the paddle.
 */
class PowerUpGrabbedPolicy
{
public:
	
	class Init
	{
	public:
		Vector2f hitNorm; // The normal where the power up hit the paddle
		
		inline void PreEmit(const Point2f &pos, uint amount) { position = pos; }
		
		inline void InitParticle(Particle &p)
		{
			p.pos = position;
			p.color = Color::White * RandomFloat(0.1f, 1.0f);
			p.color.a = 1.0f;
			p.life = RandomFloat(0.5f, 1.0f);
			p.age = 0.0f;
			p.angularVel = 0.0f;
			
			float size = RandomFloat(5.0f, 15.0f);
			p.width = size;
			p.height = size;
			
			Vector2f d = hitNorm.Perp();
			p.vel = RandomFloat(-100.0f, 100.0f) * d + RandomFloat(0.0f, 100.0f) * hitNorm;
		}
		
	private:
		Point2f position;
	};
	
	class Update
	{
	public:
		inline void PreUpdate(float dt) {}
		
		inline bool ApplyUpdate(Particle &p, float dt)
		{
			p.pos = p.pos + p.vel*dt;
			
			p.age += dt;
			
			p.color.a = 1.0f - (p.age / p.life);
			
			return p.age < p.life;
		}
	};
};

/*!
 * \class Level
 * \headerfile Level.h "Game/Level.h"
 * \brief The class that represents a single BrickOut level.
 */
class Level : public BrickGroupListener
{
public:
	Level();
	~Level();
	
	struct Attributes
	{
		// Level attributes
		std::string levelName;
		Color4f bgColor;
		
		// BrickGroups
		std::vector<BrickGroup::Attributes> brickGroupAttribs;
	};
	
	void SetAttributes(const Level::Attributes &attributes);

	void Init(const GameSettings *gameSettings);
	
	void Destroy();
	
	void Update(float dt, const IKeyboard &kb, const IMouse &mouse);

	const Paddle &GetPaddle() const { return *paddle; }
	
	const char *GetName() const { return attribs.levelName.c_str(); }
	
	bool IsComplete() const;
	
	void Draw();
	
	/*! 
	 * \brief Called by the app after all other level related drawing is done.
	 */
	void DrawGui();

	void SetListener(LevelListener *lsn) { listener = lsn; }
	
	/*!
	 * \brief Called when a powerup is grabbed by the paddle.
	 *
	 * \param[in] p the powerup that was grabbed.
	 * \param[in] hitInfo the hit information from the collision between the powerup and the paddle.
	 *
	 * NOTE: Do not destroy the powerup in this function.
	 */
	void PowerUpGrabbed(PowerUp &p, const PhysicsCollisionInfo &hitInfo);

	/*!
	 * \brief Called each time a brick is hit.
	 *
	 * \param[in] group referance to the appropriate BrickGroup.
	 * \param[in] brick referance to the Brick that was hit.
	 * \param[in] info the collision information for the brick hit.
	 * \param[in] points the number of points awarded for the hit.
	 */
	virtual void BrickHit(BrickGroup &group, Brick &brick, const PhysicsCollisionInfo &hitInfo, int points);
	
private:
	
	/*!
	 * \brief Called when a powerup ability is used.
	 *
	 * \param[in] a the powerup ability.
	 * \param[in] p the position where the ability is applied.
	 */
	void ApplyPowerUpAbility(PowerUp::Ability a, const Point2f &p);
	
	/*!
	 * \brief Determines if all bricks in the levels have been destroyed. 
	 *
	 * \return true if all bricks are destroyed, false otherwise.
	 */
	bool BricksCleared() const;
	
	const GameSettings *settings;
	
	bool launchBall;
	float launchVel;
	float launchAngle;
	float launchAngleInc;
	
	float minBallVel;
	float maxBallVel;

	Sprite arrow;
	float arrowFadeTimer;
	float arrowFadeTime;
	
	float powerUpProb; // The probability a powerup is created
	
	float pointsMultiplier;
	
	// Used to fix the ball if it gets stuck.
	float fixTimer;
	float fixTime;
	
	AABB bounds;

	PhysicsWorld *world;

	PhysicsObject levelBounds[3];
	
	Paddle *paddle;
	Ball *ball;
	
	float ballFadeInTimer;
	float ballFadeInTime;
	
	FontRef font;
	
	GUIViewport viewport;
	GUIMultiplierTimer *multiplierTimer;
	
	std::vector<BrickGroup*> brickGroups;
	std::vector<PowerUp*> powerUps;
	
	Attributes attribs;

	LevelListener *listener;
	
	ParticleSystem<200, BrickExplodePolicy::ChunksInit, BrickExplodePolicy::Update> brickExplosionChunks;
	ParticleSystem<200, BrickExplodePolicy::SparksInit, BrickExplodePolicy::Update> brickExplosionSparks;
	
	ParticleSystem<200, PowerUpGrabbedPolicy::Init, PowerUpGrabbedPolicy::Update> powerUpGrabbed;
	
	TextParticleSystem textParticles;
	
	static void TextParticleUpdateFunc(TextParticle &p, float dt)
	{
		const float fadeInTime = 0.10f;
		const float fadeOutTime = 0.25f;
		
		float endTime = p.lifeTime - fadeOutTime;
		if(p.age > endTime)
		{
			p.position.y -= 75.0f * dt;
			p.color.a = (p.lifeTime - p.age) / fadeOutTime;
		}
		else
		{
			if(p.age <= fadeInTime)
			{
				p.color.a = p.age / fadeInTime;
			}
			
			p.position.y += 25.0f * dt;
		}
	}
};

#endif
