/*!
 * \file LevelParser.h
 * \brief Implements the level parser class.
 * \author Joe Forte
 * \date 3/27/2008
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include "Common/Globals.h"
#include "Application/DataPaths.h"
#include "Game/LevelParser.h"
#include "Game/Level.h"
#include "Math/MathCommon.h"

#include "tinyxml.h"

enum LEVEL_FILE_DIRECTIVE
{
	// Top level level file directives
	DIRECTIVE_LOAD = 0,
	DIRECTIVE_BRICKGROUP,
	DIRECTIVE_LEVEL_ATTRIBUTES,

	// Level attribute directives
	DIRECTIVE_BACKGROUND_COLOR,

	// BrickGroup directives
	DIRECTIVE_BRICK_ATTRIBUTES,
	DIRECTIVE_DISTROBUTION,

	DIRECTIVE_UNKNOWN
};


// Top level level file directives
const std::string LOAD_STR              = "Load";
const std::string BRICKGROUP_STR		= "BrickGroup";
const std::string LEVEL_ATTRIBUTES_STR  = "LevelAttributes"; 

// Level attribute directives
const std::string BACKGROUND_COLOR_STR  = "BackgroundColor";

// BrickGroup directives
const std::string BRICK_ATTRIBUTES_STR	= "BrickAttributes";
const std::string DISTROBUTION_STR		= "Distrobution";

// Distrobution strings
const std::string POINT_STR				= "Point";

//--------------------------------------------------------------------
LEVEL_FILE_DIRECTIVE GetDirective(const std::string &str)
{
	if(str == LOAD_STR)                 return DIRECTIVE_LOAD;
	if(str == BRICKGROUP_STR)			return DIRECTIVE_BRICKGROUP;
	if(str == BACKGROUND_COLOR_STR)		return DIRECTIVE_BACKGROUND_COLOR;
	if(str == LEVEL_ATTRIBUTES_STR)		return DIRECTIVE_LEVEL_ATTRIBUTES;
	if(str == BRICK_ATTRIBUTES_STR)		return DIRECTIVE_BRICK_ATTRIBUTES;
	if(str == DISTROBUTION_STR)			return DIRECTIVE_DISTROBUTION;
	
	return DIRECTIVE_UNKNOWN;
}

//--------------------------------------------------------------------
bool ParsePointsDistrobution(LevelParser &parser, TiXmlElement *elm, BrickGroup::Attributes &att)
{
	TiXmlNode *curr = elm->FirstChild();
	while( curr )
	{
		if( curr->Type() == TiXmlNode::ELEMENT )
		{
			TiXmlElement *e = curr->ToElement();
			if( e->Value() == POINT_STR )
			{
				Point2f p;
				e->QueryFloatAttribute("x", &p.x);
				e->QueryFloatAttribute("y", &p.y);
				att.points.push_back(p);
			}
		}
		else
		{
			ConWarning() << "Unknown directive '" << curr->Value() << "' under '" << DISTROBUTION_STR << "'\n";
		}
	
		curr = elm->IterateChildren(curr);
	}
	
	return true;
}

//--------------------------------------------------------------------
bool ParseDistrobution(LevelParser &parser, TiXmlElement *elm, BrickGroup::Attributes &att)
{
	return ParsePointsDistrobution(parser, elm, att);
}

//--------------------------------------------------------------------
bool ParseBrickAttributes(LevelParser &parser, TiXmlElement *elm, BrickGroup::Attributes &att)
{
	int hp;
	if( !elm->QueryIntAttribute("hp", &hp) )
		att.brickHP = hp;
	else
		att.brickHP = 1;
	
	int pointsPerHit;
	if( !elm->QueryIntAttribute("pointsPerHit", &pointsPerHit) )
		att.pointsPerHit = pointsPerHit;
	else
		att.pointsPerHit = 0;
	
	int pointsPerKill;
	if( !elm->QueryIntAttribute("pointsPerKill", &pointsPerKill) )
		att.pointsPerKill = pointsPerKill;
	else
		att.pointsPerKill = 0;
	
	int width;
	if( !elm->QueryIntAttribute("width", &width) )
		att.brickWidth = width;
	else
		att.brickWidth = 64;
	
	int height;
	if( !elm->QueryIntAttribute("height", &height) )
		att.brickHeight = height;
	else
		att.brickHeight = 16;
	
	int invincible;
	if( !elm->QueryIntAttribute("invincible", &invincible) )
		att.invincible = (invincible != 0);
	else
		att.invincible = false;
	
	const char *str;
	str = elm->Attribute("image0");
	if(str)
		att.imageFiles.push_back( std::string(str) );
	else
		att.imageFiles.push_back( std::string("brick_blue.bmp") );
	
	for(int i = 1; i < att.brickHP; i++)
	{
		std::stringstream s;
		s << "image" << i;
		
		str = elm->Attribute(s.str().c_str());
		if(str)
			att.imageFiles.push_back( std::string(str) );
		else
			att.imageFiles.push_back( att.imageFiles.back().c_str() );
	}
	
	return true;
}

//--------------------------------------------------------------------
bool ParseBrickGroup(LevelParser &parser, TiXmlElement *elm, Level::Attributes &att)
{
	BrickGroup::Attributes bga;
	
	const char *str;
	str = elm->Attribute("name");
	if(str)
		bga.name = str;
	else
		bga.name = "Unspecifed";
	
	TiXmlNode *curr = elm->FirstChild();
	while( curr )
	{
		if( curr->Type() == TiXmlNode::ELEMENT )
		{
			TiXmlElement *e = curr->ToElement();
			if(e->Value() == BRICK_ATTRIBUTES_STR)
			{
				if( !ParseBrickAttributes(parser, e, bga) )
					return false;
			}
			else if(e->Value() == DISTROBUTION_STR)
			{
				if( !ParseDistrobution(parser, e, bga) )
					return false;
			}
			else
			{
				ConWarning() << "Unknown directive '" << curr->Value() << "' under '" << BRICKGROUP_STR << "'\n";
			}
		}
		
		curr = elm->IterateChildren(curr);
	}

	att.brickGroupAttribs.push_back(bga);
	
	return true;
}

//--------------------------------------------------------------------
bool ParseLevelAttributes(LevelParser &parser, TiXmlElement *elm, Level::Attributes &att)
{
	const char *str;
	str = elm->Attribute("name");
	if(str)
		att.levelName = str;
	else
		att.levelName = "Unspecified Level Name";

	TiXmlNode *curr = elm->FirstChild();
	while( curr )
	{
		if( curr->Type() == TiXmlNode::ELEMENT )
		{
			TiXmlElement *e = curr->ToElement();
			if(e->Value() == BACKGROUND_COLOR_STR)
			{
				int r, g, b;
				e->QueryIntAttribute("r", &r);
				if(r < 0 || r > 255)
				{
					ConWarning() << "The 'r' attribute of the '" << BACKGROUND_COLOR_STR << "' directive under '" 
								 << LEVEL_ATTRIBUTES_STR << "' is not in the proper range [0,255]."; 
					r = 0;
				}
				
				e->QueryIntAttribute("g", &g);
				if(g < 0 || g > 255)
				{
					ConWarning() << "The 'g' attribute of the '" << BACKGROUND_COLOR_STR << "' directive under '" 
								 << LEVEL_ATTRIBUTES_STR << "' is not in the proper range [0,255]."; 
					g = 0;
				}
				
				e->QueryIntAttribute("b", &b);
				if(b < 0 || b > 255)
				{
					ConWarning() << "The 'b' attribute of the '" << BACKGROUND_COLOR_STR << "' directive under '"
								 << LEVEL_ATTRIBUTES_STR << "' is not in the proper range [0,255]."; 
					b = 0;
				}
				
				att.bgColor = Color4f((byte)r, (byte)g, (byte)b);
			}
			else
			{
				ConWarning() << "Unknown directive '" << curr->Value() << "' under '" << LEVEL_ATTRIBUTES_STR << "'\n";
			}
		}

		curr = elm->IterateChildren(curr);
	}

	return true;
}

//--------------------------------------------------------------------
bool ParseLoad(LevelParser &parser, TiXmlElement *elm, Level::Attributes &att)
{
	const char *file;
	file = elm->Attribute("file");
	if(!file)
	{
		ConError() << "No 'file' attribute specified with the 'Load' command.\n";
		return false;
	}

	std::string path = ConcatPath(DataPaths::GetPath("Levels"), file);
	return parser.ParseAttrbutes(path.c_str(), att);
}

//--------------------------------------------------------------------
void LevelParser::Init()
{
	directiveParserMap[LOAD_STR]                = ParseLoad;
	directiveParserMap[BRICKGROUP_STR]			= ParseBrickGroup;
	directiveParserMap[LEVEL_ATTRIBUTES_STR]	= ParseLevelAttributes;
}

//--------------------------------------------------------------------
bool LevelParser::LoadLevel(const char *file, Level &lvl)
{
	Level::Attributes att;
	if( !ParseAttrbutes(file, att) )
		return false;
		
	lvl.SetAttributes(att);
	
	return true;
}

//--------------------------------------------------------------------
bool LevelParser::ParseAttrbutes(const char *file, Level::Attributes &att)
{
	TiXmlDocument doc(file);
	if( !doc.LoadFile() )
	{
		ConError() << "Level file '" << file << "' could not be parsed.\n";
		return false;
	}
	
	TiXmlHandle hDoc(&doc);
	
	TiXmlNode *Root = hDoc.FirstChildElement().Node();
	if( !Root || Root->ValueStr() != "Level")
		return false;
	
	TiXmlNode *curr = Root->FirstChild();
	
	while(curr)
	{
		if(curr->Type() == TiXmlNode::ELEMENT)
		{
			TiXmlElement *elm = curr->ToElement();
			
			std::map<std::string, fnLevelFileDirectiveParser>::iterator i;
			i = directiveParserMap.find(curr->Value());
			if(i != directiveParserMap.end())
			{
				fnLevelFileDirectiveParser func = i->second;
				if( !func(*this, elm, att) )
				{
					ConError() << "Level directive '" << curr->Value() << "' is invalid or incomplete.\n";
					return false;
				}
			}
			else
			{
				LEVEL_FILE_DIRECTIVE cmd = GetDirective(curr->Value());
				switch(cmd)
				{
					/* these cases should not appear at the top level */
					case DIRECTIVE_DISTROBUTION:
					case DIRECTIVE_BACKGROUND_COLOR:
					case DIRECTIVE_BRICK_ATTRIBUTES:
						ConWarning() << "Level file directive '" << curr->Value() << "' found at incorrect scope\n";
						break;
						
					case DIRECTIVE_UNKNOWN:
					default:
						ConWarning() << "Unknown level file directive '" << curr->Value() << "'\n";
				};
			}
		}
		
		curr = Root->IterateChildren(curr);
	}
	
	return true;
}
