/*!
 * \file MusicVisualizerEffect.h
 * \brief Declares the MusicVisualizerEffect background effect for visualizing music and sound.
 * \author Joe Forte
 * \date 2/11/2009
 */

#ifndef MUSICVISUALIZEREFFECT_H
#define MUSICVISUALIZEREFFECT_H

#include "Sound/Sound.h"
#include "Math/MathCommon.h"
#include "Graphics/ParticleSystem.h"
#include "Game/BackgroundEffect.h"

/*!
 * \class SmokeWallPolicy
 * \headerfile MusicVisualizerEffect.h "Game/MusicVisualizerEffect.h"
 * \brief Policy that creates a wall of upward moving smoke.
 */
class SmokeWallPolicy
{
public:
	
	typedef ContinuousEmitPolicy<> Emit;

	class Init
	{
	public:
		Color4f color;
		Line spawnLine;
			
		inline void PreEmit(const Point2f &pos, uint amount) {}
			
		inline void InitParticle(Particle &p)
		{
			p.pos = lerp(spawnLine.a, spawnLine.b, RandomFloat(0.0f, 1.0f));
			p.color = color;
			p.life = RandomFloat(0.5f, 2.0f);
			p.age = 0.0f;
			p.angle = 0.0f;
			p.angularVel = 0.0f;
				
			float size = RandomFloat(70.0f, 95.0f);
			p.width = size;
			p.height = size;
			
			p.vel = Vector2f(0.0f, 1.0f) * RandomFloat(100.0f, 250.0f);
			
			//p.vel = Normalize(Vector2f(RandomFloat(-1.0, 1.0f), 1.0f))  * RandomFloat(100.0f, 250.0f);
		}
	};
		
	class Update
	{
	public:
		inline void PreUpdate(float dt) {}
		
		inline bool ApplyUpdate(Particle &p, float dt)
		{
			p.pos = p.pos + p.vel*dt;
				
			p.age += dt;
			
			float amt = p.width * dt;
			p.width += amt;
			p.height += amt;
			
			p.color.a = 1.0f - (p.age / p.life);
			
			return p.age < p.life;
		}
	};
};

/*!
 * \class SmokeFountainPolicy
 * \headerfile MusicVisualizerEffect.h "Game/MusicVisualizerEffect.h"
 * \brief Policy that creates a quick fountain of upward moving smoke.
 */
class SmokeFountainPolicy
{
public:
	
	class Init
	{
	public:
		Color4f color;
		float power;
			
		inline void PreEmit(const Point2f &pos, uint amount) { position = pos; }
			
		inline void InitParticle(Particle &p)
		{
			p.pos = position;
			p.color = color;
			p.life = RandomFloat(1.5f, 2.0f);
			p.age = 0.0f;
			p.angularVel = 0.0f;
			
			float size = RandomFloat(40.0f, 75.0f);
			p.width = size;
			p.height = size;
			
			float max = lerp(50.0f, 450.0f, power);
			
			p.vel = Vector2f(0.0f, 1.0f) * RandomFloat(50.0f, max);
		}
			
	private:
		Point2f position;
	};
		
		class Update
		{
		public:
			inline void PreUpdate(float dt) {}
			
			inline bool ApplyUpdate(Particle &p, float dt)
			{
				p.pos = p.pos + p.vel*dt;
				
				p.age += dt;
				
				p.color.a = 1.0f - (p.age / p.life);
				
				return p.age < p.life;
			}
		};
	};


/*!
 * \class MusicVisualizerEffect
 * \headerfile MusicVisualizerEffect.h "Game/MusicVisualizerEffect.h"
 * \brief Music visualizer background effect.
 */
class MusicVisualizerEffect : public BackgroundEffect
{
public:
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] size the that the effect will fill.
	 */
	MusicVisualizerEffect(const Rectf &size) : BackgroundEffect(size) {}
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~MusicVisualizerEffect() {}
	
	/*!
	 * \brief Initializes the effect. Called once before Update and Draw are called.
	 *
	 * \return true if the initialization was successful, false otherwise.
	 */
	virtual bool Init();
	
	/*!
	 * \brief Updates the effect.  Called once per frame.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 */
	virtual void Update(float dt);
	
	/*!
	 * \brief Draws the effect.
	 */
	virtual void Draw();
	
protected:

	struct Bar
	{
		float val;
	};

	void DrawBar(const Bar &b, const Point2f &p);

	static const int SPECTRUM_SIZE = 42;

	Bar bars[SPECTRUM_SIZE];

	float barSize;
	float barPadding;
	float barMaxHeight;
	float barMinHeight;
	
	ParticleSystem<250, SmokeWallPolicy::Init, SmokeWallPolicy::Update, SmokeWallPolicy::Emit> smokeWall;
	//ParticleSystem<200, SmokeFountainPolicy::Init, SmokeFountainPolicy::Update> fountain;
};

#endif
