/*!
 * \file Paddle.cpp
 * \brief Implements the Paddle class.
 * \author Joe Forte
 * \date 3/18/2008
 */

#include "Application/DataPaths.h"
#include "Input/SDLKeyboard.h"
#include "Input/Mouse.h"
#include "Game/ConsoleVars.h"
#include "Game/Paddle.h"
#include "Game/GameSettings.h"
#include "Common/Globals.h"

//--------------------------------------------------------------------
Paddle::Paddle() : sticky(false), attachedBall(NULL)
{
	MakeStatic();
}

//--------------------------------------------------------------------
bool Paddle::Touch(PhysicsObject *obj, const PhysicsCollisionInfo &hitInfo)
{
	if(obj->GetObjectType() == POBJ_BALL)
	{
		AttachBall(obj, hitInfo);
	}
	
	return true;
}

//--------------------------------------------------------------------
void Paddle::SetImage(const char *file)
{
	board.Load(file);
}

//--------------------------------------------------------------------
void Paddle::SetStickyImage(const char *file)
{
	stickyBoard.Load(file);
}

//--------------------------------------------------------------------
void Paddle::SetBounds(const Rectf &r)
{	
	bounds = RectToAABB(r);
}

//--------------------------------------------------------------------
void Paddle::SetWidth(float w)
{
	// Scales the physics shape
	float width = board.GetWidth();

	Matrixf S;
	S.MakeScale(w / width, 1.0f);	
	GetShapes().MultiplyTransform(S);
	
	board.SetWidth(w);
	stickyBoard.SetWidth(w);
}

//--------------------------------------------------------------------
float Paddle::GetHeight() const
{
	return board.GetHeight();
}

//--------------------------------------------------------------------
void Paddle::Set(int w, int h)
{
	board.SetSize((float)w, (float)h);
	stickyBoard.SetSize((float)w, (float)h);
}

//--------------------------------------------------------------------
void Paddle::Move(const Point2f &amt)
{
	SetPosition( GetPosition() + amt );
}

//--------------------------------------------------------------------
AABB Paddle::GetBBox()
{
	return board.GetBBox();
}

//--------------------------------------------------------------------
void Paddle::SetPosition(const Point2f &p)
{
	PhysicsObject::SetPosition(p);

	board.SetPosition(p);
	stickyBoard.SetPosition(p);
	
	if(attachedBall)
	{
		attachedBall->SetPosition(p + ballAttachOffset);
	}
}

//--------------------------------------------------------------------
void Paddle::AddAttachment(const Point2f &pos, const Sprite &att)
{
	Attachment attach;
	attach.offset = pos;
	attach.attachment = att;
	
	attachments.push_back(attach);
}

//--------------------------------------------------------------------
void Paddle::SetSticky(bool s)
{
	if(sticky && !s)
		DetachBall();
	
	sticky = s;
}

//--------------------------------------------------------------------
void Paddle::AttachBall(PhysicsObject *ball, const PhysicsCollisionInfo &hitInfo)
{
	if(IsSticky() && ball->GetObjectType() == POBJ_BALL)
	{
		// We only grab the ball if it touches the top part of the paddle
		Vector2f n = hitInfo.contacts[0].normal;
		
		if(n.y > 0.0f)
		{
			// Computes the direction and velocity of the ball that will be used when the ball is detached.
			Vector2f v = ball->GetVelocity();
			launchVel = -(n.Reflect(v));
			
			// If this is the case then the ball cannot be saved so theres no point in delaying the inevitable.
			if(launchVel.y < 0.0f)
				return;
			
			ballAttachOffset = ball->GetPosition() -GetPosition();
			attachedBall = ball;
			
			OnAttachBall();
		}
	}
}

//--------------------------------------------------------------------
void Paddle::DetachBall()
{
	if(attachedBall)
	{
		attachedBall->SetVelocity(launchVel);
		attachedBall = NULL;
		
		OnDetachBall();
	}	
}

//--------------------------------------------------------------------
void Paddle::Reset()
{
	DetachBall();
	SetSticky(false);
}

//--------------------------------------------------------------------
void Paddle::Update(float dt, const IKeyboard &kb, const IMouse &mouse)
{
	Point2i delta = mouse.GetMoveDelta();
	Point2f dm = Point2f((float)delta.x, 0.0f);

	Point2f nextPos = GetPosition() + dm;

	// translates the bounding box to the next position.
	AABB bbox = board.GetBBox();
	bbox.max += dm;
	bbox.min += dm;

	if( bbox.min.x < bounds.min.x )
	{
		nextPos += Point2f(bounds.min.x - bbox.min.x, 0.0f);
	}
	else if( bbox.max.x > bounds.max.x )
	{
		nextPos += Point2f(bounds.max.x - bbox.max.x, 0.0f);
	}

	if(attachedBall)
	{
		AABB b = attachedBall->GetShapes().GetBoundingBox();
		
		if( b.min.x < bounds.min.x )
		{
			nextPos += Point2f(bounds.min.x - b.min.x, 0.0f);
		}
		else if( b.max.x > bounds.max.x )
		{
			nextPos += Point2f(bounds.max.x - b.max.x, 0.0f);
		}
	}
	
	SetPosition(nextPos);
	
	if(mouse.ButtonDown(Keys::MOUSE_BUTTON_LEFT) || kb.KeyDown(Keys::KEY_SPACE))
		DetachBall();
}

//--------------------------------------------------------------------
void Paddle::Draw()
{
	if(sticky && stickyBoard.CanDraw())
		stickyBoard.Draw();
	else
		board.Draw();
	
	Point2f pos = GetPosition();
	for(std::vector<Attachment>::iterator i = attachments.begin(); i != attachments.end(); i++)
	{
		i->attachment.SetPosition(pos + i->offset);
		i->attachment.Draw();
	}
	
	if(phy_drawCollisionShapes->GetInt())
		GetShapes().DrawOutline();
}

//--------------------------------------------------------------------
DefaultPaddle::DefaultPaddle(const GameSettings &settings)
{
	const char *SMOKE_IM = "smoke.png";
	const char *FIRE_IM  = "fire2.png";
	
	defWidth = settings.DefaultPaddleWidth();
	currWidth = defWidth;
	startWidth = defWidth;
	finalWidth = defWidth;
	widthTimer = 0.0f;
	
	minWidth = settings.DefaultPaddleMinWidth();
	maxWidth = settings.DefaultPaddleMaxWidth();
	
	incWidth = (maxWidth - minWidth) / (float)settings.DefaultPaddleGrowIncrements();
	
	SetImage(settings.DefaultPaddleBaseImage());
	SetStickyImage(settings.DefaultPaddleBaseStickyImage());
	
	board.SetWidthUniform(defWidth);
	stickyBoard.SetWidthUniform(defWidth);
	
	arrow.Load(settings.LaunchArrowImage());
	arrow.SetTransparency(0.0f);
	arrowFadeTime = settings.LaunchArrowFadeTime();
	arrowFadeTimer = 0.0f;
	fadeArrowIn = false;
	drawArrow = false;
	
	PhysicsCompoundShape shapes;
	if( !LoadShapeFromFile(ConcatPath(DataPaths::GetPath("Shapes"), settings.DefaultPaddleShape()).c_str(), shapes) )
	{
		ConError() << "Paddle shape file could not be loaded.\n";
	}
	
	Matrixf S, T;
	S.MakeScale(board.GetWidth(), board.GetHeight() * 2.0f);
	T.MakeTranslation(0.0f, -board.GetHeight() / 2.0f);
	shapes.MultiplyTransform(T * S);
	
	SetShape(shapes);
	
	Sprite thruster(settings.DefaultPaddleThrusterImage());
	thruster.SetSize(settings.DefaultPaddleThrusterWidth(), settings.DefaultPaddleThrusterHeight());
	Point2f thrusterSpriteOffset = Point2f(0.0f, -thruster.GetHeight() * 0.6f);
	
	//AddAttachment(thrusterSpriteOffset, thruster);
	
	//AddAttachment(THRUSTER_LEFT_OFFSET, Sprite(THRUSTER_LEFT_IM));
	//AddAttachment(THRUSTER_RIGHT_OFFSET, Sprite(THRUSTER_RIGHT_IM));
	
	thrusterSmoke.drawPolicy.image = LoadImageResource(SMOKE_IM);
	thrusterSmoke.drawPolicy.blendMode = BLENDMODE_ALPHA;
	thrusterSmoke.initPolicy.minLife = 1.0f;
	thrusterSmoke.initPolicy.maxLife = 2.0f;
	thrusterSmoke.emitPolicy.Start();
	thrusterSmoke.emitPolicy.SetFlowRate(100.0f);
	thrusterSmoke.emitPolicy.position = Point2f(0.0f, 0.0f);
	
	thrusterFire.drawPolicy.image = LoadImageResource(FIRE_IM);
	thrusterFire.drawPolicy.blendMode = BLENDMODE_ADDITIVE_WITH_ALPHA;
	thrusterFire.initPolicy.minLife = 0.1f;
	thrusterFire.initPolicy.maxLife = 0.3f;
	thrusterFire.emitPolicy.Start();
	thrusterFire.emitPolicy.SetFlowRate(100.0f);
	thrusterFire.emitPolicy.position = Point2f(0.0f, 0.0f);
	
	thrusterOffset = Point2f(0.0f, -thruster.GetHeight() * 0.4f) + thrusterSpriteOffset;
}

//--------------------------------------------------------------------
void DefaultPaddle::Grow()
{
	startWidth = currWidth;
	finalWidth = Clamp(finalWidth + incWidth, minWidth, maxWidth);
	
	widthTimer = 0.0f;
	
	SetSticky(false);
}

//--------------------------------------------------------------------
void DefaultPaddle::Shrink()
{
	startWidth = currWidth;
	finalWidth = Clamp(finalWidth - incWidth, minWidth, maxWidth);
	
	widthTimer = 0.0f;
	
	SetSticky(false);
}

//--------------------------------------------------------------------
void DefaultPaddle::Reset()
{
	startWidth = currWidth;
	finalWidth = defWidth;
	widthTimer = 0.0f;
	
	Paddle::Reset();
}

//--------------------------------------------------------------------
void DefaultPaddle::OnAttachBall()
{
	drawArrow = true;
	fadeArrowIn = true;
	arrowFadeTimer = 0.0f;
}

//--------------------------------------------------------------------
void DefaultPaddle::OnDetachBall()
{
	drawArrow = true;
	fadeArrowIn = false;
}

//--------------------------------------------------------------------
void DefaultPaddle::Update(float dt, const IKeyboard &kb, const IMouse &mouse)
{
	Point2f p = GetPosition() + thrusterOffset;
	thrusterSmoke.emitPolicy.position = p;
	thrusterFire.emitPolicy.position = p;
	
	thrusterSmoke.Update(dt);
	thrusterFire.Update(dt);
	
	if(finalWidth != currWidth)
	{
		widthTimer += dt;
		
		float low = std::min(startWidth, finalWidth);
		float high = std::max(startWidth, finalWidth);
		currWidth = Clamp(lerp(startWidth, finalWidth, widthTimer), low, high);
		SetWidth(currWidth);
	}
	
	// Fade arrow in and out
	if(drawArrow)
	{
		if(fadeArrowIn)
		{
			arrowFadeTimer = std::min(arrowFadeTimer + dt, arrowFadeTime);
		}
		else // fade arrow out
		{
			arrowFadeTimer = std::max(arrowFadeTimer - dt, 0.0f);
			
			if(arrowFadeTimer <= 0.0f)
				drawArrow = false;
		}
		
		float delta = Clamp(arrowFadeTimer / arrowFadeTime, 0.0f, 1.0f);		
		arrow.SetHeightUniform( std::max(20.0f * delta, 0.0001f) );
		arrow.SetTransparency(delta);
	}
	
	Paddle::Update(dt, kb, mouse);
}

//--------------------------------------------------------------------
void DefaultPaddle::Draw()
{
	//thrusterSmoke.Draw();
	//thrusterFire.Draw();

	if(drawArrow)
	{
		// Positions the arrow
		const PhysicsObject *ball = GetAttachedBall();
		if(ball)
		{
			AABB bbox =  ball->GetShapes().GetBoundingBox();
			float ballHeight = bbox.Height();
			
			Vector2f vel = GetLaunchVelocity();
			Vector2f ref = Normalize( Vector2f(1.0f, 0.0f) );
			float angle = ref.Angle(vel) - PI_2;
			
			Point2f pos = ball->GetPosition() + Normalize(vel) * ((ballHeight + arrow.GetHeight()) / 2.0f);
			
			arrow.SetAngle(angle);
			arrow.SetPosition(pos);
		}
		
		arrow.Draw();
	}
	
	Paddle::Draw();
}
