/*!
 * \file Paddle.h
 * \brief Declares the Paddle class that represents the user controlled game paddle.
 * \author Joe Forte
 * \date 3/18/2008
 */

#ifndef PADDLE_H
#define PADDLE_H

#include "Math/MathCommon.h"
#include "Graphics/Sprite.h"
#include "Physics/PhysicsObject.h"
#include "Graphics/ParticleSystem.h"

class GameSettings;
class IKeyboard;
class IMouse;

/*!
 * \class PaddleThrusterPolicy
 * \headerfile Paddle.h "Game/Paddle.h"
 * \brief Policy that creates a thruster ParticleSystem.
 */
class PaddleThrusterPolicy
{
public:
	
	typedef ContinuousEmitPolicy<> Emit;

	class Init
	{
	public:
		float startSize;
		float endSize;
		Color4f initColor;
		float minLife;
		float maxLife;
		
		Init() : startSize(5.0f), endSize(50.0f), initColor(Color::White), minLife(0.3f), maxLife(0.7f) {}

		inline void PreEmit(const Point2f &pos, uint amount) { position = pos; }

		inline void InitParticle(Particle &p)
		{
			p.pos = position;
			p.color = initColor * RandomFloat(0.1f, 1.0f);
			p.color.a = 0.0f;
			p.life = RandomFloat(minLife, maxLife);
			p.age = 0.0f;
			p.angle = 0.0f;
			p.angularVel = 0.0f;
			
			float size = RandomFloat(startSize, endSize);
			p.width = size;
			p.height = size;
			
			Vector2f vel(RandomFloat(-SQRT2_2, SQRT2_2), -1.0f);
			vel.Normalize();
			
			p.vel = 100.0f * vel;
		}
		
	private:
		Point2f position;
	};

	class Update
	{
	public:
		float fadeOutAge;
		float dampening;

		Update() : fadeOutAge(0.1f), dampening(80.0f) {}

		inline void PreUpdate(float dt)
		{
			Vector2f a = Vector2f(0.0f, 1.0f) * dampening;
			at = a*dt;
			att = at*dt;
		}

		inline bool ApplyUpdate(Particle &p, float dt)
		{
			p.pos = p.pos + p.vel*dt + att;
			p.vel = p.vel + at;
			
			p.age += dt;
			
			if(p.age >= fadeOutAge)
				p.color.a = 1.0f - ( (p.age - fadeOutAge) / (p.life - fadeOutAge) );
			else
				p.color.a = p.age / fadeOutAge;
			
			return p.age < p.life;
		}
		
	private:
		Vector2f at, att; // Acceleration times time, Acceleration times time squared
	};
};

/*!
 * \class PaddleThrusterFirePolicy
 * \headerfile Paddle.h "Game/Paddle.h"
 * \brief Policy that creates a fire thruster ParticleSystem.
 */
class PaddleThrusterFirePolicy
{
public:
	
	typedef ContinuousEmitPolicy<> Emit;

	class Init
	{
	public:
		float startSize;
		float endSize;
		Color4f initColor;
		float minLife;
		float maxLife;
		
		Init() : startSize(5.0f), endSize(50.0f), initColor(Color::White), minLife(0.3f), maxLife(0.7f) {}

		inline void PreEmit(const Point2f &pos, uint amount) { position = pos; }

		inline void InitParticle(Particle &p)
		{
			p.pos = position;
			p.color = initColor;
			p.color.a = 0.0f;
			p.life = RandomFloat(minLife, maxLife);
			p.age = 0.0f;
			p.angle = 0.0f;
			p.angularVel = 0.0f;
			
			float size = RandomFloat(startSize, endSize);
			p.width = size;
			p.height = size;
			
			Vector2f vel(RandomFloat(-SQRT2_2, SQRT2_2), -1.0f);
			vel.Normalize();
			
			p.vel = 100.0f * vel;
		}
		
	private:
		Point2f position;
	};

	class Update
	{
	public:
		float fadeOutAge;
		float dampening;

		Update() : fadeOutAge(0.1f), dampening(80.0f) {}

		inline void PreUpdate(float dt)
		{
			Vector2f a = Vector2f(0.0f, 1.0f) * dampening;
			at = a*dt;
			att = at*dt;
		}

		inline bool ApplyUpdate(Particle &p, float dt)
		{
			p.pos = p.pos + p.vel*dt + att;
			p.vel = p.vel + at;
			
			p.age += dt;
			
			if(p.age >= fadeOutAge)
				p.color.a = 1.0f - ( (p.age - fadeOutAge) / (p.life - fadeOutAge) );
			else
				p.color.a = p.age / fadeOutAge;
			
			return p.age < p.life;
		}

	private:
		Vector2f at, att; // Acceleration times time, Acceleration times time squared
	};
};


/*!
 * \class PaddleSparkPolicy
 * \headerfile Paddle.h "Game/Paddle.h"
 * \brief Policy that creates a Spark ParticleSystem.
 */
class PaddleSparkPolicy
{
public:
		
	typedef ContinuousEmitPolicy<> Emit;
		
	class Init
	{
	public:
		Color4f initColor;
			
		Init() : initColor(Color::Blue) {}
			
		inline void PreEmit(const Point2f &pos, uint amount) { position = pos; }
			
		inline void InitParticle(Particle &p)
		{
			p.pos = position;
			p.color = initColor * RandomFloat(0.3f, 0.7f);
			p.color.a = 0.0f;
			p.life = 5;
			p.age = 0.0f;
			
			p.angularVel = 0.0f;
			
			p.width = 50.0f;
			p.height = 20.0f;
			
			float angle = RandomFloat(0, _2PI);
			p.angle = angle;
			
			Vector2f vel(std::cos(angle), std::sin(angle));
			p.vel = 50.0f * vel;
		}
			
	private:
		Point2f position;
	};
		
	class Update
	{
	public:
		float dampening;
		
		Update() : dampening(80.0f) {}
			
		inline void PreUpdate(float dt) {}
			
		inline bool ApplyUpdate(Particle &p, float dt)
		{
			Vector2f at ;//= -p.vel * dampening * dt;
			Vector2f att;// = at*dt;
			
			p.pos = p.pos + p.vel*dt + att;
			p.vel = p.vel + at;
			
			p.age += dt;
			
			p.color.a = 1.0f - (p.age / p.life);
			
			return p.age < p.life;
		}
	};
};

/*!
 * \class Paddle
 * \headerfile Paddle.h "Game/Paddle.h"
 * \brief The user controlled paddle.
 */
class Paddle : public PhysicsObject
{
public:
	Paddle();

	virtual bool Touch(PhysicsObject *obj, const PhysicsCollisionInfo &hitInfo);

	virtual PhysicsObjectType GetObjectType() const { return POBJ_PADDLE; }
	
	/*!
	 * \brief Sets the image to use for the paddle.
	 *
	 * \param[in] file the name of the image file.
	 */
	void SetImage(const char *file);
	
	/*!
	 * \brief Sets the image to use for the paddle when the paddle is in sticky mode.
	 *
	 * \param[in] file the name of the image file.
	 */
	void SetStickyImage(const char *file);

	void SetBounds(const Rectf &r);
	
	/*!
	 * \brief Sets the width of the paddle
	 *
	 * \param[in] w the width.
	 */
	void SetWidth(float w);
	
	/*!
	 * \brief Gets the height of the paddle.
	 *
	 * \return the height.
	 */
	float GetHeight() const;
	
	void Set(int w, int h);

	void Move(const Point2f &amt);
	
	virtual void Update(float dt, const IKeyboard &kb, const IMouse &mouse);
	
	virtual void Draw();

	virtual void SetPosition(const Point2f &p);
	
	virtual AABB GetBBox();
	
	/*!
	 * \brief Grows the size of the paddle.
	 */
	virtual void Grow() = 0;
	
	/*!
	 * \brief Shrinks the size of the paddle.
	 */
	virtual void Shrink() = 0;
	
	/*!
	 * \brief Attaches an sprite to the paddle.
	 *
	 * \param[in] pos the position to place the attachment relative to the center of the paddle.
	 * \param[in] att the attachment.
	 */
	void AddAttachment(const Point2f &pos, const Sprite &att);

	/*!
	 * \brief Sets the paddles sticky mode.
	 *
	 * \param[in] s the state of the sticky mode.
	 */
	void SetSticky(bool s);
	
	/*!
	 * \brief Gets the state of the sticky mode.
	 *
	 * \return True if the paddle is sticky, false otherwise.
	 */
	bool IsSticky() const { return sticky; }
	
	/*!
	 * \brief Determines if a ball is attached to the paddle.
	 *
	 * \param[in] b the ball to check for.
	 *
	 * \return true if the given ball is attached to to paddle, false otherwise.
	 */
	bool IsBallAttached(const PhysicsObject *b) const { return GetAttachedBall() == b; }
	
	/*!
	 * \brief Resets the paddle to its default state.
	 */
	virtual void Reset();
	
protected:
	
	Sprite board;
	Sprite stickyBoard;
	
	const PhysicsObject *GetAttachedBall() const { return attachedBall; }
	Vector2f GetLaunchVelocity() const { return launchVel; }
	
	/*!
	 * \brief Called when a ball is attched to the paddle.
	 */
	virtual void OnAttachBall() {}
	
	/*!
	 * \brief Called when a ball is detached from the paddle.
	 */
	virtual void OnDetachBall() {}
	
private:
	
	/*!
	 * \brief Attaches the ball to the paddle if possible.
	 *
	 * \param[in] ball the ball to attach to the paddle.
	 * \param[in] hitInfo the collision information for the ball and paddle collision.
	 */
	void AttachBall(PhysicsObject *ball, const PhysicsCollisionInfo &hitInfo);
	
	/*!
	 * \brief Detaches the ball if it is stuck to the paddle.
	 */
	void DetachBall();
	
	struct Attachment
	{
		Point2f offset;
		Sprite attachment;
	};
	
	std::vector<Attachment> attachments;
	
	AABB bounds;

	bool sticky;
	
	PhysicsObject *attachedBall;
	Vector2f ballAttachOffset;
	Vector2f launchVel;
};

/*!
 * \class DefaultPaddle
 * \headerfile Paddle.h "Game/Paddle.h"
 * \brief The BrickOut |ReDuX| defualt paddle.
 */
class DefaultPaddle : public Paddle
{
public:
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] settings game settings that define attributes of the paddle.
	 */
	DefaultPaddle(const GameSettings &settings);
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~DefaultPaddle() {}
	
	/*!
	 * \brief Grows the size of the paddle.
	 */
	virtual void Grow();
	
	/*!
	 * \brief Shrinks the size of the paddle.
	 */
	virtual void Shrink();
	
	/*!
	 * \brief Updates tha game paddle.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 * \param[in] kb the keyboard.
	 * \param[in] mouse the mouse.
	 */
	virtual void Update(float dt, const IKeyboard &kb, const IMouse &mouse);
	
	/*!
	 * \brief Draws the paddle.
	 */
	virtual void Draw();
	
	/*!
	 * \brief Resets the paddle to its default state.
	 */
	virtual void Reset();
	
protected:
	
	/*!
	 * \brief Called when a ball is attched to the paddle.
	 */
	virtual void OnAttachBall();
	
	/*!
	 * \brief Called when a ball is detached from the paddle.
	 */
	virtual void OnDetachBall();
	
private:
	
	Sprite arrow;
	float arrowFadeTime;
	float arrowFadeTimer;
	bool fadeArrowIn;
	bool drawArrow;
	
	Point2f thrusterOffset;
	
	ParticleSystem<100, PaddleThrusterPolicy::Init, PaddleThrusterPolicy::Update, PaddleThrusterPolicy::Emit> thrusterSmoke;
	ParticleSystem<100, PaddleThrusterFirePolicy::Init, PaddleThrusterFirePolicy::Update, PaddleThrusterFirePolicy::Emit> thrusterFire;
	
	float widthTimer;
	
	float minWidth;
	float maxWidth;
	float incWidth;
	
	float defWidth;
	float currWidth;
	float startWidth;
	float finalWidth;
};

#endif
