/*!
 * \file PowerUp.cpp
 * \brief Implements the PowerUp class.
 * \author Joe Forte
 * \date 8/23/2008
 */

#include "Game/PowerUp.h"
#include "Game/Level.h"
#include "Game/ConsoleVars.h"

/* Defines the differnat images to use for each powerup ability. */
const char *PowerUpImages[PowerUp::ABILITY_COUNT] = 
{
	"Pills/green.png",  // ABILITY_STICKY
	"Pills/red.png",    // ABILITY_SHRINK
	"Pills/blue.png",   // ABILITY_GROW
	"Pills/yellow.png", // ABILITY_EXTRALIFE
	"Pills/yellow.png", // ABILITY_POINTS
	"Pills/purple.png", // ABILITY_MULTIPLIER
	"Pills/rainbow.png" // ABILITY_RANDOM
};

/* Defines the probability of choosing a powerup ability.  Must sum to 1 */
const float PowerUpProb[PowerUp::ABILITY_COUNT] =
{
	0.1f,  // ABILITY_STICKY
	0.15f, // ABILITY_SHRINK
	0.15f, // ABILITY_GROW
	0.01f, // ABILITY_EXTRALIFE
	0.3f,  // ABILITY_POINTS
	0.2f,  // ABILITY_MULTIPLIER
	0.09f  // ABILITY_RANDOM
};

//--------------------------------------------------------------------
PowerUp::PowerUp(Level *l, PhysicsWorld *w, const Point2f &p) : level(l)
{
	const float width = 30.0f;
	const float height = 10.0f;
	
	// Chooses a random ability
	PowerUp::Ability a = PowerUp::ABILITY_NONE;
	float r = RandomFloat(0.0f, 1.0f);
	for(int i = 0; i < PowerUp::ABILITY_COUNT; i++)
	{
		if(r < PowerUpProb[i])
		{
			a = (PowerUp::Ability)i;
			break;
		}
		r -= PowerUpProb[i];
	}
	
	assert(a != PowerUp::ABILITY_NONE);
	
	ability = a;
	
	pill.Load(PowerUpImages[ability]);
	pill.SetSize(width, height);
	
	AttachToWorld(w);
	SetShape(CreatePhysicsShapeBox(width, height));
	SetPosition(p);
	
	SetCollisionGroups(POWERUP_COLLISION_GROUP, PADDLE_COLLISION_GROUP | BOUNDS_COLLISION_GROUP);
	
	SetVelocity( Vector2f(0.0f, RandomFloat(-100.0f, -200.0f)) );
}

//--------------------------------------------------------------------
PowerUp::~PowerUp() {}

//--------------------------------------------------------------------
bool PowerUp::Touch(PhysicsObject *obj, const PhysicsCollisionInfo &hitInfo)
{
	assert(obj);
	
	if(level && obj->GetObjectType() == POBJ_PADDLE)
	{
		level->PowerUpGrabbed(*this, hitInfo);
	}
	
	return true;
}

//--------------------------------------------------------------------
PowerUp::Ability PowerUp::GetAbility() const
{
	if(ability == PowerUp::ABILITY_RANDOM)
	{
		// Assumes ABILITY_RANDOM is larger then all other abilities				
		return (PowerUp::Ability)(rand() % PowerUp::ABILITY_RANDOM);
	}
	
	return ability;
}

//--------------------------------------------------------------------
void PowerUp::Draw()
{
	pill.SetPosition(GetPosition());
	pill.Draw();
	
	if(phy_drawCollisionShapes->GetInt())
		GetShapes().DrawOutline();
}
