/*!
 * \file PowerUp.h
 * \brief Defines the PowerUp class that are dropped from bricks.
 * \author Joe Forte
 * \date 8/23/2008
 */

#ifndef POWERUP_H
#define POWERUP_H

#include "Graphics/Sprite.h"
#include "Physics/PhysicsObject.h"

class Level;

/*!
 * \class PowerUp
 * \headerfile PowerUp.h "Game/PowerUp.h"
 * \brief Defines the physics object that represents a movable powerup object that can be picked up by the paddle.
 */
class PowerUp : public PhysicsObject
{
public:
	
	/*!
	 * \enum Ability
	 * \brief Defines the differant powerup abilities a pill can have.
	 */
	enum Ability
	{
		ABILITY_NONE   = -1, /*!< The Null ability. Does nothing. */
		ABILITY_STICKY = 0,  /*!< The ball sticks to the paddle when it touches it. */
		ABILITY_SHRINK,      /*!< The paddle will shrink in size. */
		ABILITY_GROW,        /*!< The paddle will grow in size. */
		ABILITY_EXTRALIFE,   /*!< Gains an extra ball. */
		ABILITY_POINTS,      /*!< Grants bonus points. */
		ABILITY_MULTIPLIER,  /*!< Grants a temporary point multiplier. */
		
		/* Add new powerups here. */
		
		ABILITY_RANDOM,      /*!< Random powerup choosen when the powerup is grabbed. */
		
		ABILITY_COUNT        /*!< The number of unique abilites. */
	};
	
	/*!
	 * \brief Contructs a powerup pill with a randomly choosen ability.
	 * 
	 * \param[in] l the level the powerup belongs to.
	 * \param[in] w the physics world he powerup resides in.
	 * \param[in] p the position to create the powerup.
	 */
	PowerUp(Level *l, PhysicsWorld *w, const Point2f &p);
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~PowerUp();
	
	/*!
	 * \brief Called when another physics object collides with a PowerUp.
	 *
	 * \param[in] obj the object that collided with the PowerUp.
	 * \param[in] hitInfo the hit information for the collision.
	 *
	 * \return true if the collision is accepted, false to ignore the collision 
	 * 
	 * NOTE: both objects must return false for a collison to be ignored.
	 */
	virtual bool Touch(PhysicsObject *obj, const PhysicsCollisionInfo &hitInfo);
	
	/*!
	 * \brief Returns a type identifier for the PowerUp.
	 *
	 * \return POBJ_POWERUP
	 */
	virtual PhysicsObjectType GetObjectType() const { return POBJ_POWERUP; }
	
	/*!
	 * \brief Determines if the powerup is active or not.
	 *
	 * \return true if the powerup is active, false if the powerup is inactive.
	 */
	bool IsActive() const { return ability != ABILITY_NONE; }
	
	/*!
	 * \brief Deactiveates the powerup.
	 */
	void Deactivate() { ability = ABILITY_NONE; }
	
	/*!
	 * \brief Gets the ability of the powerup.
	 *
	 * \return the powerups ability or ABILITY_NONE if the powerup is is inactive.
	 */
	Ability GetAbility() const;
	
	/*!
	 * \brief Draws the powerup.
	 */
	void Draw();
	
protected:
	
	Ability ability;
	
	Sprite pill;
	
	Level *level;
};

#endif
