/*!
 * \file StarFieldEffect.h
 * \brief Declares the StarFieldEffect background effect. 
 * \author Joe Forte
 * \date 2/9/2009
 */

#ifndef STARFIELDEFFECT_H
#define STARFIELDEFFECT_H

#include "Graphics/ParticleSystem.h"
#include "Game/BackgroundEffect.h"

/*!
 * \class StarFieldPolicy
 * \headerfile StarFieldEffect.h "Game/StarFieldEffect.h"
 * \brief Simple policy that creates a StarField ParticleSystem.  Is all inclusive containing init and update policies.
 */
class StarFieldPolicy
{
public:
	
	typedef ContinuousEmitPolicy<> Emit;

	class Init
	{
	public:
		// init data
		float lifeTime;
		float speed;
		
		Init() : lifeTime(1.0f), speed(600.0f) {}
		
		/*!
		 * \brief Called once before new particles are emitted.
		 *
		 * \param[in] pos the requested position to emit the particles.
		 * \param[in] amount the number of particles that will be emitted.
		 */
		inline void PreEmit(const Point2f &pos, uint amount)
		{
			position = pos;
		}
		
		/*!
		 * \brief Called once on each new particle that is emitted.
		 *
		 * \param[in] p the particle to initialize.
		 */
		inline void InitParticle(Particle &p)
		{
			p.pos = position;
			p.color = Color::White;
			p.life = lifeTime;
			p.age = 0.0f;
			p.angularVel = 0.0f;
			p.width = 0.0f;
			p.height = 0.0f;
			
			float a = RandomFloat(0.0f, _2PI);
			p.angle = a;
			
			Vector2f vel(std::cos(a), std::sin(a));
			
			p.vel = speed*vel;
		}

		private:
	
		// updated after each call to PreEmit
		Point2f position;
	};

	class Update
	{
	public:
		// update data
		float endHeight;
		float endWidth;

		Update() : endHeight(2.0f), endWidth(50.0f) {}
	
		/*!
		 * \brief Called once before each update.
		 *
		 * \param[in] dt the time in seconds since the last frame.
		 */
		inline void PreUpdate(float dt) {}
		
		/*!
		 * \brief Called once before each update.
		 *
		 * \param[in,out] p the particle to apply the Policy to.
		 * \param[in] dt the time in seconds since the last frame.
		 *
		 * \return False if you want the particle to be destroyed, return true otherwise.
		 */
		inline bool ApplyUpdate(Particle &p, float dt)
		{
			p.pos = p.pos + p.vel*dt;
			p.age += dt;
			
			float delta = p.age / p.life;
			p.color = Color4f(delta, delta, delta, delta);
			p.width = endWidth * delta;
			p.height = endHeight * delta;
			
			return p.age < p.life;
		}
	};
};

/*!
 * \class StarFieldEffect
 * \headerfile StarFieldEffect.h "Game/StarFieldEffect.h"
 * \brief A moving starfield background effect.
 */
class StarFieldEffect : public BackgroundEffect
{
public:

	/*!
	 * \brief Constructor.
	 *
	 * \param[in] size the that the effect will fill.
	 */
	StarFieldEffect(const Rectf &size) : BackgroundEffect(size) {}

	/*!
	 * \brief Deconstructor.
	 */
	virtual ~StarFieldEffect() {}

	/*!
	 * \brief Initializes the effect. Called once before Update and Draw are called.
	 *
	 * \return true if the initialization was successful, false otherwise.
	 */
	virtual bool Init();

	/*!
	 * \brief Updates the effect.  Called once per frame.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 */
	virtual void Update(float dt);

	/*!
	 * \brief Draws the effect.
	 */
	virtual void Draw();

protected:

	ParticleSystem<100, StarFieldPolicy::Init, StarFieldPolicy::Update, StarFieldPolicy::Emit> starField;
	ParticleSystem<100, StarFieldPolicy::Init, StarFieldPolicy::Update, StarFieldPolicy::Emit> bigStarField;

};

#endif
