/*!
 * \file TextParticleSystem.h
 * \brief Declares the TextParticleSystem classes.
 * \author Joe Forte
 * \date 8/17/2008
 */

#ifndef TEXTPARTICLESYSTEM_H
#define TEXTPARTICLESYSTEM_H

#include <vector>
#include <string>
#include "Math/Point.h"
#include "Graphics/Font.h"
#include "Graphics/Color.h"
#include "Graphics/ParticleSystem.h"

/*!
 * \class TextParticle
 * \headerfile TextParticleSystem.h "Game/TextParticleSystem.h"
 * \brief A text particle.
 */
struct TextParticle
{
	std::string text;
	Point2f position;
	Color4f color;
	float lifeTime;
	float age;
};

/*!
 * \class TextParticleSystem
 * \headerfile TextParticleSystem.h "Game/TextParticleSystem.h"
 * \brief A simple particle system that draws each partical as a string of text.
 */
class TextParticleSystem
{
public:
	
	/*!
	 * \brief Default constructor.
	 */
	TextParticleSystem();
	
	/*!
	 * \brief Updates the active particles.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 */
	void Update(float dt);
	
	/*!
	 * \brief Adds a new text particle to the system.
	 *
	 * \param[in] str the text associated with the text particle.
	 * \param[in] pos the position to spawn the text particle.
	 * \param[in] c the color to spawn the particle.
	 */
	void Emit(const std::string &str, const Point2f &pos, const Color4f &c);
	
	/*!
	 * \brief Sets the attributes of the particle system.
	 *
	 * \patam[in] f the font to draw the text particles.
	 * \param[in] maxAge the life time of each particle.
	 */
	void SetAttributes(FontRef f, float maxAge);
	
	/*!
	 * \brief Returns the font used to draw the particle system.
	 */
	FontRef GetFont() const { return font; }
	
	/*!
	 * \brief Draws all of the active particles.
	 */
	void Draw() const;
	
	/*!
	 * \brief Clears all of the active particles.
	 */
	void Clear();
	
	/*!
	 * \typdef fnUpdateFunc
	 * \brief Defines the prototype of the particle update function that will be called for each active particle in the system.
	 *
	 * \param[in] p the particle to update.
	 * \param[in] dt the frame time in seconds since the last frame.
	 */
	typedef void (*fnUpdateFunc)(TextParticle &p, float dt);
	
	/*!
	 * \brief Sets the update function for the particle system.
	 *
	 * \param[in] f the pointer to the update function.
	 */
	void SetUpdateFunc(fnUpdateFunc f) { func = f; }
	
	/*!
	 * \brief The default update function that linearly decrease the transparancy of the particles.
	 *
	 * \param[in] p the particle to update.
	 * \param[in] dt the frame time in seconds since the last frame.
	 */
	static void DefaultUpdateFunc(TextParticle &p, float dt);
	
private:
	std::vector<TextParticle> particles;
	
	FontRef font;
	float lifeTime;
	Color4f color;
	
	fnUpdateFunc func;
};

#endif
