/*!
 * \file TransitionEffect.cpp
 * \brief Implements the common transition effects.
 * \author Joe Forte
 * \date 12/18/2009
 */

#include "Math/MathCommon.h"
#include "Common/Globals.h"
#include "Graphics/MathDraw.h"
#include "Game/TransitionEffect.h"

//--------------------------------------------------------------------
Transition::Transition()
:
transTime(0.0f),
timer(0.0f)
{}

//--------------------------------------------------------------------
Transition::Transition(float time)
:
transTime(time),
timer(0.0f)
{}

//--------------------------------------------------------------------
void Transition::SetTransitionTime(float time)
{
	transTime = time;
}

//--------------------------------------------------------------------
float Transition::GetTransitionTime()
{
	return transTime;
}

//--------------------------------------------------------------------
float Transition::GetTimeLeft() const
{
	return Clamp(transTime - timer, 0.0f, transTime);
}

//--------------------------------------------------------------------
bool Transition::IsComplete() const
{
	return GetCompletionPercent() >= 1.0f;
}

//--------------------------------------------------------------------
float Transition::GetCompletionPercent() const
{
	return Clamp(timer / transTime, 0.0f, 1.0f);
}

//--------------------------------------------------------------------
void Transition::SetCompletionPercent(float t)
{
	t = Clamp(t, 0.0f, 1.0f);
	timer = t * transTime;
}

//--------------------------------------------------------------------
void Transition::ForceCompletion()
{
	timer = transTime;
}

//--------------------------------------------------------------------
void Transition::Reset()
{
	timer = 0.0f;
}

//--------------------------------------------------------------------
void Transition::Update(float dt)
{
	if(IsComplete())
		return;

	timer = Clamp(timer + dt, 0.0f, transTime);
}

//--------------------------------------------------------------------
void Transition::Draw()
{
	OnDraw( GetCompletionPercent(), area );
}

//--------------------------------------------------------------------
void AlphaFadeInTransition::OnDraw(float t, const Rectf &r)
{
	Color4f c(color.r, color.g, color.b, 1.0f - t);
	DrawRect(r, true, c);
}

//--------------------------------------------------------------------
void AlphaFadeOutTransition::OnDraw(float t, const Rectf &r)
{
	Color4f c(color.r, color.g, color.b, t);
	DrawRect(r, true, c);
}

//--------------------------------------------------------------------
void SolidColorDelayTransition::OnDraw(float t, const Rectf &r)
{
	Color4f c(color.r, color.g, color.b, 1.0f);
	DrawRect(r, true, c);
}

//--------------------------------------------------------------------
void SlantedSlideOutTransition::OnDraw(float t, const Rectf &r)
{
	const float triW = r.w / 6.0f; // The width of the triangle
	const float w = r.w + triW; // The distance the triangle travels
	const float x = r.x - triW + (w * t); // The x coord where the triangle starts

	Point2f a(x, r.y);
	Point2f b(x + triW, r.y);
	Point2f c(x, r.y - r.h);

	DrawTriangle(a, b, c, Color::Black);

	if(x >= r.x)
	{
		Rectf q;
		q.x = r.x;
		q.w = x - r.x;
		q.y = r.y;
		q.h = r.h;

		DrawRect(q, true, Color::Black);
	}
}

//--------------------------------------------------------------------
void SlantedSlideInTransition::OnDraw(float t, const Rectf &r)
{
	const float triW = r.w / 6.0f; // The width of the triangle
	const float w = r.w + triW; // The distance the triangle travels
	const float x = r.x + (w * t); // The x coord where the triangle starts

	Point2f a(x - triW, r.y - r.h);
	Point2f b(x, r.y);
	Point2f c(x, r.y - r.h);

	DrawTriangle(a, b, c, Color::Black);

	if(x <= (r.x + r.w))
	{
		Rectf q;
		q.x = x;
		q.w = r.w - x;
		q.y = r.y;
		q.h = r.h;

		DrawRect(q, true, Color::Black);
	}
}

//--------------------------------------------------------------------
TransitionEffect::TransitionEffect()
:
listener(NULL)
{}

//--------------------------------------------------------------------
TransitionEffect::~TransitionEffect()
{
	for(std::vector<Transition*>::iterator i = active.begin(); i != active.end(); i++)
	{
		Transition *trans = *i;
		if(trans)
			delete trans;
	}

	for(std::vector<Transition*>::iterator i = complete.begin(); i != complete.end(); i++)
	{
		Transition *trans = *i;
		if(trans)
			delete trans;
	}
}

//--------------------------------------------------------------------
const Transition::TransitionType TransitionEffect::GetCurrentTransitionType() const
{
	Transition *trans = GetCurrentTransition();
	if(trans)
		return trans->GetType();

	return Transition::TRANSITION_UNKNOWN;
}

//--------------------------------------------------------------------
Transition *TransitionEffect::GetCurrentTransition() const
{
	if(!active.empty())
		return active.back();

	return NULL;
}

//--------------------------------------------------------------------
void TransitionEffect::SetArea(const Rectf &rec)
{
	area = rec;

	for(std::vector<Transition*>::iterator i = active.begin(); i != active.end(); i++)
	{
		Transition *trans = *i;
		if(trans)
			trans->SetArea(rec);
	}

	for(std::vector<Transition*>::iterator i = complete.begin(); i != complete.end(); i++)
	{
		Transition *trans = *i;
		if(trans)
			trans->SetArea(rec);
	}
}

//--------------------------------------------------------------------
bool TransitionEffect::IsComplete() const
{
	return GetCurrentTransition() == NULL;
}

//--------------------------------------------------------------------
void TransitionEffect::Reset()
{
	for(std::vector<Transition*>::reverse_iterator i = complete.rbegin(); i != complete.rend(); i++)
	{
		Transition *trans = *i;
		trans->Reset();

		active.push_back(trans);
	}

	complete.clear();
}

//--------------------------------------------------------------------
void TransitionEffect::ForceCompletion()
{
	for(std::vector<Transition*>::reverse_iterator i = active.rbegin(); i != active.rend(); i++)
	{
		Transition *trans = *i;
		trans->ForceCompletion();

		complete.push_back(trans);
	}

	active.clear();
}

//--------------------------------------------------------------------
void TransitionEffect::ResetOrder(const Transition::TransitionType typeOrder[], uint n)
{
	ForceCompletion();

	int len = Clamp<int>(n, 0, Transition::TRANSITION_COUNT);
	for(int i = len - 1; i >= 0; i--)
	{
		Transition::TransitionType type = typeOrder[i];

		std::vector<Transition*>::iterator it = complete.begin();
		while(it != complete.end())
		{
			Transition *trans = *it;
			if(trans->GetType() == type)
			{
				trans->Reset();
				active.push_back(trans);

				it = complete.erase(it);
			}
			else
			{
				it++;
			}
		}
	}
}

//--------------------------------------------------------------------
void TransitionEffect::PushTransition(Transition *trans)
{
	if(trans)
	{
		trans->Reset();
		trans->SetArea(area);
		active.push_back(trans);
	}
}
	
//--------------------------------------------------------------------
void TransitionEffect::Update(float dt)
{
	Transition *trans = GetCurrentTransition();

	if(!trans)
		return;

	if(trans->IsComplete())
	{
		if(listener)
		{
			switch(trans->GetType())
			{
			case Transition::TRANSITION_OUT:
				listener->OnTransitionOutCompleted(*this, *trans);
				break;
			case Transition::TRANSITION_DELAY:
				listener->OnTransitionDelayCompleted(*this, *trans);
				break;
			case Transition::TRANSITION_IN:
				listener->OnTransitionInCompleted(*this, *trans);
				break;
			};
		}

		complete.push_back(trans);
		active.pop_back();

		if(IsComplete())
		{
			if(listener)
				listener->OnTransitionEffectCompleted(*this);
			return;
		}
		else
		{
			trans = GetCurrentTransition();
		}
	}

	trans->Update(dt);
}

//--------------------------------------------------------------------
void TransitionEffect::Draw()
{
	Transition *trans = GetCurrentTransition();

	if(trans)
		trans->Draw();
}
