/*!
 * \file TransitionEffect.h
 * \brief Declares the TransitionEffect interface and common implementations.
 * \author Joe Forte
 * \date 12/18/2009
 */

#ifndef TRANSITIONEFFECT_H
#define TRANSITIONEFFECT_H

#include "Graphics/Color.h"

class TransitionEffect;

/*!
 * \class Transition
 * \headerfile TransitionEffect.h "Game/TransitionEffect.h"
 * \brief Base class for a transition.
 *
 * Transitions are expected to cover the entire given rectangle.
 */
class Transition
{
public:

	/*!
	 * \enum TransitionType
	 * \brief Declares the differant type of transitions.
	 */
	enum TransitionType
	{
		TRANSITION_UNKNOWN = -1,	/*!< Unknown transition type. */

		TRANSITION_OUT = 0,			/*!< Transition out. */
		TRANSITION_DELAY,			/*!< Transition delay between a transition out and in. */
		TRANSITION_IN,				/*!< Transition in. */

		TRANSITION_COUNT			/*!< The number of transition types. */
	};

	/*!
	 * \brief Constructor.
	 */
	Transition();

	/*!
	 * \brief Constructor.
	 *
	 * \param[in] time the transition time.
	 */
	Transition(float time);

	/*!
	 * \brief Deconstructor.
	 */
	virtual ~Transition() {}

	/*!
	 * \brief Returns the type of the transition.
	 */
	virtual TransitionType GetType() const { return TRANSITION_UNKNOWN; }

	/*!
	 * \brief Sets the transition time.
	 */
	void SetTransitionTime(float time);

	/*!
	 * \brief Sets the area that the transition must cover.
	 */
	void SetArea(const Rectf &rec) { area = rec; }

	/*!
	 * \brief Gets the transition time.
	 */
	float GetTransitionTime();

	/*!
	 * \brief Returns the time left in the transition.
	 */
	float GetTimeLeft() const;

	/*!
	 * \brief Returns true if the transition is complete.
	 */
	bool IsComplete() const;

	/*!
	 * \brief Returns a value in the range [0,1].
	 */
	float GetCompletionPercent() const;

	/*!
	 * \brief Sets the completion percentage.
	 *
	 * \param[in] t a value in the range [0,1].
	 */
	void SetCompletionPercent(float t);

	/*!
	 * \brief Reset the transition.
	 */
	void Reset();

	/*!
	 * \brief Forces the transition to be completed.
	 */
	void ForceCompletion();

	/*!
	 * \brief Updates the transition.
	 */
	void Update(float dt);

	/*!
	 * \brief Draws the transition.
	 */
	void Draw();

protected:

	/*!
	 * \brief Internal draw method.
	 *
	 * \param[in] t the completion percentage in the range [0,1].
	 * \param[in] r the area to cover.
	 */
	virtual void OnDraw(float t, const Rectf &r) = 0;

private:

	float transTime;
	float timer;

	Rectf area;

};

/*!
 * \class TransitionOut
 * \headerfile TransitionEffect.h "Game/TransitionEffect.h"
 * \brief Base class for out transitions.
 */
class TransitionOut : public Transition
{
public:

	/*!
	 * \brief Constructor.
	 */
	TransitionOut() {}

	/*!
	 * \brief Constructor.
 	 *
	 * \param[in] time the transition time.
	 */
	TransitionOut(float time) : Transition(time) {}

	/*!
	 * \brief Returns the type of the transition.
	 */
	virtual TransitionType GetType() const { return Transition::TRANSITION_OUT; }

};

/*!
 * \class TransitionIn
 * \headerfile TransitionEffect.h "Game/TransitionEffect.h"
 * \brief Base class for in transitions.
 */
class TransitionIn : public Transition
{
public:

	/*!
	 * \brief Constructor.
	 */
	TransitionIn() {}

	/*!
	 * \brief Constructor.
 	 *
	 * \param[in] time the transition time.
	 */
	TransitionIn(float time) : Transition(time) {}

	/*!
	 * \brief Returns the type of the transition.
	 */
	virtual TransitionType GetType() const { return Transition::TRANSITION_IN; }

};

/*!
 * \class TransitionDelay
 * \headerfile TransitionEffect.h "Game/TransitionEffect.h"
 * \brief Base class for delay transitions.
 */
class TransitionDelay : public Transition
{
public:

	/*!
	 * \brief Constructor.
	 */
	TransitionDelay() {}

	/*!
	 * \brief Constructor.
 	 *
	 * \param[in] time the transition time.
	 */
	TransitionDelay(float time) : Transition(time) {}

	/*!
	 * \brief Returns the type of the transition.
	 */
	virtual TransitionType GetType() const { return Transition::TRANSITION_DELAY; }

};

/*!
 * \class AlphaFadeOutTransition
 * \headerfile TransitionEffect.h "Game/TransitionEffect.h"
 * \brief Fades out the alpha value of a solid color.
 */
class AlphaFadeOutTransition : public TransitionOut
{
public:

	/*!
	 * \brief Constructor.
	 */
	AlphaFadeOutTransition() : color(Color::Black) {}

	/*!
	 * \brief Constructor.
 	 *
	 * \param[in] time the transition time.
	 */
	AlphaFadeOutTransition(float time) : TransitionOut(time), color(Color::Black) {}
	
	/*!
	 * \brief Sets the transition color.
	 */
	void SetColor(const Color4f &c) { color = c; }

protected:

	/*!
	 * \brief Internal draw method.
	 *
	 * \param[in] t the completion percentage in the range [0,1].
	 * \param[in] r the area to cover.
	 */
	virtual void OnDraw(float t, const Rectf &r);

private:

	Color4f color;

};

/*!
 * \class AlphaFadeInTransition
 * \headerfile TransitionEffect.h "Game/TransitionEffect.h"
 * \brief Fades in the alpha value of a solid color.
 */
class AlphaFadeInTransition : public TransitionIn
{
public:

	/*!
	 * \brief Constructor.
	 */
	AlphaFadeInTransition() : color(Color::Black) {}

	/*!
	 * \brief Constructor.
 	 *
	 * \param[in] time the transition time.
	 */
	AlphaFadeInTransition(float time) : TransitionIn(time), color(Color::Black) {}

	/*!
	 * \brief Sets the transition color.
	 */
	void SetColor(const Color4f &c) { color = c; }

protected:

	/*!
	 * \brief Internal draw method.
	 *
	 * \param[in] t the completion percentage in the range [0,1].
	 * \param[in] r the area to cover.
	 */
	virtual void OnDraw(float t, const Rectf &r);

private:

	Color4f color;

};

/*!
 * \class SolidColorDelayTransition
 * \headerfile TransitionEffect.h "Game/TransitionEffect.h"
 * \brief A solid color delay transition.
 */
class SolidColorDelayTransition : public TransitionDelay
{
public:

	/*!
	 * \brief Constructor.
	 */
	SolidColorDelayTransition() : color(Color::Black) {}

	/*!
	 * \brief Constructor.
 	 *
	 * \param[in] time the transition time.
	 */
	SolidColorDelayTransition(float time) : TransitionDelay(time), color(Color::Black) {}

	/*!
	 * \brief Sets the transition color.
	 */
	void SetColor(const Color4f &c) { color = c; }

protected:

	/*!
	 * \brief Internal draw method.
	 *
	 * \param[in] t the completion percentage in the range [0,1].
	 * \param[in] r the area to cover.
	 */
	virtual void OnDraw(float t, const Rectf &r);

private:

	Color4f color;

};


/*!
 * \class SlantedSlideOutTransition
 * \headerfile TransitionEffect.h "Game/TransitionEffect.h"
 * \brief A slanted slide out transition.
 */
class SlantedSlideOutTransition : public TransitionOut
{
public:

	/*!
	 * \brief Constructor.
	 */
	SlantedSlideOutTransition() {}

	/*!
	 * \brief Constructor.
 	 *
	 * \param[in] time the transition time.
	 */
	SlantedSlideOutTransition(float time) : TransitionOut(time) {}

protected:

	/*!
	 * \brief Internal draw method.
	 *
	 * \param[in] t the completion percentage in the range [0,1].
	 * \param[in] r the area to cover.
	 */
	virtual void OnDraw(float t, const Rectf &r);

};

/*!
 * \class SlantedSlideInTransition
 * \headerfile TransitionEffect.h "Game/TransitionEffect.h"
 * \brief A slanted slide in transition.
 */
class SlantedSlideInTransition : public TransitionIn
{
public:

	/*!
	 * \brief Constructor.
	 */
	SlantedSlideInTransition() {}

	/*!
	 * \brief Constructor.
 	 *
	 * \param[in] time the transition time.
	 */
	SlantedSlideInTransition(float time) : TransitionIn(time) {}

protected:

	/*!
	 * \brief Internal draw method.
	 *
	 * \param[in] t the completion percentage in the range [0,1].
	 * \param[in] r the area to cover.
	 */
	virtual void OnDraw(float t, const Rectf &r);

};

/*!
 * \class TransitionEffectListener
 * \headerfile TransitionEffect.h "Game/TransitionEffect.h"
 * \brief Interface for listening to transition effect events.
 *
 * Holds a stack of transitions to be executed.
 */
class TransitionEffectListener
{
public:

	/*!
	 * \brief Called when a transition out is completed.
	 *
	 * \param[in] effect the transition effect.
	 * \param[in] trans the transition that completed.
	 */
	virtual void OnTransitionOutCompleted(TransitionEffect &effect, Transition &trans) { }

	/*!
	 * \brief Called when a transition in is completed.
	 *
	 * \param[in] effect the transition effect.
	 * \param[in] trans the transition that completed.
	 */
	virtual void OnTransitionInCompleted(TransitionEffect &effect, Transition &trans) { }

	/*!
	 * \brief Called when a transition delay is completed.
	 *
	 * \param[in] effect the transition effect.
	 * \param[in] trans the transition that completed.
	 */
	virtual void OnTransitionDelayCompleted(TransitionEffect &effect, Transition &trans) { }

	/*!
	 * \brief Called when the transition effect is completed.
	 *
	 * \param[in] effect the transition effect.
	 */
	virtual void OnTransitionEffectCompleted(TransitionEffect &effect) { }

};

/*!
 * \class TransitionEffect
 * \headerfile TransitionEffect.h "Game/TransitionEffect.h"
 * \brief Interface for transition effects.
 *
 * Holds a stack of transitions to be executed.
 */
class TransitionEffect
{
public:

	/*!
	 * \brief Constructor.
	 */
	TransitionEffect();

	/*!
	 * \brief Deconstructor.
	 */
	~TransitionEffect();

	/*!
	 * \brief Sets the transition effect listener.
	 */
	void SetListener(TransitionEffectListener *l) { listener = l; }
	
	/*!
	 * \brief Returns the type of the current transition or TRANSITION_UNKNOWN if there is no transition.
	 */
	const Transition::TransitionType GetCurrentTransitionType() const;

	/*!
	 * \brief Returns the current transition or NULL if there is no transition.
	 */
	Transition *GetCurrentTransition() const;

	/*!
	 * \brief Sets the area that the transition effect must cover.
	 */
	void SetArea(const Rectf &rec);

	/*!
	 * \brief Returns true if the transition is complete.
	 */
	bool IsComplete() const;

	/*!
	 * \brief Resets the transition effect.
	 */
	void Reset();

	/*!
	 * \brief Forces the transition effec to be completed.
	 */
	void ForceCompletion();

	/*!
	 * \brief Pushes a transition onto the active stack.
	 *
	 * \param[in] trans the transition to add. NOTE: Ownership of the transititon is taken by the effect.
	 */
	void PushTransition(Transition *trans);

	/*!
	 * \brief Sets the order of transitions and resets the transition.
	 *
	 * \param[in] typeOrder the order that differant transition types will apear in the transition.
	 * \param[in] n the size of the typeOrder array. NOTE: only the first TRANSITION_COUNT items will be looked at.
	 *	
	 */
	void ResetOrder(const Transition::TransitionType typeOrder[], uint n); 
		
	/*!
	 * \brief Updates the transition.
	 */
	void Update(float dt);

	/*!
	 * \brief Draws the transition.
	 */
	void Draw();

private:

	std::vector<Transition*> active; /*!< Holds all of the transitions that are not completed. */
	std::vector<Transition*> complete; /*!< Holds all of the transitions that are completed. */

	TransitionEffectListener *listener;

	Rectf area;
	
};

#endif
