/*!
 * \file AnimatedSprite.cpp
 * \brief Implements the AnimatedSprite class.
 * \author Joe Forte
 * \date 8/30/2008
 */

#include "Common/Globals.h"
#include "Application/DataPaths.h"
#include "Graphics/AnimatedSprite.h"

#include "tinyxml.h"

//--------------------------------------------------------------------
bool LoadAnimationFrame(TiXmlElement *elm, AnimatedSprite::AnimationFrame &frame)
{
	if(elm->QueryFloatAttribute("duration", &frame.duration))
	{
		ConError() << "The AnimatedSprite Frame is missing the 'duration' attribute.\n";
		return false;
	}

	if(elm->QueryFloatAttribute("x", &frame.area.x))
	{
		ConError() << "The AnimatedSprite Frame is missing the 'x' attribute.\n";
		return false;
	}

	if(elm->QueryFloatAttribute("y", &frame.area.y))
	{
		ConError() << "The AnimatedSprite Frame is missing the 'y' attribute.\n";
		return false;
	}

	if(elm->QueryFloatAttribute("w", &frame.area.w))
	{
		ConError() << "The AnimatedSprite Frame is missing the 'w' attribute.\n";
		return false;
	}

	if(elm->QueryFloatAttribute("h", &frame.area.h))
	{
		ConError() << "The AnimatedSprite Frame is missing the 'h' attribute.\n";
		return false;
	}

	return true;
}

//--------------------------------------------------------------------
bool LoadAnimation(TiXmlElement *elm, AnimatedSprite::Animation &anim)
{
	anim.name = elm->ValueStr();
	float totalTime = 0.0f;

	TiXmlNode *curr = elm->FirstChild();
	while( curr )
	{
		if( curr->Type() == TiXmlNode::ELEMENT )
		{
			TiXmlElement *e = curr->ToElement();
			if(e->ValueStr() != "Frame")
			{
				AnimatedSprite::AnimationFrame frame;
				anim.name = e->ValueStr();

				if(!LoadAnimationFrame(e, frame))
				{
					ConError() << "The Animation '" << anim.name << "' is not formatted properly.\n";
					return false;
				}

				anim.frames.push_back(frame);
				totalTime += frame.duration;
			}
			else
			{
				ConWarning() << "The Animation '" << anim.name << "' contains and unknown directive '" 
					<< e->ValueStr() << "'.\n";
			}
		}

		curr = elm->IterateChildren(curr);
	}

	anim.duration = totalTime;

	return true;
}

//--------------------------------------------------------------------
bool LoadAnimations(TiXmlElement *elm, std::vector<AnimatedSprite::Animation> &animations)
{
	TiXmlNode *curr = elm->FirstChild();
	while( curr )
	{
		if( curr->Type() == TiXmlNode::ELEMENT )
		{
			TiXmlElement *e = curr->ToElement();

			AnimatedSprite::Animation anim;
			if(!LoadAnimation(e, anim))
			{
				ConError() << "The Animation '" << e->ValueStr() << "' is not formatted properly.\n";
				return false;
			}

			animations.push_back(anim);
		}

		curr = elm->IterateChildren(curr);
	}

	return true;
}

//--------------------------------------------------------------------
bool AnimatedSprite::LoadAnimatedSprite(const char *file)
{
	std::string fileName = ConcatPath( DataPaths::GetPath("Animations"), CreatePath(file));
	TiXmlDocument doc(fileName.c_str());
	if( !doc.LoadFile() )
	{
		ConError() << "The AnimatedSprite file '" << file << "' could not be parsed.\n";
		return false;
	}
	
	TiXmlHandle hDoc(&doc);
	
	TiXmlNode *Root = hDoc.FirstChildElement().Node();
	if( !Root || Root->ValueStr() != "AnimatedSprite")
	{
		ConError() << "The AnimatedSprite file '" << file << "' is not formatted properly.\n";
		return false;
	}
	
	TiXmlNode *curr = Root->FirstChild();
	
	while(curr)
	{
		if(curr->Type() == TiXmlNode::ELEMENT)
		{
			TiXmlElement *elm = curr->ToElement();
			
			if(elm->ValueStr() == "Image")
			{
				if(!Load(elm->GetText()))
				{
					ConError() << "The Image '" << elm->GetText() << "' in the AnimatedSprite file '" 
						<< file << "' could not be loaded.\n";
					return false;
				}
			}
			else if(elm->ValueStr() == "Animations")
			{
				if(!LoadAnimations(elm, animations))
				{
					ConError() << "The AnimatedSprite file '" << file << "' is not formatted properly.\n";
					return false;
				}
			}
		}

		curr = Root->IterateChildren(curr);
	}
	
	return true;
}

//--------------------------------------------------------------------
uint AnimatedSprite::GetNumberOfAnimations() const
{
	return (uint)animations.size();
}

//--------------------------------------------------------------------
const AnimatedSprite::Animation &AnimatedSprite::GetAnimation(uint index) const
{
	assert(index < GetNumberOfAnimations());
	return animations[index];
}

//--------------------------------------------------------------------
int AnimatedSprite::FindAnimation(const char *name) const
{
	const size_t count = animations.size();
	for(size_t i = 0; i < count; i++)
	{
		if(animations[i].name == name)
			return (int)i;
	}

	return -1;
}

//--------------------------------------------------------------------
void AnimatedSprite::DrawFrame(uint animationNumber, uint frameNumber) const
{
	const Animation &anim = GetAnimation(animationNumber);
	assert(frameNumber < (uint)anim.frames.size());

	DrawFrame(anim.frames[frameNumber]);
}

//--------------------------------------------------------------------
void AnimatedSprite::DrawFrame(const Animation &anim, uint frameNumber) const
{
	assert(frameNumber < (uint)anim.frames.size());

	DrawFrame(anim.frames[frameNumber]);
}

//--------------------------------------------------------------------
void AnimatedSprite::DrawFrame(const AnimationFrame &frame) const
{
	DrawSubArea(frame.area);
}
