/*!
 * \file AnimatedSprite.h
 * \brief Declares the AnimatedSprite class.
 * \author Joe Forte
 * \date 8/30/2008
 */

#ifndef ANIMATEDSPRITE_H
#define ANIMATEDSPRITE_H

#include <vector>
#include "Math/MathCommon.h"
#include "Graphics/Sprite.h"

/*!
 * \class AnimatedSprite
 * \headerfile AnimatedSprite.h "Graphics/AnimatedSprite.h"
 * \brief A sprite that contains animation data.
 */
class AnimatedSprite : public Sprite
{
public:
	
	/*!
	 * \struct AnimationFrame
	 * \brief Holds all of the data neccecary for rendering a frame in an animation.
	 */
	struct AnimationFrame
	{
		float duration; /*!< The amount of time the frame is active. */
		Rectf area; /*!< The rectangle defining the subarea where the frame resides. */
	};
	
	/*!
	 * \struct Animation
	 * \brief Holds all of the data neccecary for rendering a complete animation.
	 */
	struct Animation
	{
		std::string name; /*!< The name of the animation. */
		std::vector<AnimationFrame> frames; /*!< The frames in the animation. */
		float duration; /*!< The total amount of time to render the animation. */
		uint index; /*!< The index of the animation. */
	};
	
	/*!
	 * \brief Loads The animated sprite data from the specified file.
	 *
	 * \param[in] file the file containing the animation data, assumed to be relitive to the 'Animations' data paths.
	 *
	 * \returns True if the animated sprite was loaded, false otherwise.
	 */
	bool LoadAnimatedSprite(const char *file);
	
	/*!
	 * \brief Gets the number of animations the loaded sprite has.
     *
     * \return The number of animations in the Loaded sprite.
	 */
	uint GetNumberOfAnimations() const;
	
	/*!
	 * \brief Gets a spacific Animation.
     *
	 * \param[in] index the index to the Animation.
	 *
     * \return The specified animation.
	 */
	const Animation &GetAnimation(uint index) const;
	
	/*!
	 * \brief Finds an animation by name.
	 *
	 * \param[in] name the name of the animation to find.
	 *
	 * \return The index to the first animation with the given \a name, returns -1 if the animation dosn't exist.
	 */
	int FindAnimation(const char *name) const;
	
	/*!
	 * \brief Draws a spacific frame of an animation.
	 *
	 * \param[in] animationNumber the index to the animation to draw from use.
	 * \param[in] frameNumber the index to the frame to draw from the animation.
	 */
	void DrawFrame(uint animationNumber, uint frameNumber) const;
	
	/*!
	 * \brief Draws a spacific frame of an animation.
	 *
	 * \param[in] anim the animation to draw from.
	 * \param[in] frameNumber the index to the frame to draw from the animation.
	 */
	void DrawFrame(const Animation &anim, uint frameNumber) const;
	
	/*!
	 * \brief Draws a spacific frame of an animation.
	 *
	 * \param[in] frame the AnimationFrame to draw.
	 */
	void DrawFrame(const AnimationFrame &frame) const;
	
protected:
	std::vector<Animation> animations;
};

#endif
