/*!
 * \file Color.h
 * \brief Defines the Color class.
 * \author Joe Forte
 * \date 3/18/2008
 *
 * Defines the Color class for holding 32bit floating point RGBA colors
 */

#ifndef COLOR_H
#define COLOR_H

#include "Common/BaseTypes.h"

/*!
 * \class Color4f
 * \headerfile Color.h "Graphics/Color.h"
 * \brief Class for holding 32bit floating point RGBA color data.
 */
class Color4f
{
public:
	float r, g, b, a; /*!< (r)ed, (g)reen, (b)lue and (a)lpha data. */

	/*!
	 * \brief Default constructor.  Defaults to Black.
	 */
	Color4f() : r(0.0f), g(0.0f), b(0.0f), a(1.0f) {}
	
	/*!
	 * \brief Constructs a spacific color.  Alpha set to 1.
	 *
	 * \param[in] _r red value.
	 * \param[in] _g green value.
	 * \param[in] _b blue value.
	 */
	Color4f(float _r, float _g, float _b) : r(_r), g(_g), b(_b), a(1.0f) {}
	
	/*!
	 * \brief Constructs a spacific color. 
	 *
	 * \param[in] _r red value.
	 * \param[in] _g green value.
	 * \param[in] _b blue value.
	 * \param[in] _a alpha value.
	 */
	Color4f(float _r, float _g, float _b, float _a) : r(_r), g(_g), b(_b), a(_a) {}
	
	/*!
	 * \brief Constructs a spacific color from integer data.  Alpha set to 1.
	 *
	 * \param[in] _r red value.
	 * \param[in] _g green value.
	 * \param[in] _b blue value.
	 */
	Color4f(byte _r, byte _g, byte _b) : r(_r/255.0f), g(_g/255.0f), b(_b/255.0f), a(1.0f) {}
	
	/*!
	 * \brief Constructs a spacific color from integer data. 
	 *
	 * \param[in] _r red value.
	 * \param[in] _g green value.
	 * \param[in] _b blue value.
	 * \param[in] _a alpha value.
	 */
	Color4f(byte _r, byte _g, byte _b, byte _a) : r(_r/255.0f), g(_g/255.0f), b(_b/255.0f), a(_a/255.0f) {}
	
	/*!
	 * \brief Constructs a spacific color from integer data. 
	 *
	 * \param[in] c the color expected to be in the form 0xRRGGBBAA.
	 */
	Color4f(uint c) : r(((c >> 24) & 0xFF)/255.0f), g(((c >> 16) & 0xFF)/255.0f), b(((c >> 8) & 0xFF)/255.0f), a((c & 0xFF)/255.0f) {}

	/*!
	 * \brief Converts 32bit floating point color data to byte data.
	 *
	 * \param[out] _r red value.
	 * \param[out] _g green value.
	 * \param[out] _b blue value.
	 */
	void Get(byte &_r, byte &_g, byte &_b) const
	{
		_r = (byte)(r * 255.0f);
		_g = (byte)(g * 255.0f);
		_b = (byte)(b * 255.0f);
	}

	/*!
	 * \brief Converts 32bit floating point color data to byte data.
	 *
	 * \param[out] _r red value.
	 * \param[out] _g green value.
	 * \param[out] _b blue value.
	 * \param[out] _a alpha value.
	 */
	void Get(byte &_r, byte &_g, byte &_b, byte &_a) const
	{
		_r = (byte)(r * 255.0f);
		_g = (byte)(g * 255.0f);
		_b = (byte)(b * 255.0f);
		_a = (byte)(a * 255.0f);
	}

	/*!
	 * \brief Converts 32bit floating point color data into an integer of the form 0xRRGGBBAA.
	 *
	 * \return An integer of the form 0xRRGGBBAA.
	 */
	unsigned int Get() const
	{
		byte _r, _g, _b, _a;
		Get(_r, _g, _b, _a);
		
		return (_r << 24) | (_g << 16) | (_b << 8) | _a;
	}

	/*!
	 * \brief Overloaded color addition operator.
	 *
	 * \param[in] c the color to add.
	 *
	 * \return The result of adding \e this color to \a c. 
	 */
	Color4f operator+(const Color4f &c) const
	{
		return Color4f(r + c.r, g + c.g, b + c.b, a + c.a);
	}

	/*!
	 * \brief Overloaded color subtraction operator.
	 *
	 * \param[in] c the color to subtract.
	 *
	 * \return The result of subtracting \a c from \e this color. 
	 */
	Color4f operator-(const Color4f &c) const
	{
		return Color4f(r - c.r, g - c.g, b - c.b, a - c.a);
	}

	/*!
	 * \brief Overloaded constant-color multiplication operator.
	 *
	 * \param[in] c the constant to multiply by.
	 *
	 * \return The result of multiplying \a c to \e this color. 
	 */
	Color4f operator*(float c) const
	{
		return Color4f(r*c, g*c, b*c, a*c);
	}
	
	/*!
	 * \brief Overloaded color equality operator.
	 *
	 * \param[in] c the color to compare against.
	 *
	 * \return True if the r, g, b and a components of \e this color are equal to the components of \a c.
	 * False is returned otherwise.
	 */
	bool operator==(const Color4f &c) const
	{
		return (r == c.r) && (g == c.g) && (b == c.b) && (a == c.a);
	}

	/*!
	 * \brief Overloaded color-constant multiplication operator.
	 *
	 * \param[in] c the constant to multiply by.
	 * \param[in] col the color to multiply.
	 *
	 * \return The result of multiplying \a col by \a c. 
	 */
	friend Color4f operator*(float c, const Color4f &col)
	{
		return Color4f(c*col.r, c*col.g, c*col.b, c*col.a);
	}
};

/*!
 * \namespace Color
 * \brief Defines useful color constants.
 */
namespace Color
{
	const Color4f Red(1.0f, 0.0f, 0.0f);				/*!< Red */
	const Color4f Green(0.0f, 1.0f, 0.0f);				/*!< Green */
	const Color4f Blue(0.0f, 0.0f, 1.0f);				/*!< Blue */
	const Color4f Black(0.0f, 0.0f, 0.0f);				/*!< Black */
	const Color4f Yellow(1.0f, 1.0f, 0.0f);				/*!< Yellow */
	const Color4f Cyan(0.0f, 1.0f, 1.0f);				/*!< Cyan */
	const Color4f Magenta(1.0f, 0.0f, 1.0f);			/*!< Magenta */
	const Color4f White(1.0f, 1.0f, 1.0f);				/*!< White */

	const Color4f Gold(1.0f, 0.84f, 0.0f);              /*!< Gold */
	const Color4f Purple(0.5f, 0.0f, 0.5f);				/*!< Purple */
	const Color4f Orange(1.0f, 0.5f, 0.0f);				/*!< Orange */
	const Color4f Pink(1.0f, 0.7529412f, 0.7960784f);	/*!< Pink */

	const Color4f LightGray(0.2f, 0.2f, 0.2f);			/*!< LightGray */
	const Color4f Gray(0.5f, 0.5f, 0.5f);				/*!< Gray */
	const Color4f DarkGray(0.8f, 0.8f, 0.8f);			/*!< DarkGray */

	const Color4f CornFlowerBlue((byte)100, (byte)149, (byte)237); /*!< CornFlowerBlue */
};

#endif
