/*!
 * \file Curve.h
 * \brief Defines classes for representing variouse types of curves.
 * \author Joe Forte
 * \date 2/13/2009
 */

#ifndef CURVE_H
#define CURVE_H

#include <assert.h>
#include "Math/MathCommon.h"
#include "Graphics/MathDraw.h"
#include "Graphics/Color.h"

/*!
 * \class CubicSpline
 * \headerfile Curve.h "Graphics/Curve.h"
 * \brief Represents a natural cubic spline.
 */
class CubicSpline
{
public:

	/*!
	 * \brief Constructor.
	 */
	CubicSpline();

	/*!
	 * \brief Constructor.
	 *
	 * \param[in] points a vector of control points. 
	 */
	CubicSpline(const std::vector<Point2f> &points);

	/*!
	 * \brief Constructor.
	 *
	 * \param[in] points an array of control points.
	 * \param[in] n the number of control points in the array.
	 */
	CubicSpline(const Point2f *points, uint n);

	/*!
	 * \brief Sets the control points to use for the spline.
	 *
	 * \param[in] points a vector of control points. 
	 */
	void SetControlPoints(const std::vector<Point2f> &points);

	/*!
	 * \brief Sets the control points to use for the spline.
	 *
	 * \param[in] points an array of control points.
	 * \param[in] n the number of control points in the array.
	 */
	void SetControlPoints(const Point2f *points, uint n);

	/*!
	 * \brief Evaluates the spline funciton.
	 *
	 * \param[in] t the value in the range [0,1] to evaluate the spline.
	 */
	Point2f operator() (float t) const;

private:

	// Holds the coefficents for the spline function: S(t) = a(t-ti)^3 + b(t-ti)^2 + c(t-ti) + d
	struct Coefficients
	{
		float a, b, c, d;
	};

	std::vector<Coefficients> functionCoefficnets;
	std::vector<Point2f> controlPoints;
};

/*!
 * \class BezierCurve
 * \headerfile Curve.h "Graphics/Curve.h"
 * \brief Represents an nth degree bezier curve.
 *
 * \tparam n the degree of the bezier curve.
 */
template<uint N = 3>
class BezierCurve
{
public:

	/*!
	 * \brief Constructs an nth degree Bezier Curve.
	 */
	BezierCurve();

	/*!
	 * \brief Deconstructor.
	 */
	~BezierCurve() {}

	/*!
	 * \brief Sets the ith control point of the curve.
	 *
	 * \param[in] i the index, must be in the range [0, N]
	 * \param[in] p the control point.
	 */
	void SetControlPoint(uint i, const Point2f &p);

	/*!
	 * \brief Gets the ith control point of the curve.
	 *
	 * \param[in] i the index, must be in the range [0, N]
	 *
	 * \return the ith control point.
	 */
	Point2f GetControlPoint(uint i) const;

	/*!
	 * \brief Sets the detail level that will be used to draw the curve.
	 *
	 * \param[in] detail higher values produce smoother curves but are slower to compute.
	 */
	void SetDetail(uint detail);

	/*!
	 * \brief Draws the curve.
	 *
	 * \param[in] color the color to draw the curve. Defaults to white.
	 * \param[in] width the width of the curve. Defaults to 2.
	 */
	void Draw(const Color4f &color = Color::White, float width = 2.0f);

	/*!
	 * \brief Draws the control points with connecting lines.
	 */
	void DrawControlPoints() const;

private:

	void ComputeDrawPoints();

	Point2f EvaluateBezierFunction(float t);

	std::vector<Point2f> drawPoints;

	Point2f controlPoints[N + 1];
	
	float constants[N + 1];

	float tIncrement;
};

//--------------------------------------------------------------------
template<uint N>
BezierCurve<N>::BezierCurve() : tIncrement(1.0f / 100.0f) 
{
	// Computes the constant factors in the bezier function
	constants[0] = 1.0f;
	constants[N] = 1.0f;

	uint mid = (N + 1) / 2;
	for(uint i = 1; i < mid; i++)
	{
		constants[i] = (float)(N * i);
	}

	for(uint i = 0; i < mid; i++)
	{
		constants[i + mid] = constants[mid - i - 1];
	}

}

//--------------------------------------------------------------------
template<uint N>
void BezierCurve<N>::SetControlPoint(uint i, const Point2f &p)
{
	assert(i <= N);
	controlPoints[i] = p;

	drawPoints.clear();
}

//--------------------------------------------------------------------
template<uint N>
Point2f BezierCurve<N>::GetControlPoint(uint i) const
{
	assert(i <= N);
	return controlPoints[i];
}

//--------------------------------------------------------------------
template<uint N>
void BezierCurve<N>::SetDetail(uint detail)
{
	tIncrement = 1.0f / (float)detail;

	drawPoints.clear();
}

//--------------------------------------------------------------------
template<uint N>
void BezierCurve<N>::Draw(const Color4f &color, float width)
{
	if(drawPoints.empty())
		ComputeDrawPoints();

	DrawLineStrip(drawPoints, color, width);
}

//--------------------------------------------------------------------
template<uint N>
void BezierCurve<N>::DrawControlPoints() const
{
	DrawLineStrip(controlPoints, N + 1, Color::Green, 1.0f);
	for(uint i = 0; i <= N; i++)
	{
		DrawPoint(controlPoints[i], Color::Red, 4.0f);
	}
}

//--------------------------------------------------------------------
template<uint N>
void BezierCurve<N>::ComputeDrawPoints()
{
	if(tIncrement <= 0.0f)
		tIncrement = 1.0f / 100.0f;

	float t = 0.0f;

	while(t < 1.0f)
	{
		drawPoints.push_back( EvaluateBezierFunction(t) );
		t = t + tIncrement;
	}
}

//--------------------------------------------------------------------
template<uint N>
Point2f BezierCurve<N>::EvaluateBezierFunction(float t)
{
	// Precomputes leading constant
	float tminus[N + 1];
	float tpower[N + 1];

	tminus[N] = 1.0f; 
	tpower[0] = 1.0f;

	const float mt = (1.0f - t);
	for(uint i = 1; i <= N; i++)
	{
		tminus[N - i] = mt * tminus[N - i + 1];
		tpower[i] = t * tpower[i - 1];
	}

	Point2f p(0.0f, 0.0f);
	for(uint i = 0; i <= N; i++)
	{		
		p = p + constants[i] * tpower[i] * tminus[i] * controlPoints[i];
	}

	return p;
}

#endif
