/*********************************************************************

	File: Font.cpp
	Data: 3/8/2008
	Desc: Implements the screen Font class.

**********************************************************************/

#include <assert.h>
#include <sstream>
#include <iostream>
#include <cmath>
#include <SDL.h>
#include <SDL_opengl.h>
#include <FTGL/ftgl.h>
#include "Common/Globals.h"
#include "Common/FileSystem.h"
#include "Graphics/Font.h"

const std::string Font::DEFAULT_FONT_PATH = CreatePath("VeraMono.ttf");
const int Font::DEFAULT_FONT_SIZE = 12;

//--------------------------------------------------------------------
Font::Font()
{
	font = NULL;
}

//--------------------------------------------------------------------
Font::~Font()
{
	Destroy();
}

//--------------------------------------------------------------------
IMPLEMENT_COMPLEX_RESOURCE_LOAD_FUNCTION(Font, int, size)
{
	Destroy();

	// Workaround for FTGL crashing when it gets a path that doesn't point to an actual file.
	if(!FileExists(filePath))
		return false;

	int pointSize = DEFAULT_FONT_SIZE;
	if(size)
		pointSize = *size;

	font = new FTGLTextureFont(filePath);

	font->Attach(filePath);
	font->FaceSize(pointSize);

	ptSize = pointSize;

	return true;
}

//--------------------------------------------------------------------
void Font::Destroy()
{
	if(font)
	{
		delete font;
		font = NULL;
	}
}

//--------------------------------------------------------------------
float Font::StringWidth(const char *str) const
{
	return font->Advance(str);
}

//--------------------------------------------------------------------
float Font::StringHeight(const char *str) const
{
	float w, h;
	StringDimensions(str, w, h);
	return h;
}

//--------------------------------------------------------------------
void Font::StringDimensions(const char *str, float &w, float &h) const
{
	float llx, lly, llz;
	float urx, ury, urz;
	font->BBox(str, llx, lly, llz, urx, ury, urz);

	w = std::abs(llx - urx);
	h = std::abs(lly - ury);
}

//--------------------------------------------------------------------
float Font::LineHeight() const
{
	return font->LineHeight();
}

//--------------------------------------------------------------------
Point2f Font::GetTextPosition(const char *str, const Rectf &area, TextAlignment align, const Point2f &offset) const
{
	float w, h;
	StringDimensions(str, w, h);

	float w2 = w / 2.0f;
	float h2 = h / 2.0f;
	float aw2 = area.w / 2.0f;
	float ah2 = area.h / 2.0f;

	// calculates the position of the string assuming the area is centered at the origin
	Point2f pos;
	switch (align)
	{
		case ALIGN_TOP:
			pos = Point2f(-w2, ah2);
			break;
		case ALIGN_TOP_LEFT:
			pos = Point2f(-aw2, ah2);
			break;
		case ALIGN_TOP_RIGHT:
			pos = Point2f(aw2 - w, ah2);
			break;
		case ALIGN_CENTER:
			pos = Point2f(-w2, h2);
			break;
		case ALIGN_LEFT:
			pos = Point2f(-aw2, h2);
			break;
		case ALIGN_RIGHT:
			pos = Point2f(aw2 - w, h2);
			break;
		case ALIGN_BOTTOM:
			pos = Point2f(-w2 ,-ah2 + h);
			break;
		case ALIGN_BOTTOM_LEFT:
			pos = Point2f(-aw2, -ah2 + h);
			break;
		case ALIGN_BOTTOM_RIGHT:
			pos = Point2f(aw2 - w, -ah2 + h);
			break;
	};

	// transforms the point to the true area of the
	Point2f center(area.x + (area.w / 2.0f), area.y - (area.h / 2.0f));

	return pos + center + offset - Point2f(0.0f, h);
}

//--------------------------------------------------------------------
void Font::DrawTextStr(const char *str, const Color4f &c, const Point2f &p) const
{
	Draw(str, c, Point2f(p.x, p.y - LineHeight()));
}

//--------------------------------------------------------------------
void Font::DrawTextStr(const char *str, const Color4f &c, const Rectf &area, TextAlignment align, const Point2f &offset) const
{
	Draw(str, c, GetTextPosition(str, area, align, offset));
}

//--------------------------------------------------------------------
void Font::Draw(const char *str, const Color4f &c, const Point2f &p) const
{
	assert(font && str);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(p.x, p.y, 0.0);

	glColor4f(c.r, c.g, c.b, c.a);

	font->Render(str);

	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	glPopMatrix();

	// FTGL neglects to unbind the font texture
	glBindTexture(GL_TEXTURE_2D, 0);
}

//--------------------------------------------------------------------
FontRef LoadFontResource(const char *file, int size)
{
	FontMgr *mgr = GetFontManager();
	assert(mgr);

	FontRef ref = mgr->LoadResource(file, &size);
	if(!ref.IsValid())
	{
		ConError() << "The Font resource |c00FF00FF'" << file << "'|r' could not be loaded.\n";
	}

	return ref;
}
