/*!
 * \file Font.h
 * \brief Defines the Font resource class.
 * \author Joe Forte
 * \date 3/8/2008
 *
 * Defines the Font resource class for loading and drawing True Type fonts.
 */

#ifndef FONT_H
#define FONT_H

#include <string>
#include "Memory/ResourceManager.h"
#include "Graphics/Color.h"
#include "Math/MathCommon.h"
#include "Common/BaseTypes.h"

class FTFont;

/*!
 * \class Font
 * \headerfile Font.h "Graphics/Font.h"
 * \brief The Font resource class.
 *
 * A resource type that is used for loading and drawing true type fonts.  All coordinates given or created by
 * the Font class are assumed to be in absolute screen coordinates unless otherwise noted.  
 */
class Font : public Resource<int>
{
public:
	Font();
	~Font();
	
	/*!< The path to the default font. */
	static const std::string DEFAULT_FONT_PATH;
	
	/*!< The default font size. */
	static const int DEFAULT_FONT_SIZE;

	/*
	 * Should only be called by the FontMgr class.
	 */
	DECLARE_COMPLEX_RESOURCE_LOAD_FUNCTION(int, size);
	
	/*
	 * Should only be called by the FontMgr class.
	 */
	virtual void Destroy();
	
	/*!
	 * \enum TextAlignment
	 * \brief defines the differant types of font alignment when drawing a string.
	 */
	enum TextAlignment
	{
		ALIGN_NONE = 0,		/*!< No alignment */
		ALIGN_TOP,			/*!< Top center alignment */
		ALIGN_TOP_LEFT,		/*!< Top left alignemnt */
		ALIGN_TOP_RIGHT,	/*!< Top right alignment */
		ALIGN_CENTER,		/*!< Center alignment */
		ALIGN_LEFT,			/*!< Left center alignment */
		ALIGN_RIGHT,		/*!< Right center alignment */
		ALIGN_BOTTOM,		/*!< Bottom center alignment */
		ALIGN_BOTTOM_LEFT,	/*!< Bottom left alignment */
		ALIGN_BOTTOM_RIGHT	/*!< Bottom right alignment */
	};

	/*!
	 * \brief Gets the width of the given string in absolute screen coordinates.
	 *
	 * \param[in] str string to determine the with of.
	 *
	 * \return The width of \a str in absolute screen coordinates.
	 */
	float StringWidth(const char *str) const;
	
	/*!
	 * \brief Gets the height of the given string in absolute screen coordinates.
	 *
	 * \param[in] str string to determine the height of.
	 *
	 * \return The height of \a str in absolute screen coordinates.
	 */
	float StringHeight(const char *str) const;
	
	/*!
	 * \brief Gets the width and height of the given string in absolute screen coordinates.
	 *
	 * \param[in] str string to determine the with and height of.
	 * \param[out] w is set to the width of \a str in absolute screen coordinates.
	 * \param[out] h is set to the height of \a str in absolute screen coordinates.
	 */
	void StringDimensions(const char *str, float &w, float &h) const;
	
	/*!
	 * \brief Gets the maximum height for all characters in the font.
	 *
	 * \return The maximum height for all characters in the font in absolute screen coordinates.
	 */
	float LineHeight() const;
	
	/*!
	 * \brief Determines the position to draw a string given some alignment and area criterion. 
	 *
	 * \param[in] str the string of text.
	 * \param[in] area the rectance where the text is to sit.
	 * \param[in] align the type of alignment to use.
	 * \param[in] offset the amount to offset from the final computed position.
	 *
	 * \return The position in absolute screen coordinates where \a str should be drawn to meet the alignment and area criterion.
	 */
	Point2f GetTextPosition(const char *str, const Rectf &area, TextAlignment align, const Point2f &offset = Point2f(0.0f, 0.0f)) const; 

	/*!
	 * \brief Draws a string of text.
	 *
	 * \param[in] str the string to draw.
	 * \param[in] c the color to draw this string.
	 * \param[in] p the top left corner where \a str will be drawn.
	 */
	void DrawTextStr(const char *str, const Color4f &c, const Point2f &p) const;
	
	/*!
	 * \brief Draws a string of aligned text.
	 *
	 * \param[in] str the string to draw.
	 * \param[in] c the color to draw this string.
	 * \param[in] area the Rect \a str is expected to be drawn into.
	 * \param[in] align the alignment of \a str within \a area.
	 * \param[in] offset the ammount to offset \a str from its given alignment.
	 */
	void DrawTextStr(const char *str, const Color4f &c, const Rectf &area, TextAlignment align, const Point2f &offset = Point2f(0.0f, 0.0f)) const;
	
private:
	
	void Draw(const char *str, const Color4f &c, const Point2f &p) const;
	
	int ptSize;

	FTFont *font;
};

DECLARE_COMPLEX_RESOURCE_TYPE(Font, int);

/*!
 * \brief Loads an Font using the global FontManager.
 * 
 * \param[in] file the filename of the Font to load assumed to be relative to the 'Fonts' data path.
 * \param[in] size the point size of the loaded Font.
 *
 * \return The FontRef of the loaded Font.
 *
 * NOTE: This should only be called after the global enviroment has been initialized.
 */
FontRef LoadFontResource(const char *file, int size);

#endif

