/*!
 * \file Image.h
 * \brief Implements the Image resource class.
 * \author Joe Forte
 * \date 3/8/2008
 */

#include <iostream>
#include <SDL_opengl.h>
#include <SOIL.h>
#include "Common/Path.h"
#include "Common/Globals.h"
#include "Graphics/Image.h"

const std::string Image::ERROR_IMAGE_PATH = CreatePath("error.png");

//--------------------------------------------------------------------
Image::Image() : texID(0) {}

//--------------------------------------------------------------------
Image::~Image()
{
	Destroy();
}

//--------------------------------------------------------------------
IMPLEMENT_RESOURCE_LOAD_FUNCTION(Image)
{
	Destroy();
	
	int channels = 0;
	int w = 0;
	int h = 0;
	unsigned char *data = SOIL_load_image(filePath, &w, &h, &channels, SOIL_LOAD_AUTO);
	if(!data)
		return false;
	
	texID = SOIL_create_OGL_texture(data, w, h, channels, SOIL_CREATE_NEW_ID, 0);
	if(texID == 0)
	{
		const char *err = SOIL_last_result();
		delete data;
		return false;
	}
	
	delete data;
	
	width = w;
	height = h;
	hasAlpha = channels == 4;
	
	return true;
}

//--------------------------------------------------------------------
void Image::Destroy()
{	
	if(texID)
		glDeleteTextures(1, &texID);

	texID = 0;
}

//--------------------------------------------------------------------
uint Image::GetDataUsageInBytes() const 
{
	if(HasAlpha())
	{
		return width * height * 4; 
	}
	else
	{
		return width * height * 3;
	}
}

//--------------------------------------------------------------------
int Image::Width() const
{
	return width;
}

//--------------------------------------------------------------------
int Image::Height() const
{
	return height;
}

//--------------------------------------------------------------------
bool Image::HasAlpha() const
{
	return hasAlpha; 
}

//--------------------------------------------------------------------
void Image::Draw(float alpha, DrawLocation loc) const
{
	Rectf r;
	r.x = 0.0f;
	r.y = 0.0f;
	r.w = (float)Width();
	r.h = (float)Height();
	
	DrawSubImage(r, alpha, loc);
}

//--------------------------------------------------------------------
void Image::DrawSubImage(const Rectf &r, float alpha, DrawLocation loc) const
{
	if(texID)
	{
		float imw = (float)Width();
		float imh = (float)Height();
		
		// Computes the texture coords
		float u0 = (r.x)       / imw; Clamp(u0, 0.0f, 1.0f);
		float u1 = (r.x + r.w) / imw; Clamp(u1, 0.0f, 1.0f);
		float v0 = (r.y)       / imh; Clamp(v0, 0.0f, 1.0f);
		float v1 = (r.y + r.h) / imh; Clamp(v1, 0.0f, 1.0f);
		
		float w = r.w;
		float h = r.h;
		
		glBindTexture(GL_TEXTURE_2D, texID);
		
		glBegin(GL_QUADS);
		
		glColor4f(1.0f, 1.0f, 1.0f, alpha);
		
		switch(loc)
		{
			case CENTER:
			{
				float w_2 = w / 2.0f;
				float h_2 = h / 2.0f;
				
				glTexCoord2f(u0, v1); glVertex3f(-w_2, -h_2, 0.0f);
				glTexCoord2f(u1, v1); glVertex3f( w_2, -h_2, 0.0f);
				glTexCoord2f(u1, v0); glVertex3f( w_2,  h_2, 0.0f);
				glTexCoord2f(u0, v0); glVertex3f(-w_2,  h_2, 0.0f);
				break;
			}
				
			case TOPLEFT:
				glTexCoord2f(0.0f, 1.0f); glVertex3f(0.0f,   -h, 0.0f);
				glTexCoord2f(1.0f, 1.0f); glVertex3f(   w,   -h, 0.0f);
				glTexCoord2f(1.0f, 0.0f); glVertex3f(   w, 0.0f, 0.0f);
				glTexCoord2f(0.0f, 0.0f); glVertex3f(0.0f, 0.0f, 0.0f);
				break;
				
			case TOPRIGHT:
				glTexCoord2f(0.0f, 1.0f); glVertex3f(  -w,   -h, 0.0f);
				glTexCoord2f(1.0f, 1.0f); glVertex3f(0.0f,   -h, 0.0f);
				glTexCoord2f(1.0f, 0.0f); glVertex3f(0.0f, 0.0f, 0.0f);
				glTexCoord2f(0.0f, 0.0f); glVertex3f(  -w, 0.0f, 0.0f);
				break;
				
			case BOTTOMLEFT:
				glTexCoord2f(0.0f, 1.0f); glVertex3f(0.0f, 0.0f, 0.0f);
				glTexCoord2f(1.0f, 1.0f); glVertex3f(   w, 0.0f, 0.0f);
				glTexCoord2f(1.0f, 0.0f); glVertex3f(   w,    h, 0.0f);
				glTexCoord2f(0.0f, 0.0f); glVertex3f(0.0f,    h, 0.0f);
				break;
				
			case BOTTOMRIGHT:
				glTexCoord2f(0.0f, 1.0f); glVertex3f(  -w, 0.0f, 0.0f);
				glTexCoord2f(1.0f, 1.0f); glVertex3f(0.0f, 0.0f, 0.0f);
				glTexCoord2f(1.0f, 0.0f); glVertex3f(0.0f,    h, 0.0f);
				glTexCoord2f(0.0f, 0.0f); glVertex3f(   w,    h, 0.0f);
				break;
		};
		
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		
		glEnd();
		
		glBindTexture(GL_TEXTURE_2D, 0);
	}
}

//--------------------------------------------------------------------
ImageRef LoadImageResource(const char *file)
{
	ImageMgr *mgr = GetImageManager();
	assert(mgr);
	
	ImageRef ref = mgr->LoadResource(file);
	if(!ref.IsValid())
	{
		ConError() << "The Image resource |c00FF00FF'" << file << "'|r' could not be loaded.\n";
	}
	
	return ref;
}
