/*!
 * \file Image.h
 * \brief Defines the Image resource class.
 * \author Joe Forte
 * \date 3/8/2008
 *
 * Defines the Image resource class for loading and drawing image files.
 */

#ifndef IMAGE_H
#define IMAGE_H

#include <string>
#include "Common/BaseTypes.h"
#include "Math/MathCommon.h"
#include "Memory/ResourceManager.h"

/*!
 * \class Image
 * \headerfile Image.h "Graphics/Image.h"
 * \brief The Image resource class.
 *
 * A resource type that is used for loading and drawing image files.  The following image formats are supported:
 * -# PNG
 * -# JPG
 * -# TGA
 * -# BMP
 * -# DDS
 */
class Image : public Resource<>
{
public:
	
	/*!< The path to an image that will always exist. Used when loading an image that does not exist. */
	static const std::string ERROR_IMAGE_PATH;
	
	Image();
	~Image();

	/*
	 * Should only be called by the FontMgr class.
	 */
	DECLARE_RESOURCE_LOAD_FUNCTION();

	/*
	 * Should only be called by the FontMgr class.
	 */
	virtual void Destroy();

	/*!
	 * \brief Estimates the number of bytes used by the loaded image resource.
	 */
	virtual uint GetDataUsageInBytes() const;

	/*!
	 * \brief Gets the width in pixels of the currently loaded image.
	 *
	 * \return The width in pixels of the image.
	 */
	int Width() const;
	
	/*!
	 * \brief Gets the height in pixels of the currently loaded image.
	 *
	 * \return The height in pixels of the image.
	 */
	int Height() const;

	/*!
	 * \brief Determines if the loaded image has an alpha channel.
	 *
	 * \return True if the loaded image has an alpha channel, false otherwise.
	 */
	bool HasAlpha() const;
	
	/*!
	 * \enum DrawLocation
	 * \brief Defines the differant location to draw the image from.
	 */
	enum DrawLocation
	{
		CENTER = 0, /*!< The center of the image is drawn at the origin */
		TOPLEFT,	/*!< The top left corner of the image is drawn at the origin */
		TOPRIGHT,	/*!< The top right corner of the image is drawn at the origin */
		BOTTOMLEFT,	/*!< The bottom left corner of the image is drawn at the origin */
		BOTTOMRIGHT	/*!< The bottom right corner of the image is drawn at the origin */
	};
	
	/*!
	 * \brief Draws the image.
	 *
	 * \param[in] alpha the opacity to apply uniformly to the image. Defaults to 1 (fully opaque).
	 * \param[in] loc the location to draw the image. Defaults to CENTER.
	 * \see DrawSubImage
	 *
	 * The image is drawn at the origin with the same width and height as the actual image.  The
	 * DrawLocation is used to determine the place where the image is drawn at the origin.  For example,
	 * CENTER draws the image center at the origin whereas TOPLEFT draws the image with the top left corner
	 * of the image at the origin.
	 */
	void Draw(float alpha = 1.0f, DrawLocation loc = CENTER) const;
	
	/*!
	 * \brief Draws a portion of the image.
	 *
	 * \param[in] r the area of the image to draw.  Will be clamped to the actual bounds of the image.  The top
	 * left corner of the image is at (0,0) and the bottom right position is at (width,height).
	 * \param[in] alpha the opacity to apply uniformly to the image. Defaults to 1 (fully opaque).
	 * \param[in] loc the location to draw the image. Defaults to CENTER.
	 * \see Draw
	 *
	 * The subimage is drawn at the origin with the same width and height as specified by \a r.  
	 * The DrawLocation is used to determine the place where the image is drawn at the origin.  For example,
	 * CENTER draws the image center at the origin whereas TOPLEFT draws the image with the top left corner
	 * of the image at the origin.
	 */
	void DrawSubImage(const Rectf &r, float alpha = 1.0f, DrawLocation loc = CENTER) const;
	
	/*!
	 * \brief Returns the texture ID generated by opengl.
	 *
	 * \return The opengl texture id for the image.
	 */
	uint GetOpenGLID() const { return texID; }

private:
	uint texID;
	
	uint width;
	uint height;
	bool hasAlpha;
};

DECLARE_RESOURCE_TYPE(Image);

/*!
 * \brief Loads an image using the global ImageManager.
 * 
 * \param[in] file the filename of the image to load assumed to be relative to the 'Textures' data path.
 *
 * \return The ImageRef of the loaded image.
 *
 * NOTE: This should only be called after the global enviroment has been initialized.
 */
ImageRef LoadImageResource(const char *file);

#endif

