/*********************************************************************
 
	File: MathDraw.cpp
	Data: 3/18/2008
	Desc: Implements functions for drawing common math types
 
**********************************************************************/

#include <SDL_opengl.h>
#include "Common/Globals.h"
#include "Graphics/MathDraw.h"

//--------------------------------------------------------------------
void DrawLine(const Line &l, const Color4f &color, float width)
{
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	glLineWidth(width);

	glBegin(GL_LINES);

		glColor4f(color.r, color.g, color.b, color.a);
		glVertex3f(l.a.x, l.a.y, 0.0f);
		glVertex3f(l.b.x, l.b.y, 0.0f);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	glEnd();
	
	glPopMatrix();
}

//--------------------------------------------------------------------
void DrawLineStrip(const std::vector<Point2f> &points, const Color4f &color, float width)
{
	uint num = (uint)points.size();
	if(num < 2)
		return;

	DrawLineStrip(&points[0], num, color, width);
}

//--------------------------------------------------------------------
void DrawLineStrip(const Point2f *points, uint n, const Color4f &color, float width)
{
	if(points == NULL || n < 2)
		return;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	glLineWidth(width);

	glBegin(GL_LINE_STRIP);

		glColor4f(color.r, color.g, color.b, color.a);

		for(uint i = 0; i < n; i++)
		{
			glVertex3f(points[i].x, points[i].y, 0.0f);
		}

		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	glEnd();
	
	glPopMatrix();
}

//--------------------------------------------------------------------
void DrawPoint(const Point2f &p, const Color4f &color, float size)
{
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	glPointSize(size);

	glBegin(GL_POINTS);

		glColor4f(color.r, color.g, color.b, color.a);
		glVertex3f(p.x, p.y, 0.0f);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	glEnd();
	
	glPopMatrix();
}

//--------------------------------------------------------------------
void DrawVector(const Vector2f &v, const Point2f &origin, const Color4f &color, float width)
{
	Point2f end = origin + v;
	
	// determines where to place the lines to create the arrow head
	Point2f xd(-SQRT2_2, SQRT2_2);
	Point2f yd(-SQRT2_2, -SQRT2_2);

	Matrixf R, T, S;

	float a = v.Angle( Vector2f(1.0f, 0.0f) );
	if( Sign(v.y) < 0 ) // Accounts for the range limitation of arccos
		a = -a;

	R.MakeRotation(a);
	T.MakeTranslation(end.x, end.y);

	float s = v.Length() / 10.f;
	S.MakeScale(s, s);

	Matrixf trans = T * R * S;
	yd = trans.Transform( yd );
	xd = trans.Transform( xd );
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	// Draws the arrow using lines
	glLineWidth(width);
	glBegin(GL_LINES);

		glColor4f(color.r, color.g, color.b, color.a);
		glVertex3f(origin.x, origin.y, 0.0f);
		glVertex3f(end.x, end.y, 0.0f);

		glVertex3f(end.x, end.y, 0.0f);
		glVertex3f(yd.x, yd.y, 0.0f);
		
		glVertex3f(end.x, end.y, 0.0f);
		glVertex3f(xd.x, xd.y, 0.0f);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	glEnd();
	
	glPopMatrix();
}

//--------------------------------------------------------------------
void DrawCircle(const Circle &c, const Color4f &color, float width)
{
	const int detail = 64;
	const float inc = _2PI / (float)detail;
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	glLineWidth(width);
	glBegin(GL_LINE_LOOP);
	
		glColor4f(color.r, color.g, color.b, color.a);
		
		float a = 0.0f;
	
		// iterates one less then detail because we are using a line loop
		for(int i = 0; i < (detail - 1); i++)
		{
			float x = c.r * std::cos(a) + c.center.x;
			float y = c.r * std::sin(a) + c.center.y;
			
			glVertex3f(x, y, 0.0f);
			
			a = a + inc;
		}

		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	
	glEnd();
	
	glPopMatrix();
}

//--------------------------------------------------------------------
void DrawRay(const Ray &r, float len, const Color4f &color, float width)
{
	DrawVector(r.dir * len, r.start, color, width);
}

//--------------------------------------------------------------------
void DrawAABB(const AABB &aabb, const Matrixf &m, const Color4f &color, float width)
{
	float w = aabb.Width();
	float h = aabb.Height();
	
	glLineWidth(width);

	float glMat[16];
	m.GetOpenGLMatrix(glMat);
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMultMatrixf(glMat);
	
	glBegin(GL_LINE_LOOP);
	
		glColor4f(color.r, color.g, color.b, color.a);
		glVertex3f(aabb.min.x, aabb.min.y, 0.0f);
		glVertex3f(aabb.min.x + w, aabb.min.y, 0.0f);
	
		glVertex3f(aabb.max.x, aabb.max.y, 0.0f);
		glVertex3f(aabb.max.x - w, aabb.max.y, 0.0f);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	
	glEnd();
	
	glPopMatrix();
}

//--------------------------------------------------------------------
void DrawAABB(const AABB &aabb, const Color4f &color, float width)
{
	Matrixf I;
	I.MakeIdentity();

	DrawAABB(aabb, I, color, width);
}

//--------------------------------------------------------------------
void DrawRect(const Rectf &r, bool fill, const Color4f &color, float width)
{
	Point2f c = RectToAABB<float>(r).CenterPosition();
	
	DrawQuad(c, r.w, r.h, fill, color, width);
}

//--------------------------------------------------------------------
void DrawQuad(const Point2f &center, float w, float h, bool fill, const Color4f &color, float width)
{
	Matrixf I;
	I.MakeIdentity();

	DrawQuad(center, w, h, I, fill, color, width);
}

//--------------------------------------------------------------------
void DrawQuad(const Point2f &center, float w, float h, const Matrixf &m, bool fill, const Color4f &color, float width)
{
	if(!fill)
	{
		DrawAABB( AABB(center, w, h), m, color, width );
	}
	else
	{
		Matrixf T, M;
		T.MakeTranslation(center.x, center.y);
		M = T*m;

		float glMat[16];
		M.GetOpenGLMatrix(glMat);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glMultMatrixf(glMat);

		w = w / 2.0f;
		h = h / 2.0f;
		glBegin(GL_QUADS);

			glColor4f(color.r, color.g, color.b, color.a);
			glVertex3f(-w, -h, 0.0f);
			glVertex3f( w, -h, 0.0f);
			glVertex3f( w,  h, 0.0f);
			glVertex3f(-w,  h, 0.0f);
			glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

		glEnd();

		glPopMatrix();
	}
}

//--------------------------------------------------------------------
void DrawTriangle(const Point2f &a, const Point2f &b, const Point2f &c, const Color4f &color)
{
	glBegin(GL_TRIANGLES);

	glColor4f(color.r, color.g, color.b, color.a);

	glVertex3f(a.x, a.y, 0.0f);
	glVertex3f(b.x, b.y, 0.0f);
	glVertex3f(c.x, c.y, 0.0f);

	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	glEnd();
}

//--------------------------------------------------------------------
void DrawPoly(const Point2f *verts, int n, const Matrixf &m, const Color4f c, float width)
{
	if(n <= 0)
		return;
	
	float glMat[16];
	m.GetOpenGLMatrix(glMat);
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMultMatrixf(glMat);
	
	glLineWidth(width);
	
	glBegin(GL_LINES);
	
	glColor4f(c.r, c.g, c.b, c.a);
	
	uint len = n - 1;
	for(uint i = 0; i < len; i++)
	{
		glVertex3f(verts[i].x, verts[i].y, 0.0f);
		glVertex3f(verts[i + 1].x, verts[i + 1].y, 0.0f);
	}
	
	glVertex3f(verts[0].x, verts[0].y, 0.0f);
	glVertex3f(verts[len].x, verts[len].y, 0.0f);
	
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	
	glEnd();
	
	glPopMatrix();
}

//--------------------------------------------------------------------
void DrawPoly(const std::vector<Point2f> &verts, const Point2f &p, const Color4f c, float width)
{
	if(verts.empty())
		return;

	Matrixf T;
	T.MakeTranslation(p.x, p.y);

	DrawPoly(&(verts[0]), (uint)verts.size(), T, c, width);
}

//--------------------------------------------------------------------
void DrawPoly(const Point2f *verts, int n, const Point2f &p, const Color4f c, float width)
{
	if(n <= 0)
		return;

	Matrixf T;
	T.MakeTranslation(p.x, p.y);

	DrawPoly(verts, n, T, c, width);
}

//--------------------------------------------------------------------
void DrawFullscreenQuad(const Color4f &color)
{
	uint w, h;
	GetEnv()->ScreenDim(w, h);
	
	DrawQuad( Point2f(0.0f, 0.0f), (float)w, (float)h, true, color);
}
