/*!
 * \file MathDraw.h
 * \brief Defines functions for drawing common math objects.
 * \author Joe Forte
 * \date 3/8/2008
 */

#include <vector>
#include "Graphics/Color.h"
#include "Math/MathCommon.h"

#ifndef MATHDRAW_H
#define MATHDRAW_H

/*!
 * \brief Draws a Line.
 *
 * \param[in] l the Line to draw.
 * \param[in] color the color to draw the line.  Defaults to Gray.
 * \param[in] width the width of the line should be drawn.  Defaults to 2.
 */
void DrawLine(const Line &l, const Color4f &color = Color::Gray, float width = 2.0f);

/*!
 * \brief Draws a sequence of connected Lines.
 *
 * \param[in] points the points that define the sequence of lines.
 * \param[in] color the color to draw the line.  Defaults to Gray.
 * \param[in] width the width of the line should be drawn.  Defaults to 2.
 */
void DrawLineStrip(const std::vector<Point2f> &points, const Color4f &color = Color::Gray, float width = 2.0f);

/*!
 * \brief Draws a sequence of connected Lines.
 *
 * \param[in] points the points that define the sequence of lines.
 * \param[in] n the number of points.
 * \param[in] color the color to draw the line.  Defaults to Gray.
 * \param[in] width the width of the line should be drawn.  Defaults to 2.
 */
void DrawLineStrip(const Point2f *points, uint n, const Color4f &color = Color::Gray, float width = 2.0f);

/*!
 * \brief Draws a Point.
 *
 * \param[in] p the Point to draw.
 * \param[in] color the color to draw the Point.  Defaults to Red.
 * \param[in] size the size of the Point should be drawn.  Defaults to 5.
 */
void DrawPoint(const Point2f &p, const Color4f &color = Color::Red, float size = 5.0f);

/*!
 * \brief Draws a Vector.
 *
 * \param[in] v the Vector to draw.
 * \param[in] origin the origin of the coordinate system the vector is in.  Defaults to (0,0).
 * \param[in] color the color to draw the vector.  Defaults to CornFlowerBlue.
 * \param[in] width the width of the Vector should be drawn.  Defaults to 2.
 */
void DrawVector(const Vector2f &v, const Point2f &origin = Point2f(0.0f, 0.0f), const Color4f &color = Color::CornFlowerBlue, float width = 2.0f);

/*!
 * \brief Draws a Circle.
 *
 * \param[in] c the Circle to draw.
 * \param[in] color the color to draw the Circle.  Defaults to Yellow.
 * \param[in] width the width of the Circle should be drawn.  Defaults to 2.
 */
void DrawCircle(const Circle &c, const Color4f &color = Color::Yellow, float width = 2.0f);

/*!
 * \brief Draws a Ray.
 *
 * \param[in] r the Ray to draw.
 * \param[in] len the length of the ray.  Defaults to 10.
 * \param[in] color the color to draw the Ray.  Defaults to Blue.
 * \param[in] width the width of the Ray should be drawn.  Defaults to 2.
 */
void DrawRay(const Ray &r, float len = 10.0f, const Color4f &color = Color::Blue, float width = 2.0f);

/*!
 * \brief Draws an AABB.
 *
 * \param[in] aabb the AABB to draw.
 * \param[in] color the color to draw the AABB.  Defaults to Green.
 * \param[in] width the width of the AABB should be drawn.  Defaults to 2.
 */
void DrawAABB(const AABB &aabb, const Color4f &color = Color::Green, float width = 2.0f);

/*!
 * \brief Draws an AABB.
 *
 * \param[in] aabb the AABB to draw.
 * \param[in] m the transformation to apply to the AABB.
 * \param[in] color the color to draw the AABB.  Defaults to Green.
 * \param[in] width the width of the AABB should be drawn.  Defaults to 2.
 */
void DrawAABB(const AABB &aabb, const Matrixf &m, const Color4f &color = Color::Green, float width = 2.0f);

/*!
 * \brief Draws a Rect.
 *
 * \param[in] r the Rect to draw.
 * \param[in] fill set to true to fill the Rect with the given color.  Set to false to only draw the outline of the Rect. 
 * \param[in] color the color to draw the Rect.  Defaults to Green.
 * \param[in] width the width of the Rect should be drawn.  Defaults to 2.  Ignored if \a fill equals true.
 */
void DrawRect(const Rectf &r, bool fill = true, const Color4f &color = Color::Green, float width = 2.0f);

/*!
 * \brief Draws a Quad.
 *
 * \param[in] center the center position of the Quad to draw.
 * \param[in] w the width of the Quad to Draw.
 * \param[in] h the height of the Quad to Draw.
 * \param[in] fill set to true to fill the Quad with the given color.  Set to false to only draw the outline of the Quad. 
 * \param[in] color the color to draw the Quad.  Defaults to Green.
 * \param[in] width the width of the Rect should be drawn.  Defaults to 2.  Ignored if \a fill equals true.
 */
void DrawQuad(const Point2f &center, float w, float h, bool fill = true, const Color4f &color = Color::Green, float width = 2.0f);

/*!
 * \brief Draws a Quad.
 *
 * \param[in] center the center position of the Quad to draw.
 * \param[in] w the width of the Quad to Draw.
 * \param[in] h the height of the Quad to Draw.
 * \param[in] m the transformation to apply to the quad.
 * \param[in] fill set to true to fill the Quad with the given color.  Set to false to only draw the outline of the Quad. 
 * \param[in] color the color to draw the Quad.  Defaults to Green.
 * \param[in] width the width of the Rect should be drawn.  Defaults to 2.  Ignored if \a fill equals true.
 */
void DrawQuad(const Point2f &center, float w, float h, const Matrixf &m, bool fill = true, const Color4f &color = Color::Green, float width = 2.0f);

/*!
 * \brief Draws a solid Triangle.
 *
 * \param[in] a the first point.
 * \param[in] b the second point.
 * \param[in] c the third point.
 * \param[in] color the color to draw the Triangle.
 */
void DrawTriangle(const Point2f &a, const Point2f &b, const Point2f &c, const Color4f &color);

/*!
 * \brief Draws a Polygon.
 *
 * \param[in] verts the vertices for the polygon where the first and the last vertices are connected.
 * \param[in] p the position to draw the polygon.  Defaults to the origin.
 * \param[in] c the color to draw the polygon.
 * \param[in] width the width of the lines used to draw the polygon.
 */
void DrawPoly(const std::vector<Point2f> &verts, const Point2f &p = Point2f(0.0f, 0.0f), const Color4f c = Color::Green, float width = 2.0f);

/*!
 * \brief Draws a Polygon.
 *
 * \param[in] verts the vertices for the polygon where the first and the last vertices are connected.
 * \param[in] n the number of vertices.
 * \param[in] m the transformation Matrix to apply to the poly.
 * \param[in] c the color to draw the polygon.
 * \param[in] width the width of the lines used to draw the polygon.
 */
void DrawPoly(const Point2f *verts, int n, const Matrixf &m, const Color4f c = Color::Green, float width = 2.0f);

/*!
 * \brief Draws a Polygon.
 *
 * \param[in] verts the vertices for the polygon where the first and the last vertices are connected.
 * \param[in] n the number of vertices.
 * \param[in] p the position to draw the polygon.  Defaults to the origin.
 * \param[in] c the color to draw the polygon.
 * \param[in] width the width of the lines used to draw the polygon.
 */
void DrawPoly(const Point2f *verts, int n, const Point2f &p = Point2f(0.0f, 0.0f), const Color4f c = Color::Green, float width = 2.0f);

/*!
 * \brief Draws a Quad that fills the whole screen.
 * 
 * \param[in] color the color to draw the fullscreen Quad.
 */
void DrawFullscreenQuad(const Color4f &color);

#endif
