/*!
 * \file ParticleSystem.cpp
 * \brief Implements the ParticleSystem and associated classes.
 * \author Joe Forte
 * \date 7/17/2008
 */

#include <gl/glew.h>
#include "Graphics/ParticleSystem.h"
#include "Common/Globals.h"

struct ParticleVertex
{
	float x, y, z;
	float r, g, b, a;
	float u, v;
};

const uint PARTICLEVERTEX_COLOR_OFFSET = sizeof(float) * 3;
const uint PARTICLEVERTEX_TEX_OFFSET = sizeof(float) * 7;

//--------------------------------------------------------------------
DefaultParticleDrawPolicy::~DefaultParticleDrawPolicy()
{
	if(vbo)
		glDeleteBuffers(1, (const GLuint*)&vbo);
	
	if(ibo)
		glDeleteBuffers(1, (const GLuint*)&ibo);
}

//--------------------------------------------------------------------
void DefaultParticleDrawPolicy::CreateBuffers()
{
	glGenBuffers(1, (GLuint*)&vbo);
	
	// Creates the static index buffer
	const uint numIndices = maxParticles*6;
	uint *indices = new uint[numIndices];
	
	uint v = 0;
	uint i = 0;
	while(i < numIndices)
	{
		// Triangle 1
		indices[i++] = v;
		indices[i++] = v + 3;
		indices[i++] = v + 2;
		
		// Triangle 2
		indices[i++] = v;
		indices[i++] = v + 2;
		indices[i++] = v + 1;
		
		v = v + 4;
	}
	
	glGenBuffers(1, (GLuint*)&ibo);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(uint)*numIndices, (const GLvoid*)indices, GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	delete [] indices;
}

//--------------------------------------------------------------------
void DefaultParticleDrawPolicy::Draw(const std::vector<size_t> &active, const std::vector<Particle> &particleStore)
{
	// Vertex buffer method
	
	const uint numParticles = (uint)active.size();
	if(numParticles == 0)
		return;
	
	if(!vbo)
		CreateBuffers();
	
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(GL_ARRAY_BUFFER, sizeof(ParticleVertex)*numParticles*4, NULL, GL_STREAM_DRAW);
	
	ParticleVertex *vert = (ParticleVertex*)glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
	assert(vert);
	
	for(uint i = 0; i < numParticles; i++)
	{
		const Particle &prt = particleStore[ active[i] ];
		const Color4f &col = prt.color;
		const Point2f &p = prt.pos;
		
		float w = prt.width / 2.0f;
		float h = prt.height / 2.0f;
		
		float s = std::sin(prt.angle);
		float c = std::cos(prt.angle);
		
		float wc = w*c; float ws = w*s;
		float hs = h*s; float hc = h*c;
		
		vert[0].x = p.x - wc + hs; vert[0].y = p.y - ws - hc; vert[0].z = 0.0f; // Bottom Left
		vert[1].x = p.x + wc + hs; vert[1].y = p.y + ws - hc; vert[1].z = 0.0f; // Bottom Right
		vert[2].x = p.x + wc - hs; vert[2].y = p.y + ws + hc; vert[2].z = 0.0f; // Top Right
		vert[3].x = p.x - wc - hs; vert[3].y = p.y - ws + hc; vert[3].z = 0.0f; // Top Left
		
		vert[0].r = col.r; vert[0].g = col.g; vert[0].b = col.b; vert[0].a = col.a;
		vert[1].r = col.r; vert[1].g = col.g; vert[1].b = col.b; vert[1].a = col.a;
		vert[2].r = col.r; vert[2].g = col.g; vert[2].b = col.b; vert[2].a = col.a;
		vert[3].r = col.r; vert[3].g = col.g; vert[3].b = col.b; vert[3].a = col.a;
		
		vert[0].u = 0.0f; vert[0].v = 0.0f;
		vert[1].u = 1.0f; vert[1].v = 0.0f;
		vert[2].u = 1.0f; vert[2].v = 1.0f;
		vert[3].u = 0.0f; vert[3].v = 1.0f;
		
		vert = vert + 4;
	}
	
	glUnmapBuffer(GL_ARRAY_BUFFER);

	// Draw particles
	if(image.IsValid())
		glBindTexture(GL_TEXTURE_2D, image->GetOpenGLID());
	
	GetScreen()->BeginBlendMode(blendMode);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	{
		
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_COLOR_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		
		glBindBuffer(GL_ARRAY_BUFFER, vbo);
		glVertexPointer(3, GL_FLOAT, sizeof(ParticleVertex), NULL);
		glColorPointer(4, GL_FLOAT, sizeof(ParticleVertex), (const void*)PARTICLEVERTEX_COLOR_OFFSET);
		glTexCoordPointer(2, GL_FLOAT, sizeof(ParticleVertex), (const void*)PARTICLEVERTEX_TEX_OFFSET);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
		
		//glDrawElements(GL_TRIANGLES, numParticles*2, GL_UNSIGNED_INT, 0);
		
		glDrawRangeElements(GL_TRIANGLES, 0, numParticles*6, numParticles*6, GL_UNSIGNED_INT, 0);
		
		//glDrawArrays(GL_QUADS, 0, numParticles);
		
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glDisableClientState(GL_COLOR_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);
		
	}
	glPopMatrix();
	glBindTexture(GL_TEXTURE_2D, 0);
	
	GetScreen()->EndBlendMode();
	
/*
	// Immediate mode method
	
	if(image.IsValid())
		glBindTexture(GL_TEXTURE_2D, image->GetOpenGLID());
	
	GetScreen()->BeginBlendMode(blendMode);
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glBegin(GL_QUADS);
	
	uint len = (uint)active.size();
	for(uint i = 0; i < len; i++)
	{
		const Particle &prt = particleStore[ active[i] ];
		const Point2f &p = prt.pos;
		
		float w = prt.width / 2.0f;
		float h = prt.height / 2.0f;

		float s = std::sin(prt.angle);
		float c = std::cos(prt.angle);

		float wc = w*c; float ws = w*s;
		float hs = h*s; float hc = h*c;
		
		glColor4f(prt.color.r, prt.color.g, prt.color.b, prt.color.a);
		
		glTexCoord2f(0.0f, 0.0f);
		glVertex2f(p.x - wc + hs, p.y - ws - hc);
		
		glTexCoord2f(1.0f, 0.0f);
		glVertex2f(p.x + wc + hs, p.y + ws - hc);
		
		glTexCoord2f(1.0f, 1.0f);
		glVertex2f(p.x + wc - hs, p.y + ws + hc);
		
		glTexCoord2f(0.0f, 1.0f);
		glVertex2f(p.x - wc - hs, p.y - ws + hc);
	}
	
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	
	glEnd();

	glPopMatrix();
	glBindTexture(GL_TEXTURE_2D, 0);
	
	GetScreen()->EndBlendMode();
*/
}
