/*!
 * \file ParticleSystem.h
 * \brief Declares the policy based ParticleSystem and is associated classes.
 * \author Joe Forte
 * \date 7/17/2008
 */

#ifndef PARTICLESYSTEM_H
#define PARTICLESYSTEM_H

#include <assert.h>
#include <vector>
#include "Graphics/Screen.h"
#include "Graphics/Image.h"
#include "Graphics/Color.h"
#include "Math/MathCommon.h"

/*!
 * \struct Particle
 * \headerfile ParticleSystem.h "Graphics/ParticleSystem.h"
 * \brief A basic struct for Particles, should be sufficient for most particle systems.
 */
struct Particle
{
	Point2f pos;      /*<! the center position of the particle. */
	Vector2f vel;     /*<! the linear velocity of the particle. */
	
	float angle;      /*<! the current angle of the particle. */
	float angularVel; /*<! the angular velocity of the particle. */
	
	float width;      /*<! the width of the particle. */
	float height;     /*<! the height of the particle. */
	
	Color4f color;    /*<! the color and transparency of the particle. */
	
	float age;        /*<! the current age of the particle in seconds. */
	float life;       /*<! the amount of time in seconds this particle should remain active. */
};

/*!
 * \class NullInitPolicy
 * \headerfile ParticleSystem.h "Graphics/ParticleSystem.h"
 * \brief A policy that does nothing and can be used as any policy template paramater in the \e InitPolicy class.  
 *
 * \tparam ParticleType the type of particle to use in the particle system.  The ParticleType is expected to inheret from
 * the Particle class (or at least have duplicate members).
 */
template<class ParticleType = Particle>
class NullInitPolicy
{
public:
	inline void PreEmit(const Point2f &pos, uint amount) {}
	inline void InitParticle(ParticleType &p) {}
};

/*!
 * \class NullUpdatePolicy
 * \headerfile ParticleSystem.h "Graphics/ParticleSystem.h"
 * \brief A policy that does nothing and can be used as any policy template paramater in the \e UpdatePolicy class.  
 *
 * \tparam ParticleType the type of particle to use in the particle system.  The ParticleType is expected to inheret from
 * the Particle class (or at least have duplicate members).
 */
template<class ParticleType = Particle>
class NullUpdatePolicy
{
public:
	inline void PreUpdate(float dt) {}
	inline bool ApplyUpdate(ParticleType &p, float dt) { return true; }
};

/*!
 * \class CompositeInitPolicy
 * \headerfile ParticleSystem.h "Graphics/ParticleSystem.h"
 * \brief A policy that combines any two distict policies with the same methods as found in the \e InitPolicy class.
 * NOTE: Policy1 and Policy2 should not have the same names for any of its members.
 *
 * \tparam Policy1 the first policy.
 * \tparam Policy2 the second policy.
 * \tparam ParticleType the type of particle to use in the particle system.  The ParticleType is expected to inheret from
 * the Particle class (or at least have duplicate members).
 */
template<class Policy1, class Policy2, class ParticleType = Particle>
class CompositeInitPolicy : public Policy1, Policy2
{
public:
	/*!
	 * \brief Called once before new particles are emitted.
	 *
	 * \param[in] pos the requested position to emit the particles.
	 * \param[in] amount the number of particles that will be emitted.
	 */
	inline void PreEmit(const Point2f &pos, uint amount)
	{
		Policy1::PreEmit(pos, amount);
		Policy2::PreEmit(pos, amount);
	}
	
	/*!
	 * \brief Called once on each new particle that is emitted.
	 *
	 * \param[in] p the particle to initialize.
	 */
	inline void InitParticle(ParticleType &p)
	{
		Policy1::InitParticle(p);
		Policy2::InitParticle(p);
	}
};

/*!
 * \class CompositePolicy
 * \headerfile ParticleSystem.h "Graphics/ParticleSystem.h"
 * \brief A policy that combines any two distict policies with the same methods as found in the \e UpdatePolicy class.
 * NOTE: Policy1 and Policy2 should not have the same names for any of its members.
 *
 * \tparam Policy1 the first policy.
 * \tparam Policy2 the second policy.
 * \tparam ParticleType the type of particle to use in the particle system.  The ParticleType is expected to inheret from
 * the Particle class (or at least have duplicate members).
 */
template<class Policy1, class Policy2, class ParticleType = Particle>
class CompositeUpdatePolicy : public Policy1, Policy2
{
public:
	/*!
	 * \brief Called once before each update.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 */
	inline void PreUpdate(float dt) 
	{
		Policy1::PreUpdate(dt);
		Policy2::PreUpdate(dt);
	}
	
	/*!
	 * \brief Called once before each update.
	 *
	 * \param[in,out] p the partical to apply the Policy to.
	 * \param[in] dt the time in seconds since the last frame.
	 *
	 * \return False if you want the particle to be destroyed, return true otherwise.
	 */
	inline bool ApplyUpdate(ParticleType &p, float dt)
	{
		if( Policy1::ApplyUpdate(p, dt) &&
		    Policy2::ApplyUpdate(p, dt) )
		{
			return false;
		}

		return true;
	}
};

/*!
 * \class InitPolicy
 * \headerfile ParticleSystem.h "Graphics/ParticleSystem.h"
 * \brief The complete Policy template for initializing particles, used by the ParticleSystem class.
 *
 * \tparam PositionPolicy the policy for updating particle position, expected to have the same methods as the Policy template.
 * \tparam VelocityPolicy the policy for updating particle velocity, expected to have the same methods as the Policy template.
 * \tparam AnglePolicy the policy for updating particle angle and angular velocity, expected to have the same methods as the Policy template.
 * \tparam SizePolicy the policy for updating particle size, expected to have the same methods as the Policy template.
 * \tparam ColorPolicy the policy for updating particle color, expected to have the same methods as the Policy template.
 * \tparam AgePolicy the policy for updating particle age, expected to have the same methods as the Policy template.
 * \tparam UserPolict an extra policy that can do whatever the user wants.  Defaults to the NullPolicy.
 * \tparam ParticleType the type of particle to use in the particle system.  The ParticleType is expected to inheret from
 * the Particle class (or at least have duplicate members).
 */
template<class PositionPolicy, class VelocityPolicy, class AnglePolicy, 
         class SizePolicy, class ColorPolicy, class AgePolicy,
         class UserPolicy = NullInitPolicy<Particle>, class ParticleType = Particle>
class InitPolicy
{
public:
	PositionPolicy positionPolicy; /*!< The position policy */
	VelocityPolicy velocityPolicy; /*!< The velocity policy */
	AnglePolicy anglePolicy;       /*!< The angle and agular velocity policy */
	SizePolicy sizePolicy;         /*!< The size policy */
	ColorPolicy colorPolicy;       /*!< The color policy */
	AgePolicy agePolicy;           /*!< The age policy */
	UserPolicy userPolicy;         /*!< The user policy */
		
	/*!
	 * \brief Called once before new particles are emitted.
	 *
	 * \param[in] pos the requested position to emit the particles.
	 * \param[in] amount the number of particles that will be emitted.
	 */
	inline void PreEmit(const Point2f &pos, uint amount)
	{
		positionPolicy.PreEmit(pos, amount);
		velocityPolicy.PreEmit(pos, amount);
		anglePolicy.PreEmit(pos, amount);
		sizePolicy.PreEmit(pos, amount);
		colorPolicy.PreEmit(pos, amount);
		agePolicy.PreEmit(pos, amount);
		userPolicy.PreEmit(pos, amount);
	}
		
	/*!
	 * \brief Called once on each new particle that is emitted.
	 *
	 * \param[in] p the particle to initialize.
	 */
	inline void InitParticle(ParticleType &p)
	{
		positionPolicy.InitParticle(p);
		velocityPolicy.InitParticle(p);
		anglePolicy.InitParticle(p);
		sizePolicy.InitParticle(p);
		colorPolicy.InitParticle(p);
		agePolicy.InitParticle(p);
		userPolicy.InitParticle(p);
	}
};

/*!
 * \class UpdatePolicy
 * \headerfile ParticleSystem.h "Graphics/ParticleSystem.h"
 * \brief The complete Policy template for updating particles, used by the ParticleSystem class.
 *
 * \tparam MotionPolicy the policy that moves particles, expected to have the same methods as the Policy template.
 * \tparam PositionPolicy the policy for updating particle position, expected to have the same methods as the Policy template.
 * \tparam VelocityPolicy the policy for updating particle velocity, expected to have the same methods as the Policy template.
 * \tparam AnglePolicy the policy for updating particle angle and angular velocity, expected to have the same methods as the Policy template.
 * \tparam SizePolicy the policy for updating particle size, expected to have the same methods as the Policy template.
 * \tparam ColorPolicy the policy for updating particle color, expected to have the same methods as the Policy template.
 * \tparam AgePolicy the policy for updating particle age, expected to have the same methods as the Policy template.
 * \tparam UserPolict an extra policy that can do whatever the user wants.  Defaults to the NullPolicy.
 * \tparam ParticleType the type of particle to use in the particle system.  The ParticleType is expected to inheret from
 * the Particle class (or at least have duplicate members).
 */
template<class MotionPolicy, class PositionPolicy, class VelocityPolicy,
         class AnglePolicy, class SizePolicy, class ColorPolicy, class AgePolicy,
         class UserPolicy = NullUpdatePolicy<Particle>, class ParticleType = Particle>
class UpdatePolicy
{
public:
	MotionPolicy motionPolicy;     /*!< The motion policy */
	PositionPolicy positionPolicy; /*!< The position policy */
	VelocityPolicy velocityPolicy; /*!< The velocity policy */
	AnglePolicy anglePolicy;       /*!< The angle and agular velocity policy */
	SizePolicy sizePolicy;         /*!< The size policy */
	ColorPolicy colorPolicy;       /*!< The color policy */
	AgePolicy agePolicy;           /*!< The age policy */
	UserPolicy userPolicy;         /*!< The user policy */
	
	/*!
	 * \brief Called once before each update.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 */
	inline void PreUpdate(float dt)
	{
		motionPolicy.PreUpdate(dt);
		positionPolicy.PreUpdate(dt);
		velocityPolicy.PreUpdate(dt);
		anglePolicy.PreUpdate(dt);
		sizePolicy.PreUpdate(dt);
		colorPolicy.PreUpdate(dt);
		agePolicy.PreUpdate(dt);
		userPolicy.PreUpdate(dt);
	}
	
	/*!
	 * \brief Called once before each update.
	 *
	 * \param[in,out] p the particle to apply the Policy to.
	 * \param[in] dt the time in seconds since the last frame.
	 *
	 * \return False if you want the particle to be destroyed, return true otherwise.
	 */
	inline bool ApplyUpdate(ParticleType &p, float dt)
	{
		if( motionPolicy.ApplyUpdate(p, dt) &&
		    positionPolicy.ApplyUpdate(p, dt) &&
		    velocityPolicy.ApplyUpdate(p, dt) &&
		    anglePolicy.ApplyUpdate(p, dt) &&
		    sizePolicy.ApplyUpdate(p, dt) &&
		    colorPolicy.ApplyUpdate(p, dt) &&
 		    agePolicy.ApplyUpdate(p, dt) &&
		    userPolicy.ApplyUpdate(p, dt) )
		{
			return true;
		}
		
		return false;
	}
};

/*!
 * \class DefaultEmitPolicy
 * \headerfile ParticleSystem.h "Graphics/ParticleSystem.h"
 * \brief Defines the default emit policy that never emits particles.
 *
 * \tparam ParticleType the type of particle to use in the particle system.  The ParticleType is expected to inheret from
 * the Particle class (or at least have duplicate members).
 */
template<class ParticleType = Particle>
class DefaultEmitPolicy
{
public:	
	/*!
	 * \brief Determines the number of particles to emit on each update.  NOTE: All EmitPolicies must
     * implement this function.
	 *
	 * \param[in] dt the frame time.
	 *
	 * \return The number of particles to emit.
	 */
	inline uint AmountToEmit(float dt)
	{
		return 0;
	}
	
	/*!
	 * \brief Gets the positon to emit particles from.  NOTE: All EmitPolicies must
     * implement this function.  Called once before particles are emitted. 
	 *
	 * \return The position to emit particles.
	 */
	inline Point2f GetEmitPosition()
	{
		return Point2f(0.0f, 0.0f);
	}
};

/*!
 * \class DefaultParticleDraw
 * \headerfile ParticleSystem.h "Graphics/ParticleSystem.h"
 * \brief The default policy for drawing particles of type \e Particle.
 */
class DefaultParticleDrawPolicy
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	DefaultParticleDrawPolicy() : blendMode(BLENDMODE_ALPHA), vbo(0), ibo(0) {}
	
	/*!
	 * \brief Deconstructor.
	 */
	~DefaultParticleDrawPolicy();
	
	/*!
	 * \brief Called one at initialization.
	 *
	 * \param[in] max the maximum number of particles that will ever be drawn at a time.
	 */
	void Init(uint max) { maxParticles = max; }
	
	/*!
	 * \brief Draws the active set of particles.
	 *
	 * \param[in] active a vector of indices into the \a particleStore of the particles to draw.
	 * \param[in] particleStore a vector holding all of the particles (active and dead).
	 */
	void Draw(const std::vector<size_t> &active, const std::vector<Particle> &particleStore);
	
	ImageRef image;
	BlendMode blendMode; /*<! The BlendMode to use when rendering the particles.  Defaults to BLENDMODE_ALPHA. */
	
private:
	
	void CreateBuffers();
	
	uint maxParticles;
	
	uint vbo; // vertex buffer handle	
	uint ibo; // index buffer handle
};

/*!
 * \class ParticleSystem
 * \headerfile ParticleSystem.h "Graphics/ParticleSystem.h"
 * \brief The base class for spawning and managing particles.
 *
 * \tparam MaxParticles the maximum number of particles that can be active in the particle system.
 * \tparam InitPolicy the \e Policy used to initialize new particles.
 * \tparam UpdatePolicy the \e Policy used on each active particle during an update.
 * \tparam EmitPolicy the policy used for emitting particles on each update, expected to have the same methods as the \a DefaultEmitPolicy class.
 * \tparam DrawPolicy the policy used for drawing the particles, expected to have the same methods as the \a DefaultParticleDrawPolicy class.
 * \tparam ParticleType the type of particle to use in the particle system.  The ParticleType is expected to inheret from
 * the Particle class (or at least have duplicate members).
 */
template<uint MaxParticles, 
         class InitPolicy, 
         class UpdatePolicy, 
         class EmitPolicy = DefaultEmitPolicy<>,
         class DrawPolicy = DefaultParticleDrawPolicy, 
         class ParticleType = Particle>
class ParticleSystem
{
public:
	InitPolicy initPolicy;     /*!< The initialization policy */
	UpdatePolicy updatePolicy; /*!< The update policy */
	EmitPolicy emitPolicy;     /*!< The emit policy */
	DrawPolicy drawPolicy;     /*!< The draw policy */
	
	/*!
	 * \brief Default constructor.
	 */
	ParticleSystem() : particleStore(MaxParticles)
	{
		for(size_t i = 0; i < MaxParticles; i++)
		{
			dead.push_back(i);
		}
		
		drawPolicy.Init(MaxParticles);
	}
	
	/*!
	 * \brief Determines the total number of active particles in the system.
	 *
	 * \return The number of active particles.
	 */
	uint GetParticleCount() const
	{
		return (uint)active.size();
	}
	
	/*!
	 * \return The maximum number of active particles the system can handle.
	 */
	uint GetMaxNumberParticles() const
	{
		return MaxParticles;
	}
	
	/*!
	 * \brief Emits the requested number of particles at the given position.
	 *
	 * \param[in] pos the position to emit the particles.
	 * \param[in] amt the number of particles to emit.  NOTE: The particle system will only emit particles if the number of active particles
	 * is less then MaxParticles.
	 */
	void Emit(const Point2f &pos, uint amt)
	{
		// calcuates the actual number of particles to emit.
		amt = std::min(amt, MaxParticles - GetParticleCount());
		
		assert(amt <= (uint)dead.size());
		
		initPolicy.PreEmit(pos, amt);
		
		for(uint i = 0; i < amt; i++)
		{
			size_t index = dead.back();
			
			dead.pop_back();
			active.push_back(index);
			
			initPolicy.InitParticle(particleStore[index]);
		}
	}
	
	/*!
	 * \brief Updates the particle system.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 */
	void Update(float dt)
	{
		uint toEmit = emitPolicy.AmountToEmit(dt);
		if(toEmit > 0)
			Emit(emitPolicy.GetEmitPosition(), toEmit);
		
		uint count = GetParticleCount();
		if(count == 0)
			return;
		
		updatePolicy.PreUpdate(dt);
		
		for(int i = count - 1; i >= 0; i--)
		{
			ParticleType &p = particleStore[ active[i] ];
			
			if( !updatePolicy.ApplyUpdate(p, dt) )
			{
				size_t index = active[i];
				active.erase(active.begin() + i);
				dead.push_back(index);
			}
		}
	}
	
	/*!
	 * \brief Draws all of the currently active particles.
	 */
	void Draw()
	{
		drawPolicy.Draw(active, particleStore);
	}

	/*!
	 * \brief Removes all particles from the system.
	 */
	void Clear()
	{
		for(std::vector<size_t>::iterator i = active.begin(); i != active.end(); i++)
		{
			dead.push_back(*i);
		}
		
		active.clear();
	}
	
private:
	
	// active list and dead List both hold indices into the particleStore
	std::vector<size_t> active;
	std::vector<size_t> dead;
	std::vector<ParticleType> particleStore;
};

/*!
 * \class ContinuousEmitPolicy
 * \headerfile ParticleSystem.h "Graphics/ParticleSystem.h"
 * \brief Defines the an emit policy that continuously emits particles with some flowrate.
 *
 * \tparam ParticleType the type of particle to use in the particle system.  The ParticleType is expected to inheret from
 * the Particle class (or at least have duplicate members).
 */
template<class ParticleType = Particle>
class ContinuousEmitPolicy
{
public:
	ContinuousEmitPolicy() : emit(false), flowRate(0.0f), emitted(0), timer(0.0f) {}
		
	/*!
	 * \brief Determines the number of particles to emit on each update.  NOTE: All EmitPolicies must
	 * implement this function.
	 *
	 * \param[in] dt the frame time.
	 *
	 * \return The number of particles to emit.
	 */
	inline uint AmountToEmit(float dt)
	{
		if(emit)
		{
			timer += dt;
			uint spawn = (uint)std::max(0.0f, std::min(std::floor(timer * flowRate) - emitted, flowRate));
			emitted += spawn;
			
			if(timer >= 1.0f)
			{
				timer = 0.0f;
				emitted = 0;
			}
			
			return spawn;
		}
		
		return 0;
	}
		
	/*!
	 * \brief Gets the positon to emit particles from.  NOTE: All EmitPolicies must
	 * implement this function.  Called once before particles are emitted. 
	 *
	 * \return The position to emit particles.
	 */
	inline Point2f GetEmitPosition()
	{
		return position;
	}
	
	/*!
	 * \brief Starts emitting particles.
	 */
	inline void Start() 
	{
		emit = true;
	}
	
	/*!
	 * \brief Stops the policy from emitting particles.
	 */
	inline void Stop() 
	{
		emit = false;
	}
	
	/*!
	 * \brief Sets the flow rate of particles.
	 *
     * \param[in] rate the flow rate in particles per second.
	 */
	inline void SetFlowRate(float rate)
	{
		flowRate = rate;
	}
	
	Point2f position;
	
private:
	bool emit;
	
	float flowRate;
	float timer;
	
	uint emitted;
};

#endif
