/*!
 * \file SDLScreen.cpp
 * \brief Implements the SDLScreen class.
 * \author Joe Forte
 * \date 3/8/2008
 */

#include <assert.h>
#include <sstream>
#include <iostream>
#include <SDL.h>
#include <SDL_opengl.h>
#include <SOIL.h>
#include "Common/Error.h"
#include "Graphics/Font.h"
#include "Graphics/SDLScreen.h"

// Platform spacific includes used by the CenterScreen() method.
#ifdef _WIN32
	#include <Windows.h>
	#include <SDL_syswm.h>
#elif defined __APPLE__
	#include <Carbon/Carbon.h>
#include <SDL_syswm.h>
#endif

//--------------------------------------------------------------------
SDLScreen::SDLScreen() : screen(NULL), blendModeSet(false), blendMode(BLENDMODE_ALPHA), textQSize(0) {}

//--------------------------------------------------------------------
SDLScreen::SDLScreen(int w, int h, const Color4f &backGroundColor, const std::string &title, bool center)
{
	if( !SDL_WasInit( SDL_INIT_VIDEO ) )
	{
		THROW_ERROR( "SDL could not be initialized." ); 
	}

	screen = NULL;
	textQSize = 0;

	Set(w, h, backGroundColor, title);

	if(center)
		CenterScreen();
}

//--------------------------------------------------------------------
SDLScreen::~SDLScreen() {}

//--------------------------------------------------------------------
bool SDLScreen::Set(int w, int h)
{
	if( !SDL_WasInit( SDL_INIT_VIDEO ) )
	{
		std::cerr << "ERROR: SDL has not been initialized." << std::endl;
		return false;
	}
	
	width = w;
	height = h;

	// Sets up OpenGL using SDL
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);

	screen = SDL_SetVideoMode(width, height, 32, SDL_OPENGL);

	if(!screen)
		return false;

	SDL_WM_SetCaption(windowTitle.c_str(), windowTitle.c_str());

	return InitOpenGL();
}

//--------------------------------------------------------------------
bool SDLScreen::InitOpenGL()
{
	// Sets up openGL
	glViewport(0, 0, width, height);
	
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	
	const float w_2 = width / 2.0f;
	const float h_2 = height / 2.0f;	
	glOrtho(-w_2, w_2, -h_2, h_2, -1.0, 1.0);
	
	glMatrixMode(GL_MODELVIEW);
	
	return true;
}

//--------------------------------------------------------------------
bool SDLScreen::Set(int w, int h, const Color4f &backGroundColor, const std::string &title)
{
	windowTitle = title;
	bgColor = backGroundColor;

	return Set(w, h);
}

//--------------------------------------------------------------------
bool SDLScreen::CenterScreen()
{
	if(!screen)
		return false;

#ifdef _WIN32

	// Gets the current window handle.
	HWND window = NULL;

	SDL_SysWMinfo info;
	SDL_VERSION(&info.version);
	if(SDL_GetWMInfo(&info))
		window = info.window;
	else
		return false;

	// Computes the position of the window with respect to the size of the desktop
	HWND desktop = GetDesktopWindow();
	if(!desktop)
		return false;

	RECT desktopRect, windowRect;
	if(GetWindowRect(desktop, &desktopRect) == FALSE || 
	   GetWindowRect(window, &windowRect) == FALSE)
	{
		return false;
	}

	int dw = std::abs(desktopRect.right - desktopRect.left);
	int dh = std::abs(desktopRect.bottom - desktopRect.top);

	int ww = std::abs(windowRect.right - windowRect.left);
	int wh = std::abs(windowRect.bottom - windowRect.top);

	int x = (dw - ww) / 2;
	int y = (dh - wh) / 2;

	return SetWindowPos(window, HWND_TOP, x, y, 0, 0, SWP_NOSIZE) == TRUE;

#elif defined __APPLE__
	
	return false;
	
#else
	
	return false;

#endif
}



//--------------------------------------------------------------------
Rectf SDLScreen::GetRect() const
{
	Rectf r;
	r.x = -width / 2.0f;
	r.y = height / 2.0f;
	r.w = (float)width;
	r.h = (float)height;
	return r;
}

//--------------------------------------------------------------------
bool SDLScreen::SaveScreenshot(const std::string &fileName) const
{
	return SOIL_save_screenshot((fileName + ".tga").c_str(), SOIL_SAVE_TYPE_TGA, 0, 0, width, height) == 1;
}

//--------------------------------------------------------------------
bool SDLScreen::BeginBlendMode(BlendMode mode)
{
	if(blendModeSet)
		return false;
	
	blendModeSet = true;
	blendMode = mode;
	
	switch(mode) 
	{
		// NOTE: Alpha is the default blend mode and is already set.
		case BLENDMODE_ADDITIVE:
			glBlendFunc(GL_ONE, GL_ONE);
			break;
		case BLENDMODE_ADDITIVE_WITH_ALPHA:
			glBlendFunc(GL_SRC_ALPHA, GL_ONE);
			break;
		case BLENDMODE_MASK:
			glBlendFunc(GL_DST_COLOR, GL_ZERO);
			break;
		case BLENDMODE_NONE:
			glBlendFunc(GL_ONE, GL_ZERO);
			break;
	};
	
	return true;
}

//--------------------------------------------------------------------
bool SDLScreen::EndBlendMode()
{
	if(!blendModeSet)
		return false;
	
	blendModeSet = false;
	blendMode = BLENDMODE_ALPHA;
	
	// We always revert to default alpha blending
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	return true;
}

//--------------------------------------------------------------------
BlendMode SDLScreen::GetBlendMode() const
{
	return blendMode;
}

//--------------------------------------------------------------------
bool SDLScreen::EnableDrawRegion(Recti area)
{
	drawRegion = area;
	
	Rectf r = GetRect();
	
	/*
	glScissor((GLint)area.x, 
			  (GLint)(area.y - area.h), 
			  (GLsizei)area.w,
			  (GLsizei)area.h);
	*/
	glScissor((GLint)(area.x - r.x), 
			  (GLint)(area.y - area.h + r.y), 
			  (GLsizei)area.w,
			  (GLsizei)area.h);
	
	glEnable(GL_SCISSOR_TEST);
	
	return glGetError() == GL_NO_ERROR;
}

//--------------------------------------------------------------------
bool SDLScreen::DisableDrawRegion()
{
	glDisable(GL_SCISSOR_TEST);
	
	drawRegion.x = 0;
	drawRegion.y = 0;
	drawRegion.w = (int)width;
	drawRegion.h = (int)height;
	
	return glGetError() == GL_NO_ERROR;
}

//--------------------------------------------------------------------
bool SDLScreen::BeginFrame()
{
	assert(screen);
	return Clear(bgColor);
}

//--------------------------------------------------------------------
bool SDLScreen::EndFrame()
{
	assert(screen);
	
	DrawQueuedText();

	SDL_GL_SwapBuffers();

	return true;
}

//--------------------------------------------------------------------
void SDLScreen::NormToScreenCoords(float xin, float yin, float &x, float &y) const
{
	x = xin * (width / 2.0f);
	y = yin * (height / 2.0f);
}

//--------------------------------------------------------------------
void SDLScreen::ScreenToNormCoords(float xin, float yin, float &x, float &y) const
{
	x = (2.0f * xin) / width;
	y = (2.0f * yin) / height;
}

//--------------------------------------------------------------------
bool SDLScreen::Clear(const Color4f &color)
{
	assert(screen);
	
	glClearColor(color.r, color.g, color.b, color.a);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	
	return true;
}

//--------------------------------------------------------------------
bool SDLScreen::QueueText(FontRef f, const char *str, const Color4f &color, const Point2f &p)
{
	TextQInfo info;
	info.f = f;
	info.p = p;
	info.color = color;
	info.txt = str;
	info.align = Font::ALIGN_NONE;
	
	textQ.push_back(info);
	textQSize++;
	
	return true;
}

//--------------------------------------------------------------------
bool SDLScreen::QueueText(FontRef f, const char *str, const Color4f &color, Font::TextAlignment align, const Point2f &offset)
{
	TextQInfo info;
	info.f = f;
	info.p = offset;
	info.color = color;
	info.txt = str;
	info.align = align;
	
	textQ.push_back(info);
	textQSize++;
	
	return true;
}

//--------------------------------------------------------------------
bool SDLScreen::DrawQueuedText()
{
	Rectf r = GetRect();
	
	for(uint i = 0; i < textQSize; i++)
	{
		TextQInfo &info = textQ[i];
		if( info.f.IsValid() )
		{
			if(info.align == Font::ALIGN_NONE)
				info.f->DrawTextStr(info.txt.c_str(), info.color, ::NormToScreenCoords(info.p));
			else
				info.f->DrawTextStr(info.txt.c_str(), info.color, r, info.align, ::NormToScreenCoords(info.p));
		}
	}
	
	textQ.clear();
	textQSize = 0;

	return true;
}
