/*!
 * \file SDLScreen.h
 * \brief Declares the Screen interface that represents the main drawable area of the game.
 * \author Joe Forte
 * \date 3/8/2008
 */

#ifndef SDLSCREEN_H
#define SDLSCREEN_H

#include "Graphics/Screen.h"

struct SDL_Surface;

/*!
 * \class SDLScreen
 * \headerfile SDLScreen.h "Graphics/SDLScreen.h"
 * \brief Implementation of the IScreen interface using SDL.
 */
class SDLScreen : public IScreen
{
public:
	/*!
	 * \brief Constructor.
	 */
	SDLScreen();
	
	/*!
	 * \brief Constructs an SDLScreen.
	 *
	 * \param[in] w the width of the drawable area of the screen.
	 * \param[in] h the height of the drawable area of the screen.
	 * \param[in] backGroundColor the color of the drawable area of the screen.
	 * \param[in] title the title of the window that houses the screen.
	 * \param[in] center set to true to have the screen centered on the desktop.  Defaults to false.
	 */
	SDLScreen(int w, int h, const Color4f &backGroundColor, const std::string &title, bool center = false);
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~SDLScreen();

	/*!
	 * \brief Sets the size of the drawable area of the screen.
	 *
	 * \param[in] w the width of the drawable area of the screen.
	 * \param[in] h the height of the drawable area of the screen.
	 */
	bool Set(int w, int h);
	
	/*!
	 * \brief Sets the parameters of the screen.
	 *
	 * \param[in] w the width of the drawable area of the screen.
	 * \param[in] h the height of the drawable area of the screen.
	 * \param[in] backGroundColor the color of the drawable area of the screen.
	 * \param[in] title the title of the window that houses the screen.
	 */
	bool Set(int w, int h, const Color4f &backGroundColor, const std::string &title);

	/*!
	 * \brief Centers the window on the desktop. NOTE: Only usable in windowed mode.
	 *
	 * \return true if the screen was centered, false otherwise.
	 *
	 * NOTE: Currently only supported for WIN32.
	 */
	bool CenterScreen();
	
	/*!
	 * \brief Sets the screen to a state where it can be drawn too.
	 */
	bool BeginFrame();
	
	/*!
	 * \brief Stops the ability to draw to the screen and displays the current contents of the screen.
	 */
	bool EndFrame();

	/*!
	 * \brief Gets the width of the drawable screen area.
	 *
	 * \return The width of the drawable screen area.
	 */
	virtual int Width() const { return width; }
	
	/*!
	 * \brief Gets the height of the drawable screen area.
	 *
	 * \return The height of the drawable screen area.
	 */
	virtual int Height() const { return height; }
	
	/*!
	 * \brief Gets the Rect of the drawable screen area.
	 *
	 * \return The Rect of the drawable screen area.
	 */
	virtual Rectf GetRect() const;
	
	/*!
	 * \brief Saves an image of the current screens buffer.
	 *
	 * \param[in] fileName the name of the image file to save.
	 *
	 * \return True if the image was saved, false otherwise.
	 */
	virtual bool SaveScreenshot(const std::string &fileName) const;

	/*!
	 * \brief Gets the title of the window that houses the screen.
	 *
	 * \return The title of the window that houses the screen.
	 */
	virtual std::string GetWindowTitle() const { return windowTitle; } 
	
	/*!
	 * \brief Converts normalized device coordinates to absolute screen coordinates.
	 *
	 * \param[in] xin the normalized device coordinate in x to be converted.
	 * \param[in] yin the normalized device coordinate in y to be converted.
	 * \param[out] x the converted value of \a xin.
	 * \param[out] y the converted value of \a yin.
	 * \see ScreenToNormCoords
	 *
	 * In normalized device coordinates the top left corner of the screen is at position (-1,-1), the bottom right 
	 * corner of the screen is at position (1,1), and the center of the screen is at position (0,0).  In absolute 
	 * screen coordinates the top left corner of the screen is at position (-width/2,-height/2), the bottom right
	 * corner of the screen is at position (width/2,height/2), and the center of the screen is at position (0,0).
	 */
	virtual void NormToScreenCoords(float xin, float yin, float &x, float &y) const;
	
	/*!
	 * \brief Converts absolute screen coordinates to normalized device coordinates to
	 *
	 * \param[in] xin the absolute screen coordinate in x to be converted.
	 * \param[in] yin the absolute screen coordinate in y to be converted.
	 * \param[out] x the converted value of \a xin.
	 * \param[out] y the converted value of \a yin.
	 * \see NormToScreenCoords
	 *
	 * In normalized device coordinates the top left corner of the screen is at position (-1,-1), the bottom right 
	 * corner of the screen is at position (1,1), and the center of the screen is at position (0,0).  In absolute 
	 * screen coordinates the top left corner of the screen is at position (-width/2,-height/2), the bottom right
	 * corner of the screen is at position (width/2,height/2), and the center of the screen is at position (0,0).
	 */
	virtual void ScreenToNormCoords(float xin, float yin, float &x, float &y) const;

	/*!
	 * \brief Queues text to be drawn a the end of the frame.
	 *
	 * \param[in] f the font used to draw the text.
	 * \param[in] str the text to draw.
	 * \param[in] color the color to draw \a str.
	 * \param[in] p the position to draw \a str.  Assumed to be in normalized device coordinates.
	 */
	virtual bool QueueText(FontRef f, const char *str, const Color4f &color, const Point2f &p);
	
	/*!
	 * \brief Queues aligned text to be drawn a the end of the frame.
	 *
	 * \param[in] f the font used to draw the text.
	 * \param[in] str the text to draw.
	 * \param[in] color the color to draw \a str.
	 * \param[in] align the text alignment to use when drawing \a str.  The screen is used as the area to align to.
	 * \param[in] offset the amount to offset the text.  Assumed to be in normalized device coordinates.  Defaults to (0,0).
	 */
	virtual bool QueueText(FontRef f, const char *str, const Color4f &color, Font::TextAlignment align, const Point2f &offset = Point2f(0.0f, 0.0f));
	
	/*!
	 * \brief Sets the BlendMode all further draw calls  Default is BLENDMODE_ALPHA. NOTE: BeginBlendMode/EndBlendMode cannot be nested.
	 *
	 * \param[in] mode the BlendMode to use.
	 *
	 * \return True if the BlendMode was successfully set.  Returns false if the blend mode could not be set.  This can occure if
	 * BeginBlendMode has already been called but EndBlendMode has not.
	 */
	virtual bool BeginBlendMode(BlendMode mode);
	
	/*!
	 * \brief Ends the BlendMode mode being used and reverts back to BLENDMODE_ALPHA.
	 *
	 * \return True the BlendMode has been set back to BLENDMODE_ALPHA, returns false if EndBlendMode is called without first calling BeginBlendMode.
	 */
	virtual bool EndBlendMode();
	
	/*!
	 * \brief Gets the currently active blend mode
	 *
	 * \returns the active BlendMode.
	 */
	virtual BlendMode GetBlendMode() const;
	
	/*!
	 * \brief Sets the screen area where all subsequent draw calls will be limited to.
	 *
	 * \param[in] area the screen rectangle.
	 * 
	 * \return true if the draw region was created, false otherwise.
	 *
	 * NOTE: This can be called without a matching call to DisableDrawRegion().
	 */
	virtual bool EnableDrawRegion(Recti area);
	
	/*!
	 * \brief Disables using the draw engine.
	 *
	 * \return true if the draw region was disabled, false otherwise.
	 */
	virtual bool DisableDrawRegion();
	
	/*!
	 * \brief Gets the current draw region.
	 *
	 * \return the draw region.
	 */
	virtual Recti GetDrawRegion() const { return drawRegion; }
	
	/*!
	 * \brief Draws the currently queued text.  Called in EndFrame().
	 */
	bool DrawQueuedText();

	/*!
	 * \brief Clears the drawable area of the screen to the specified color.  Called in BeginFrame().
	 *
	 * \param[in] color the color to clear the screen to.
	 */
	bool Clear(const Color4f &color);
	
private:
	bool InitOpenGL();

	int width;
	int height;
	Color4f bgColor;
	std::string windowTitle;
	
	Recti drawRegion;
	
	BlendMode blendMode;
	bool blendModeSet; /*!< True between BeginBlendMode() and EndBlendMode(). */

	SDL_Surface *screen;
	
	struct TextQInfo
	{
		std::string txt;
		Point2f p;
		Color4f color;
		
		Font::TextAlignment align;
		
		FontRef f;
	};
	
	std::vector<TextQInfo> textQ;
	uint textQSize;
};

#endif
