/*!
 * \file Screen.h
 * \brief Declares the Screen interface that represents the main drawable area of the game.
 * \author Joe Forte
 * \date 3/8/2008
 */

#ifndef SCREEN_H
#define SCREEN_H

#include <string>
#include <vector>
#include "Graphics/Font.h"
#include "Graphics/Color.h"
#include "Math/MathCommon.h"
#include "Common/BaseTypes.h"

/*!
 * \enum BlendMode
 * \brief Defines the different methods for blending the image with the rest of the scene. 
 */
enum BlendMode
{
	BLENDMODE_ALPHA = 0, /*!< Blends using standard alpha blending. */
	BLENDMODE_ADDITIVE, /*!< Blends by adding the source (image) color withe the destination (framebuffer) color. */
	BLENDMODE_ADDITIVE_WITH_ALPHA, /*!< Blends by modulating the source color with its alpha value and adding it to the destination color. */ 
	BLENDMODE_MASK, /*!< Renders the image as a mask over the framebuffer. */  
	BLENDMODE_NONE /*!< No blending will be done. */
};

/*!
 * \class IScreen
 * \headerfile Screen.h "Graphics/Screen.h"
 * \brief Interface class for interacting and manipulating the screen.
 */
class IScreen
{
public:

	/*!
	 * \brief Centers the window on the desktop. NOTE: Only usable in windowed mode.
	 *
	 * \return true if the screen was centered, false otherwise.
	 *
	 * NOTE: Currently only supported for WIN32.
	 */
	virtual bool CenterScreen() = 0;

	/*!
	 * \brief Gets the width of the drawable screen area.
	 *
	 * \return The width of the drawable screen area.
	 */
	virtual int Width() const = 0;
	
	/*!
	 * \brief Gets the height of the drawable screen area.
	 *
	 * \return The height of the drawable screen area.
	 */
	virtual int Height() const = 0;
	
	/*!
	 * \brief Gets the Rect of the drawable screen area.
	 *
	 * \return The Rect of the drawable screen area.
	 */
	virtual Rectf GetRect() const = 0;

	/*!
	 * \brief Gets the title of the window that houses the screen.
	 *
	 * \return The title of the window that houses the screen.
	 */
	virtual std::string GetWindowTitle() const = 0;
	
	/*!
	 * \brief Saves an image of the current screens buffer.
	 *
	 * \param[in] fileName the name of the image file to save.
	 *
	 * \return True if the image was saved, false otherwise.
	 */
	virtual bool SaveScreenshot(const std::string &fileName) const = 0;
	
	/*!
	 * \brief Converts normalized device coordinates to absolute screen coordinates.
	 *
	 * \param[in] xin the normalized device coordinate in x to be converted.
	 * \param[in] yin the normalized device coordinate in y to be converted.
	 * \param[out] x the converted value of \a xin.
	 * \param[out] y the converted value of \a yin.
	 * \see ScreenToNormCoords
	 *
	 * In normalized device coordinates the top left corner of the screen is at position (-1,-1), the bottom right 
	 * corner of the screen is at position (1,1), and the center of the screen is at position (0,0).  In absolute 
	 * screen coordinates the top left corner of the screen is at position (-width/2,-height/2), the bottom right
	 * corner of the screen is at position (width/2,height/2), and the center of the screen is at position (0,0).
	 */
	virtual void NormToScreenCoords(float xin, float yin, float &x, float &y) const = 0;
	
	/*!
	 * \brief Converts absolute screen coordinates to normalized device coordinates to
	 *
	 * \param[in] xin the absolute screen coordinate in x to be converted.
	 * \param[in] yin the absolute screen coordinate in y to be converted.
	 * \param[out] x the converted value of \a xin.
	 * \param[out] y the converted value of \a yin.
	 * \see NormToScreenCoords
	 *
	 * In normalized device coordinates the top left corner of the screen is at position (-1,-1), the bottom right 
	 * corner of the screen is at position (1,1), and the center of the screen is at position (0,0).  In absolute 
	 * screen coordinates the top left corner of the screen is at position (-width/2,-height/2), the bottom right
	 * corner of the screen is at position (width/2,height/2), and the center of the screen is at position (0,0).
	 */
	virtual void ScreenToNormCoords(float xin, float yin, float &x, float &y) const = 0;

	/*!
	 * \brief Queues text to be drawn a the end of the frame.
	 *
	 * \param[in] f the font used to draw the text.
	 * \param[in] str the text to draw.
	 * \param[in] color the color to draw \a str.
	 * \param[in] p the position to draw \a str.  Assumed to be in normalized device coordinates.
	 */
	virtual bool QueueText(FontRef f, const char *str, const Color4f &color, const Point2f &p) = 0;
	
	/*!
	 * \brief Queues aligned text to be drawn a the end of the frame.
	 *
	 * \param[in] f the font used to draw the text.
	 * \param[in] str the text to draw.
	 * \param[in] color the color to draw \a str.
	 * \param[in] align the text alignment to use when drawing \a str.  The screen is used as the area to align to.
	 * \param[in] offset the amount to offset the text.  Assumed to be in normalized device coordinates.  Defaults to (0,0).
	 */
	virtual bool QueueText(FontRef f, const char *str, const Color4f &color, Font::TextAlignment align, const Point2f &offset = Point2f(0.0f, 0.0f)) = 0;

	/*!
	 * \brief Sets the BlendMode all further draw calls  Default is BLENDMODE_ALPHA. NOTE: BeginBlendMode/EndBlendMode cannot be nested.
	 *
	 * \param[in] mode the BlendMode to use.
	 *
	 * \return True if the BlendMode was successfully set.  Returns false if the blend mode could not be set.  This can occure if
	 * BeginBlendMode has already been called but EndBlendMode has not.
	 */
	virtual bool BeginBlendMode(BlendMode mode) = 0;
	
	/*!
	 * \brief Ends the BlendMode mode being used and reverts back to BLENDMODE_ALPHA.
	 *
	 * \return True the BlendMode has been set back to BLENDMODE_ALPHA, returns false if EndBlendMode is called without first calling BeginBlendMode.
	 */
	virtual bool EndBlendMode() = 0;
	
	/*!
	 * \brief Gets the currently active blend mode
	 *
	 * \returns the active BlendMode.
	 */
	virtual BlendMode GetBlendMode() const = 0;
	
	/*!
	 * \brief Sets the screen area where all subsequent draw calls will be limited to.
	 *
	 * \param[in] area the screen rectangle.
	 * 
	 * \return true if the draw region was created, false otherwise.
	 *
	 * NOTE: This can be called without a matching call to DisableDrawRegion().
	 */
	virtual bool EnableDrawRegion(Recti area) = 0;
	
	/*!
	 * \brief Disables using the draw engine.
	 *
	 * \return true if the draw region was disabled, false otherwise.
	 */
	virtual bool DisableDrawRegion() = 0;
	
	/*!
	 * \brief Gets the current draw region.
	 *
	 * \return the draw region.
	 */
	virtual Recti GetDrawRegion() const = 0;
};

#endif
