/*********************************************************************
 
	File: Sprite.cpp
	Data: 3/18/2008
	Desc: Implements the Sprite class.
 
**********************************************************************/

#include <algorithm>
#include <SDL_opengl.h>
#include "Math/MathCommon.h"
#include "Common/Globals.h"
#include "Graphics/Sprite.h"

//--------------------------------------------------------------------
Sprite::Sprite() : angle(0.0f), alpha(1.0f), scaleX(1.0f), scaleY(1.0f) {}

//--------------------------------------------------------------------
Sprite::Sprite(const char *file) : angle(0.0f), alpha(1.0f), scaleX(1.0f), scaleY(1.0f)
{
	Load(file);
}

//--------------------------------------------------------------------
bool Sprite::Load(const char *file)
{
	angle = 0.0f;
	alpha = 1.0f;
	scaleX = 1.0f;
	scaleY = 1.0f;

	ImageRef image = LoadImageResource(file);
	if(!image.IsValid())
		return false;

	im = image;
	return true;
}

//--------------------------------------------------------------------
void Sprite::SetAngle(float a)
{
	angle = a;
	if(a > _2PI || a < -_2PI)
	{
		a = a - (_2PI / a) * _2PI;
	}
}

//--------------------------------------------------------------------
void Sprite::Rotate(float a)
{
	angle += a;
	if(a > _2PI || a < -_2PI)
	{
		a = a - (_2PI / a) * _2PI;
	}
}

//--------------------------------------------------------------------
void Sprite::SetPosition(const Point2f &p)
{
	pos = p;
}

//--------------------------------------------------------------------
void Sprite::Move(const Point2f &p)
{
	pos += p;
}

//--------------------------------------------------------------------
void Sprite::SetTransparency(float t)
{
	alpha = t;
	Clamp(t, 0.0f, 1.0f);
}

//--------------------------------------------------------------------
void Sprite::Fade(float t)
{
	alpha += t;
	Clamp(t, 0.0f, 1.0f);
}

//--------------------------------------------------------------------
void Sprite::SetSize(float w, float h)
{
	SetWidth(w);
	SetHeight(h);
}

//--------------------------------------------------------------------
void Sprite::SetWidth(float w)
{
	assert(w > 0.0f);

	scaleX = w / im->Width();
}

//--------------------------------------------------------------------
void Sprite::SetHeight(float h)
{
	assert(h > 0.0f);

	scaleY = h / im->Height();
}

//--------------------------------------------------------------------
void Sprite::SetWidthUniform(float w)
{
	SetWidth(w);
	scaleY = scaleX; 
}

//--------------------------------------------------------------------
void Sprite::SetHeightUniform(float h)
{
	SetHeight(h);
	scaleX = scaleY;
}

//--------------------------------------------------------------------
void Sprite::SetScale(float s)
{
	scaleX = s;
	ClampLow(scaleX, 0.0f);
	scaleY = scaleX;
}

//--------------------------------------------------------------------
void Sprite::Grow(float s)
{
	scaleX += s;
	ClampLow(scaleX, 0.0f);
	scaleY = scaleX;
}

//--------------------------------------------------------------------
void Sprite::SetScale(float sx, float sy)
{
	scaleX = sx;
	scaleY = sy;
	ClampLow(scaleX, 0.0f);
	ClampLow(scaleY, 0.0f);
}

//--------------------------------------------------------------------
void Sprite::Grow(float sx, float sy)
{
	scaleX += sx;
	scaleY += sy;
	ClampLow(scaleX, 0.0f);
	ClampLow(scaleY, 0.0f);
}

//--------------------------------------------------------------------
AABB Sprite::GetBBox() const
{	
	Matrixf S, T, R;
	S.MakeScale(scaleX, scaleY);
	T.MakeTranslation(pos.x, pos.y);
	R.MakeRotation(angle);
	
	Matrixf trans = T * R * S;
	
	const float w_2 = im->Width() / 2.0f;
	const float h_2 = im->Height() / 2.0f;
	
	Point2f p1 = trans.Transform( Point2f(-w_2, -h_2) );
	Point2f p2 = trans.Transform( Point2f( w_2, -h_2) );
	Point2f p3 = trans.Transform( Point2f( w_2,  h_2) );
	Point2f p4 = trans.Transform( Point2f(-w_2,  h_2) );

	AABB aabb;
	aabb.min.x = std::min( std::min(p1.x, p2.x), std::min(p3.x, p4.x) );
	aabb.min.y = std::min( std::min(p1.y, p2.y), std::min(p3.y, p4.y) );
	
	aabb.max.x = std::max( std::max(p1.x, p2.x), std::max(p3.x, p4.x) );
	aabb.max.y = std::max( std::max(p1.y, p2.y), std::max(p3.y, p4.y) );
	return aabb;
}

//--------------------------------------------------------------------
bool Sprite::CanDraw() const
{
	return im.IsValid();
}

//--------------------------------------------------------------------
void Sprite::Draw(Image::DrawLocation loc) const
{
	Draw(pos, loc);
}

//--------------------------------------------------------------------
void Sprite::Draw(const Point2f &p, Image::DrawLocation loc) const
{
	if(im)
	{
		Rectf area;
		area.x = 0.0f;
		area.y = 0.0f;
		area.w = (float)im->Width();
		area.h = (float)im->Height();
		DrawSubArea(p, area, loc);
	}
}

//--------------------------------------------------------------------
void Sprite::DrawSubArea(const Rectf &area, Image::DrawLocation loc) const
{
	DrawSubArea(pos, area, loc);
}

//--------------------------------------------------------------------
void Sprite::DrawSubArea(const Point2f &p, const Rectf &area, Image::DrawLocation loc) const
{
	if(im)
	{
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glTranslatef(p.x, p.y, 0.0f);
		glRotatef(RadToDeg(angle), 0.0f, 0.0f, 1.0f);
		glScalef(scaleX, scaleY, 1.0f);
	
		im->DrawSubImage(area, alpha, loc);
	
		glPopMatrix();
	}
}
