/*!
 * \file Sprite.h
 * \brief Declares the sprite class.
 * \author Joe Forte
 * \date 3/18/2008
 */

#ifndef SPRITE_H
#define SPRITE_H

#include "Math/MathCommon.h"
#include "Graphics/Image.h"

/*!
 * \class Sprite
 * \headerfile Sprite.h "Graphics/Sprite.h"
 * \brief Class that allows for the movment, scaling and rotation of 2D images.
 */
class Sprite
{
public:
	/*!
	 * \brief Default constructor.
	 */
	Sprite();
	
	/*!
	 * \brief Constructs a Sprite using the given file. 
	 *
	 * \param[in] file the file name of the image to load.
	 */
	Sprite(const char *file);
	
	/*!
	 * \brief Sets the image the Sprite takes.
	 *
	 * \param[in] file the file name of the image to load.
	 *
	 * \return True if the image was loaded, false otherwise.
	 */
	bool Load(const char *file);
	
	/*!
	 * \brief Sets angle of rotation for the Sprite.
	 *
	 * \param[in] a the angle in radians.
	 */
	void SetAngle(float a);
	
	/*!
	 * \brief Increments the angle of rotation for the Sprite.
	 *
	 * \param[in] a the angle in radians.
	 */
	void Rotate(float a);
	
	/*!
	 * \brief Gets the current angle of the Sprite.
	 *
	 * \return The angle of the Sprite in radians.
	 */
	float Angle() const { return angle; }
	
	/*!
	 * \brief Sets the position of the Sprite.
	 *
	 * \param[in] p the position of the Sprite.
	 */
	void SetPosition(const Point2f &p);
	
	/*!
	 * \brief Moves the position of the Sprite.
	 *
	 * \param[in] p the amount to move the Sprite.
	 */
	void Move(const Point2f &p);
	
	/*!
	 * \brief Gets the current position of the Sprite.
	 *
	 * \return The current position of the Sprite.
	 */
	Point2f Position() const { return pos; }
	
	/*!
	 * \brief Sets the opacity of the Sprite.
	 *
	 * \param[in] t the transparency to set the Sprite. \a t should be in the range [0,1] 
	 * where 0 is fully transparent and 1 is fully opaque.
	 */
	void SetTransparency(float t);
	
	/*!
	 * \brief Fades the Sprite in and out.
	 *
	 * \param[in] t the amount to fade the transparency of the Sprite.
	 */
	void Fade(float t);
	
	/*!
	 * \brief Gets the current transparency of the Sprite.
	 *
	 * \return The current transparency of the Sprite.
	 */
	float Transparency() const { return alpha; }

	/*!
	 * \brief Sets the size of the Sprite.
	 *
	 * \param[in] w the desired width of the sprite
	 * \param[in] h the desired height of the sprite
	 *
	 * Computes the scale factors to make the loaded image have the given width and height.
	 */
	void SetSize(float w, float h);

	/*!
	 * \brief Sets the width of the Sprite.
	 *
	 * \param[in] w the desired width of the sprite
	 *
	 * Computes the scale factor that make the loaded image have the given width.
	 */
	void SetWidth(float w);

	/*!
	 * \brief Sets the height of the Sprite.
	 *
	 * \param[in] h the desired height of the sprite
	 *
	 * Computes the scale factor that make the loaded image have the given width.
	 */
	void SetHeight(float h);

	/*!
	 * \brief Sets the width of the Sprite and scales the height uniformly.
	 *
	 * \param[in] w the desired width of the sprite
	 *
	 * Computes the scale factor that make the loaded image have the given width.
	 */
	void SetWidthUniform(float w);

	/*!
	 * \brief Sets the height of the Sprite and scales the width uniformly.
	 *
	 * \param[in] h the desired height of the sprite
	 *
	 * Computes the scale factor that make the loaded image have the given height.
	 */
	void SetHeightUniform(float h);
	
	/*!
	 * \brief Sets the uniform scale of the Sprite.
	 *
	 * \param[in] s the uniform scaling factor, \f$s > 0\f$.
	 */
	void SetScale(float s);
	
	/*!
	 * \brief Sets the nonuniform scale of the Sprite.
	 *
	 * \param[in] sx the scaling factor along the x-axis, \f$s_x > 0\f$.
	 * \param[in] sy the scaling factor along the y-axis, \f$s_y > 0\f$.
	 */
	void SetScale(float sx, float sy);
	
	/*!
	 * \brief Scales the Sprite uniformly.
	 *
	 * \param[in] s the amount to uniformly scaling the Sprite.
	 */
	void Grow(float s);
	
	/*!
	 * \brief Scales the Sprite non-uniformly.
	 *
	 * \param[in] sx the amount to scale the Sprite along the x-axis.
	 * \param[in] sy the amount to scale the Sprite along the y-axis.
	 */
	void Grow(float sx, float sy);
	
	/*!
	 * \brief Gets the scaling factors for the Sprite.
	 *
	 * \param[out] sx the scaling factor along the x-axis.
	 * \param[out] sy the scaling factor along the y-axis.
	 */
	void Scale(float &sx, float &sy) const { sx = scaleX; sy = scaleY; }
	
	/*!
	 * \brief Retrives the width and height of the Sprite.
	 *
	 * \param[out] w the width of the Sprite.
	 * \param[out] h the height of the Sprite.
	 */
	void GetDimensions(float &w, float &h) const { w = GetWidth(); h = GetHeight(); }

	/*!
	 * \brief Retrives the height of the Sprite.
	 *
	 * \return The height of the Sprite.
	 */
	float GetHeight() const { return (float)im->Height() * scaleY; }

	/*!
	 * \brief Retrives the Width of the Sprite.
	 *
	 * \return The Width of the Sprite.
	 */
	float GetWidth() const { return (float)im->Width() * scaleX; }
	
	/*!
	 * \brief Computes the bounding box for the Sprite.
	 *
	 * \return The AABB that bounds the Sprite.
	 */
	AABB GetBBox() const;

	/*!
	 * \brief Determines if the sprite is in a state where it can be drawn.
	 *
	 * \return True if the sprite can be drawn, flase otherwise.
	 */
	bool CanDraw() const;
	
	/*!
	 * \brief Draws the sprite.
	 *
	 * \param[in] loc the locaation to draw the fixed point of the Sprite. Defaults to CENTER.
	 * \see Image
	 */
	void Draw(Image::DrawLocation loc = Image::CENTER) const;

	/*!
	 * \brief Draws the sprite.
	 *
	 * \param[in] pos the position to draw the sprite.  Supersedes the position of the sprite.
	 * \param[in] loc the locaation to draw the fixed point of the Sprite. Defaults to CENTER.
	 * \see Image
	 */
	void Draw(const Point2f &p, Image::DrawLocation loc = Image::CENTER) const;

	/*!
	 * \brief Draws a portion of the sprite.
	 *
	 * \param[in] area the subregioin to draw from the sprite. 
	 * \param[in] loc the locaation to draw the fixed point of the Sprite. Defaults to CENTER.
	 * \see Image
	 */
	void DrawSubArea(const Rectf &area, Image::DrawLocation loc = Image::CENTER) const;

	/*!
	 * \brief Draws a portion of the sprite.
	 *
	 * \param[in] p the position to draw the sprite.  Supersedes the position of the sprite.
	 * \param[in] area the subregioin to draw from the sprite. 
	 * \param[in] loc the locaation to draw the fixed point of the Sprite. Defaults to CENTER.
	 * \see Image
	 */
	void DrawSubArea(const Point2f &p, const Rectf &area, Image::DrawLocation loc = Image::CENTER) const;

	/*!
	 * \brief Gets the image associated with the Sprite.
	 *
	 * \return The ImageRef used by the Sprite.
	 */
	ImageRef GetImage() { return im; }
	
protected:
	ImageRef im;
	
	float alpha;
	float angle;
	float scaleX;
	float scaleY;
	Point2f pos;
};

#endif
