/*!
 * \file BufferedKeyboardInput.cpp
 * \brief Implements the BufferedKeyboardInput class.
 * \author Joe Forte
 * \date 10/17/2009
 */

#include <assert.h>
#include <limits>
#include <algorithm>

#include "Math/MathCommon.h"
#include "Input/BufferedKeyboardInput.h"

//--------------------------------------------------------------------
BufferedKeyboardInput::BufferedKeyboardInput() : cursorPos(0) {}

//--------------------------------------------------------------------
void BufferedKeyboardInput::Update(const IKeyboard &kb)
{
	uint size = kb.InputQueueSize();
	for(uint i = 0; i < size; i++)
	{
		BufferKeyInput( kb.InputQueueGet(i) );
	}
}

//--------------------------------------------------------------------
void BufferedKeyboardInput::BufferKeyInput(const KeyData &data)
{
	if( !FilterKeyInput(data) )
	{
		HandleKeyInput(data);
	}
}

//--------------------------------------------------------------------
void BufferedKeyboardInput::ClearInput()
{
	input.clear();
	SetCursorPos(0);
}

//--------------------------------------------------------------------
void BufferedKeyboardInput::SetInput(const char *str)
{
	SetInput( std::string(str) );
}

//--------------------------------------------------------------------
void BufferedKeyboardInput::SetInput(const std::string &str)
{
	input = str;
	ClampCursorPos();
}

//--------------------------------------------------------------------
const std::string &BufferedKeyboardInput::GetInput() const
{
	return input;
}

//--------------------------------------------------------------------
uint BufferedKeyboardInput::GetInputLength() const
{
	return (uint)input.size();
}

//--------------------------------------------------------------------
void BufferedKeyboardInput::SetCursorPos(uint p)
{
	cursorPos = p;
	ClampCursorPos();
}

//--------------------------------------------------------------------
void BufferedKeyboardInput::MoveCursorToEnd()
{
	SetCursorPos( GetInputLength() );
}

//--------------------------------------------------------------------
void BufferedKeyboardInput::ClampCursorPos()
{
	cursorPos = Clamp(cursorPos, 0u, GetInputLength());
}

//--------------------------------------------------------------------
bool BufferedKeyboardInput::FilterKeyInput(const KeyData &data)
{
	if( data.state == Keys::KEYSTATE_RELEASED || data.state == Keys::KEYSTATE_UP )
		return true;
	
	return false;
}

//--------------------------------------------------------------------
bool BufferedKeyboardInput::ShouldInsertUnicodeChar(uint unicode) const
{
	return unicode > 0 && (unicode & 0xFF80) == 0; // Only the displayable ascii characters are considered
}

//--------------------------------------------------------------------
void BufferedKeyboardInput::HandleKeyInput(const KeyData &data)
{
	if(data.key == Keys::KEY_BACKSPACE)
	{
		if( cursorPos > 0 )
		{
			input.erase(cursorPos - 1, 1);
			cursorPos--;
		}
	}
	else if(data.key == Keys::KEY_DELETE)
	{
		if( !input.empty() && cursorPos < input.size() )
		{
			input.erase(cursorPos, 1);
		}
	}
	else if(data.key == Keys::KEY_LEFT)
	{
		if(cursorPos > 0)
			cursorPos--;
	}
	else if(data.key == Keys::KEY_RIGHT)
	{
		cursorPos++;
		ClampCursorPos();
	}
	else if( ShouldInsertUnicodeChar(data.unicodeChar) )
	{
		input.insert(cursorPos, 1, (char)data.unicodeChar);
		cursorPos++;
	}
}

//--------------------------------------------------------------------
KeyFilteredBufferedKeyboardInput::KeyFilteredBufferedKeyboardInput(const Key *keys, uint n)
{
	for(uint i = 0; i < n; i++)
	{
		AddFilteredKey(keys[i]);
	}
}

//--------------------------------------------------------------------
void KeyFilteredBufferedKeyboardInput::AddFilteredKey(Key k)
{
	if(IsKeyFiltered(k))
		return;
	
	filteredKeys.push_back(k);
	std::sort(filteredKeys.begin(), filteredKeys.end());
}

//--------------------------------------------------------------------
bool KeyFilteredBufferedKeyboardInput::IsKeyFiltered(Key k) const
{
	return std::binary_search(filteredKeys.begin(), filteredKeys.end(), k);
}

//--------------------------------------------------------------------
bool KeyFilteredBufferedKeyboardInput::FilterKeyInput(const KeyData &data)
{
	if(IsKeyFiltered(data.key))
		return true;
	
	return BufferedKeyboardInput::FilterKeyInput(data);
}

//--------------------------------------------------------------------
LengthLimitedBufferedKeyboardInput::LengthLimitedBufferedKeyboardInput() 
:
maxLength(std::numeric_limits<uint>::max()) 
{}

//--------------------------------------------------------------------
LengthLimitedBufferedKeyboardInput::LengthLimitedBufferedKeyboardInput(uint len) 
:
maxLength(len)
{}

//--------------------------------------------------------------------
LengthLimitedBufferedKeyboardInput::LengthLimitedBufferedKeyboardInput(uint len, const Key *keys, uint n)
:
KeyFilteredBufferedKeyboardInput(keys, n),
maxLength(len)
{}

//--------------------------------------------------------------------
void LengthLimitedBufferedKeyboardInput::SetMaxLength(uint len)
{
	maxLength = len;
	
	if(GetInputLength() > maxLength)
	{
		std::string input = GetInput();
		SetInput(input.substr(0, maxLength));
		ClampCursorPos();
	}
}

//--------------------------------------------------------------------
uint LengthLimitedBufferedKeyboardInput::GetMaxLength() const
{
	return maxLength;
}

//--------------------------------------------------------------------
bool LengthLimitedBufferedKeyboardInput::ShouldInsertUnicodeChar(uint unicode) const
{
	if( GetInputLength() == GetMaxLength() )
		return false;
	
	assert( GetInputLength() < GetMaxLength() );
	
	return KeyFilteredBufferedKeyboardInput::ShouldInsertUnicodeChar(unicode);
}
