/*!
 * \file BufferedKeyboardInput.h
 * \brief Declares the BufferedKeyboardInput class that buffers text input from a keyboard.
 * \author Joe Forte
 * \date 10/17/2009
 */

#ifndef BUFFEREDKEYBOARDINPUT_H
#define BUFFEREDKEYBOARDINPUT_H

#include "Input/Keyboard.h"
#include "Common/BaseTypes.h"

#include <string>

/*!
 * \class BufferedKeyboardInput
 * \headerfile BufferedKeyboardInput.h "Input/BufferedKeyboardInput.h"
 * \brief Class that buffers text input from a keyboard.
 */
class BufferedKeyboardInput
{
public:

	/*!
	 * \brief Constructor.
	 */
	BufferedKeyboardInput();

	/*!
	 * \brief Updates the buffer with new input from the keyboard.
	 *
	 * \param[in] kb the keyboard to get input data from.
	 */
	void Update(const IKeyboard &kb);

	/*!
	 * \brief Buffers a single input.
	 *
	 * \param[in] data the key input.
	 */
	void BufferKeyInput(const KeyData &data);

	/*!
	 * \brief Clears the buffered text input.
	 */
	void ClearInput();
	
	/*!
	 * \brief Sets the buffered text input sting.
	 *
	 * \param[in] str the string to set the text input to.
	 */
	void SetInput(const char *str);
	
	/*!
	 * \brief Sets the buffered text input sting.
	 *
	 * \param[in] str the string to set the text input to.
	 */
	void SetInput(const std::string &str);
	
	/*!
	 * \brief Gets the buffered text input string.
	 *
	 * \return the text input string.
	 */
	const std::string &GetInput() const;
	
	/*!
	 * \brief Returns the length of the input text.
	 */
	uint GetInputLength() const;
	
	/*!
	 * \brief Gets the buffered text input cursor position.
	 *
	 * \return the buffered text input cursor position.
	 */
	int GetCursorPos() const { return cursorPos; }
	
	/*!
	 * \brief Moves the cursor to the end if the buffered input.
	 */
	void MoveCursorToEnd();
	
	/*!
	 * \brief Sets the buffered text input cursor position.
	 *
	 * \param[in] p the buffered text input cursor position.
	 */
	void SetCursorPos(uint p);

protected:
	
	/*!
	 * \brief Determines if the input should be ignored.
	 *
	 * \param[in] data the input to filter.
	 *
	 * \return true if the input should be ignored, false otherwise.
	 */
	virtual bool FilterKeyInput(const KeyData &data);
	
	/*!
	 * \brief Determines if the given unicode char should be inserted into the input.
	 */
	virtual bool ShouldInsertUnicodeChar(uint unicode) const;
	
	/*!
	 * \brief Inserts a single key input into the test input.
	 *
	 * \param[in] data the input to insert into the bufferd text input.
	 */
	virtual void HandleKeyInput(const KeyData &data);
	
	/*!
	 * \brief Clamps the cursor into a valid range.
	 */
	void ClampCursorPos();
	
private:

	std::string input;
	uint cursorPos;

};

/*!
 * \class KeyFilteredBufferedKeyboardInput
 * \headerfile BufferedKeyboardInput.h "Input/BufferedKeyboardInput.h"
 * \brief Class that filters keys before buffering text input from a keyboard.
 */
class KeyFilteredBufferedKeyboardInput : public BufferedKeyboardInput
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	KeyFilteredBufferedKeyboardInput() {}
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] keys an array of keys to filter.
	 * \param[in] n the length of the keys array.
	 */
	KeyFilteredBufferedKeyboardInput(const Key *keys, uint n);
	
	/*!
	 * \brief Adds a key to be filtered.
	 *
	 * \param[in] k the key to be added.
	 */
	void AddFilteredKey(Key k);
	
	/*!
	 * \brief Returns true if the input key is filtered, false otherwise.
	 */
	bool IsKeyFiltered(Key k) const;
	
protected:
	
	/*!
	 * \brief Determines if the input should be ignored.
	 *
	 * \param[in] data the input to filter.
	 *
	 * \return true if the input should be ignored, false otherwise.
	 */
	virtual bool FilterKeyInput(const KeyData &data);
	
private:
	
	std::vector<Key> filteredKeys;
	
};

/*!
 * \class LengthLimitedBufferedKeyboardInput
 * \headerfile BufferedKeyboardInput.h "Input/BufferedKeyboardInput.h"
 * \brief Class that limits the length of input from the keyboard.
 */
class LengthLimitedBufferedKeyboardInput : public KeyFilteredBufferedKeyboardInput
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	LengthLimitedBufferedKeyboardInput();
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] len the max length of the input.
	 */
	LengthLimitedBufferedKeyboardInput(uint len);
	
	/*!
	 * \brief Constructor.
	 *
	 * \param[in] len the max length of the input.
	 * \param[in] keys an array of keys to filter.
	 * \param[in] n the length of the keys array.
	 */
	LengthLimitedBufferedKeyboardInput(uint len, const Key *keys, uint n);
	
	/*!
	 * \brief Sets the max length.
	 */
	void SetMaxLength(uint len);
	
	/*!
	 * \brief Returns the max length.
	 */
	uint GetMaxLength() const;
	
protected:
	
	/*!
	 * \brief Determines if the given unicode char should be inserted into the input.
	 */
	virtual bool ShouldInsertUnicodeChar(uint unicode) const;
	
private:
	
	uint maxLength;
	
};

#endif
