/*!
 * \file Keyboard.h
 * \brief Declares the Keyboard interface.
 * \author Joe Forte
 * \date 9/12/2008
 */

#ifndef KEYBOARD_H
#define KEYBOARD_H

#include <string>
#include "Common/BaseTypes.h"
#include "Input/Keys.h"

/*!
 * \class KeyData
 * \headerfile Keyboard.h "Input/Keyboard.h"
 * \brief Defines all of the information needed to fully describe a single key input.
 */
struct KeyData
{
	Key key;
	KeyMod mod;
	KeyState state;
	
	int unicodeChar;
	
	bool repeated; /*!< Set to true if the key is repeated. */
};

/*!
 * \class IKeyboard
 * \headerfile Keyboard.h "Input/Keyboard.h"
 * \brief Defines the interface that a Keyboard implementation must implement.
 */
class IKeyboard
{
public:
	
	/*!
	 * \brief Determines if the given Key is down.
	 *
	 * \param[in] k the Key to check.
	 *
	 * \return True if the Key \a k is down, false otherwise.
	 */
	virtual bool KeyDown(Key k) const = 0;
	
	/*!
	 * \brief Determines if the given Key was just pressed.
	 *
	 * \param[in] k the Key to check.
	 *
	 * \return True if the Key \a k was just pressed, false otherwise.
	 */
	virtual bool KeyPressed(Key k) const = 0;
	
	/*!
	 * \brief Determines if the given Key was just pressed or is being repeated.
	 *
	 * \param[in] k the Key to check.
	 *
	 * \return True if the Key \a k was just pressed or is being repeated, false otherwise.
	 */
	virtual bool KeyPressedRepeat(Key k) const = 0;
	
	/*!
	 * \brief Determines if the given Key is being repeated.
	 *
	 * \param[in] k the Key to check.
	 *
	 * \return True if the Key \a k is being repeated, false otherwise.
	 */
	virtual bool KeyRepeat(Key k) const = 0;

	/*!
	 * \brief Returns the size of the input queue.
	 */
	virtual uint InputQueueSize() const = 0;

	/*!
	 * \brief Returns the key at the top of the input queue.
	 */
	virtual const KeyData &InputQueuePeek() const = 0;

	/*!
	 * \brief Returns the ith key in the input queue.
	 */
	virtual const KeyData &InputQueueGet(uint i) const = 0;
};

#endif
