/*!
 * \file Mouse.h
 * \brief Declares the mouse interface and mouse data types.
 * \author Joe Forte
 * \date 9/5/2008
 */

#ifndef MOUSE_H
#define MOUSE_H

#include "Math/Point.h"
#include "Input/Keys.h"

/*!
 * \class IMouse
 * \headerfile Mouse.h "Input/Mouse.h"
 * \brief Defines the interface that a Mouse implementation must implement.
 */
class IMouse
{
public:
	
	/*!
	 * \brief Determines if the given mouse button was just pressed.
	 *
	 * \param[in] button the MouseButton to check.
	 * \param[in] pos pointer that will hold the position where the \a button was pressed.  Defaults to NULL.
	 *
	 * \return True if the given button has been pressed.  Optionally outputs the position where the button was pressed.
	 */
	virtual bool ButtonPressed(MouseButton button, Point2i *pos = NULL) const = 0;
	
	/*!
	 * \brief Determines if the given mouse button was just released.
	 *
	 * \param[in] button the MouseButton to check.
	 * \param[in] pos pointer that will hold the position where the \a button was released.  Defaults to NULL.
	 *
	 * \return True if the given button wass just released.
	 */
	virtual bool ButtonReleased(MouseButton button, Point2i *pos = NULL) const = 0;
	
	/*!
	 * \brief Determines if the given mouse button is down.
	 *
	 * \param[in] button the MouseButton to check.
	 *
	 * \return True if the given button has been pressed and is now being held down.
	 */
	virtual bool ButtonDown(MouseButton button) const = 0;
	
	/*!
	 * \brief Determines if the mouse wheel has been moved up.
	 *
	 * \return True if the mouse wheel has been moved up. 
	 */
	virtual bool WheelUp() const = 0;
	
	/*!
	 * \brief Determines if the mouse wheel has been moved down.
	 *
	 * \return True if the mouse wheel has been moved down. 
	 */
	virtual bool WheelDown() const = 0;
	
	/*!
	 * \brief Determines if the mouse has screen focus.
	 *
	 * \return True if the mouse has screen focus, false otherwise.
	 */
	virtual bool HasFocus() const = 0;
	
	/*!
	 * \brief Determines if the mouse is in fixed position mode.
	 *
	 * \return True if the mouse is in fixed position mode, false otherwise.
	 */
	virtual bool InFixedPositionMode() const = 0;
	
	/*!
	 * \brief Gets the current position of the mouse.
	 *
	 * \return the current mouse position.
	 */
	virtual Point2i GetPos() const = 0;
	
	/*!
	 * \brief Gets the mouse position in normalized screen coordinates.
	 *
	 * \return the current mouse position in normalized screen coordinates.
	 */
	virtual Point2f GetPosNorm() const = 0;
	
	/*!
	 * \brief Gets the amount the mouse has moved since the last update.
	 *
	 * \return the amount the mouse has moved since the last update.
	 */
	virtual Point2i GetMoveDelta() const = 0;
};

#endif
