/*!
 * \file Keyboard.cpp
 * \brief Implements the SDLKeyboard class.
 * \author Joe Forte
 * \date 9/12/2008
 */

#include <assert.h>
#include <algorithm>
#include <SDL.h>

#include "Input/SDLKeyboard.h"

//--------------------------------------------------------------------
SDLKeyboard::SDLKeyboard()
{	
	KeyData data;
	data.key			= Keys::KEY_UNKNOWN;
	data.mod			= Keys::KEYMOD_NONE;
	data.unicodeChar	= 0;
	data.repeated		= false;
	
	std::fill_n(keys, (int)SDLK_LAST, data);
	
	SDL_EnableUNICODE(1);
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
}

//--------------------------------------------------------------------
void SDLKeyboard::HandleEvent(const SDL_KeyboardEvent &event)
{		
	KeyData data;
	data.key			= (Key)event.keysym.sym;
	data.mod			= (KeyMod)event.keysym.mod;
	data.unicodeChar	= (int)event.keysym.unicode;
	data.repeated		= false;
	
	if(event.state == SDL_PRESSED)
	{
		data.state = Keys::KEYSTATE_PRESSED;
		
		KeyState currState = keys[event.keysym.sym].state;
		if(currState == Keys::KEYSTATE_PRESSED || currState == Keys::KEYSTATE_DOWN)
			data.repeated = true;
	}
	else
	{
		data.state = Keys::KEYSTATE_RELEASED;
	}
	
	keys[event.keysym.sym] = data;
	inputQueue.push_back(data);
}

//--------------------------------------------------------------------
bool SDLKeyboard::KeyDown(Key k) const
{	
	return keys[k].state == Keys::KEYSTATE_DOWN || keys[k].state == Keys::KEYSTATE_PRESSED; 
}

//--------------------------------------------------------------------
bool SDLKeyboard::KeyPressed(Key k) const
{
	return keys[k].state == Keys::KEYSTATE_PRESSED && !keys[k].repeated;
}

//--------------------------------------------------------------------
bool SDLKeyboard::KeyPressedRepeat(Key k) const
{
	return keys[k].state == Keys::KEYSTATE_PRESSED;
}

//--------------------------------------------------------------------
bool SDLKeyboard::KeyRepeat(Key k) const
{
	return keys[k].state == Keys::KEYSTATE_PRESSED && keys[k].repeated;
}

//--------------------------------------------------------------------
void SDLKeyboard::Update(float dt)
{
	InputQueueClear();
	
	for(int k = 0; k < SDLK_LAST; k++)
	{
		if(keys[k].state == Keys::KEYSTATE_PRESSED)
		{
			keys[k].state = Keys::KEYSTATE_DOWN;
		}
		else if(keys[k].state == Keys::KEYSTATE_RELEASED)
		{
			keys[k].state = Keys::KEYSTATE_UP;
			keys[k].repeated = false;
		}
	}
}
