/*!
 * \file Keyboard.h
 * \brief Declares the SDL implementation of the IKeyboard interface.
 * \author Joe Forte
 * \date 9/12/2008
 */

#ifndef SDLKEYBOARD_H
#define SDLKEYBOARD_H

#include "Input/Keyboard.h"
#include "Common/BaseTypes.h"

#include <vector>

struct SDL_KeyboardEvent;

/*!
 * \class SDLKeyboard
 * \headerfile SDLKeyboard.h "Input/SDLKeyboard.h"
 * \brief The SDL implementation of the IKeyboard interface.
 */
class SDLKeyboard : public IKeyboard
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	SDLKeyboard();
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~SDLKeyboard() {}
	
	/*!
	 * \brief Determines if the given Key is down.
	 *
	 * \param[in] k the Key to check.
	 *
	 * \return True if the Key \a k is down, false otherwise.
	 */
	virtual bool KeyDown(Key k) const;
	
	/*!
	 * \brief Determines if the given Key was just pressed.
	 *
	 * \param[in] k the Key to check.
	 *
	 * \return True if the Key \a k was just pressed, false otherwise.
	 */
	virtual bool KeyPressed(Key k) const;
	
	/*!
	 * \brief Determines if the given Key was just pressed or is being repeated.
	 *
	 * \param[in] k the Key to check.
	 *
	 * \return True if the Key \a k was just pressed or is being repeated, false otherwise.
	 */
	virtual bool KeyPressedRepeat(Key k) const;
	
	/*!
	 * \brief Determines if the given Key is being repeated.
	 *
	 * \param[in] k the Key to check.
	 *
	 * \return True if the Key \a k is being repeated, false otherwise.
	 */
	virtual bool KeyRepeat(Key k) const;

	/*!
	 * \brief Returns the size of the input queue.
	 */
	virtual uint InputQueueSize() const { return (uint)inputQueue.size(); }

	/*!
	 * \brief Returns the key at the top of the input queue.
	 */
	virtual const KeyData &InputQueuePeek() const { return inputQueue.front(); }

	/*!
	 * \brief Returns the ith key in the input queue.
	 */
	virtual const KeyData &InputQueueGet(uint i) const { return inputQueue[i]; }

	/*!
	 * \brief Clears the iput queue.
	 */
	void InputQueueClear() { inputQueue.clear(); }
	
	/*!
	 * \brief Call once a frame.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 */
	void Update(float dt);
	
	/*!
	 * \brief Called each time an SDL keyboard event occures.
	 *
	 * \param[in] event the SDL keyboard event.
	 */
	void HandleEvent(const SDL_KeyboardEvent &event);
	
private:
	
	std::vector<KeyData> inputQueue;
	
	KeyData keys[Keys::KEY_COUNT];
};

#endif
