/*!
 * \file SDLMouse.cpp
 * \brief Implements the SDLMouse class.
 * \author Joe Forte
 * \date 9/5/2008
 */

#include <SDL.h>
#include <SDL_opengl.h>
#include "Input/SDLMouse.h"
#include "Math/MathCommon.h"
#include "Common/Globals.h"

//--------------------------------------------------------------------
SDLMouse::SDLMouse() : hasFocus(true), mouseWarped(false), fixedPositionMode(false)
{
	SDL_ShowCursor(SDL_DISABLE);
	
	for(int i = 0; i < Keys::MOUSE_BUTTON_COUNT; i++)
	{
		buttonStates[i] = BUTTON_UP;
		buttonPositions[i] = Point2i(0, 0);
	}

	int x, y;
	SDL_GetMouseState(&x, &y);
	lastPosition = currPosition = Point2i(x, y);
}

//--------------------------------------------------------------------
SDLMouse::~SDLMouse() {}

//--------------------------------------------------------------------
void SDLMouse::GainedFocus()
{ 
	hasFocus = true;
}

//--------------------------------------------------------------------
void SDLMouse::LostFocus()
{ 
	hasFocus = false;;
}

//--------------------------------------------------------------------
bool SDLMouse::HasFocus() const
{ 
	return hasFocus;
}

//--------------------------------------------------------------------
void SDLMouse::StartFixedPositionMode()
{
	fixedPositionMode = true;
}

//--------------------------------------------------------------------
void SDLMouse::EndFixedPositionMode()
{
	fixedPositionMode = false;
}

//--------------------------------------------------------------------
bool SDLMouse::InFixedPositionMode() const
{
	return fixedPositionMode;
}

//--------------------------------------------------------------------
void SDLMouse::HandleEvent(const SDL_MouseButtonEvent &e)
{
	if(e.button == SDL_BUTTON_WHEELUP)
	{
		wheelUp = true;
		return;
	}
	else if(e.button == SDL_BUTTON_WHEELDOWN)
	{
		wheelDown = true;
		return;
	}

	// Determines what button was pressed
	MouseButton button = Keys::MOUSE_BUTTON_UNKNOWN;
	switch(e.button)
	{
	case SDL_BUTTON_LEFT:
		button = Keys::MOUSE_BUTTON_LEFT;
		break;
	case SDL_BUTTON_MIDDLE:
		button = Keys::MOUSE_BUTTON_MIDDLE;
		break;
	case SDL_BUTTON_RIGHT:
		button = Keys::MOUSE_BUTTON_RIGHT;
		break;
	};

	if(e.type == SDL_MOUSEBUTTONDOWN)
		buttonStates[button] = BUTTON_PRESSED;
	else
		buttonStates[button] = BUTTON_RELEASED;

	buttonPositions[button] = Point2i((int)e.x, (int)e.y);
}

//--------------------------------------------------------------------
bool SDLMouse::ButtonPressed(MouseButton button, Point2i *pos) const
{
	if(button < 0 || button >= Keys::MOUSE_BUTTON_COUNT)
		return false;

	if(pos)
		*pos = buttonPositions[button];
	
	return buttonStates[button] == BUTTON_PRESSED;
}

//--------------------------------------------------------------------
bool SDLMouse::ButtonReleased(MouseButton button, Point2i *pos) const
{
	if(button < 0 || button >= Keys::MOUSE_BUTTON_COUNT)
		return false;
	
	if(pos)
		*pos = buttonPositions[button];
	
	return buttonStates[button] == BUTTON_RELEASED;
}

//--------------------------------------------------------------------
bool SDLMouse::ButtonDown(MouseButton button) const
{
	if(button < 0 || button >= Keys::MOUSE_BUTTON_COUNT)
		return false;

	return buttonStates[button] == BUTTON_DOWN;
}

//--------------------------------------------------------------------
bool SDLMouse::WheelUp() const
{
	return wheelUp;
}

//--------------------------------------------------------------------
bool SDLMouse:: WheelDown() const
{
	return wheelDown;
}

//--------------------------------------------------------------------
void SDLMouse::Update()
{
	// Ignores the mouse movment if it was since the last update warped.
	if(mouseWarped)
	{
		mouseWarped = false;
	}
	else
	{
		Point2i pos;
		SDL_GetMouseState(&pos.x, &pos.y);
		
		if(fixedPositionMode)
		{
			MoveTo(Point2i(0,0));
			currPosition = pos;
		}
		else
		{
			lastPosition = currPosition;
			currPosition = pos;
		}
	}

	// Resets wheel state
	wheelUp = false;
	wheelDown = false;
	
	for(int i = 0; i < Keys::MOUSE_BUTTON_COUNT; i++)
	{
		if( buttonStates[i] == BUTTON_PRESSED )
			buttonStates[i] = BUTTON_DOWN;
		else if( buttonStates[i] == BUTTON_RELEASED )
			buttonStates[i] = BUTTON_UP;
	}
}

//--------------------------------------------------------------------
void SDLMouse::MoveTo(const Point2i &pos)
{
	uint w, h;
	GetEnv()->ScreenDim(w, h);

	// Convertes pos to the coordinate system used by SDL
	Point2i p((w / 2) + pos.x, (h / 2) - pos.y);
	
	SDL_WarpMouse((Uint16)p.x, (Uint16)p.y);
	currPosition = lastPosition = p;

	mouseWarped = true;
}

//--------------------------------------------------------------------
Point2i SDLMouse::GetPos() const
{
	uint w, h;
	GetEnv()->ScreenDim(w, h);

	return Point2i(currPosition.x - (w / 2), (h / 2) - currPosition.y);
}

//--------------------------------------------------------------------
Point2f SDLMouse::GetPosNorm() const
{
	Point2i tmp = GetPos();
	Point2f p((float)tmp.x, (float)tmp.y);
	return ScreenToNormCoords(p);
}

//--------------------------------------------------------------------
Point2i SDLMouse::GetMoveDelta() const
{
	int dx = currPosition.x - lastPosition.x;
	int dy = currPosition.y - lastPosition.y;
	return Point2i(dx, dy);
}

//--------------------------------------------------------------------
void SDLMouse::SetCursor(const char *file)
{
	cursor.Load(file);
	cursor.SetHeightUniform(24.0f);
}

//--------------------------------------------------------------------
Sprite &SDLMouse::GetCursor()
{
	return cursor;
}

//--------------------------------------------------------------------
void SDLMouse::Draw()
{
	if(HasFocus())
	{
		Point2i p = GetPos();

		cursor.SetPosition( Point2f((float)p.x, (float)p.y) );
		cursor.Draw(Image::TOPLEFT);
	}
}
