/*!
 * \file SDLMouse.h
 * \brief Declares the SDL implementation of the IMouse interface.
 * \author Joe Forte
 * \date 9/5/2008
 */

#ifndef SDLMOUSE_H
#define SDLMOUSE_H

#include "Math/Point.h"
#include "Graphics/Sprite.h"
#include "Input/Mouse.h"

struct SDL_MouseButtonEvent;

/*!
 * \class SDLMouse
 * \headerfile SDLMouse.h "Input/SDLMouse.h"
 * \brief The SDL implementation of the IMouse interface.
 */
class SDLMouse : public IMouse
{
public:

	/*!
	 * \brief Constructor.
	 */
	SDLMouse();

	/*!
	 * \brief Deconstructor.
	 */
	virtual ~SDLMouse();

	/*!
	 * \brief Determines if the given mouse button was just pressed.
	 *
	 * \param[in] button the MouseButton to check.
	 * \param[in] pos pointer that will hold the position where the \a button was pressed.  Defaults to NULL.
	 *
	 * \return True if the given button has been pressed.  Optionally outputs the position where the button was pressed.
	 */
	virtual bool ButtonPressed(MouseButton button, Point2i *pos = NULL) const;

	/*!
	 * \brief Determines if the given mouse button was just released.
	 *
	 * \param[in] button the MouseButton to check.
	 * \param[in] pos pointer that will hold the position where the \a button was released.  Defaults to NULL.
	 *
	 * \return True if the given button wass just released.
	 */
	virtual bool ButtonReleased(MouseButton button, Point2i *pos = NULL) const;

	/*!
	 * \brief Determines if the given mouse button is down.
	 *
	 * \param[in] button the MouseButton to check.
	 *
	 * \return True if the given button has been pressed and is now being held down.
	 */
	virtual bool ButtonDown(MouseButton button) const;

	/*!
	 * \brief Determines if the mouse wheel has been moved up.
	 *
	 * \return True if the mouse wheel has been moved up.
	 */
	virtual bool WheelUp() const;

	/*!
	 * \brief Determines if the mouse wheel has been moved down.
	 *
	 * \return True if the mouse wheel has been moved down.
	 */
	virtual bool WheelDown() const;

	/*!
	 * \brief Informs the mouse that it has has gained screen focus.
	 */
	virtual void GainedFocus();

	/*!
	 * \brief Informs the mouse that it has has lost screen focus.
	 */
	virtual void LostFocus();

	/*!
	 * \brief Determines if the mouse has screen focus.
	 *
	 * \return True if the mouse has screen focus, false otherwise.
	 */
	virtual bool HasFocus() const;

	/*!
	 * \brief updates the state of the mouse.  Should be called once at the beginning
	 * or the end of each frame.
	 */
	virtual void Update();

	/*!
	 * \brief Moves the mouse to the given position.
	 *
	 * \param[in] pos the position to move the mouse to.
	 */
	virtual void MoveTo(const Point2i &pos);

	/*!
	 * \brief Places the mouse into fixed position mode where the mouse is fixed at the origin and only relative mouse movement is considered.
	 */
	virtual void StartFixedPositionMode();

	/*!
	 * \brief Takes the mouse out of fixed position mode.
	 */
	virtual void EndFixedPositionMode();

	/*!
	 * \brief Determines if the mouse is in fixed position mode.
	 *
	 * \return True if the mouse is in fixed position mode, false otherwise.
	 */
	virtual bool InFixedPositionMode() const;

	/*!
	 * \brief Gets the current position of the mouse.
	 *
	 * \return the current mouse position.
	 */
	virtual Point2i GetPos() const;

	/*!
	 * \brief Gets the mouse position in normalized screen coordinates.
	 *
	 * \return the current mouse position in normalized screen coordinates.
	 */
	virtual Point2f GetPosNorm() const;

	/*!
	 * \brief Gets the amount the mouse has moved since the last update.
	 *
	 * \return the amount the mouse has moved since the last update.
	 */
	virtual Point2i GetMoveDelta() const;

	/*!
	 * \brief Set the cursor image.
	 *
	 * \param[in] file the name of the cursor image file.
	 */
	virtual void SetCursor(const char *file);

	/*!
	 * \brief Gets the cursor image.
	 *
	 * \return The sprite used to represent the cursor.
	 */
	virtual Sprite &GetCursor();

	/*!
	 * \brief Draws the mouse cursor.
	 */
	virtual void Draw();

	/*!
	 * \brief Called each time an SDL mouse event occures.
	 *
	 * \param[in] e the SDL mouse button event.
	 */
	void HandleEvent(const SDL_MouseButtonEvent &e);

protected:
	Point2i lastPosition;
	Point2i currPosition;

	bool hasFocus;

	bool wheelUp;
	bool wheelDown;

	bool mouseWarped; // Set to true when the mouse is moved with a call to MoveTo.

	bool fixedPositionMode;

	enum BUTTON_STATE
	{
		BUTTON_UP = 0,
		BUTTON_PRESSED,
		BUTTON_DOWN,
		BUTTON_RELEASED
	};

	BUTTON_STATE buttonStates[Keys::MOUSE_BUTTON_COUNT];
	Point2i buttonPositions[Keys::MOUSE_BUTTON_COUNT]; // The positions where the buttons were pressed/released

	Sprite cursor;
};

#endif
