/*!
 * \file MathCommon.cpp
 * \brief Implements common/useful functions.
 * \author Joe Forte
 * \date 3/27/2008
 */

#include <assert.h>
#include <ctime>
#include <SDL_opengl.h>
#include "Math/MathCommon.h"
#include "Common/Globals.h"

/** NOTE: A fair number of the collision related functions have been
    adapted from the book "Real-Time Collision Detection" by Christer
	Ericson.  I note the page numbers where applicable in the code.
	Its a great referance material and comes highly recommended. **/

//--------------------------------------------------------------------
Point2f Ray::PointOnLine(float t) const
{
	return start + dir * t;
}

//--------------------------------------------------------------------
Rectf AABB::ToRect() const
{
	return AABBToRect<float>(*this);
}

//--------------------------------------------------------------------
Point2f AABB::CenterPosition() const
{
	float w = (max.x - min.x) / 2.0f;
	float h = (max.y - min.y) / 2.0f;
	
	return Point2f(min.x + w, min.y + h);
}

//--------------------------------------------------------------------
void AABB::Extends(float &halfWidth, float &halfHeight) const
{
	Point2f c = CenterPosition();
	halfWidth = std::abs( max.x - c.x );
	halfHeight = std::abs( max.y - c.y );
}

//--------------------------------------------------------------------
float AABB::Width() const
{
	return std::abs( max.x - min.x );
}

//--------------------------------------------------------------------
float AABB::Height() const
{
	return std::abs( max.y - min.y );
}

//--------------------------------------------------------------------
bool AABB::Overlap(const AABB &b2) const
{
	if( min.x < b2.max.x && max.x > b2.min.x &&
        min.y < b2.max.y && max.y > b2.min.y )
	{
		return true;
	}
	
	return false;
}

//--------------------------------------------------------------------
bool AABB::Overlap(const Circle &c) const
{
	return SquaredDistanceTo(c.center) <= c.r * c.r;
}

//--------------------------------------------------------------------
int AABB::Intersect(const Line &l, float &enterTime, float &exitTime) const
{	
	Vector2f d = l.b - l.a;
	Ray r(l.a, d);

	int i = Intersect(r, enterTime, exitTime);

	if(i > 0 && exitTime > d.Length())
	{
		enterTime = exitTime;
		return i - 1;
	}
	
	return i;
}

//--------------------------------------------------------------------
int AABB::Intersect(const Ray &r, float &enterTime, float &exitTime) const
{	
	/** pg 181 **/
	float tmin = 0.0f;
	float tmax = FLT_MAX;

	for(int i = 0; i < 2; i++)
	{
		if( std::abs(r.dir[i]) < EPSILON )
		{
			// Ray is parallel.  Early exit if the start position is not within the slab.
			if(r.start[i] < min[i] || r.start[i] > max[i])
				return 0;
		}
		else
		{
			float ood = 1.0f / r.dir[i];
			float t1 = (min[i] - r.start[i]) * ood;
			float t2 = (max[i] - r.start[i]) * ood;

			if(t1 > t2)
				Swap(t1, t2);

			tmin = std::max(tmin, t1);
			tmax = std::min(tmax, t2);

			if(tmin > tmax)
				return 0;
		}
	}

	if(tmin == tmax || tmin <= 0.0f)
	{
		enterTime = exitTime = tmax;
		return 1;
	}

	enterTime = tmin;
	exitTime = tmax;
	return 2;
}

//--------------------------------------------------------------------
bool AABB::Inside(const Point2f &p) const
{
	if( p.x < max.x && p.x > min.x &&
        p.y < max.y && p.y > min.y )
	{
		return true;
	}
	
	return false;
}

//--------------------------------------------------------------------
Point2f AABB::ClosestPoint(const Point2f &p) const
{
	/** pg 130 **/
	Point2f out;
	if( Inside(p) )
	{
		for(int i = 0; i < 2; i++)
		{
			float v = p[i];
			v = std::min(v, min[i]);
			v = std::max(v, max[i]);
			out[i] = v;
		}
	}
	for(int i = 0; i < 2; i++)
	{
		float v = p[i];
		v = std::max(v, min[i]);
		v = std::min(v, max[i]);
		out[i] = v;
	}

	return out;
}

//--------------------------------------------------------------------
Vector2f AABB::GetClosestNormal(const Point2f &p) const
{
	Point2f c = ClosestPoint(p);
	float w = Width();
	
	if(c.y >= max.y)
	{
		if(c.x >= max.x)
			return Vector2f(SQRT2_2, SQRT2_2);
		
		if(c.x <= max.x - w)
			return Vector2f(-SQRT2_2, SQRT2_2);
		
		return Vector2f(0.0f, 1.0f);
	}
	
	if(c.y <= min.y)
	{
		if(c.x >= max.x)
			return Vector2f(SQRT2_2, -SQRT2_2);
		
		if(c.x <= max.x - w)
			return Vector2f(-SQRT2_2, -SQRT2_2);
		
		return Vector2f(0.0f, -1.0f);
	}
	
	if(c.x >= max.x)
		return Vector2f(1.0f, 0.0f);
	
	return Vector2f(-1.0f, 0.0f);
}

//--------------------------------------------------------------------
float AABB::SquaredDistanceTo(const Point2f &p) const
{
	/** pg 131 **/
	float dist = 0;
	for(int i = 0; i < 2; i++)
	{
		float v = p[i];
		if(v < min[i])
			dist += (min[i] - v) * (min[i] - v);

		if(v > max[i])
			dist += (v - max[i]) * (v - max[i]);
	}

	return dist;
}

//--------------------------------------------------------------------
float ClosestPoints(const Line &l1, const Line &l2, Point2f &out1, Point2f &out2)
{
	/** pg 149 **/
	Vector2f d1 = l1.b - l1.a;
	Vector2f d2 = l2.b - l2.a;
	Vector2f r = l1.a - l2.a;

	float a = d1.SquaredLength();
	float e = d2.SquaredLength();
	float f = d2.Dot(r);

	float s, t;
	if(a <= EPSILON && e <= EPSILON)
	{
		// Both lines degenerate into points
		out1 = l1.a;
		out2 = l2.a;
		return r.SquaredLength();
	}
	if(a <= EPSILON)
	{
		// First line is a point
		s = 0.0f;
		t = f / e;
		Clamp(t, 0.0f, 1.0f);
	}
	else
	{
		float c = d1.Dot(r);
		if(e <= EPSILON)
		{
			// Second line is a point
			t = 0.0f;
			s = -c / a;
			Clamp(s, 0.0f, 1.0f);
		}
		else
		{
			// neither line is degenerate
			float b = d1.Dot(d2);
			float denom = a*e - b*b;

			if(denom != 0.0f)
			{
				s = (b*f - c*e) / denom;
				Clamp(s, 0.0f, 1.0f);
			}
			else
			{
				s = 0.0f;
			}

			t = (b*s + f) / e;
			if(t < 0.0f)
			{
				t = 0.0f;
				s = -c / a;
				Clamp(s, 0.0f, 1.0f);
			}
			else if(t > 1.0f)
			{
				t = 1.0f;
				s = (b - c) / a;
				Clamp(s, 0.0f, 1.0f);
			}
		}
	}

	out1 = l1.a + d1 * s;
	out2 = l2.a + d2 * t;

	Vector2f tmp = out1 - out2;
	return tmp.SquaredLength();
}

//--------------------------------------------------------------------
int Intersect(const Ray &r, const Circle &c, float &t1, float &t2)
{
	/** pg 178 **/
	Vector2f m = r.start - c.center;
	float b = m.Dot( r.dir );
	float d = m.SquaredLength() - c.r * c.r;
	if(d > 0.0f && b > 0.0f) // ray is outside the sphere
		return 0;

	float discr = b*b - d;
	if(discr < 0.0f) // MISS
		return 0;

	float sq = std::sqrt(discr);
	t1 = -b - sq;
	t2 = -b + sq;

	if( t1 < 0.0f ) // the ray is inside the sphere
	{
		t1 = t2;
		return 1;
	}

	return 2;
}

//--------------------------------------------------------------------
int Intersect(const Line &l, const Circle &c, float &t1, float &t2)
{
	Vector2f d = l.b - l.a;
	Ray r(l.a, d);

	int i = Intersect(r, c, t1, t2);

	if(i > 0 && t2 > d.Length())
		return i - 1;
	return i;
}

//--------------------------------------------------------------------
Point2f ScreenToNormCoords(const Point2f &p)
{
	Point2f out;
	GetScreen()->ScreenToNormCoords(p.x, p.y, out.x, out.y);
	return out;
}

//--------------------------------------------------------------------
Point2f NormToScreenCoords(const Point2f &p)
{
	Point2f out;
	GetScreen()->NormToScreenCoords(p.x, p.y, out.x, out.y);
	return out;
}

//--------------------------------------------------------------------
uint TimeSeed()
{
	/* Hashes the bytes of time_t. */
	time_t t = time(NULL);
	byte *b = (byte*)&t;

	uint seed = 0;
	for(size_t i = 0; i < sizeof(time_t); i++)
	{
		seed = seed * (std::numeric_limits<byte>::max() + 2) + b[i];
	}

	return seed;
}
