/*!
 * \file MathCommon.h
 * \brief Defines and includes common Math data types and functions.
 * \author Joe Forte
 * \date 3/27/2008
 */

#ifndef MATHCOMMON_H
#define MATHCOMMON_H

/*********************************************************************
 *  Common Includes
 *********************************************************************/
#include <cmath>
#include <algorithm>
#include <vector>
#include <limits>
#include <float.h>

#ifndef _POINT_H
	#include "Math/Point.h"
#endif

#ifndef _VECTOR_H_
	#include "Math/Vector.h"
#endif

#ifndef _MATRIX_H
	#include "Math/Matrix.h"
#endif

/*********************************************************************
 *  Common Constants
 *********************************************************************/
const float PI			= 3.141592653589793f;	/*!< \f$\pi\f$ */
const float PI_2		= PI / 2.0f;			/*!< \f$\frac{\pi}{2}\f$ */
const float PI_4		= PI / 4.0f;			/*!< \f$\frac{\pi}{4}\f$ */
const float _2PI		= 2.0f * PI;			/*!< \f$2\pi\f$ */
const float INV_PI		= 1.0f / PI;			/*!< \f$\frac{1}{\pi}\f$ */
const float E			= 2.718281828459045f;	/*!< \f$e\f$ */
const float SQRT2       = 1.414213562373095f;	/*!< \f$\sqrt{2}\f$ */
const float INV_SQRT2   = 1.0f / SQRT2;			/*!< \f$\frac{1}\sqrt{2}\f$ */
const float SQRT2_2     = 0.707106781186547f;	/*!< \f$\frac{\sqrt{2}}{2}\f$ */
const float RAD_TO_DEG	= 180.0f / PI;			/*!< Radians to Degrees constant \f$\frac{180}{\pi}\f$ */
const float DEG_TO_RAD	= PI / 180.0f;			/*!< Degrees to Radians constant \f$\frac{\pi}{180}\f$ */
const float EPSILON     = 0.0001f;				/*!< Small floating point value. */

/*********************************************************************
 *  Common Macros
 *********************************************************************/

/*!
 * \brief Constructs an integer with one bit set.
 *
 * \param[in] x the xth bit in the integer to set.
 */
#define BIT(x) (1 << (x))

/*********************************************************************
 *  Common Types
 *********************************************************************/

/*!
 * \struct Rect_t
 * \headerfile MathCommon.h "Math/MathCommon.h"
 * \brief A 2D rectangle.
 *
 * \tparam the data type for the rectangle data. Usually \c float \c double or \c int.
 */
template<class T>
struct Rect_t
{
	T x, y; /*!< The (x,y) position for the top left corner of the rectangle */
	T w, h; /*!< The (w)idth and (h)eight of the rectangle. */ 
};

/*!
 * \typedef Rectf
 * \brief A Template instantiation of the Rect_t template using the \c float data type.
 */
typedef Rect_t<float> Rectf;

/*!
 * \typedef Rectd
 * \brief A Template instantiation of the Rect_t template using the \c double data type.
 */
typedef Rect_t<double> Rectd;

/*!
 * \typedef Recti
 * \brief A Template instantiation of the Rect_t template using the \c int data type.
 */
typedef Rect_t<int> Recti;

/*!
 * \struct Line
 * \headerfile MathCommon.h "Math/MathCommon.h"
 * \brief A 2D Line segment.
 */
struct Line
{
	Point2f a; /*!< Staring position for the line segment. */
	Point2f b; /*!< End position for the line segment. */
};

/*!
 * \struct Circle
 * \headerfile MathCommon.h "Math/MathCommon.h"
 */
struct Circle
{
	Point2f center; /*!< Center positing of the circle. */
	float r; /*!< The radius of the circle. */
};

/*!
 * \struct Poly
 * \headerfile MathCommon.h "Math/MathCommon.h"
 */
struct Poly
{
	std::vector<Point2f> vertices;
};

/*!
 * \class Ray
 * \headerfile MathCommon.h "Math/MathCommon.h"
 * \brief A 2D Ray.
 */
class Ray
{
public:
	Point2f start; /*!< Starting position for the ray. */
	Vector2f dir; /*!< Direction of the ray.  Must always be normalized. */
	
	/*!
	 * \brief Default constructor.  The ray should not be used until after setting the directions. 
	 */
	Ray() {}
	
	/*!
	 * \brief Constructs a specific ray.
	 *
	 * \param[in] s the starting position.
	 * \param[in] d the direction (should never be the zero vector), does not have to be normalized.
	 */
	Ray(const Point2f &s, const Vector2f &d) : start(s), dir(d) { dir.Normalize(); }
	
	/*!
	 * \brief Constructs a ray from a line segment directed from the start of the line to the end.
	 *
	 * \param[in] l the line to construct the Ray from.  NOTE: \a l.a should not equal \a l.b
	 */
	Ray(const Line &l)
	{
		start = l.a;
		dir = l.b - l.a;
		dir.Normalize();
	}

	/*!
	 * \brief Sets the starting position of the Ray.
	 *
	 * \param[in] p the new starting position.
	 */
	void Set(const Point2f &p) { start = p; }
	
	/*!
	 * \brief Sets the direction of the Ray.
	 *
	 * \param[in] v the new direction (should never be the zero vector), does not have to be normalized.
	 */
	void Set(const Vector2f &v) { dir = v; dir.Normalize(); }
	
	/*!
	 * \brief Sets the starting position and the direction of the ray.
	 *
	 * \param[in] p the new starting position.
	 * \param[in] v the new direction (should never be the zero vector), does not have to be normalized.
	 */
	void Set(const Point2f &p, const Vector2f &v) { start = p; dir = v; dir.Normalize(); }
	
	/*!
	 * \brief Computes the point on the Ray at some given time.
	 *
	 * \param[in] t the time.
	 *
	 * \return The point on the Ray at time \a t.
	 */
	Point2f PointOnLine(float t) const;
};

/*!
 * \class AABB
 * \headerfile MathCommon.h "Math/MathCommon.h"
 * \brief 2D Axis Aligned Bounding Box.
 */
class AABB
{
public:
	Point2f min; /*!< Bottom left corner of the bounding box. */
	Point2f max; /*!< Top right corner of the bounding box. */
	
	/*!
	 * \brief Default constructor. 
	 */
	AABB() {}
	
	/*!
	 * \brief Constructs a specific AABB.
	 *
	 * \param[in] MIN the min position of the AABB.
	 * \param[in] MAX the max position of the AABB.
	 */
	AABB(const Point2f &MIN, const Point2f &MAX) : min(MIN), max(MAX) {}
	
	/*!
	 * \brief Constructs a specific AABB.
	 *
	 * \param[in] minx the x value of the min position of the AABB.
	 * \param[in] miny the y value of the min position of the AABB.
	 * \param[in] maxx the x value of the max position of the AABB.
	  *\param[in] maxy the y value of the max position of the AABB.
	 */
	AABB(float minx, float miny, float maxx, float maxy) : min(minx, miny), max(maxx, maxy) {}
	
	/*!
	 * \brief Constructs a specific AABB.
	 *
	 * \param[in] center the center position of the AABB.
	 * \param[in] w the width of the AABB.
	 * \param[in] h the height of the AABB.
	 */
	AABB(const Point2f &center, float w, float h)
	{
		w = w / 2.0f;
		h = h / 2.0f;

		min.x = center.x - w;
		min.y = center.y - h;
		max.x = center.x + w;
		max.y = center.y + h;
	}

	/*!
	 * \brief Converts the AABB to a Rect.
	 *
	 * \return The Rectf from of the AABB.
	 */
	Rectf ToRect() const;
	
	/*!
	 * \brief Computes the width of the AABB.
	 *
	 * \return The width of the AABB.
	 */
	float Width() const;
	
	/*!
	 * \brief Computes the height of the AABB.
	 *
	 * \return The height of the AABB.
	 */
	float Height() const;
	
	/*!
	 * \brief Computes the area of the AABB.
	 *
	 * \return the area.
	 */
	float Area() const { return Width() * Height(); }
	
	/*!
	 * \brief Computes the center position of the AABB.
	 *
	 * \return The center position of the AABB.
	 */
	Point2f CenterPosition() const;

	/*!
	 * \brief Gets the half width and height.
	 *
	 * \param[out] halfWidth the half width.
	 * \param[out] halfHeight the half height.
	 */
	void Extends(float &halfWidth, float &halfHeight) const;

	/*!
	 * \brief Determines if the given AABB overlaps with \e this AABB.
	 *
	 * \param[in] b2 the AABB to check for overlap against.
	 *
	 * \return True if \a b2 overlaps \e this AABB.  False if there is no overlap.
	 */
	bool Overlap(const AABB &b2) const;

	/*!
	 * \brief Determines if the given Circle overlaps with \e this AABB.
	 *
	 * \param[in] c the circle to check for overlap against.
	 *
	 * \return True if \a c overlaps \e this AABB.  False if there is no overlap.
	 */
	bool Overlap(const Circle &c) const;
	
	/*!
	 * \brief Intersects the given Lines with \e this AABB.
	 *
	 * \param[in] l The line to intersect against.
	 * \param[out] enterTime the time the ray enters the AABB. Only valid when 1 or 2 is returned.
	 * \param[out] exitTime the time the ray exits the AABB. Only valid when 1 or 2 is returned.
	 *
	 * \return The number of intersection points with \e this AABB, either 0, 1 or 2.  When 0 is
	 * returned there was no intersection and enterTime and exitTime are undifined.  When 1 is returned
	 * there is one intersection and enterTime will be equal to exitTime.  When 2 is returned there are
	 * two points of intersection and enterTime is strictly less then exit time.
	 */
	int Intersect(const Line &l, float &enterTime, float &exitTime) const;
	
	/*!
	 * \brief Intersects the given Ray with \e this AABB.
	 *
	 * \param[in] r The Ray to intersect against.
	 * \param[out] enterTime the time the ray enters the AABB. Only valid when 1 or 2 is returned.
	 * \param[out] exitTime the time the ray exits the AABB. Only valid when 1 or 2 is returned.
	 *
	 * \return The number of intersection points with \e this AABB, either 0, 1 or 2.  When 0 is
	 * returned there was no intersection and enterTime and exitTime are undifined.  When 1 is returned
	 * there is one intersection and enterTime will be equal to exitTime.  When 2 is returned there are
	 * two points of intersection and enterTime is strictly less then exit time.
	 */
	int Intersect(const Ray &r, float &enterTime, float &exitTime) const;	
	
	/*!
	 * \brief Determines if the given point is inside the AABB.
	 *
	 * \param[in] p the point to check against.
	 *
	 * \return True if \a p is inside the AABB. False is \a p is outside the AABB.
	 */
	bool Inside(const Point2f &p) const;

	/*!
	 * \brief Computes a point on the AABB closes to the given point.
	 *
	 * \param[in] p the point.
	 *
	 * \return A point on \e this AABB closes to \a p. 
	 */
	Point2f ClosestPoint(const Point2f &p) const;
	
	/*!
	 * \brief Determines the surface of the AABB closest to the given point and returns the normal vector of that surface.
	 *
	 * \param[in] p the point.
	 *
	 * \return The normal of the surface closest to the point \a p. 
	 */
	Vector2f GetClosestNormal(const Point2f &p) const; 

	/*!
	 * \brief Computes the squared distance between the given point and the closest point on the AABB.
	 *
	 * \param[in] the point.
	 *
	 * \return The squared distacne between \a p and the closest point to \a p on the AABB.
	 */
	float SquaredDistanceTo(const Point2f &p) const;
};

/*********************************************************************
 *  Collision Functions
 *********************************************************************/

/*!
 * \brief Computes the closest points between the two given lines 
 * 
 * \param[in] l1 the first line.
 * \param[in] l2 the second line.
 * \param[out] out1 the closest point on the first line.
 * \param[out] out2 the closest point on the second line. 
 *
 * \return The distance between \a out1 and \a out2.
 */
float ClosestPoints(const Line &l1, const Line &l2, Point2f &out1, Point2f &out2);

/*!
 * \brief Intersects a Ray with a Circle and finds the enter and exit time
 * 
 * \param[in] r the Ray.
 * \param[in] c the circle.
 * \param[out] t1 the time the Ray \a r enters the Circle \a c.
 * \param[out] t2 the time the Ray \a r exits the Circle \a c. 
 *
 * \return The number of intersections the Ray \a r has with the Circle \a c.  
 * If 0 is returned there where was no intersection. If 1 is returned then there 
 * was one intersection and \a t1 holds the time.  If 2 is returned then there were 
 * two intersections and the times are in \a t1 and \a t2 where \a t1 < \a t2.
 */
int Intersect(const Ray &r, const Circle &c, float &t1, float &t2);

/*!
 * \brief Intersects a Line with a Circle and finds the enter and exit time
 * 
 * \param[in] l the Line.
 * \param[in] c the circle.
 * \param[out] t1 the time the Line \a l enters the Circle \a c.
 * \param[out] t2 the time the Line \a l exits the Circle \a c. 
 *
 * \return The number of intersections the Line \a l has with the Circle \a c.  
 * If 0 is returned there where was no intersection. If 1 is returned then there 
 * was one intersection and \a t1 holds the time.  If 2 is returned then there were 
 * two intersections and the times are in \a t1 and \a t2 where \a t1 < \a t2.
 */
int Intersect(const Line &l, const Circle &c, float &t1, float &t2);

/*********************************************************************
 *  Helpful Functions
 *********************************************************************/

/*!
 * \brief Linear interpolates between two objects of type \a T.
 *
 * \tparam T the data type to linearly interpolate between.  The type must support the following operators:
 * -# T operator+(const &T) const
 * -# T operator-(const &T) const
 * -# T operator*(const &U) const
 * \tparam U the data type time takes.  Usually \c float or \c double.
 *
 * \param[in] t1 the starting point to interpolate from.
 * \param[in] t2 the ending point to interpolate to.
 * \param[in] t the time to interpolate to.
 *
 * \return The interpolated position between \a t1 and \a t2 such that \a t1 is returned when \a t = 0
 * and \a t2 is returned when \a t = 1.
 */
template<class T, class U>
T lerp(const T &t1, const T &t2, U t)
{
	return t1 + (t2 - t1) * t;
}

/*!
 * \brief Converts a Rect to an AABB.
 *
 * \tparam T the data type of the Rect.  If T is not \c float data may be truncated.
 *
 * \param[in] r The Rect to convert.
 * \see AABBToRect
 *
 * \return An AABB equivalent to \a r.
 */
template<class T>
AABB RectToAABB(const Rect_t<T> &r)
{
	AABB aabb;
	aabb.min.x = (float)(r.x);
	aabb.min.y = (float)(r.y - r.h);
	
	aabb.max.x = (float)(r.x + r.w);
	aabb.max.y = (float)(r.y);
	
	return aabb;
}

/*!
 * \brief Converts an AABB to a Rect.
 *
 * \tparam T the data type of the Rect.  If T is not \c float data may be truncated.
 *
 * \param[in] aabb The AABB to convert.
 * \see RectToAABB
 *
 * \return A Rect equivalent to \a aabb.
 */
template<class T>
Rect_t<T> AABBToRect(const AABB &aabb)
{
	Rect_t<T> r;
	r.w = (T)(aabb.max.x - aabb.min.x);
	r.h = (T)(aabb.max.y - aabb.min.y);
	
	r.x = (T)aabb.min.x;
	r.y = (T)aabb.max.y;
	return r;
}

/*!
 * \brief Expands (or contracts) the size of the given rectangle about the center.
 *
 * \tparam T the data type of the Rect.  If T is not \c float data may be truncated.
 *
 * \param[in] r the Rect to expand.
 * \param[in] dx the amount to expand in the x direction.
 * \param[in] dy the amount to expand in the y direction.
 *
 * \return The expanded Rect.
 */
template<class T>
Rect_t<T> ExpandRect(const Rect_t<T> &r, float dx, float dy)
{
	Rect_t<T> out;
	out.w = r.w + dx;
	out.h = r.h + dy;
	out.x = r.x - (dx / 2.0f);
	out.y = r.y + (dy / 2.0f);
	return out;
}

/*!
 * \brief Tests if the given point is inside the given rect
 *
 * \tparam T the data type of the Rect and the Point2
 *
 * \param[in] r the Rect to check against.
 * \param[in] p the point to test.
 *
 * \return True if the point \a p is inside Rect \a r.
 */
template<class T>
bool InsideRect(const Rect_t<T> &r, const Point2<T> &p)
{
	if( p.x >= r.x && 
		p.x <= (r.x + r.w) &&
		p.y <= r.y &&
		p.y >= (r.y - r.h) )
	{
		return true;
	}

	return false;
}

/*!
 * \brief Clamps a given value into a given range.
 *
 * \tparam T the data type of used.
 *
 * \param[in,out] val the value to clamp.  Will be set to be in the range [low,high] 
 * \param[in] low the low end of the range.
 * \param[in] high the hight end of the range.
 * \see ClampLow
 * \see ClampHigh
 */
template<class T>
void ClampInPlace(T &val, const T &low, const T &high)
{
	if(val < low)
		val = low;
	else if(val > high)
		val = high;
}

/*!
 * \brief Clamps a given value into a given range.
 *
 * \tparam T the data type of used.
 *
 * \param[in] val the value to clamp. 
 * \param[in] low the low end of the range.
 * \param[in] high the hight end of the range.
 * \see ClampLow
 * \see ClampHigh
 *
 * \return The result of clamping \a val into the range [low,high]
 */
template<class T>
T Clamp(const T &val, const T &low, const T &high)
{
	if(val < low)
		return low;
	else if(val > high)
		return high;
	return val;
}

/*!
 * \brief Clamps a given value into a given range.
 *
 * \tparam T the data type of used.
 *
 * \param[in,out] val the value to clamp.  Will be set to be in the range [low,infinity) 
 * \param[in] low the low end of the range.
 * \see Clamp
 * \see ClampHigh
 */
template<class T>
void ClampLow(T &val, const T &low)
{
	if(val < low)
		val = low;
}

/*!
 * \brief Clamps a given value into a given range.
 *
 * \tparam T the data type of used.
 *
 * \param[in,out] val the value to clamp.  Will be set to be in the range (infinity,high] 
 * \param[in] high the hight end of the range.
 * \see Clamp
 * \see ClampLow
 */
template<class T>
void ClampHigh(T &val, const T &high)
{
	if(val > high)
		val = high;
}

/*!
 * \brief Determines the sign of the given value.
 *
 * \tparam T the data type of used.
 *
 * \param[in] val the value to determine the sign of.
 * \see SignZero
 *
 * \return -1 if \a val < 0 or 1 if \a val >= 0.
 */
template<class T>
T Sign(const T &val)
{
	if(val < (T)0)
		return (T)-1;
	return (T)1;
}

/*!
 * \brief Determines the sign of the given value.
 *
 * \tparam T the data type of used.
 *
 * \param[in] val the value to determine the sign of.
 * \see Sign
 *
 * \return -1 if \a val < 0, 0 if \a val == 0 or 1 if \a val > 0.
 */
template<class T>
T SignZero(const T &val)
{
	if(val < (T)0)
		return (T)-1;
	else if(val > (T)0)
		return (T)1;
	return (T)0;
}

/*!
 * \brief Determines the minimum of two elements.
 *
 * \tparam T the data type of used.
 *
 * \param[in] v1 the first value.
 * \param[in] v2 the second value.
 *
 * \return \a v1 if \a v1 <= \a v2, or \a v2 if \a v1 > \a v2.
 */
template<class T>
T Minimum(const T &v1, const T &v2)
{
	if(v2 < v1)
		return v2;
	return v1;
}

/*!
 * \brief Determines the maximum of two elements.
 *
 * \tparam T the data type of used.
 *
 * \param[in] v1 the first value.
 * \param[in] v2 the second value.
 *
 * \return \a v1 if \a v1 >= \a v2, or \a v2 if \a v1 < \a v2.
 */
template<class T>
T Maximum(const T &v1, const T &v2)
{
	if(v2 > v1)
		return v2;
	return v1;
}

/*!
 * \brief Swaps the given two values.
 *
 * \tparam T the data type of the values to Swap.
 *
 * \param[in,out] v1 will be set to \a v2.
 * \param[in,out] v2 will be set to \a v1.
 */
template<class T>
void Swap(T &v1, T &v2)
{
	T tmp = v1;
	v1 = v2;
	v2 = tmp;
}

/*!
 * \brief Converts an angle in Radians to Degrees.
 *
 * \param[in] a the angle in Radians to convert.
 * \see DegToRad
 * 
 * \return The angle \a a in Degrees.
 */
inline float RadToDeg(float a)
{
	return a * RAD_TO_DEG;
}

/*!
 * \brief Converts an angle in Degrees to Radians.
 *
 * \param[in] a the angle in Degrees to convert.
 * \see RadToDeg
 * 
 * \return The angle \a a in Radians.
 */
inline float DegToRad(float a)
{
	return a * DEG_TO_RAD;
}

/*!
 * \brief Tests if two floating point values are within a given tolerance.
 *
 * \param[in] a the first value.
 * \param[in] b the second value.
 * \param[in] tol the tolerance. Defaults to 0.00001f.
 *
 * \return true if the given values are within the given tolerance from each other.
 */
inline bool Close(float a, float b, float tol = 0.00001f)
{
	return std::abs(b - a) <= tol * std::abs(a);
}

/*********************************************************************
 *  Misc. Functions
 *********************************************************************/

/*!
 * \brief Computes a random float in a given range.
 *
 * \param[in] min the minimum end of the range.
 * \param[in] max the maximum end of the range.
 *
 * \return A random float in the range [min,max]
 */
inline float RandomFloat(float min, float max)
{
	return (max - min)*((float)rand()/(float)RAND_MAX) + min;
}

/*!
 * \brief Computes a pretty good seed for a random number generator using the system time.
 *
 * \return A seed value.
 */
uint TimeSeed();

/*!
 * \brief Transforms the given polygon.
 *
 * \param[in,out] p the polygon to be transformed.
 * \param[in] m the transformation matrix.
 */
inline void Transform(Poly &p, const Matrixf &m)
{
	std::vector<Point2f>::iterator i;
	for(i = p.vertices.begin(); i != p.vertices.end(); i++)
	{
		*i = m.Transform(*i);
	}
}

/*!
 * \brief Transforms the given points.
 *
 * \param[in,out] p the points to be transformed.
 * \param[in] n the number of points.
 * \param[in] m the transformation matrix.
 */
inline void Transform(Point2f *p, uint n, const Matrixf &m)
{
	for(uint i = 0; i < n; i++)
	{
		p[i] = m.Transform(p[i]);
	}
}

/*!
 * \brief Converts a point in absolute Screen coordinates to normalized device coordinates.
 *
 * \param[in] p the point to convert.
 * \see NormToScreenCoords
 * 
 * \return The point \a p in normalized device coordinates.
 */
Point2f ScreenToNormCoords(const Point2f &p);

/*!
 * \brief Converts a point in normalized device coordinates to absolute Screen coordinates.
 *
 * \param[in] p the point to convert.
 * \see ScreenToNormCoords
 * 
 * \return The point \a p in absolute Screen coordinates.
 */
Point2f NormToScreenCoords(const Point2f &p);

/*!
 * \brief Converts an AABB in normalized device coordinates to absolute Screen coordinates.
 *
 * \param[in] a the AABB to convert.
 * 
 * \return The AABB \a a in absolute Screen coordinates.
 */
inline AABB ScreenToNormCoords(const AABB &a)
{
	AABB out;
	out.min = ScreenToNormCoords(a.min);
	out.max = ScreenToNormCoords(a.max);
	return out;
}

/*!
 * \brief Converts an AABB in normalized device coordinates to absolute Screen coordinates.
 *
 * \param[in] a the AABB to convert.
 * 
 * \return The AABB \a a in absolute Screen coordinates.
 */
inline AABB NormToScreenCoords(const AABB &a)
{
	AABB out;
	out.min = NormToScreenCoords(a.min);
	out.max = NormToScreenCoords(a.max);
	return out;
}

#endif
