/*!
 * \file Matrix.h
 * \brief Defines a 3x3 matrix template class.
 * \author Joe Forte
 * \date 3/27/2008
 *
 * Defines a 3x3 Matrix vector template along with the most common template instansiations.
 */

#ifndef MATRIX_H
#define MATRIX_H

#include <cmath>
#include "Math/Point.h"
#include "Math/Vector.h"

/*!
 * \class Matrix
 * \headerfile Matrix.h "Math/Matrix.h"
 * \brief A 3x3 matrix class suited for 2D graphics. 
 *
 * All Matrix methods and operations are assumed to be using a left-handed coordinate system unless otherwise noted.
 *
 * \tparam T the data type of the matrix data.  Usually \c float or \c double.
 */
template<class T>
class Matrix
{
public:

	/*!
	 * Data can be accessed either by name as _RC where R and C are in the range [1,3],
	 * or by array access as m[R][C] where R and C are in the range [0,2].
	 */
	union
	{
		struct
		{
			T _11, _12, _13;
			T _21, _22, _23;
			T _31, _32, _33;
		};
		T m[3][3];
	};	
	
	/*!
	 * \brief Sets all of the matrix fields to 0.
	 */
	void MakeZero()
	{
		for(int i = 0; i < 3; i++)
		{
			for(int j = 0; j < 3; j++)
			{
				m[i][j] = (T)0.0;
			}
		}
	}
	
	/*!
	 * \brief Sets the matrix to the identity matrix.
	 */
	void MakeIdentity()
	{
		MakeZero();
		_11 = _22 = _33 = (T)1.0;
	}
	
	/*!
	 * \brief Turns the matrix into a translation matrix.
	 *
	 * \param[in] x the x position to translate to.
	 * \param[in] y the y position to translate to.
	 */
	void MakeTranslation(T x, T y)
	{
		MakeIdentity();
		_13 = x;
		_23 = y;
	}
	
	/*!
	 * \brief Turns the matrix into a scale matrix.
	 *
	 * \param[in] x the amount to scale by in the x direction.
	 * \param[in] y the amount to scale by in the y direction.
	 */
	void MakeScale(T x, T y)
	{
		MakeIdentity();
		_11 = x;
		_22 = y;
	}
	
	/*!
	 * \brief Turns the matrix into a rotation matrix about the z-axis (suitable for 2D graphics).
	 *
	 * \param[in] angle the angle to rotate by in radians.
	 */
	void MakeRotation(T angle)
	{
		const T s = std::sin(angle);
		const T c = std::cos(angle);
		
		MakeIdentity();
		_11 = c; _12 = -s;
		_21 = s; _22 = c;
	}
	
	/*!
	 * \brief Computes the determinent of the matrix.
	 *
	 * \return The determinent.
	 */
	T Det() const
	{
		const T d1 = _11 * (_22*_33 - _23*_32);
		const T d2 = _12 * (_21*_33 - _23*_31);
		const T d3 = _13 * (_21*_32 - _22*_31);
		
		return d1 - d2 + d3;
	}
	
	/*!
	 * \brief Inverts the matrix if possible.
	 *
	 * \return True if the inverise exists. False is the inverse doesn't exits, in which case the matrix was not changed.
	 */
	bool Invert()
	{
		const T d = Det();
		
		if( d == (T)0.0 )
			return false;
	
		const T t = (T)1.0 / d;
		
		Matrix a;
		a._11 = (_33*_22 - _32*_23) * t;	
		a._12 = (_32*_13 - _33*_12) * t;	
		a._13 = (_23*_12 - _22*_13) * t;
		
		a._21 = (_31*_23 - _33*_21) * t;	
		a._22 = (_33*_11 - _31*_13) * t;	
		a._23 = (_21*_13 - _23*_11) * t;
		
		a._11 = (_32*_21 - _31*_22) * t;	
		a._12 = (_31*_12 - _32*_11) * t;	
		a._13 = (_22*_11 - _21*_12) * t;
		
		*this = a;
	}
	
	/*!
	 * \brief Transforms the given 2D point by the matrix taking the z commpont of the point to be 1.
	 *
	 * \param[in] p the point to transform.
	 *
	 * \return The transformed point.
	 */
	Point2<T> Transform(const Point2<T> &p) const
	{
		Point2<T> out;
		out.x = p.x*_11 + p.y*_12 + _13;
		out.y = p.x*_21 + p.y*_22 + _23;
		return out;
	}

	/*!
	 * \brief Transforms the given 2D vector by the matrix taking the z commpont of the point to be 0.
	 *
	 * \param[in] v the vector to transform.
	 *
	 * \return The transformed vector.
	 */
	Vector2<T> Transform(const Vector2<T> &v) const
	{
		Vector2<T> out;
		out.x = v.x*_11 + v.y*_12;
		out.y = v.x*_21 + v.y*_22;
		return out;
	}
	
	/*!
	 * \brief Overloaded matrix addition operator.
	 *
	 * \param[in] b the matrix to add to.
	 *
	 * \return The result of adding \a b to \e this matrix.
	 */
	Matrix operator+(const Matrix &b) const
	{
		Matrix c;
		for(int i = 0; i < 3; i++)
		{
			for(int j = 0; j < 3; j++)
			{
				c.m[i][j] = m[i][j] + b.m[i][j];
			}
		}
		return c;
	}
	
	/*!
	 * \brief Overloaded matrix multiplicatioin operator.
	 *
	 * \param[in] b the matrix to multiply by.
	 *
	 * \return The matrix result of multiply \e this matrix by \a b.
	 */
	Matrix operator*(const Matrix &b) const
	{
		Matrix c;
		for(int i = 0; i < 3; i++)
		{
			for(int j = 0; j < 3; j++)
			{
				T sum = (T)0.0;
				for(int k = 0; k < 3; k++)
				{
					sum += m[i][k] * b.m[k][j];
				}
				c.m[i][j] = sum;
			}
		}
		return c;
	}
	
	/*!
	 * \brief Overloaded matrix-constant multiplication operator.
	 *
	 * \param[in] k the constaant to multiply by.
	 *
	 * \return The matrix result of multiplying \e this matrix by \a k.
	 */
	Matrix operator*(T k) const
	{
		Matrix c;
		for(int i = 0; i < 3; i++)
		{
			for(int j = 0; j < 3; j++)
			{
				c.m[i][j] = k * m[i][j];
			}
		}
		return c;
	}
	
	/*!
	 * \brief Overloaded constant-matrix multiplication operator.
	 *
	 * \param[in] k the constant to multiply by.
	 * \param[in] a the matrix to multiply by.
	 *
	 * \return The matrix result of multiplying \a k by \a a.
	 */
	friend Matrix operator*(T k, const Matrix &a)
	{
		return a*k;
	}
	
	/*!
	 * \brief Converts the matrix to a form suitible for OpenGL
	 *
	 * \param[in] m a 16 element array in the form OpenGL expects.
	 */
	void GetOpenGLMatrix(T m[16]) const
	{
		// OpenGL expects matrices in column-major order
		m[0] = _11;  m[4] = _12;  m[8]  = (T)0; m[12] = _13;
		m[1] = _21;  m[5] = _22;  m[9]  = (T)0; m[13] = _23;
		m[2] = _31;  m[6] = _32;  m[10] = (T)0; m[14] = _33;
		m[3] = (T)0; m[7] = (T)0; m[11] = (T)0; m[15] = (T)1;
	}
};

/*!
 * \typedef Matrixf
 * \brief A Template instansiation of the 3x3 Matrix template using the \c float data type.
 */
typedef Matrix<float> Matrixf;

/*!
 * \typedef Matrixd
 * \brief A Template instansiation of the 3x3 Matrix template using the \c double data type.
 */
typedef Matrix<double> Matrixd;

#endif
