/*!
 * \file Point.h
 * \brief Defines a 2D Point template class.
 * \author Joe Forte
 * \date 3/27/2008
 *
 * Defines a 2D Point template along with the most common template instansiations.
 */

#ifndef POINT_H
#define POINT_H

#include "Common/BaseTypes.h"
#include "Math/Vector.h"

/*! 
 * \class Point2
 * \headerfile Point.h "Math/Point.h"
 * \brief A 2D Point template.
 *
 * \tparam T the type that \e x and \e y take.  Usually \c float, \c double, \c int or \c byte. 
 *
 * A 2D Point template that provides useful methods for manipulating 2D Points.
 */
template<class T>
class Point2
{
public:
	/*!
	 * \brief The (x,y) point data.
	 */
	T x, y;
	
	/*!
	 * \brief Defualt constructer.  Constructs the origin (0,0). 
	 */
	Point2() : x( (T)0.0 ), y( (T)0.0 ) {}
	
	/*!
	 * \brief Constructs a spacific point.
	 *
	 * \param[in] _x the x position.
	 * \param[in] _y the y position.
	 */
	Point2(T _x, T _y) : x(_x), y(_y) {}
	
	/*!
	 * \brief Constructer to convert a 2D vector to a point.
	 *
	 * \param[in] v the vector to convert.
	 */
	Point2(const Vector2<T> &v) : x(v.x), y(v.y) {}
	
	/*!
	 * \brief Sets the (x,y) data of th point.
	 *
	 * \param[in] _x the value to set \e x to.
	 * \param[in] _y the value to set \e y to.
	 */
	void Set(T _x, T _y) { x = _x; y = _y; }
	
	/*!
	 * \brief Overloaded point addition operator.
	 *
	 * \param[in] p the point to add to.
	 *
	 * \return The Point result of adding \e this point and \a p.
	 */
	Point2 operator+(const Point2 &p) const
	{
		return Point2(x + p.x, y + p.y);
	}
	
	/*!
	 * \brief Overloaded point-vector addition operator.
	 *
	 * \param[in] v the vector to add to.
	 *
	 * \return The Point result of adding \e this point to the vector \a v.
	 */
	Point2 operator+(const Vector2<T> &v) const
	{
		return Point2(x + v.x, y + v.y);
	}
	
	/*!
	 * \brief Overloaded sum assignment operator.  Adds the given Point to \e this Point.
	 *
	 * \param[in] p the point to add.
	 *
	 * \return A referance to \e this Point.
	 */
	Point2 &operator+=(const Point2 &p)
	{
		x += p.x;
		y += p.y;
		return *this;
	}
	
	/*!
	 * \brief Overloaded Point subtraction operator.
	 *
	 * \param[in] p the point to subtract from.
	 *
	 * \return The Vector result of subtracting \a p from \e this point.
	 */
	Vector2<T> operator-(const Point2 &p) const
	{
		return Vector2<T>(x - p.x, y - p.y);
	}

	/*!
	 * \brief Overloaded Point negation operator.
	 *
	 * \return The negation of \e this point.
	 */
	Point2<T> operator-() const
	{
		return Point2<T>(-x, -y);
	}
	
	/*!
	 * \brief Overloaded subtraction assignment operator.  Subtracts the given Point from \e this Point.
	 *
	 * \param[in] p the point to subtract.
	 *
	 * \return A referance to \e this Point.
	 */
	Point2 &operator-=(const Point2 &p)
	{
		x -= p.x;
		y -= p.y;
		return *this;
	}

	/*!
	 * \brief Overloaded array access operator.
	 *
	 * \param[in] i must be either 0 for \e x or 1 for \e y
	 *
	 * \return Either a referance to \e x if \a i = 0 or a referance to \e y if \a i = 1
	 */
	T& operator[](int i)
	{
		return ((T*)this)[i];
	}

	/*!
	 * \brief Overloaded array access operator.
	 *
	 * \param[in] i must be either 0 for \e x or 1 for \e y
	 *
	 * \return Either \e x if \a i = 0 or \e y if \a i = 1
	 */
	T operator[](int i) const
	{
		return ((T*)this)[i];
	}
	
	/*!
	 * \brief Overloaded constant multiplication operator.
	 *
	 * \param[in] c the constant to multiply by.
	 *
	 * \return The point result of multiply \e this point by \a c.
	 */
	Point2 operator*(T c) const
	{
		return Point2(c*x, c*y);
	}
	
	/*!
	 * \brief Overloaded constant-point multiplication operator.
	 *
	 * \param[in] c the constant.
	 * \param[in] p the point.
	 *
	 * \return The Point result \a c*\a p. 
	 */
	friend Point2 operator*(float c, const Point2 &p)
	{
		return Point2(c*p.x, c*p.y);
	}
};

/*!
 * \typedef Point2f
 * \brief A Template instansiation of the Point2 template using the \c float data type.
 */
typedef Point2<float> Point2f;

/*!
 * \typedef Point2d
 * \brief A Template instansiation of the Point2 template using the \c double data type.
 */
typedef Point2<double> Point2d;

/*!
 * \typedef Point2i
 * \brief A Template instansiation of the Point2 template using the \c int data type.
 */
typedef Point2<int> Point2i;

/*!
 * \typedef Point2b
 * \brief A Template instansiation of the Point2 template using the \c byte data type.
 */
typedef Point2<byte> Point2b;

/*!
 * \brief Converts a Point2 object of one type to another
 *
 * \tparam Tin the data type used by the point thats being converted.  Tin should overload the following:
 * -# operator Tout (const Tin &)
 * \tparam Tout the data type to convert to.
 *
 * \param[in] in the point to convert.
 *
 * \return Point converted to use the type Tout.
 */
template<class Tin, class Tout>
Point2<Tout> ConvertPoint2(const Point2<Tin> &in)
{
	return Point2<Tout>( (Tout)in.x, (Tout)in.y );
}

/*!
 * \brief Converts a Point2 object of one type to another
 *
 * \tparam Tin the data type used by the point thats being converted.  Tin should overload the following:
 * -# operator Tout (const Tin &)
 * \tparam Tout the data type to convert to.
 *
 * \param[in] in the point to convert.
 * \param[out] out point where converted point is stored.
 */
template<class Tin, class Tout>
void ConvertPoint2(const Point2<Tin> &in, Point2<Tout> &out)
{
	out.x = (Tout)in.x;
	out.y = (Tout)in.y;
}

#endif
