/*!
 * \file Vector.h
 * \brief Defines a 2D vector class.
 * \author Joe Forte
 * \date 3/27/2008
 *
 * Defines a 2D vector template along with the most common template instansiations.
 */

#ifndef VECTOR_H_
#define VECTOR_H_

#include <cmath>

/*! 
 * \class Vector2
 * \headerfile Vector.h "Math/Vector.h"
 * \brief A 2D vector template.
 *
 * \tparam T the type that \a x and \a y take.  Usually \c float or \c double. 
 *
 * A 2D vector template that provides useful methods for manipulating 2D vectors.
 */
template<class T>
class Vector2
{
public:
	/*!
	 * \brief The (x,y) vector data.
	 */
	T x, y;
	
	/*!
	 * \brief Default constructor. Constructs the zero vector.
	 */
	Vector2() : x((T)0), y((T)0) {}
	
	/*!
	 * \brief Constructs a spacific vector.
	 *
	 * \param[in] _x the value \a x is set to.
	 * \param[in] _y the value \a y is set to.
	 */
	Vector2(T _x, T _y) : x(_x), y(_y) {}

	/*!
	 * \brief Sets the \a x and \a y values of the vector.
	 *
	 * \param[in] _x the value \a x is set to.
	 * \param[in] _y the value \a y is set to.
	 */
	void Set(T _x, T _y) { x = _x; y = _y; }
	
	/*!
	 * \brief Computes the dot product.
	 *
	 * \param[in] v the vector to compute the dot product with.
	 *
	 * \return The dot product between \e this vector and the input vector \a v. \f$v_1\cdot v_2 = x_1*x_2 + y_1*y_2\f$ 
	 */
	T Dot(const Vector2 &v) const
	{
		return x*v.x + y*v.y;
	}
	
	/*!
	 * \brief Computes the length of \e this vector.
	 *
	 * \return the length of this vector \f$\sqrt{x^2 + y^2}\f$
	 */
	T Length() const
	{
		return std::sqrt( x*x + y*y );
	}
	
	/*!
	 * \brief Computes the squared length of \e this vector.
	 *
	 * \return The squared length of this vector \f$x^2 + y^2\f$
	 */
	T SquaredLength() const
	{
		return x*x + y*y;
	}

	/*!
	 * \brief Computes the angle in radians between \e this vector and \a v.
	 *
	 * \param[in] v the vector used to compute and angle between.
	 *
	 * \return The angle in radians between \e this vector and \a v.
	 */
	T Angle(const Vector2 &v) const
	{
		T denom = Length() * v.Length();
		if( denom == (T)0 )
			return (T)0;

		return std::acos( Dot(v) / denom );
	}
	
	/*!
	 * \brief Normalizes \e this the vector.
	 *
	 * Changes \a x and \a y such that the length of \e this vector is 1 and the direction is preserved. 
	 */
	void Normalize()
	{
		T len = Length();
		x = x / len;
		y = y / len;
	}

	/*!
	 * \brief Reflects \e this vector about the vector \a n.
	 *
	 * \param[in] n the vector that to reflect against.
	 *
	 * \return The reflected vector. \f$v_2 = v_1 - 2n(v_1\cdot n)\f$
	 */
	const Vector2 Reflect(const Vector2 &n) const
	{
		T d = ((T)2) * Dot(n);
		return (*this) - (n * d);
	}

	/*!
	 * \brief Refracts \e this vector about the vector \a n with index of refraction \a index.
	 *
	 * \param[in] n the vector that to refract against.
	 * \param[in] index the index of refraction.
	 *
	 * \return The refracted vector.
	 */
	const Vector2 Refract(const Vector2 &n, T index = 1.0f) const
	{
		Vector2 np = n * Dot(-n);
		Vector2 s = index * ((*this) + np);

		return -n * std::sqrt( (T)1 - s.SquaredLength() ); 
	}

	/*!
	 * \brief Projects \e this vector onto the vector \a L.
	 *
	 * \param[in] L the vector to project onto.
	 *
	 * \return The projected vector \f$L(v\cdot L)\f$
	 */
	const Vector2 Project(const Vector2 &L) const
	{
		return L * Dot(L);
	}

	/*!
	 * \brief Computes a vector perpendicular to \e this vector.
	 *
	 * \return The perpendicular vector.
	 */
	const Vector2 Perp() const
	{
		return Vector2(y, -x);
	}
	
	/*!
	 * \brief Overloaded vector addition operator.
	 *
	 * \param[in] v the vector to add.
	 *
	 * \return The vector result of adding \e this vector to \a v.
	 */
	const Vector2 operator+(const Vector2 &v) const
	{
		return Vector2(x + v.x, y + v.y);
	}
	
	/*!
	 * \brief Overloaded vector subtraction operator.
	 *
	 * \param[in] v the vector to subtract.
	 *
	 * \return The vector result of subtracting \a v font \e this vector.
	 */
	const Vector2 operator-(const Vector2 &v) const
	{
		return Vector2(x - v.x, y - v.y);
	}

	/*!
	 * \brief Overloaded vector negation operator.
	 *
	 * \return The negation of \e this vector.
	 */
	Vector2<T> operator-() const
	{
		return Vector2<T>(-x, -y);
	}
	
	/*!
	 * \brief Overloaded constant multiplication operator.
	 *
	 * \param[in] c the constant to multiply by.
	 *
	 * \return The vector result of multiply \e this vector by \a c.
	 */
	const Vector2 operator*(T c) const
	{
		return Vector2(x*c, y*c);
	}
	
	/*!
	 * \brief Overloaded constant division operator.
	 *
	 * \param[in] c the constant to divide by.
	 *
	 * \return The vector result of divideing \e this vector by \a c.
	 */
	const Vector2 operator/(T c) const
	{
		return Vector2(x/c, y/c);
	}
	
	/*!
	 * \brief Overloaded array access operator.
	 *
	 * \param[in] i must be either 0 for \e x or 1 for \e y
	 *
	 * \return Either a referance to \e x if \a i = 0 or a referance to \e y if \a i = 1
	 */
	T& operator[](int i)
	{
		return ((T*)this)[i];
	}

	/*!
	 * \brief Overloaded array access operator.
	 *
	 * \param[in] i must be either 0 for \e x or 1 for \e y
	 *
	 * \return Either \e x if \a i = 0 or \e y if \a i = 1
	 */
	T operator[](int i) const
	{
		return ((T*)this)[i];
	}

	/*!
	 * \brief Overloaded constant-vector multiplication operator.
	 *
	 * \param[in] c the constant.
	 * \param[in] v the vector.
	 *
	 * \return The vector result \a c*\a v. 
	 */
	friend const Vector2 operator*(T c, const Vector2 &v)
	{
		return Vector2(v.x*c, v.y*c);
	}
};

/*!
 * \brief Normalizes the given vector.
 *
 * \tparam T the data type of the vector.  Usually \c float or \c double
 * \param[in] v the vector to normalize.
 *
 * \return The normalized vector \a v. 
 */
template<class T>
Vector2<T> Normalize(const Vector2<T> &v)
{
	T len = v.Length();
	return Vector2<T>(v.x / len, v.y / len);
}

/*!
 * \brief Computes the dot product between two vectors.
 *
 * \tparam T the data type of the vector.  Usually \c float or \c double
 * \param[in] v1 the first vector.
 * \param[in] v2 the second vector.
 *
 * \return The dot product of \a v1 and \a v2 \f$v_1\cdot v_2\f$ 
 */
template<class T>
T Dot(const Vector2<T> &v1, const Vector2<T> &v2)
{
	return v1.x*v2.x + v1.y*v2.y;
}

/*!
 * \brief Computes the length of the given vector.
 *
 * \tparam T the data type of the vector.  Usually \c float or \c double
 * \param[in] v the vector to compute the length of.
 *
 * \return The length of the vector \a v \f$sqrt{x^2 + y^2}\f$ 
 */
template<class T>
T Length(const Vector2<T> &v)
{
	return std::sqrt(v.x*v.x + v.y*v.y);
}

/*!
 * \typedef Vector2f
 * \brief A Template instansiation of the Vector2 template using the \c float data type.
 */
typedef Vector2<float> Vector2f;

/*!
 * \typedef Vector2d
 * \brief A Template instansiation of the Vector2 template using the \c double data type.
 */
typedef Vector2<double> Vector2d;

#endif
