/*!
 * \file ResourceManager.h
 * \brief Defines classes and types for loading/creating managed resources.
 * \author Joe Forte
 * \date 5/19/2008
 *
 * Defines the Resource class that all managed resources must inherit from.  The
 * ResourceMgr class responsible for managing resources is also defined.
 */

#ifndef RESOURCEMANAGER_H
#define RESOURCEMANAGER_H

#include <assert.h>
#include <iostream>
#include <string>
#include <vector>
#include "Common/BaseTypes.h"
#include "Common/Path.h"

template<class resType, class initType> class ResourceMgr;
template<class resType, class initType> class ResourceRef;

/*!
 * \typedef ResourceID
 * \brief The type used to uniquely identify managed resources of a certain type.
 */
typedef int ResourceID;

/*!
 * \brief NULL ResourceID.  Serves the same purpose as NULL but is used with ResourceIDs.
 */
const ResourceID NULL_RESOURCEID = -1;

/*!
 * \brief Can be used as init data to force the Resource manager to never share a resource.
 */
class ForceNoResourceSharing
{
public:
	bool operator==(const ForceNoResourceSharing &d) const { return false; }
};

/*
 * Used internally for resources that do not need extra initialzation data.
 */
class VoidResourceInitData
{
public:
	bool operator==(const VoidResourceInitData &d) const { return true; }
};

/*!
 * \class Resource
 * \headerfile ResourceManager.h "Memory/ResourceManager.h"
 * \brief The base type for all managed resources.
 *
 * \tparam initType The data type of the object passed into the Load function.  initType must
 * have a default constructor and have the following operators overloaded:
 * -# bool operator==(const initType&) const
 */
template<class initType = VoidResourceInitData>
class Resource
{
public:
	Resource() : refCount(0) {}
	virtual ~Resource() {}

	/*!
	 * \brief Gets the file name for the resource.
	 *
	 * \returns The file name for \a this resource.
	 */
	const std::string &GetFileName() const { return fileName; }

	/*!
	 * \brief Gets the ResourceID for the resource.
	 *
	 * \returns The ResoueceID for \a this resource or NULL_RESOURCEID if the resource is not valid.
	 */
	ResourceID GetID() const { return id; }

	/*!
	 * \brief Loads the specified resource.
	 *
	 * \param[in] filePath the full path to the resource including the file name.
	 * \param[in] fileName the name of the resource file.
	 * \param[in] initData a pointer to the initialization data, can be NULL.
	 *
	 * \see Image
	 * \see Font
	 *
	 * \return True if the resource was loaded succesfully, false otherwise.
	 *
	 * This function must be implemented by all resources but should never be called outside of the
	 * ResourceMgr class.  The helper macro pairs DECLARE_RESOURCE_LOAD_FUNCTION / IMPLEMENT_RESOURCE_LOAD_FUNCTION
	 * and DECLARE_COMPLEX_RESOURCE_LOAD_FUNCTION / IMPLEMENT_COMPLEX_RESOURCE_LOAD_FUNCTION can be used
	 * in child classes.
	 */
	virtual bool Load(const char *filePath, const char *fileName, const initType *initData) = 0;

	/*!
	 * \brief Destroys the resource.
	 *
	 * Deallocates all data allocated when the resource was loaded.  Must be implemented by all classes
	 * that inherit from the Resource class but should never be called outside of the ResourceMgr class.
	 */
	virtual void Destroy() = 0;

	/*!
	 * \brief Estimates the number of bytes used by the loaded resource.
	 */
	virtual uint GetDataUsageInBytes() const { return 0; }

	/*
	 * For use by the ResourceMgr and ResourceRef classes only.
	 */
	uint GetRefCount() const { return refCount; }
	void IncRef() { refCount++; }
	void DecRef() { assert(refCount > 0); refCount--; }
	const initType &GetInitData() const { return initData; }
	void Set(const std::string &filename, ResourceID newID, const initType &loadData)
	{
		fileName = filename;
		id = newID;
		initData = loadData;
	}

private:
	std::string fileName;
	ResourceID id;

	initType initData;

	uint refCount;
};

/*!
 * \class ResourceRef
 * \headerfile ResourceManager.h "Memory/ResourceManager.h"
 * \brief The class used to access resource data.
 *
 * \tparam resType The type of the Resource being managed.  Must inherit from the Resource class.
 * \tparam initType The data type of the object passed into the Load function.  initType must
 * have a default constructor and have the following operators overloaded:
 * -# bool operator==(const initType&) const
 *
 * A ResourceRef is returned by the Resource Manager when a resource is loaded.  All operation on resource data should go throught
 * this class. This class should be treated much like a pointer.  Never keep a pointer to a ResourceRef, if you need to share a
 * resource simply copy the ResourceRef.  When passing a ResourceRef to a function, always pass by value.
 */
template<class resType, class initType>
class ResourceRef
{
public:

	/*!
	 * \brief Default constructor.  Constructs an invalid referance.
	 */
	ResourceRef() : id(NULL_RESOURCEID), mgr(NULL) {}
	ResourceRef(ResourceID i, ResourceMgr<resType, initType> *m) : id(i), mgr(m) { IncRefCount(); }

	/*!
	 * \brief Copy constructor.
	 */
	ResourceRef(const ResourceRef &rf) { id = rf.id; mgr = rf.mgr; IncRefCount(); }
	~ResourceRef() { DecRefCount(); }

	/*!
	 * \brief Determines if the Resource is Valid and can be used.
	 *
	 * \return True if the resource exists and can be used, false if the resource is invalid should not be used.
	 */
	bool IsValid() const { return Get() != NULL; }

	/*!
	 * \brief Overloaded boolean negation operator.
	 *
	 * \see IsValid
	 *
	 * \return True if the resource is invalid should not be used, false if the resource exists and can be used.
	 */
	bool operator!() const { return Get() == NULL; }

	/*!
	 * \brief Overloaded bool cast operator.
	 *
	 * \see IsValid
	 *
	 * \return True if the resource exists and can be used, false if the resource is invalid should not be used.
	 */
	operator bool() const { return Get() != NULL; }

	/*!
	 * \brief Overloaded equality operator.
	 *
	 * \param[in] ref the ResourceRef to compare against.
	 *
	 * \return True if \e this ResourceRef and \e ref are equal, false otherwise.
	 */
	bool operator==(const ResourceRef &ref) const { return id == ref.id && mgr == ref.mgr; }

	/*!
	 * \brief Overloaded equality operator.
	 *
	 * \param[in] ref the ResourceRef to compare against.
	 *
	 * \return True if \e this ResourceRef and \e ref are not equal, false otherwise.
	 */
	bool operator!=(const ResourceRef &ref) const { return id != ref.id || mgr != ref.mgr; }

	/*!
	 * \brief Gets the ResourceID.
	 *
	 * \return The ResourceID for \e this ResourceRef, or NULL_RESOURCE if \e this ResourceRef is invalid.
	 */
	ResourceID GetID() const { return id; }

	/*!
	 * \brief Overloaded pointer dereferance operator.
	 *
	 * \return A pointer to the Resource data object.
	 */
	resType *operator->() { return Get(); }

	/*!
	 * \brief Constant overloaded pointer dereferance operator.
	 *
	 * \return A const pointer to the Resource data object.
	 */
	const resType *operator->() const { return Get(); }

	/*!
	 * \brief Invalidates the ResourceRef setting it to NULL_REF in the process.  Equivalent to setting the ResourceRef to NULL_REF.
	 */
	void Invalidate()
	{
		DecRefCount();
		id  = NULL_RESOURCEID;
		mgr = NULL;
	}

	/*!
	 * \brief Overloaded assignment operator.
	 *
	 * \param[in] rhs the ResourceRef to set \e this ResourceRef to.
	 *
	 * \return A referance to \e this ResourceRef.
	 */
	ResourceRef &operator=(const ResourceRef &rhs)
	{
		if(this == &rhs)
			return *this;

		DecRefCount();
		id  = rhs.id;
		mgr = rhs.mgr;
		IncRefCount();

		return *this;
	}

private:

	resType *Get() const;

	void DecRefCount()
	{
		resType *res = Get();

		if(res)
		{
			assert(mgr);

			res->DecRef();
			if(res->GetRefCount() == 0)
			{
				mgr->DestroyResource(id);
				id  = NULL_RESOURCEID;
				mgr = NULL;
			}
		}
	}

	void IncRefCount()
	{
		resType *res = Get();

		if(res)
			res->IncRef();
	}

	ResourceID id;

	ResourceMgr<resType, initType> *mgr;
};

/*!
 * \class ResourceMgr
 * \headerfile ResourceManager.h "Memory/ResourceManager.h"
 * \brief Loads and manages Resource data of a spacific type.
 *
 * \tparam resType The type of the Resource being managed.  Must inherit from the Resource class.
 * \tparam initType The data type of the object passed into the Load function.  initType must
 * have a default constructor and have the following operators overloaded:
 * -# bool operator==(const initType&) const
 */
template <class resType, class initType>
class ResourceMgr
{
public:
	typedef resType resource_type;
	typedef initType initialization_type;
	typedef ResourceRef<resType, initType> ref_type;

	/*!
	 * \brief Constructor.
	 */
	ResourceMgr();
	
	/*!
	 * \brief Deconstructor.
	 */
	~ResourceMgr();
	
	/*!
	 * \brief Sets the default resource to use when a resource is loaded that doesn't exist.
	 *
	 * \param[in] defaultResourcePath the path to a default resource that must exist or NULL if no default resource is needed.
	 * \param[in] initData the initialization data for the default resource. Defaults to NULL.
	 */
	void SetDefaultResource(const char *defaultResourcePath, const initType *initData = NULL);

	/*!
	 * \brief Loads a resource.
	 *
	 * \param[in] fileName name of the resource file to load.
	 * \param[in] initData pointer to the initialization data.  Defaults to NULL.
	 * \param[in] useResourcePath set to true to use the currently set resource path.  Defaults to true.
	 *
	 * \return A Resource Referance to the loaded data or an invalid if the file couldn't be loaded.
	 */
	ref_type LoadResource(const char *fileName, const initType *initData = NULL, bool useResourcePath = true);

	/*!
	 * \brief Destroys all resouces.  All resource managed by \e this class will become invalid.
	 */
	void DestoryAllResources();

	/*!
	 * \brief Estimates the total number of bytes used by all managed resources.
	 *
	 * \return The estimated number of bytes used.
	 */
	uint GetDataUsageInBytes() { return bytesUsed; }

	/*!
	 * \brief Gets the number of resources being managed.
	 *
	 * \return The number of managed resources.
	 */
	uint GetNumberOfResourcesLoaded() { return resourcesLoaded; }

	/*!
	 * \brief Sets a path to be concatenated onto all file names when loading resources.
	 *
	 * \param[in] p the path.
	 */
	void SetResourcePath(const char *p) { path = p; }

	/*!
	 * \brief Gets the resource data.
	 *
	 * \return The vector of resources.
	 */
	std::vector<resType*> GetResources() const { return resources; }

	/*
	 * Currently unused and not needed.
	 */
	void GarbageCollect();

private:
	void DestroyResource(ResourceID id);
	resType *GetResource(ResourceID id);

	std::vector<resType*> resources;
	std::vector<ResourceID> idStack;

	ResourceID lastID;

	std::string path;

	uint bytesUsed;
	uint resourcesLoaded;
	
	ref_type defaultResource;

	friend class ResourceRef<resType, initType>;
};

/*!
 * \def DECLARE_RESOURCE_TYPE
 * \brief Defines two types for the given Resourse class.
 *
 * \param[in] className the name of the resource class.
 *
 * For instance, given a resource class Mesh the following types are defined
 * -# MeshRef (a ResourceRef type used for accessing the Mesh resource)
 * -# MeshMgr (the Resource Manager class for managing mesh resources)
 */
#define DECLARE_RESOURCE_TYPE(className) \
	typedef ResourceRef<className, VoidResourceInitData> className##Ref; \
	typedef ResourceMgr<className, VoidResourceInitData> className##Mgr;

/*!
 * \def DECLARE_COMPLEX_RESOURCE_TYPE
 * \brief Defines two types for the given Resouese class.
 *
 * \param[in] className the name of the resource class.
 * \param[in] loadDataType the data type used when loading a resource.
 *
 * For instance, given a resource class Mesh the following types are defined
 * -# MeshRef (a ResourceRef type used for accessing the Mesh resource)
 * -# MeshMgr (the Resource Manager class for managing mesh resources)
 */
#define DECLARE_COMPLEX_RESOURCE_TYPE(className, loadDataType) \
	typedef ResourceRef<className, loadDataType> className##Ref; \
	typedef ResourceMgr<className, loadDataType> className##Mgr;

/*!
 * \def DECLARE_RESOURCE_LOAD_FUNCTION
 * \brief Helpful macro for declaring the Resource Load function in classes that derive from the Resource class.
 *
 * \see Image
 */
#define DECLARE_RESOURCE_LOAD_FUNCTION() \
	virtual bool Load(const char *filePath, const char *fileName, const VoidResourceInitData *void_data_do_not_use_);

/*!
 * \def DECLARE_COMPLEX_RESOURCE_LOAD_FUNCTION
 * \brief Helpful macro for declaring the Resource Load function in classes that derive from the Resource class that
 * uses user defined initialization data.
 *
 * \param[in] loadDataType the data type of the inialization data.
 * \param[in] initDataName the name of the function paramater for the pointer to the initialization data.
 *
 * \see Font
 */
#define DECLARE_COMPLEX_RESOURCE_LOAD_FUNCTION(loadDataType, initDataName) \
	virtual bool Load(const char *filePath, const char *fileName, const loadDataType *initDataName);

/*!
 * \def IMPLEMENT_RESOURCE_LOAD_FUNCTION
 * \brief Helpful macro for implemeting the Resource Load function.  Should be used in conjuction with the
 * DECLARE_RESOURCE_LOAD_FUNCTION macro.
 *
 * \param[in] className the name of the resource class.
 *
 * \see Image
 * \see DECLARE_RESOURCE_LOAD_FUNCTION
 */
#define IMPLEMENT_RESOURCE_LOAD_FUNCTION(className) \
	bool className::Load(const char *filePath, const char *fileName, const VoidResourceInitData *_void_data_do_not_use_)

/*!
 * \def IMPLEMENT_COMPLEX_RESOURCE_LOAD_FUNCTION
 * \brief Helpful macro for implemeting the Resource Load function.  Should be used in conjuction with the
 * DECLARE_COMPLEX_RESOURCE_LOAD_FUNCTION macro.
 *
 * \param[in] className the name of the resource class.
 * \param[in] loadDataType the data type of the inialization data.
 * \param[in] initDataName the name of the function paramater for the pointer to the initialization data.
 *
 * \see Image
 * \see DECLARE_COMPLEX_RESOURCE_LOAD_FUNCTION
 */
#define IMPLEMENT_COMPLEX_RESOURCE_LOAD_FUNCTION(className, loadDataType, initDataName) \
	bool className::Load(const char *filePath, const char *fileName, const loadDataType *initDataName)

//--------------------------------------------------------------------
template<class resType, class initType>
ResourceMgr<resType, initType>::ResourceMgr() : lastID(NULL_RESOURCEID), bytesUsed(0), resourcesLoaded(0) {}

//--------------------------------------------------------------------
template<class resType, class initType>
ResourceMgr<resType, initType>::~ResourceMgr()
{
	DestoryAllResources();
}

//--------------------------------------------------------------------
template<class resType, class initType>
void ResourceMgr<resType, initType>::SetDefaultResource(const char *defaultResourcePath, const initType *initData)
{
	if(defaultResourcePath)
	{
		defaultResource = LoadResource(defaultResourcePath, initData);
	}
}

//--------------------------------------------------------------------
template<class resType, class initType>
ResourceRef<resType, initType> ResourceMgr<resType, initType>::LoadResource(const char *fileName, const initType *initData, bool useResourcePath)
{
	initType loadData;
	if(initData)
		loadData = *initData;
	else
		loadData = initType();

	size_t index;
	size_t len = resources.size();
	for(index = 0; index < len; index++)
	{
		if( resources[index] != NULL &&
		    resources[index]->GetFileName() == fileName &&
			resources[index]->GetInitData() == loadData )
		{
			return ResourceRef<resType, initType>((ResourceID)index, this);
		}
	}

	// The resource has not already been loaded
	ResourceID newID;
	if( !idStack.empty() )
	{
		newID = idStack.back();
		idStack.pop_back();
		resources[newID] = new resType;
	}
	else
	{
		newID = (int)resources.size();
		resources.push_back( new resType );
	}

	resources[newID]->Set(fileName, newID, loadData);
	
	std::string filePath;
	if(useResourcePath)
		filePath = ConcatPath(path, std::string(fileName));
	else
		filePath = fileName;
	
	if( !resources[newID]->Load(filePath.c_str(), fileName, initData) )
	{
		delete resources[newID];
		resources[newID] = NULL;
		idStack.push_back(newID);
		
		return defaultResource;
	}

	bytesUsed += resources[newID]->GetDataUsageInBytes();
	resourcesLoaded++;

	return ResourceRef<resType, initType>(newID, this);
}

//--------------------------------------------------------------------
template<class resType, class initType>
void ResourceMgr<resType, initType>::GarbageCollect()
{
	size_t len = resources.size();
	for(size_t i = 0; i < len; i++)
	{
		if( resources[i] && resources[i]->GetRefCount() == 0 )
		{
			DestroyResource( resources[i]->GetID() );
		}
	}
}

//--------------------------------------------------------------------
template<class resType, class initType>
void ResourceMgr<resType, initType>::DestroyResource(ResourceID id)
{
	if( id == NULL_RESOURCEID		||
	    id <= 0						||
	    id >= (int)resources.size() ||
	    resources[id] == NULL )
	{
		return;
	}

	uint byteCount = resources[id]->GetDataUsageInBytes();
	assert(byteCount <= bytesUsed);
	bytesUsed -= byteCount;
	resourcesLoaded--;

	resources[id]->Destroy();
	delete resources[id];
	resources[id] = NULL;

	idStack.push_back(id);
}

//--------------------------------------------------------------------
template<class resType, class initType>
resType *ResourceMgr<resType, initType>::GetResource(ResourceID id)
{
	if(id == NULL_RESOURCEID || (uint)id >= resources.size() || (uint)id < 0)
		return NULL;
	return resources[id];
}

//--------------------------------------------------------------------
template<class resType, class initType>
void ResourceMgr<resType, initType>::DestoryAllResources()
{
	size_t len = resources.size();
	for(size_t i = 0; i < len; i++)
	{
		if(resources[i])
			delete resources[i];
	}

	resources.clear();
	idStack.clear();
	lastID = NULL_RESOURCEID;
}

//--------------------------------------------------------------------
template<class resType, class initType>
resType *ResourceRef<resType, initType>::Get() const
{
	if(mgr && id != NULL_RESOURCEID)
		return mgr->GetResource(id);
	return NULL;
}

#endif

