/*!
 * \file PhysicsJoint.cpp
 * \brief Implements the physics joints used to constrain physics objects.
 * \author Joe Forte
 * \date 1/21/2008
 */

#include "Physics/PhysicsJoint.h"

#include <Box2D.h>

//--------------------------------------------------------------------------------------
PhysicsJoint::PhysicsJoint() : world(NULL), obj1(NULL), obj2(NULL), joint(NULL), allowCollisions(false)
{
}

//--------------------------------------------------------------------------------------
PhysicsJoint::~PhysicsJoint()
{
}

//--------------------------------------------------------------------------------------
void PhysicsJoint::Attach1(PhysicsObject *obj)
{
	obj1 = obj;
}

//--------------------------------------------------------------------------------------
void PhysicsJoint::Attach2(PhysicsObject *obj)
{
	obj2 = obj;
}

//--------------------------------------------------------------------------------------
PhysicsObject *PhysicsJoint::GetAttached1() const
{
	return obj1;
}

//--------------------------------------------------------------------------------------
PhysicsObject *PhysicsJoint::GetAttached2() const
{
	return obj2;
}

//--------------------------------------------------------------------------------------
void PhysicsJoint::EnableCollisions(bool yn)
{
	allowCollisions = yn;
}

//--------------------------------------------------------------------------------------
bool PhysicsJoint::AreCollisionsAllowed() const
{
	return allowCollisions;
}

//--------------------------------------------------------------------------------------
bool PhysicsJoint::CreateJoint(PhysicsWorld *w)
{
	if(!w)
		return false;
	
	DestroyJoint();
	
	world = w;
	
	joint = CreateBox2DJoint();
	return joint != NULL;
}

//--------------------------------------------------------------------------------------
PhysicsWorld *PhysicsJoint::GetWorld() const
{
	return world;
}

//--------------------------------------------------------------------------------------
b2Joint *PhysicsJoint::GetBox2DJoint() const
{
	return joint;
}

//--------------------------------------------------------------------------------------
bool PhysicsJoint::IsJointValid() const
{
	return joint != NULL;
}

//--------------------------------------------------------------------------------------
void PhysicsJoint::DestroyJoint()
{
}

//--------------------------------------------------------------------------------------
PhysicsDistanceJoint::PhysicsDistanceJoint() : length(0.0f)
{
}
		
//--------------------------------------------------------------------------------------
PhysicsDistanceJoint::~PhysicsDistanceJoint()
{
}
		
//--------------------------------------------------------------------------------------
void PhysicsDistanceJoint::SetAnchorPoint1(const Point2f &p)
{
	anchorPoint1 = p;
}

//--------------------------------------------------------------------------------------
void PhysicsDistanceJoint::SetAnchorPoint2(const Point2f &p)
{
	anchorPoint2 = p;
}

//--------------------------------------------------------------------------------------
void PhysicsDistanceJoint::SetLength(float len)
{
	length = len;
}

//--------------------------------------------------------------------------------------
b2Joint *PhysicsDistanceJoint::CreateBox2DJoint()
{
	b2World *w = GetWorld()->GetBox2DWorld();
	if(!w)
		return NULL;
	
	b2Body *b1 = NULL;
	PhysicsObject *obj1 = GetAttached1();
	if(obj1)
	{
		const PhysicsBody *body = obj1->GetBody();
		if(body)
			b1 = body->GetBox2DBody();
	}
	
	b2Body *b2 = NULL;
	PhysicsObject *obj2 = GetAttached1();
	if(obj2)
	{
		const PhysicsBody *body = obj2->GetBody();
		if(body)
			b2 = body->GetBox2DBody();
	}
	
	
	b2DistanceJointDef def;
	def.body1 = b1;
	def.body2 = b2;
	def.localAnchor1 = b2Vec2(anchorPoint1.x, anchorPoint1.y);
	def.localAnchor2 = b2Vec2(anchorPoint2.x, anchorPoint2.y);
	def.length = length;
	
	return w->CreateJoint(&def);
}
