/*!
 * \file PhysicsJoint.h
 * \brief Defines the physics joints used to constrain physics objects.
 * \author Joe Forte
 * \date 1/21/2008
 */

#ifndef PHYSICSJOINT_H
#define PHYSICSJOINT_H

#include "Physics/PhysicsObject.h"

// Box2D types
class b2Joint;

/*!
 * \class PhysicsJoint
 * \headerfile PhysicsJoint.h "Physics/PhysicsJoint.h"
 * \brief Base class for representing physics objects.
 */
class PhysicsJoint
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	PhysicsJoint();
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~PhysicsJoint();
	
	/*!
	 * \brief Attaches an object to the joint.
	 *
	 * \param[in] obj the object to attach to the joint.
	 */
	void Attach1(PhysicsObject *obj);
	
	/*!
	 * \brief Attaches an object to the joint.
	 *
	 * \param[in] obj the object to attach to the joint.
	 */
	void Attach2(PhysicsObject *obj);
	
	/*!
	 * \brief Gets the first object attached to the joint.
	 *
	 * \return the first object attached to the joint, or NULL if there is no attachment.
	 */
	PhysicsObject *GetAttached1() const;
	
	/*!
	 * \brief Gets the second object attached to the joint.
	 *
	 * \return the first object attached to the joint, or NULL if there is no attachment.
	 */
	PhysicsObject *GetAttached2() const;
	
	/*!
	 * \brief Enables/Disable the ability for attached objects to collide.  Defaults to deisabled. 
	 *
	 * \param[in] yn set to true to enable collisions or false to disable collisions.
	 */
	void EnableCollisions(bool yn);
	
	/*!
	 * \brief Determines if attached objects are allowed to collide.
	 *
	 * \return true if the attached objects can collide, false otherwise.
	 */
	bool AreCollisionsAllowed() const;
	
	/*!
	 * \brief Creates the actual physics joint.  NOTE: Call this function once the joint is setup.
	 *
	 * \param[in] w the physics world to create the joint in.
	 *
	 * \return true if the joint was create, false otherwise.
	 */
	bool CreateJoint(PhysicsWorld *w);
	
	/*!
	 * \brief Gets the pointer to the physics world the joint belongs to.
	 *
	 * \return the physics world.
	 */
	PhysicsWorld *GetWorld() const;
	
	/*!
	 * \brief Gets the pointer to the Box2D joint.
	 *
	 * \return the Box2d joint.
	 */
	b2Joint *GetBox2DJoint() const;
	
protected:
	
	/*!
	 * \brief Determines if the joint has been created.
	 *
	 * \return true if the joint has been created, false otherwise.
	 */
	bool IsJointValid() const;
	
	/*!
	 * \brief Frees the resources associated with the joints.
	 */
	void DestroyJoint();
	
	/*!
	 * \brief Creates the actual Box2D physics joint.
	 *
	 * \param[in] world the physics world to create the joint in.
	 *
	 * \return a pointer to the created joint or NULL if the joint could not be created.
	 *
	 * NOTE: This function should be implemented by each joint class.
	 */
	virtual b2Joint *CreateBox2DJoint() = 0;
	
private:
	
	PhysicsWorld *world;
	
	PhysicsObject *obj1;
	PhysicsObject *obj2;
	
	b2Joint *joint;
	
	bool allowCollisions;
	
};

/*!
 * \class PhysicsDistanceJoint
 * \headerfile PhysicsJoint.h "Physics/PhysicsJoint.h"
 * \brief Represent a simple joint that maintains a constant distance between two objects.
 */
class PhysicsDistanceJoint : public PhysicsJoint
{
public:
	
	/*!
	 * \brief Constructor.
	 */
	PhysicsDistanceJoint();
	
	/*!
	 * \brief Deconstructor.
	 */
	virtual ~PhysicsDistanceJoint();
	
	/*!
	 * \brief Sets the position (local to the first object) to anchor the joint.
	 *
	 * \param[in] p the anchor point.
	 */
	void SetAnchorPoint1(const Point2f &p);
	
	/*!
	 * \brief Sets the position (local to the second object) to anchor the joint.
	 *
	 * \param[in] p the anchor point.
	 */
	void SetAnchorPoint2(const Point2f &p);
	
	/*!
	 * \brief Sets the length for the joint to maintain between the two objects.
	 *
	 * \param[in] len the length.
	 */
	void SetLength(float len);
	
protected:
	
	/*!
	 * \brief Creates the actual Box2D physics joint.
	 *
	 * \param[in] world the physics world to create the joint in.
	 *
	 * \return a pointer to the created joint or NULL if the joint could not be created.
	 *
	 * NOTE: This function should be implemented by each joint class.
	 */
	virtual b2Joint *CreateBox2DJoint();
	
private:
	
	Point2f anchorPoint1;
	Point2f anchorPoint2;
	
	float length;
};

#endif
