/*********************************************************************
 
	File: PhysicsObject.cpp
	Data: 3/18/2008
	Desc: Implements the PhysicsObject class.
 
**********************************************************************/

#include <assert.h>
#include "Physics/PhysicsWorld.h"
#include "Physics/PhysicsObject.h"

#include <Box2D.h>

//--------------------------------------------------------------------------------------
PhysicsObject::PhysicsObject() : friction(0.0f), bounce(1.0f), mass(1.0f), world(NULL)
{
}

//--------------------------------------------------------------------------------------
PhysicsObject::PhysicsObject(PhysicsWorld *w) : friction(0.0f), bounce(1.0f), mass(1.0f), world(NULL)
{
	AttachToWorld(w);
}

//--------------------------------------------------------------------------------------
PhysicsObject::~PhysicsObject()
{
	Destroy();
}
	
//--------------------------------------------------------------------------------------
bool PhysicsObject::Touch(PhysicsObject *obj, const PhysicsCollisionInfo &hitInfo)
{
	return true;
}
	
//--------------------------------------------------------------------------------------
void PhysicsObject::Moved()
{
}

//--------------------------------------------------------------------------------------
void PhysicsObject::SetCollisionGroups(CollisionGroup belongsTo, CollisionGroup collidesWith)
{
	shapes.SetCollisionGroups(belongsTo, collidesWith);
}

//--------------------------------------------------------------------------------------
void PhysicsObject::AttachToWorld(PhysicsWorld *w)
{
	if(world)
	{
		DetachFromWorld();
	}

	world = w;
	body = world->CreateBody(this, mass);
	shapes.SetBody(&body);

	//body.GetBox2DBody()->SetMassFromShapes();
}

//--------------------------------------------------------------------------------------
void PhysicsObject::DetachFromWorld()
{
	Destroy();
	world = NULL;
}

//--------------------------------------------------------------------------------------
void PhysicsObject::SetConstants(float newFriction, float newBounce)
{
	friction = newFriction;
	bounce = newBounce;

	if(shapes.IsValid())
	{
		shapes.Destroy();
		shapes.SetBody(&body);
	}
}

//--------------------------------------------------------------------------------------
void PhysicsObject::GetConstants(float &f, float &b)
{
	f = friction;
	b = bounce;
}

//--------------------------------------------------------------------------------------
bool PhysicsObject::IsStatic() const
{
	return body.GetBox2DBody()->IsStatic();
}

//--------------------------------------------------------------------------------------
void PhysicsObject::SetShape(const PhysicsShape &s)
{
	if(shapes.IsValid())
	{
		shapes.Destroy();
	}

	shapes.AddShape(s);

	if(body.IsValid())
		shapes.SetBody(&body);
}

//--------------------------------------------------------------------------------------
void PhysicsObject::SetShape(const PhysicsCompoundShape &s)
{
	if(shapes.IsValid())
	{
		shapes.Destroy();
	}
	
	shapes = s;
	
	if(body.IsValid())
		shapes.SetBody(&body);
}

//--------------------------------------------------------------------------------------
void PhysicsObject::AddShape(const PhysicsShape &s)
{
	PhysicsShape shape = s;
	if(body.IsValid())
		shape.SetBody(&body);
	
	shapes.AddShape(shape);
}

//--------------------------------------------------------------------------------------
void PhysicsObject::ApplyForce(const Vector2f &f)
{
	assert(body.IsValid());
	return body.ApplyForce(f);
}

//--------------------------------------------------------------------------------------
void PhysicsObject::ApplyImpulse(const Vector2f &im)
{
	assert(body.IsValid());
	return body.ApplyImpulse(im);
}

//--------------------------------------------------------------------------------------
Point2f PhysicsObject::GetPosition() const
{
	assert(body.IsValid());
	return body.GetPosition();
}

//--------------------------------------------------------------------------------------
void PhysicsObject::SetPosition(const Point2f &p)
{
	assert(body.IsValid());
	return body.SetPosition(p);
}

//--------------------------------------------------------------------------------------
Vector2f PhysicsObject::GetVelocity() const
{
	assert(body.IsValid());
	return body.GetVelocity();
}

//--------------------------------------------------------------------------------------
void PhysicsObject::SetVelocity(const Vector2f &v)
{
	assert(body.IsValid());
	body.SetVelocity(v);
}

//--------------------------------------------------------------------------------------
void PhysicsObject::SetMass(float m, float i)
{
	assert(m >= 0.0f);
	
	mass = m;
	
	if(body.IsValid())
	{
		b2Body *b = body.GetBox2DBody();
		
		b2MassData data;
		data.mass = mass;
		data.center = b->GetPosition();
		data.I = i;
		
		b->SetMass(&data);
		//b->SetMassFromShapes();
	}
}

//--------------------------------------------------------------------------------------
float PhysicsObject::GetMass() const
{
	return mass;
}

//--------------------------------------------------------------------------------------
void PhysicsObject::MakeStatic()
{
	SetMass(0.0f);
}

//--------------------------------------------------------------------------------------
float PhysicsObject::GetAngle() const
{
	assert(body.IsValid());
	return body.GetAngle();
}

//--------------------------------------------------------------------------------------
void PhysicsObject::Destroy()
{
	if(world && body.IsValid())
	{
		world->DestroyBody(this);
		shapes.InValidate();
		body.InValidate();
	}
}
