/*********************************************************************
 
	File: PhysicsObject.h
	Data: 3/18/2008
	Desc: Declares the PhysicsObject class.
 
**********************************************************************/

#ifndef PHYSICSOBJECT_H
#define PHYSICSOBJECT_H

#include "Math/MathCommon.h"
#include "Physics/PhysicsObjectTypes.h"
#include "Physics/PhysicsBody.h"
#include "Physics/PhysicsShape.h"
#include "Physics/PhysicsWorld.h"

class PhysicsWorld;
class PhysicsBody;
class PhysicsShape;

struct PhysicsCollisionInfo;

class PhysicsObject
{
public:
	PhysicsObject();
	PhysicsObject(PhysicsWorld *w);
	virtual ~PhysicsObject();
		
	/* called when this object is touched by another object.  Return true if you want the
	   physics engine to handle the collision.  Return false if you want the collision to 
	   be ignored. */
	virtual bool Touch(PhysicsObject *obj, const PhysicsCollisionInfo &hitInfo);
		
	/* called by the physics engine every frame after a timestep.  (Only called when the 
	   physics object is active not static). */
	virtual void Moved();
	
	/*!
	 * \brief Returns the Physics Type of this PhysicsObject.
	 *
	 * \return The PhysicsObjectType of the object.
	 */
	virtual PhysicsObjectType GetObjectType() const { return POBJ_UNKNOWN; }
		
	void AttachToWorld(PhysicsWorld *w);
	void DetachFromWorld();
		
	void SetConstants(float newFriction, float newBounce);
	void GetConstants(float &friction, float &bounce);
		
	bool IsStatic() const;
	bool IsCollidable() const;
		
	/* Applys the given force to this objects center of mass. */
	virtual void ApplyForce(const Vector2f &f);
	
	/*!
	 * \brief Applys an impulse to the objects center of mass.
	 *
	 * \param[in] im the impulse.
	 */
	virtual void ApplyImpulse(const Vector2f &im);
		
	void SetShape(const PhysicsShape &s);
	void SetShape(const PhysicsCompoundShape &s);
	void AddShape(const PhysicsShape &s);
	
	/*!
	 * \brief Sets the collision groups that the object will belong to and will collide with.
	 *
	 * \param[in] belongsTo the groups that the object belongs to.
	 * \param[in] collidesWith the groups that the object collides with.
	 */
	void SetCollisionGroups(CollisionGroup belongsTo, CollisionGroup collidesWith);
	
	Point2f GetPosition() const;
	virtual void SetPosition(const Point2f &p);
		
	Vector2f GetVelocity() const;
	virtual void SetVelocity(const Vector2f &v);
	
	/*!
	 * \brief Sets the mass and inertia of the object.
	 *
	 * \param[in] m the mass of the object.
	 * \param[in] i the inertia of the object. Defaults to 1.
	 */
	void SetMass(float m, float i = 1.0f); // NOTE: static objects have mass = 0
	
	float GetMass() const;
	
	void MakeStatic();
	
	float GetAngle() const;
	
	const PhysicsBody *GetBody() const { return &body; }
	
	uint GetNumberOfShapes() const { return shapes.GetShapeCount(); }
	const PhysicsShape *GetShape(uint i = 0) const { return &shapes.Get(i); }
	
	const PhysicsCompoundShape &GetShapes() const { return shapes; }
	
	PhysicsCompoundShape &GetShapes() { return shapes; }
		
	PhysicsWorld *GetWorld() const { return world; }
		
	void Destroy();
		
private:
	float friction; // Static and kinetic friction is assumed to be the same.
	float bounce;
	
	float mass;
		
	PhysicsBody body;
	PhysicsCompoundShape shapes;
		
	PhysicsWorld *world;
		
	PhysicsObject(const PhysicsObject &);
	PhysicsObject &operator=(const PhysicsObject &);
};

#endif
