/*!
 * \file PhysicsShape.h
 * \brief Implements the PhysicsShape and PhysicsCompoundShape classes aswell as the Shape file format.
 * \author Joe Forte
 * \date 3/18/2008
 */

#include <assert.h>
#include <fstream>
#include "Common/Globals.h"
#include "Graphics/Color.h"
#include "Graphics/MathDraw.h"
#include "Physics/PhysicsObject.h"
#include "Physics/PhysicsBody.h"
#include "Physics/PhysicsWorld.h"
#include "Physics/PhysicsShape.h"

#include <Box2D.h>

//--------------------------------------------------------------------------------------
void PhysicsShape::SetBody(PhysicsBody *b)
{
	assert(b);
	
	body = b;
	
	InitShape();
}

//--------------------------------------------------------------------------------------
void PhysicsShape::InitShape()
{
	if(body)
	{
		if(shape)
			Destroy();
		
		switch(type)
		{
			case PSHAPE_CIRCLE:
				shape = CreateBox2DCircle(body);
				break;
			case PSHAPE_BOX:
				shape = CreateBox2DBox(body);
				break;
			case PSHAPE_POLY:
				shape = CreateBox2DPoly(body);
				break;
			default:
				shape = NULL;
				break;
		};
		
		SetCollisionFilters();
	}
}

//--------------------------------------------------------------------------------------
void PhysicsShape::SetCollisionGroups(CollisionGroup belongsTo, CollisionGroup collidesWith)
{
	belongsToGroups = belongsTo;
	collidesWithGroups = collidesWith;
	
	SetCollisionFilters();
}

//--------------------------------------------------------------------------------------
void PhysicsShape::SetCollisionFilters()
{
	if(shape)
	{
		b2FilterData filter;
		filter.categoryBits = (uint16)belongsToGroups;
		filter.maskBits = (uint16)collidesWithGroups;
		filter.groupIndex = 0;
		shape->SetFilterData(filter);
		
		PhysicsObject *obj = GetOwner();
		obj->GetWorld()->GetBox2DWorld()->Refilter(shape);
	}
}

//--------------------------------------------------------------------------------------
void PhysicsShape::Destroy()
{
	assert(body && shape);

	body->GetBox2DBody()->DestroyShape(shape);
	shape = NULL;
}

//--------------------------------------------------------------------------------------
void PhysicsShape::SetData(const PhysicsShapeData &d)
{
	data = d;
}

//--------------------------------------------------------------------------------------
void PhysicsShape::SetType(PhysicsShapeType t)
{
	type = t;
}

//--------------------------------------------------------------------------------------
void PhysicsShape::DrawOutline() const
{
	Point2f p(0.0f, 0.0f);
	float a = 0.0f;
	if(body)
	{
		p = body->GetPosition();
		a = body->GetAngle();
	}

	Matrixf T, R, M;
	T.MakeTranslation(p.x, p.y);
	R.MakeRotation(a);
	M = T*data.transform*R;

	switch(type)
	{
		case PSHAPE_CIRCLE:
			{
				Circle c;
				c.r = data.r;
				c.center = p + Point2f(data.transform._13, data.transform._23);
				DrawCircle(c, Color::Green);
			}
			break;
		case PSHAPE_BOX:
			DrawQuad(Point2f(0.0f, 0.0f), data.w, data.h, M, false, Color::Green);
			break;
		case PSHAPE_POLY:
			DrawPoly(data.verts, data.vertCount, M, Color::Green);
			break;
	};
}

//--------------------------------------------------------------------------------------
void PhysicsShape::MultiplyTransform(const Matrixf &m)
{
	data.transform = m*data.transform;
	InitShape();
}

//--------------------------------------------------------------------------------------
PhysicsObject *PhysicsShape::GetOwner() const
{
	if(body)
		return body->GetOwner();
	return NULL;
}

//--------------------------------------------------------------------------------------
AABB PhysicsShape::GetBoundingBox() const
{
	AABB bbox;

	Point2f center(0.0f, 0.0f);
	PhysicsObject *owner = GetOwner();
	if(owner)
		center = owner->GetPosition();
	center = data.transform.Transform(center);

	switch(type)
	{
		case PSHAPE_CIRCLE:
			bbox.max.x = center.x + data.r;
			bbox.max.y = center.y + data.r;
			bbox.min.x = center.x - data.r;
			bbox.min.y = center.y - data.r;
			break;

		case PSHAPE_BOX:
		case PSHAPE_POLY:
			{
				assert(data.vertCount <= PSHAPE_POLY_MAX_VERTS);

				Matrixf T, M;
				T.MakeTranslation(center.x, center.y);
				M = T * data.transform;

				float minx = FLT_MAX; float maxx = FLT_MIN;
				float miny = FLT_MAX; float maxy = FLT_MIN;
				for(uint i = 0; i < data.vertCount; i++)
				{
					Point2f p = M.Transform(data.verts[i]);
					minx = std::min(p.x, minx); maxx = std::max(p.x, maxx);
					miny = std::min(p.y, miny); maxy = std::max(p.y, maxy);
				}

				bbox.min.x = minx; bbox.max.x = maxx;
				bbox.min.y = miny; bbox.max.y = maxy;
			}
			break;
	};

	return bbox;
}

//--------------------------------------------------------------------------------------
b2Shape *PhysicsShape::CreateBox2DBox(PhysicsBody *b)
{
	assert(b);
	b2Body *b2b = b->GetBox2DBody();
	assert(b2b);

	PhysicsObject *obj = b->GetOwner();
	assert(obj);

	float friction, restitution;
	obj->GetConstants(friction, restitution);

	b2PolygonDef def;
	def.isSensor		= false;
	def.friction		= friction;
	def.restitution		= restitution;
	def.vertexCount     = 4;
	
	float w2 = data.w / 2.0f;
	float h2 = data.h / 2.0f;
	
	Point2f points[4];
	points[0].x = -w2; points[0].y = -h2;
	points[1].x =  w2; points[1].y = -h2;
	points[2].x =  w2; points[2].y =  h2;
	points[3].x = -w2; points[3].y =  h2;
	Transform(points, 4, data.transform);
	
	for(int i = 0; i < 4; i++)
	{
		def.vertices[i].Set(points[i].x, points[i].y);
	}
	
	b2Shape *s = b2b->CreateShape(&def);
	assert(s);

	return s;
}

//--------------------------------------------------------------------------------------
b2Shape *PhysicsShape::CreateBox2DCircle(PhysicsBody *b)
{
	assert(b);
	b2Body *b2b = b->GetBox2DBody();
	assert(b2b);

	PhysicsObject *obj = b->GetOwner();
	assert(obj);

	float friction, restitution;
	obj->GetConstants(friction, restitution);
	
	Point2f center = data.transform.Transform(Point2f(0.0f, 0.0f));
	b2Vec2 pos(center.x, center.y);

	b2CircleDef def;
	def.radius			= data.r;
	def.isSensor		= false;
	def.localPosition	= pos;
	def.friction		= friction;
	def.restitution		= restitution;

	b2Shape *s = b2b->CreateShape(&def);
	assert(s);

	return s;
}

//--------------------------------------------------------------------------------------
b2Shape *PhysicsShape::CreateBox2DPoly(PhysicsBody *b)
{
	assert(b);
	b2Body *b2b = b->GetBox2DBody();
	assert(b2b);
	
	PhysicsObject *obj = b->GetOwner();
	assert(obj);
	
	float friction, restitution;
	obj->GetConstants(friction, restitution);
	
	b2PolygonDef def;
	def.isSensor		= false;
	def.friction		= friction;
	def.restitution		= restitution;
	def.vertexCount     = data.vertCount;
	
	for(uint i = 0; i < data.vertCount; i++)
	{
		Point2f pos = data.transform.Transform(data.verts[i]);
		def.vertices[i].Set(pos.x, pos.y);
	}
	
	b2Shape *s = b2b->CreateShape(&def);
	assert(s);
	
	return s;
}

//--------------------------------------------------------------------------------------
void PhysicsCompoundShape::SetBody(PhysicsBody *b)
{
	std::vector<PhysicsShape>::iterator i;
	for(i = shapes.begin(); i != shapes.end(); i++)
	{
		i->SetBody(b);
	}
}

//--------------------------------------------------------------------------------------
void PhysicsCompoundShape::AddShape(const PhysicsShape &s)
{
	shapes.push_back(s);
}

//--------------------------------------------------------------------------------------
uint PhysicsCompoundShape::GetShapeCount() const
{
	return (uint)shapes.size();
}

//--------------------------------------------------------------------------------------
PhysicsShape &PhysicsCompoundShape::Get(uint i)
{
	assert(i < GetShapeCount());
	return shapes[i];
}

//--------------------------------------------------------------------------------------
const PhysicsShape &PhysicsCompoundShape::Get(uint i) const
{
	assert(i < GetShapeCount());
	return shapes[i];
}

//--------------------------------------------------------------------------------------
AABB PhysicsCompoundShape::GetBoundingBox() const
{
	AABB bbox(FLT_MAX, FLT_MAX, -FLT_MAX, -FLT_MAX);
	
	std::vector<PhysicsShape>::const_iterator i;
	for(i = shapes.begin(); i != shapes.end(); i++)
	{
		AABB b = i->GetBoundingBox();
		
		bbox.max.x = std::max(bbox.max.x, b.max.x);
		bbox.max.y = std::max(bbox.max.y, b.max.y);
		
		bbox.min.x = std::min(bbox.min.x, b.min.x);
		bbox.min.y = std::min(bbox.min.y, b.min.y);
	}
	
	return bbox;
}

//--------------------------------------------------------------------------------------
bool PhysicsCompoundShape::IsValid() const 
{ 
	std::vector<PhysicsShape>::const_iterator i;
	for(i = shapes.begin(); i != shapes.end(); i++)
	{
		if(!i->IsValid())
			return false;
	}
	return true;
}

//--------------------------------------------------------------------------------------
void PhysicsCompoundShape::InValidate()
{
	std::vector<PhysicsShape>::iterator i;
	for(i = shapes.begin(); i != shapes.end(); i++)
	{
		i->InValidate();
	}
}

//--------------------------------------------------------------------------------------
void PhysicsCompoundShape::DrawOutline() const
{
	std::vector<PhysicsShape>::const_iterator i;
	for(i = shapes.begin(); i != shapes.end(); i++)
	{
		i->DrawOutline();
	}
}

//--------------------------------------------------------------------------------------
void PhysicsCompoundShape::SetCollisionGroups(CollisionGroup belongsTo, CollisionGroup collidesWith)
{
	std::vector<PhysicsShape>::iterator i;
	for(i = shapes.begin(); i != shapes.end(); i++)
	{
		i->SetCollisionGroups(belongsTo, collidesWith);
	}
}

//--------------------------------------------------------------------------------------
void PhysicsCompoundShape::MultiplyTransform(const Matrixf &m)
{
	std::vector<PhysicsShape>::iterator i;
	for(i = shapes.begin(); i != shapes.end(); i++)
	{
		i->MultiplyTransform(m);
	}
}

//--------------------------------------------------------------------------------------
void PhysicsCompoundShape::Destroy()
{
	std::vector<PhysicsShape>::iterator i;
	for(i = shapes.begin(); i != shapes.end(); i++)
	{
		i->Destroy();
	}
	shapes.clear();
}

//--------------------------------------------------------------------------------------
bool LoadCircleFromFile(std::fstream &file, PhysicsCompoundShape &shape)
{
	float r;
	if( (file >> r).fail())
		return false;
	
	float x, y;
	if( (file >> x >> y).fail() )
		return false;
	
	shape.AddShape(CreatePhysicsShapeCircle(r, x, y));
	return true;
}

//--------------------------------------------------------------------------------------
bool LoadBoxFromFile(std::fstream &file, PhysicsCompoundShape &shape)
{
	float w, h;
	if( (file >> w >> h).fail() )
		return false;
	
	float x, y, sx, sy, a;
	if( (file >> x >> y >> sx >> sy >> a).fail() )
		return false;

	if(sx == 0.0f || sy == 0.0f)
	{
		ConError() << "Degenerate scale values specified.\n";
		return false;
	}
	
	shape.AddShape(CreatePhysicsShapeBox(w, h, x, y, sx, sy, a));
	return true;
}

//--------------------------------------------------------------------------------------
bool LoadPolyFromFile(std::fstream &file, uint count, PhysicsCompoundShape &shape)
{	
	Point2f points[PSHAPE_POLY_MAX_VERTS];
	
	float tx, ty, sx, sy, a;
	if( (file >> tx >> ty >> sx >> sy >> a).fail() )
		return false;

	if(sx == 0.0f || sy == 0.0f)
	{
		ConError() << "Degenerate scale values specified.\n";
		return false;
	}
	
	for(uint i = 0; i < count; i++)
	{
		float x, y;
		if( (file >> x >> y).fail() )
			return false;
		
		points[i].x = x;
		points[i].y = y;
 	}
	
	shape.AddShape(CreatePhysicsShapePoly(points, count, tx, ty, sx, sy, a));
	return true;
}

//--------------------------------------------------------------------------------------
bool LoadShapeFromFile(const char *file, PhysicsCompoundShape &shape)
{
	std::fstream f(file);
	
	if(!f.is_open())
	{
		ConError() << "Shape file '|c00FF00FF" << file << "|r could not be opened.\n";
		return false;
	}
	
	std::string str;
	while(f >> str)
	{		
		if(str == "poly")
		{
			uint n = 0;
			if( (f >> n).fail() || n > PSHAPE_POLY_MAX_VERTS)
			{
				ConError() << "Shape file '|c00FF00FF" << file << "|r' specifies an invalid number of points in a '|c00FF00FFpoly|r' command.\n";
				return false;
			}
			
			if(!LoadPolyFromFile(f, n, shape))
			{
				ConError() << "The '|c00FF00FFpoly|r' command is improperly formatted in the shape file '|c00FF00FF" << file << "|r'.\n";
				return false;
			}
		}
		else if(str == "box")
		{
			if(!LoadBoxFromFile(f, shape))
			{
				ConError() << "The '|c00FF00FFbox|r' command is improperly formatted in the shape file '|c00FF00FF" << file << "|r'.\n";
				return false;
			}
		}
		else if(str == "circle")
		{
			if(!LoadCircleFromFile(f, shape))
			{
				ConError() << "The '|c00FF00FFcircle|r' command is improperly formatted in the shape file '|c00FF00FF" << file << "|r'.\n";
				return false;
			}
		}
		else
		{
			ConWarning() << "Invalid command '|c00FF00FF" << str << "|r' found in the shape file '|c00FF00FF" << file << "|r'.\n";
		}
	}
	
	return true;
}

//--------------------------------------------------------------------------------------
PhysicsShape CreatePhysicsShapeCircle(float r, float x, float y)
{
	PhysicsShape::PhysicsShapeData d;
	d.r = r;
	d.transform.MakeTranslation(x, y);
	
	PhysicsShape shape;
	shape.SetData(d);
	shape.SetType(PhysicsShape::PSHAPE_CIRCLE);
	
	return shape;
}

//--------------------------------------------------------------------------------------
PhysicsShape CreatePhysicsShapeBox(float w, float h, float x, float y, float sx, float sy, float a)
{
	assert(sx > 0.0f && sy > 0.0f);
	
	PhysicsShape::PhysicsShapeData d;
	d.w = w;
	d.h = h;
	
	Matrixf R, T, S;
	S.MakeScale(sx, sy);
	R.MakeRotation(a);
	T.MakeTranslation(x, y);
	d.transform = T*R*S;
	
	PhysicsShape shape;
	shape.SetData(d);
	shape.SetType(PhysicsShape::PSHAPE_BOX);
	
	return shape;
}

//--------------------------------------------------------------------------------------
PhysicsShape CreatePhysicsShapePoly(const Point2f *verts, uint n, float x, float y, float sx, float sy, float a)
{
	assert(sx > 0.0f && sy > 0.0f);
	
	PhysicsShape::PhysicsShapeData d;
	
	Matrixf R, T, S;
	S.MakeScale(sx, sy);
	R.MakeRotation(a);
	T.MakeTranslation(x, y);
	d.transform = T*R*S;
	
	n = std::min(n, PSHAPE_POLY_MAX_VERTS);
	for(uint i = 0; i < n; i++)
	{
		d.verts[i] = verts[i];
	}
	
	d.vertCount = n;
	
	PhysicsShape shape;
	shape.SetData(d);
	shape.SetType(PhysicsShape::PSHAPE_POLY);
	
	return shape;
}
