/*!
 * \file PhysicsShape.h
 * \brief Declares the PhysicsShape and PhysicsCompoundShape classes aswell as the Shape file format.
 * \author Joe Forte
 * \date 3/18/2008
 */

#ifndef PHYSICSSHAPE_H
#define PHYSICSSHAPE_H

#include <vector>
#include "Math/Matrix.h"
#include "Math/Vector.h"
#include "Math/Point.h"
#include "Common/BaseTypes.h"

class PhysicsBody;
class PhysicsObject;

// Box2D types
class b2Shape;

/*!
 * \typedef CollisionGroup
 *
 * \brief Defines the group data type.
 */
typedef uint16 CollisionGroup;

const CollisionGroup CG_GROUP1  = (CollisionGroup)BIT(1);  /*!< Group 1. The default group objects belong to and collide with.  */
const CollisionGroup CG_GROUP2  = (CollisionGroup)BIT(2);  /*!< Group 2  */
const CollisionGroup CG_GROUP3  = (CollisionGroup)BIT(3);  /*!< Group 3  */
const CollisionGroup CG_GROUP4  = (CollisionGroup)BIT(4);  /*!< Group 4  */
const CollisionGroup CG_GROUP5  = (CollisionGroup)BIT(5);  /*!< Group 5  */
const CollisionGroup CG_GROUP6  = (CollisionGroup)BIT(6);  /*!< Group 6  */
const CollisionGroup CG_GROUP7  = (CollisionGroup)BIT(7);  /*!< Group 7  */
const CollisionGroup CG_GROUP8  = (CollisionGroup)BIT(8);  /*!< Group 8  */
const CollisionGroup CG_GROUP9  = (CollisionGroup)BIT(9);  /*!< Group 9  */
const CollisionGroup CG_GROUP10 = (CollisionGroup)BIT(10); /*!< Group 10 */
const CollisionGroup CG_GROUP11 = (CollisionGroup)BIT(11); /*!< Group 11 */
const CollisionGroup CG_GROUP12 = (CollisionGroup)BIT(12); /*!< Group 12 */
const CollisionGroup CG_GROUP13 = (CollisionGroup)BIT(13); /*!< Group 13 */
const CollisionGroup CG_GROUP14 = (CollisionGroup)BIT(14); /*!< Group 14 */
const CollisionGroup CG_GROUP15 = (CollisionGroup)BIT(15); /*!< Group 15 */
const CollisionGroup CG_GROUP16 = (CollisionGroup)BIT(16); /*!< Group 16 */
const CollisionGroup CG_ALL     = (CollisionGroup)0xFFFF;  /*!< All possible collision groups */

/*!
 * \brief The maximum number of vertices that can be used in a single polygon shape.
 */
static const uint PSHAPE_POLY_MAX_VERTS = 8;

/*!
 * \class PhysicsShape
 * \headerfile PhysicsShape.h "Physics/PhysicsShape.h"
 * \brief Represents a single physics shape used for collision detections.
 */
class PhysicsShape
{
public:
	
	/*!
	 * \enum PhysicsShapeType
	 * \brief Defines the differant shape types allowed by the PhysicsShape class.
	 */
	enum PhysicsShapeType
	{
		PSHAPE_UNKNOWN = -1, /*!< The shape type is not defined. */
		PSHAPE_CIRCLE = 0, /*!< Circle shape. */
		PSHAPE_BOX, /*!< Rectangular box shape. */
		PSHAPE_POLY, /*!< Arbitrary polygon shape. */
		
		PSHAPE_COUNT /*!< The number of allowable shape types. */
	};
	
	/*!
	 * \struct PhysicsShapeData
	 * \headerfile PhysicsShape.h "Physics/PhysicsShape.h"
	 * \brief Holds all of the data needed to fully descibe a PhysicsShape.
	 */
	struct PhysicsShapeData
	{
		float r; /*!< The radius of the circle.  Only valid for PSHAPE_CIRCLE shapes. */
		
		float w; /*!< The width of the box. Only valid for PSHAPE_BOX shapes. */
		float h; /*!< The height of the box. Only valid for PSHAPE_BOX shapes. */
		
		Point2f verts[PSHAPE_POLY_MAX_VERTS]; /*!< The vertices of the polygon. Only valid for PSHAPE_POLY shapes. */
		uint vertCount; /*!< The number of vertices in the polygon. Only valid for PSHAPE_POLY shapes. */
		
		Matrixf transform; /*!< The local transformation for the shape. Valid for all shape types. */
	};
	
	/*!
	 * \brief Constructor.
	 */
	PhysicsShape() : shape(NULL), body(NULL), type(PSHAPE_UNKNOWN), belongsToGroups(CG_GROUP1), collidesWithGroups(CG_GROUP1) {}

	/*!
	 * \brief Deconstructor.
	 */
	~PhysicsShape() {}

	void SetBody(PhysicsBody *b);
	
	void SetData(const PhysicsShapeData &d);
	
	void SetType(PhysicsShapeType t);
	
	PhysicsShapeType GetType() const { return type; }
	
	// Call when you want the shape to be rebuilt. NOTE: called internally in SetBody()
	void InitShape();

	void Destroy();
	
	bool IsValid() const { return shape != NULL; }
	
	void InValidate() { shape = NULL; body = NULL; }
	
	void DrawOutline() const;
	
	/*!
	 * \brief Sets the collision groups that the shape will belong to and will collide with.
	 *
	 * \param[in] belongsTo the groups that the shape belongs to.
	 * \param[in] collidesWith the groups that the shaoe collides with.
	 */
	void SetCollisionGroups(CollisionGroup belongsTo, CollisionGroup collidesWith);
	
	/*!
	 * \brief Transforms the shape by the given Matrix.
	 *
	 * \param[in] m the Matrix to transform the shape by.
	 */
	void MultiplyTransform(const Matrixf &m);

	/*!
	 * \brief Computes the bounding box for the shaped.
	 *
	 * \return An axis aligned bounding box containting the shape.
	 */
	AABB GetBoundingBox() const;

	/*!
	 * \brief Gets the PhysicsObject associated with this shape.
	 *
	 * \return A pointer to the physics object that owns this object or NULL if no such object exists.
	 */
	PhysicsObject *GetOwner() const;

	b2Shape *GetBox2DShape() const { return shape; }

protected:
	
	/*!
	 * \brief Sets the collison filters used by Box2D.
	 */
	void SetCollisionFilters();
	
	b2Shape *CreateBox2DCircle(PhysicsBody *b);
	b2Shape *CreateBox2DBox(PhysicsBody *b);
	b2Shape *CreateBox2DPoly(PhysicsBody *b);
	
	CollisionGroup belongsToGroups;
	CollisionGroup collidesWithGroups;
	
	PhysicsShapeData data;

	PhysicsShapeType type;
	
	PhysicsBody *body;
	
	b2Shape *shape;
};

class PhysicsCompoundShape
{
public:
	void SetBody(PhysicsBody *b);
	
	void AddShape(const PhysicsShape &s);
	
	uint GetShapeCount() const;
	
	PhysicsShape &Get(uint i);
	
	const PhysicsShape &Get(uint i) const;
	
	bool IsValid() const;
	
	void InValidate();
	
	void DrawOutline() const;
	
	/*!
	 * \brief Sets the collision groups that the shape will belong to and will collide with.
	 *
	 * \param[in] belongsTo the groups that the shape belongs to.
	 * \param[in] collidesWith the groups that the shaoe collides with.
	 */
	void SetCollisionGroups(CollisionGroup belongsTo, CollisionGroup collidesWith);
	
	/*!
	 * \brief Transforms the shapes by the given Matrix.
	 *
	 * \param[in] m the Matrix to transform the shapes by.
	 */
	void MultiplyTransform(const Matrixf &m);

	/*!
	 * \brief Computes the bounding box for the shape.
	 *
	 * \return An axis aligned bounding box containting the compound shape.
	 */
	AABB GetBoundingBox() const;
	
	void Destroy();
	
private:
	std::vector<PhysicsShape> shapes;
};

/*!
 * \brief Loads shape data from the specified shape file.
 *
 * \param[in] file the path to shape file.
 * \param[out] shape the PhysicsCompoundShape that the shape data specified in \a file will be added to.
 *
 * \return True if the shape file was succesfully loaded, false otherwise.
 *
 * Shape File format (*.shape)
 *     The shape file format is a simple human readable ASCII text file that describes one or
 * more shapes.  There are three shape types that can be defined as follows:
 *
 * Circle shape:
 *   A circle can be defined by writing 'circle' followed by the radius 'r', the center positions 
 *   'x' and 'y' coordinates as follows:
 *
 *       circle r x y
 *
 *    Example:
 *       circle 5.0 0.0 0.0
 *
 * Box shape:
 *   A box can be defined by writing 'box' followd by its width 'w', height 'h', its center position 
 *   'x' and 'y', its scale in the x and y direction 'sx' and 'sy' and its angle in radians 'a' as
 *   follows:
 *
 *       box w h x y sx sy a
 *
 *    Example:
 *       box 5.0 5.0 0.0 0.0 1.0 1.0 0.0
 *
 * Polygon shape:
 *   A polygon can be defined by writing 'poly' followed by the number of vertices in the polygon 'n',
 *   the amount to translate the polygon 'tx' and 'ty', the amount to scale the polygon 'sx' and 'sy',
 *   the angle of the polygon 'a' and 'n' xy coordinate pairs for each verex in the polygon as follows:
 *
 *     poly n tx ty sx sy a
 *     x1 y1
 *     x2 y2
 *     ...
 *     xn yn
 *
 *    Example:
 *        poly 3 0.0 0.0 1.0 1.0 0.0
 *        -1.0 -1.0
 *        0.0 1.0
 *        1.0 1.0
 *
 * Shape file constrints:
 *   -# All angles are in radians, and all rotaions are done about the origin (0,0).
 *   -# Scaling factors do not have tobe uniform but they can never equal 0.
 *   -# The number of vertecies in a polygon is constrained to PSHAPE_POLY_MAX_VERTS (currently set to 8).
 *   -# Vertices must be in Counter Clockwise Winding Order.
 *   -# An arbitrary number of shapes can be defined in one shape file.  Note that performance may be hindered
 *      by a large number of shapes.
 */
bool LoadShapeFromFile(const char *file, PhysicsCompoundShape &shape);

PhysicsShape CreatePhysicsShapeCircle(float r, 
									  float x = 0.0f, 
									  float y = 0.0f);

PhysicsShape CreatePhysicsShapeBox(float w, 
								   float h, 
								   float x = 0.0f, 
								   float y = 0.0f,
								   float sx = 1.0f,
								   float sy = 1.0f,
								   float a = 0.0f);

PhysicsShape CreatePhysicsShapePoly(const Point2f *verts,
									uint n,
									float x = 0.0f,
									float y = 0.0f, 
									float sx = 1.0f, 
									float sy = 1.0f, 
									float a = 0.0f); // n must be less then PSHAPE_POLY_MAX_VERTS

#endif
