/*********************************************************************
 
	File: PhysicsWorld.cpp
	Data: 3/18/2008
	Desc: Implements the PhysicsWorld class that maintains the physics
	bodies and simulation.
 
**********************************************************************/

#include <assert.h>
#include "Physics/PhysicsWorld.h"
#include "Physics/PhysicsObject.h"
#include "Physics/PhysicsBody.h"

#include "Common/Globals.h"
#include "Game/ConsoleVars.h"

#include <Box2D.h>

//--------------------------------------------------------------------------------------
void PhysicsContactListner::Add(const b2ContactPoint* point)
{
	ContactPair s(point->shape1, point->shape2);
	
	b2ContactPoint p = *point;
	data[s].push_back(p);
}

//--------------------------------------------------------------------------------------
void PhysicsContactListner::Persist(const b2ContactPoint* point)
{
}
		
//--------------------------------------------------------------------------------------
void PhysicsContactListner::Remove(const b2ContactPoint* point)
{
}
		
//--------------------------------------------------------------------------------------
void PhysicsContactListner::Result(const b2ContactResult* point)
{
}

//--------------------------------------------------------------------------------------
void PhysicsContactListner::ClearContactData()
{
	data.clear();
}

//--------------------------------------------------------------------------------------
PhysicsWorld::PhysicsWorld(const AABB &bounds)
{
	b2AABB aabb;
	aabb.lowerBound.Set(bounds.min.x, bounds.min.y);
	aabb.upperBound.Set(bounds.max.x, bounds.max.y);
	
	b2Vec2 g = b2Vec2_zero;
	
	bool allowSleep = true;
	
	world = new b2World(aabb, g, allowSleep);

	assert(world);
	
	world->SetContactListener(&contactListner);
}

//--------------------------------------------------------------------------------------
PhysicsWorld::~PhysicsWorld()
{
	// The destruction of a Box2D world will destroy any bodies and shapes still in the world
	delete world;
}

//--------------------------------------------------------------------------------------
void PhysicsWorld::SetGravity(const Vector2f &g)
{
	assert(world);

	gravity = g;
	
	b2Vec2 grav(g.x, g.y);
	world->SetGravity(grav);
}

//--------------------------------------------------------------------------------------
void PhysicsWorld::GetGravity(Vector2f &g) const
{
	g = gravity;
}

//--------------------------------------------------------------------------------------
PhysicsBody PhysicsWorld::GetStaticWorldBody() const
{
	assert(world);
	return PhysicsBody( world->GetGroundBody() );
}

//--------------------------------------------------------------------------------------
void PhysicsWorld::Update(float timeDelta)
{
	assert(world);
	
	float scale = phy_timeScale->GetFloat();
	if(scale < 0.01f)
	{
		ConWarning() << "'phy_timeScale' must be greater then 0.01, setting to 0.01.\n";
		phy_timeScale->SetFloat(0.01f);
		scale = 0.01f;
	}

	const int iterations = 10;
	const float timeInc = 1.0f / 60.0f;
	float dt = 0.0f;
	do
	{
		contactListner.ClearContactData();
		
		world->Step(scale*timeInc, iterations);
		dt += timeInc;
		
		InformContacts();
		
	} while(dt < timeDelta);
	
	std::map<b2Body*, PhysicsObject*>::iterator i;
	for(i = bodyMap.begin(); i != bodyMap.end(); i++)
	{
		b2Body *b = i->first;
		PhysicsObject *obj = i->second;

		assert(b && obj);

		if( !b->IsSleeping() )
		{
			obj->Moved();
		}
	}
}

//--------------------------------------------------------------------------------------
PhysicsBody PhysicsWorld::CreateBody(PhysicsObject *owner, bool fastMoving)
{
	return CreateBody(owner, Point2f(0.0f, 0.0f), 1.0f, fastMoving);
}

//--------------------------------------------------------------------------------------
PhysicsBody PhysicsWorld::CreateBody(PhysicsObject *owner, float mass, bool fastMoving)
{
	return CreateBody(owner, Point2f(0.0f, 0.0f), mass, fastMoving);
}

//--------------------------------------------------------------------------------------
PhysicsBody PhysicsWorld::CreateBody(PhysicsObject *owner, const Point2f &p, bool fastMoving)
{
	return CreateBody(owner, p, 1.0f, fastMoving);
}

//--------------------------------------------------------------------------------------
PhysicsBody PhysicsWorld::CreateBody(PhysicsObject *owner, const Point2f &p, float mass, bool fastMoving)
{
	assert(world && owner);

	b2BodyDef def;
	def.userData = (void*)owner;
	def.isBullet = fastMoving;
	def.linearDamping = 0.0f;
	def.angularDamping = 0.1f;

	b2Vec2 pos(p.x, p.y);
	def.position = pos;
	def.massData.center = pos;
	def.massData.I = (mass == 0.0f ? 0.0f : 1.0f);
	def.massData.mass = mass;
	
	b2Body *b = world->CreateBody(&def);

	assert(b);

	bodyMap[b] = owner;
	
	return PhysicsBody(b);
}

//--------------------------------------------------------------------------------------
void PhysicsWorld::DestroyBody(const PhysicsObject *owner)
{
	assert(owner && owner->GetBody());

	b2Body *b = owner->GetBody()->GetBox2DBody();
	
	assert(b && b->GetUserData() == owner && owner->GetWorld() == this);

	world->DestroyBody(b);
	bodyMap.erase(b);
}

//--------------------------------------------------------------------------------------
void ConvertContactData(const PhysicsContactListner::ContactList &data, PhysicsCollisionInfo &info)
{
	uint size = (uint)data.size();
	
	for(uint i = 0; i < size; i++)
	{
		const b2ContactPoint &cp = data[i];
		
		PhysicsCollisionInfo::ContactInfo c;
		c.position = Point2f( cp.position.x, cp.position.y );
		c.normal = Vector2f( cp.normal.x, cp.normal.y );
		c.penetration = cp.separation;
		
		info.contacts.push_back(c);
	}
}

//--------------------------------------------------------------------------------------
void PhysicsWorld::InformContacts()
{
	PhysicsContactListner::ContactData::iterator i;
	for(i = contactListner.data.begin(); i != contactListner.data.end(); i++)
	{
		const PhysicsContactListner::ContactPair &p = i->first;
		const PhysicsContactListner::ContactList &lst = i->second;
		
		PhysicsObject *obj1 = reinterpret_cast<PhysicsObject*>( p.first->GetBody()->GetUserData() );
		PhysicsObject *obj2 = reinterpret_cast<PhysicsObject*>( p.second->GetBody()->GetUserData() );
		
		PhysicsCollisionInfo info;
		ConvertContactData(lst, info);
		
		obj1->Touch(obj2, info);
		obj2->Touch(obj1, info);
	}
}
