/*********************************************************************
 
	File: PhysicsWorld.h
	Data: 3/18/2008
	Desc: Declares the PhysicsWorld class that maintains the physics
    bodies and simulation.
 
**********************************************************************/

#ifndef PHYSICSWORLD_H
#define PHYSICSWORLD_H

#include <map>
#include <vector>

#include "Math/MathCommon.h"

#include <Box2D.h>

class PhysicsObject;
class PhysicsBody;

// Box2D types
class b2World;
class b2Body;

class PhysicsContactListner : public b2ContactListener
{
public:
	/// Called when a contact point is added. This includes the geometry
	/// and the forces.
	virtual void Add(const b2ContactPoint* point);
	
	/// Called when a contact point persists. This includes the geometry
	/// and the forces.
	virtual void Persist(const b2ContactPoint* point);
	
	/// Called when a contact point is removed. This includes the last
	/// computed geometry and forces.
	virtual void Remove(const b2ContactPoint* point);
	
	/// Called after a contact point is solved.
	virtual void Result(const b2ContactResult* point);
	
	void ClearContactData();
	
	typedef std::pair<b2Shape*, b2Shape*> ContactPair;
	typedef std::vector<b2ContactPoint> ContactList;
	typedef std::map<ContactPair, ContactList> ContactData;
	
	ContactData data;
};

struct PhysicsCollisionInfo
{	
	struct ContactInfo
	{
		Point2f position;
		Vector2f normal;
		float penetration;
	};
	
	std::vector<ContactInfo> contacts;
};

class PhysicsWorld
{
public:
	PhysicsWorld(const AABB &bounds);
	~PhysicsWorld();
	
	void SetGravity(const Vector2f &g);
	void GetGravity(Vector2f &g) const;
	
	virtual void Update(float timeDelta);
	
	uint GetBodyCount() const { return (uint)bodyMap.size(); }

	PhysicsBody GetStaticWorldBody() const;
	
	PhysicsBody CreateBody(PhysicsObject *owner, bool fastMoving = false);
	PhysicsBody CreateBody(PhysicsObject *owner, float mass, bool fastMoving = false);
	PhysicsBody CreateBody(PhysicsObject *owner, const Point2f &p, bool fastMoving = false);
	PhysicsBody CreateBody(PhysicsObject *owner, const Point2f &p, float mass, bool fastMoving = false);

	void DestroyBody(const PhysicsObject *owner);

	b2World *GetBox2DWorld() const { return world; }
	
protected:
	
	void InformContacts();
	
	b2World *world;
	
	Vector2f gravity;
	
	PhysicsContactListner contactListner;
	
	std::map<b2Body*, PhysicsObject*> bodyMap; // maps ode body ids to a PhysicsObjects
	
	// updates all of the objects after a world update
	void UpdateBodyMap();
	
	// destroys all of the objects in the body map
	void DestroyBodyMap();	
};

#endif
