/*!
 * \file MusicPlayer.cpp
 * \brief Implements the MusicPlayer class.
 * \author Joe Forte
 * \date 6/14/2009
 */

#include "Common/Globals.h"
#include "Application/DataPaths.h"
#include "Sound/MusicPlayer.h"

#include "tinyxml.h"

const std::string PLAYLIST_ELEMENT = "Playlist";
const std::string SONG_ELEMENT = "Song";
const std::string NAME_ELEMENT = "Name";
const std::string FILE_ELEMENT = "File";
const std::string ARTIST_ELEMENT = "Artist";
const std::string PATH_ATTRIBUTE = "path";
const std::string PATH_ATTRIBUTE_RELATIVE = "relative"; // The file path is relative to the playlist file.
const std::string PATH_ATTRIBUTE_ABSOLUTE = "absolute"; // The file path is a complete path.
const std::string PATH_ATTRIBUTE_RESOURCE = "resource"; // The file path is relative to the Music resource path. This is the defualt.

/*!
 * \brief Parses the Song element from a play list XML file.
 *
 * \param[in] songNode the song XML node.
 * \param[in] list the playlist to fill with songs.
 * \patam[in] relativePath the path to the XML file.
 * \param[in] xmlFileName the name of the XML file (for debug purposus only).
 *
 * \return true if the Song element was parsed, false otherwise.
 */
bool PlaylistXMLLoadSong(TiXmlNode *songNode, Playlist &list, const std::string &relativePath, const char *xmlFileName);

//--------------------------------------------------------------------
bool LoadPlaylistFromXMLFile(const std::string &fileName, Playlist &list)
{
	return LoadPlaylistFromXMLFile(fileName.c_str(), list);
}

//--------------------------------------------------------------------
bool LoadPlaylistFromXMLFile(const char *fileName, Playlist &list)
{
	TiXmlDocument doc(fileName);
	if( !doc.LoadFile() )
	{
		ConError() << "Playlist file '" << fileName << "' could not be parsed.\n";
		return false;
	}
	
	TiXmlHandle hDoc(&doc);
	
	TiXmlNode *Root = hDoc.FirstChildElement().Node();
	if( !Root || Root->ValueStr() != PLAYLIST_ELEMENT)
		return false;
	
	std::string path = CreatePath(fileName);
	std::string relativePath = GetPathPart(path);
	
	TiXmlNode *curr = Root->FirstChild();
	
	while(curr)
	{
		if(curr->Type() == TiXmlNode::ELEMENT)
		{
			if(SONG_ELEMENT == curr->Value())
			{
				if(!PlaylistXMLLoadSong(curr, list, relativePath, fileName))
					return false;
			}
			else
			{
				ConWarning() << "Unsupported XML element '" << curr->Value() << "' in playlist file " << fileName << ".\n";
			}
		}
		
		curr = Root->IterateChildren(curr);
	}
	
	return true;
}

//--------------------------------------------------------------------
bool PlaylistXMLLoadSong(TiXmlNode *songNode, Playlist &list, const std::string &relativePath, const char *xmlFileName)
{
	std::string songName = "Unknown Song";
	std::string songArtist = "Unknown Artist";
	std::string fileName;
	
	TiXmlNode *curr = songNode->FirstChild();
	while(curr)
	{
		if(curr->Type() == TiXmlNode::ELEMENT)
		{
			if(NAME_ELEMENT == curr->Value())
			{
				TiXmlElement *elm = curr->ToElement();
				TiXmlNode *child = elm->FirstChild();
				if(child->Type() == TiXmlNode::TEXT)
				{
					TiXmlText *text = child->ToText();
					songName = text->Value();
				}
				
			}
			else if(FILE_ELEMENT == curr->Value())
			{
				TiXmlElement *elm = curr->ToElement();
				TiXmlNode *child = elm->FirstChild();
				if(child->Type() == TiXmlNode::TEXT)
				{
					TiXmlText *text = child->ToText();
					fileName = CreatePath(text->Value());
					
					const char *pathType = elm->Attribute(PATH_ATTRIBUTE.c_str());
					
					if(pathType == NULL || 
					   PATH_ATTRIBUTE_RESOURCE == pathType || 
					   (PATH_ATTRIBUTE_RELATIVE != pathType && PATH_ATTRIBUTE_ABSOLUTE != pathType))
					{
						fileName = ConcatPath(DataPaths::GetPath("Music"), fileName);
					}
					else if(PATH_ATTRIBUTE_RELATIVE == pathType)
					{
						fileName = ConcatPath(relativePath, fileName);
					}
				}
			}
			else if(ARTIST_ELEMENT == curr->Value())
			{
				TiXmlElement *elm = curr->ToElement();
				TiXmlNode *child = elm->FirstChild();
				if(child->Type() == TiXmlNode::TEXT)
				{
					TiXmlText *text = child->ToText();
					songArtist = text->Value();
				}
			}
			else
			{
				ConWarning() << "Unsupported XML element '" << curr->Value() << "' in playlist file " << xmlFileName << ".\n";
			}
		}
		
		curr = songNode->IterateChildren(curr);
	}
	
	Playlist::Song song;
	song.name = songName;
	song.file = fileName;
	song.artist = songArtist;
	song.length = -1;
	
	list.AddSongToBack(song);
	
	return true;
}

//--------------------------------------------------------------------
MusicPlayer::MusicPlayer() 
:
volume(128),
delay(1.0f),
delayTimer(0.0f),
fadeOut(false),
paused(false),
stopped(false),
repeatMode(REPEAT_OFF)
{}

//--------------------------------------------------------------------
MusicPlayer::MusicPlayer(const Playlist &pl) 
: 
volume(128),
delay(1.0f),
delayTimer(0.0f),
fadeOut(false),
paused(false),
stopped(false),
repeatMode(REPEAT_OFF)
{
	SetPlaylist(pl);
}

//--------------------------------------------------------------------
byte MusicPlayer::GetVolume() const
{
	return volume;
}

//--------------------------------------------------------------------
void MusicPlayer::SetVolume(byte v)
{
	volume = v;
	
	if(currSong.IsValid())
		currSong.SetVolume(volume);
}

//--------------------------------------------------------------------
void MusicPlayer::MoveToStart()
{
	currSongIterator = playlist.Begin();
	LoadCurrentSongAndPlay();
}

//--------------------------------------------------------------------
void MusicPlayer::MoveToEnd()
{
	currSongIterator = playlist.Back();
	LoadCurrentSongAndPlay();
}

//--------------------------------------------------------------------
void MusicPlayer::Next()
{
	currSongIterator++;
	LoadCurrentSongAndPlay();
}

//--------------------------------------------------------------------
void MusicPlayer::Previous()
{
	currSongIterator--;
	LoadCurrentSongAndPlay();
}

//--------------------------------------------------------------------
bool MusicPlayer::Play()
{
	stopped = false;
	
	if(currSong.IsValid())
	{
		currSong.SetVolume(volume);
		currSong.Play();
			
		paused = false;
			
		ConMsg() << "Playing \"" << currSongIterator->name << "\" by " << currSongIterator->artist << "\n";
		
		return currSong.IsPlaying();
	}
	
	return false;
}

//--------------------------------------------------------------------
void MusicPlayer::LoadCurrentSongAndPlay()
{
	if(currSongIterator)
	{
		if(currSong.IsValid())
			currSong.Stop();
		
		currSong.Load(currSongIterator->file.c_str(), false);
		
		if(!IsPaused() && !IsStopped())
			Play();
	}
	
}

//--------------------------------------------------------------------
void MusicPlayer::Stop()
{
	if(currSong.IsValid())
		currSong.Stop();
	stopped = true;
}

//--------------------------------------------------------------------
void MusicPlayer::Pause()
{
	currSong.Pause();
	paused = true;
}

//--------------------------------------------------------------------
bool MusicPlayer::IsPaused()
{
	return paused;
}

//--------------------------------------------------------------------
bool MusicPlayer::IsStopped()
{
	return stopped;
}

//--------------------------------------------------------------------
bool MusicPlayer::IsSongDone()
{
	return !IsPaused() && !IsStopped() && currSong.IsValid() && !currSong.IsPlaying();
}

//--------------------------------------------------------------------
void MusicPlayer::FadeOut(float time)
{
	fadeOut = true;
	fadeOutTime = time;
	fadeOutTimer = 0.0f;
}

//--------------------------------------------------------------------
void MusicPlayer::Update(float dt)
{
	if(fadeOut)
	{
		fadeOutTimer += dt;
		if(fadeOutTimer >= fadeOutTime)
		{
			fadeOut = false;
			Pause();
		}
		
		float delta = 1.0f - (fadeOutTimer / fadeOutTime);
		byte newVol = (byte)(volume * delta);
		if(currSong.IsValid())
			currSong.SetVolume(newVol);
		
	}
	else if(!currSong.IsValid() || IsSongDone())
	{
		delayTimer += dt;
		if(delayTimer > delay)
		{
			if(GetRepeatMode() == REPEAT_SINGLE)
			{
				Stop();
				Play();
			}
			else
			{
				Next();
				if(currSongIterator == playlist.End() && GetRepeatMode() == REPEAT_ON)
					MoveToStart();
				
				delayTimer = 0.0f;
			}
		}
	}
}
