/*!
 * \file MusicPlayer.h
 * \brief Defines the MusicPlayer class for playing custom playlists.
 * \author Joe Forte
 * \date 6/14/2009
 */

#ifndef MUSICPLAYER_H
#define MUSICPLAYER_H

#include <vector>
#include <string>
#include "Sound/Sound.h"

class Playlist;

/*!
 * \brief Loads a play list from an XML file.
 *
 * \param[in] fileName the name if the XML file to load.
 * \param[in] list the playlist to fill with songs.
 *
 * \return true if the XML file was loaded, false otherwise.
 */
bool LoadPlaylistFromXMLFile(const char *fileName, Playlist &list);
bool LoadPlaylistFromXMLFile(const std::string &fileName, Playlist &list);

/*!
 * \class Playlist
 * \headerfile MusicPlayer.h "Sound/MusicPlayer.h"
 * \brief The Playlist class holds a list of songs to play.
 */
class Playlist
{
public:
	
	/*!
	 * \class Song
	 * \brief The information held for each song in the playlist.
	 */
	struct Song
	{
		std::string file;
		std::string name;
		std::string artist;
		
		int length; /*!< The duration of the song in seconds. */
	};
	
	/*!
	 * \brief Returns the number of songs in the playlist.
	 */
	uint Length() const { return (uint)songList.size(); }
	
	/*!
	 * \brief Adds a song to the end of the playlist. NOTE: Only invalidates the end iterator.
	 */
	void AddSongToBack(const Song &song) { songList.push_back(song); }
	
	/*!
	 * \brief Reorders songs in the playlist.  NOTE: Does not invalidate Iterators (althogh they may no longer refer to the same song).
	 */
	void Shuffle();
	
	/*!
	 * \class Iterator
	 * \brief A forward and backward (constant) iterator for walking through the list of songs.
	 */
	class Iterator
	{
	public:
		
		/*! 
		 * \brief Creates an invalid iterator. 
		 */
		Iterator() : curr(-1), list(NULL) {}
		
		/*! 
		 * \brief Increment operator. 
		 */
		void operator++() { curr++; }
		void operator++(int i) { curr++; }
		
		/*! 
		 * \brief Decrement operator. 
		 */
		void operator--() { curr--; }
		void operator--(int i) { curr--; }
		
		/*! 
		 * \brief Boolean equality. 
		 */
		bool operator==(const Iterator &other) const { return list == other.list && curr == other.curr; }

		/*! 
		 * \brief Boolean inequality. 
		 */
		bool operator!=(const Iterator &other) const { return !(*this == other); }
		
		/*! 
		 * \brief Boolean negation. 
		 */
		bool operator!() const { return !IsValid(); }

		/*! 
		 * \brief Boolean type cast. 
		 */
		operator bool() const { return IsValid(); }
		
		/*! 
		 * \brief Dereference operator. 
		 */
		const Song &operator*() const { return (*list)[curr]; }

		/*! 
		 * \brief Dereference operator. 
		 */
		const Song *operator->() const { return &(*list)[curr]; }
		
		/*! 
		 * \brief Checks for iterator validity.
		 */
		bool IsValid() const { return list && curr >= 0 && (size_t)curr < list->size(); }
		
		/*! 
		 * \brief Returns true if the iterator is a valid end iterator.
		 */
		bool IsEnd() const { return list && (size_t)curr == list->size(); }
		
	private:
		
		/*! 
		 * \brief Private constructor.  Only the Playlist class is allowed to create valid iterators.
		 */
		Iterator(int at, const std::vector<Song> *lst) : curr(at), list(lst) {}
		
		int curr;
		const std::vector<Song> *list;
		
		friend class Playlist;
	};
	
	/*!
	 * \brief Constant iterator.
	 */
	typedef Iterator ConstIterator;
	
	/*!
	 * \brief Returns an iterator to the start of the playlist.
	 */
	ConstIterator Begin() const { return ConstIterator(0, &songList); }
	
	/*!
	 * \brief Returns an iterator to one past the end of the playlist.
	 */
	ConstIterator End() const { return ConstIterator(Length(), &songList); }

	/*!
	* \brief Returns an iterator to the last song in the playlist.
	*/
	ConstIterator Back() const { return ConstIterator(Length() - 1, &songList); }
	
	/*!
	 * \brief Returns an iterator to the first occurrence of the song, or the end iterator if it does not exist.
	 */
	ConstIterator Find(const char *name) const { return Find(std::string(name)); }
	
	/*!
	 * \brief Returns an iterator to the first occurrence of the song, or the end iterator if it does not exist.
	 */
	ConstIterator Find(const std::string &name) const
	{
		ConstIterator end = End();
		for(ConstIterator i = Begin(); i != end; i++)
		{
			if(name == i->name)
				return i;
		}
		
		return end;
	}
	
private:
	
	std::vector<Song> songList;
};

/*!
 * \class MusicPlayer
 * \headerfile MusicPlayer.h "Sound/MusicPlayer.h"
 * \brief The MusicPlayer class.
 */
class MusicPlayer
{
public:
	
	/*!
	 * \enum RepeatMode
	 * \brief Defines the differant repeat modes the music player supports.
	 */
	enum RepeatMode
	{
		REPEAT_OFF = 0, /*!< No repeating is done. */
		REPEAT_ON,      /*!< Repeats the playlist. */
		REPEAT_SINGLE   /*!< Repeates on a single song. */
	};

	/*!
	 * \brief Default Constructor.
	 */
	MusicPlayer();

	/*!
	 * \brief Constructor.
	 *
	 * \param[in] pl the playlist to use.
	 */
	MusicPlayer(const Playlist &pl);

	/*!
	 * \brief Sets the Music players playlist.
	 *
	 * \param[in] pl the playlist to use.
	 */
	void SetPlaylist(const Playlist &pl) { playlist = pl; MoveToStart(); }
	
	/*!
	 * \brief Returns the current playlist.
	 */
	const Playlist &GetPlaylist() const { return playlist; }
	
	/*!
	 * \brief Returns a constant playlist iterator to the current song.
	 */
	Playlist::ConstIterator GetCurrentSongIterator() const { return currSongIterator; }
	
	/*!
	 * \brief Returns the currently set repeat mode.
	 */
	RepeatMode GetRepeatMode() const { return repeatMode; }
	
	/*!
	 * \brief Sets the music players repeat mode.
	 */
	void SetRepeatMode(RepeatMode mode) { repeatMode = mode; }
	
	/*!
	 * \brief Gets the volume of the music player.
	 */
	byte GetVolume() const;
	
	/*!
	 * \brief Sets the volume of the music player.
	 */
	void SetVolume(byte v);

	/*!
	 * \brief Moves to the start of the playlist.
	 */
	void MoveToStart();

	/*!
	* \brief Moves to the end of the playlist.
	*/
	void MoveToEnd();

	/*!
	 * \brief Advances to the next song in the playlist.
	 */
	void Next();

	/*!
	 * \brief Advances to the previous song in the playlist.
	 */
	void Previous();

	/*!
	 * \brief Plays the currently active song.
	 *
	 * \return true if the song is playing, false otherwise.
	 */
	bool Play();

	/*!
	 * \brief Stops playing the current song.
	 */
	void Stop();

	/*!
	 * \brief Pauses the currently active song.
	 */
	void Pause();

	/*!
	 * \brief Returns true if the current song is paused.
	 */
	bool IsPaused();

	/*!
	 * \brief Returns true if the music player is stopped.
	 */
	bool IsStopped();

	/*!
	 * \brief Returns true if the currently playing song is done.
	 */
	bool IsSongDone();
	
	/*!
	 * \brief Fades out the currently playing song in the specified amount of time.
	 *
	 * \param[in] time the time in seconds to fade out the song.
	 *
	 * Once the fade out time is reached the song is paused.
	 */
	void FadeOut(float time);

	/*!
	 * \brief Called each frame.
	 *
	 * \param[in] dt the time in seconds since the last frame.
	 *
	 * NOTE: Only needed if you want the music player to manage moving from song to song.
	 */
	void Update(float dt);
	
private:
	
	/*!
	 * \brief Loads the current song and plays it if applicable.
	 */
	void LoadCurrentSongAndPlay();
	
	Playlist playlist;
	Playlist::Iterator currSongIterator;
	
	Music currSong;

	byte volume;
	
	RepeatMode repeatMode;

	float delay;
	float delayTimer;
	
	float fadeOutTime;
	float fadeOutTimer;
	bool fadeOut;

	bool paused;
	bool stopped;
};

#endif
