/*!
 * \file Sound.cpp
 * \brief Implements the Sound resource class.
 * \author Joe Forte
 * \date 7/22/2008
 */

#include <assert.h>
#include "Common/Globals.h"
#include "Sound/Sound.h"
#include "fmod.h"

//--------------------------------------------------------------------
bool InitSoundSubSystem()
{
	if(FSOUND_Init(44100, 32, 0) != 1)
		return false;
	
	// Enable the ability to get full spectrum information from the fmod mixing unit
	FSOUND_DSP_SetActive( FSOUND_DSP_GetFFTUnit(), 1 );
	return true;
}

//--------------------------------------------------------------------
void GetSoundSpectrum(float *spectrum, uint n)
{
	assert(spectrum);

	float *s = FSOUND_DSP_GetSpectrum();
	assert(s);

	// The number of elements to combine from the fmod buffer
	uint num = SOUND_SPECTRUM_MAX_BUFFER_SIZE / n;
	uint index = 0;
	for(uint i = 0; i < n; i++)
	{
		float sum = 0.0f;
		for(uint j = 0; j < num; j++)
		{
			sum = sum + s[index];
			index = index + 1;
		}

		spectrum[i] = std::min(1.0f, sum / 1.5f);
	}
}

//--------------------------------------------------------------------
SoundSample::SoundSample() : sample(NULL) {}

//--------------------------------------------------------------------
SoundSample::~SoundSample() {}

//--------------------------------------------------------------------
IMPLEMENT_RESOURCE_LOAD_FUNCTION(SoundSample)
{
	sample = FSOUND_Sample_Load(FSOUND_UNMANAGED, filePath, FSOUND_LOOP_OFF, 0, 0); 
	return sample != NULL;
}

//--------------------------------------------------------------------
void SoundSample::Destroy()
{
	if(sample)
	{
		FSOUND_Sample_Free(sample);
		sample = NULL;
	}
}

//--------------------------------------------------------------------
uint SoundSample::GetDataUsageInBytes() const
{
	return (uint)FSOUND_Sample_GetLength(sample);
}

//--------------------------------------------------------------------
bool SoundSample::IsPlaying(int channel)
{
	return FSOUND_IsPlaying(channel) == 1;
}

//--------------------------------------------------------------------
int SoundSample::Play()
{
	return FSOUND_PlaySound(FSOUND_FREE, sample);
}

//--------------------------------------------------------------------
void SoundSample::Stop(int channel)
{
	FSOUND_StopSound(channel);
}

//--------------------------------------------------------------------
SoundStream::SoundStream() : stream(NULL), volume(128), pauseTime(0), channel(0) {}

//--------------------------------------------------------------------
SoundStream::~SoundStream() {}

//--------------------------------------------------------------------
IMPLEMENT_COMPLEX_RESOURCE_LOAD_FUNCTION(SoundStream, ForceNoResourceSharing, not_needed)
{
	stream = FSOUND_Stream_Open(filePath, FSOUND_LOOP_OFF, 0, 0); 
	return stream != NULL;
}

//--------------------------------------------------------------------
void SoundStream::Destroy()
{
	if(stream)
	{
		FSOUND_Stream_Close(stream);
		stream = NULL;
	}
}

//--------------------------------------------------------------------
uint SoundStream::GetDataUsageInBytes() const
{
	return (uint)FSOUND_Stream_GetLength(stream);
}

//--------------------------------------------------------------------
void SoundStream::SetVolume(byte newVol)
{
	volume = newVol;
	if(IsPlaying())
		FSOUND_SetVolume(channel, (int)volume);
}

//--------------------------------------------------------------------
void SoundStream::GetLevels(float &left, float &right)
{
	float l, r;
	FSOUND_GetCurrentLevels(channel, &l, &r);

	left = l;
	right = r;
}

//--------------------------------------------------------------------
bool SoundStream::IsPlaying()
{
	return FSOUND_IsPlaying(channel) == 1;
}

//--------------------------------------------------------------------
void SoundStream::Play()
{
	FSOUND_Stream_SetTime(stream, pauseTime);
	channel = FSOUND_Stream_Play(channel, stream);
	FSOUND_SetVolume(channel, (int)volume);
}

//--------------------------------------------------------------------
void SoundStream::Stop()
{
	FSOUND_Stream_Stop(stream);
	pauseTime = 0;
}

//--------------------------------------------------------------------
void SoundStream::Pause()
{
	pauseTime = FSOUND_Stream_GetTime(stream);
	FSOUND_Stream_Stop(stream);
}

//--------------------------------------------------------------------
Sound::Sound() : channel(0), volume(128) {}

//--------------------------------------------------------------------
Sound::~Sound() {}

//--------------------------------------------------------------------
void Sound::Load(const char *file, bool useResourcePath)
{
	sample = LoadSoundSampleResource(file, useResourcePath);
}
	
//--------------------------------------------------------------------
bool Sound::IsPlaying()
{
	assert(sample.IsValid());
	return sample->IsPlaying(channel);
}

//--------------------------------------------------------------------
void Sound::Play()
{
	if(sample.IsValid())
	{
		channel = sample->Play();
	}
}

//--------------------------------------------------------------------
void Sound::Stop()
{
	if(sample.IsValid())
		sample->Stop(channel);
}

//--------------------------------------------------------------------
void Sound::SetVolume(byte newVol)
{
	volume = newVol;
	if(channel)
		FSOUND_SetVolume(channel, (int)volume);
}

//--------------------------------------------------------------------
byte Sound::GetVolume() const
{
	return volume;
}

//--------------------------------------------------------------------
void Music::Load(const char *file, bool useResourcePath)
{
	stream = LoadSoundStreamResource(file, useResourcePath);
}

//--------------------------------------------------------------------
void Music::SetVolume(byte newVol)
{
	if(stream)
		stream->SetVolume(newVol);
}

//--------------------------------------------------------------------
byte Music::GetVolume() const 
{ 
	assert(stream);
	return stream->GetVolume();
}

//--------------------------------------------------------------------
void Music::GetLevels(float &left, float &right)
{
	assert(stream);
	return stream->GetLevels(left, right);
}

//--------------------------------------------------------------------
bool Music::IsPlaying()
{
	if(stream)
		return stream->IsPlaying();
	return false;
}

//--------------------------------------------------------------------
void Music::Play()
{
	if(stream)
		stream->Play();
}

//--------------------------------------------------------------------
void Music::Stop()
{
	if(stream)
		stream->Stop();
}

//--------------------------------------------------------------------
void Music::Pause()
{
	if(stream)
		stream->Pause();
}

//--------------------------------------------------------------------
SoundSampleRef LoadSoundSampleResource(const char *file,  bool useResourcePath)
{
	SoundSampleMgr *mgr = GetSoundSampleManager();
	assert(mgr);
	
	SoundSampleRef ref = mgr->LoadResource(file, NULL, useResourcePath);
	if(!ref.IsValid())
	{
		ConError() << "The SoundSample resource '" << file << "' could not be loaded.\n";
	}
	
	return ref;
}

//--------------------------------------------------------------------
SoundStreamRef LoadSoundStreamResource(const char *file, bool useResourcePath)
{
	SoundStreamMgr *mgr = GetSoundStreamManager();
	assert(mgr);
	
	SoundStreamRef ref = mgr->LoadResource(file, NULL, useResourcePath);
	if(!ref.IsValid())
	{
		ConError() << "The SoundStream resource '" << file << "' could not be loaded.\n";
	}
	
	return ref;
}
