/*!
 * \file Sound.h
 * \brief Defines the Sound resource class.
 * \author Joe Forte
 * \date 7/22/2008
 *
 * Defines the Sound resource class for loading and playing sound effects.
 */

#ifndef SOUND_H
#define SOUND_H

#include "Memory/ResourceManager.h"

struct FSOUND_SAMPLE;
struct FSOUND_STREAM;

/*!
 * \brief Initializes the sound system (fmod).  Call this function once before creating any sounds.
 *
 * \return True if the sound sub system (fmod) was initialized.  Returns false if initialization failed.
 * NOTE: Do not create any sounds if false is returned. 
 */
bool InitSoundSubSystem();

/*! The number of elements in the sound spectrum buffer. */
const int SOUND_SPECTRUM_MAX_BUFFER_SIZE = 512;

/*!
 * \brief Gets the frequancy information for all currently playing audio.
 *
 * \param[out] spectrum an array of floats that will hold the frequancy data.
 * \param[in] n the number of floating point values, should be less then SOUND_SPECTRUM_MAX_BUFFER_SIZE.
 */
void GetSoundSpectrum(float *spectrum, uint n);

/*!
 * \class SoundSample
 * \headerfile Sound.h "Sound/Font.h"
 * \brief The SoundSample resource class.
 *
 * A resource type that is used for loading and playing Sound effects.  The following sound formats are 
 * supported:
 * -# WAV
 * -# MP2
 * -# MP3
 * -# OGG
 * -# RAW 
 */
class SoundSample : public Resource<>
{
public:
	SoundSample();
	~SoundSample();

	/*
	 * Should only be called by the SoundStreamMgr class.
	 */
	DECLARE_RESOURCE_LOAD_FUNCTION();

	/*
	 * Should only be called by the FontMgr class.
	 */
	virtual void Destroy();
	
	/*!
	 * \brief Estimates the number of bytes used by the loaded resource.
	 */
	virtual uint GetDataUsageInBytes() const;

	/*!
	 * \brief Determines if the sound is currently playing.
	 *
	 * \param[in] channel the channel handle returned by Play.
	 *
	 * \return True if the sound is playing, false otherwise.
	 */
	bool IsPlaying(int channel);

	/*!
	 * \brief Plays the currently loaded sound.
	 *
	 * \returns an fmod splacific hanndle to the channel that the sound is being played on.
	 */
	int Play();
	
	/*!
	 * \brief Stops playing the sound.
	 *
	 * \param[in] channel the channel handle returned by Play.
	 */
	void Stop(int channel);

private:
	FSOUND_SAMPLE *sample; // fmod sound handle
};

/*!
 * \class SoundStream
 * \headerfile Sound.h "Sound/Font.h"
 * \brief The SoundStream resource class.
 *
 * An unsharable mutable resource type that is used for streaming and playing music files.  
 * The following sound formats are supported:
 * -# WAV
 * -# MP2
 * -# MP3
 * -# OGG
 * -# RAW
 *
 * NOTE: SoundStreams are not shared by the SoundStreamMgr.  The only way to share a SoundStream is by 
 * using the SoundStreamRef returned by the SoundStreamMgr.  Call LoadResource again will always return 
 * a freshly loaded resource.
 */
class SoundStream : public Resource<ForceNoResourceSharing>
{
public:
	SoundStream();
	~SoundStream();

	/*
	 * Should only be called by the SoundStreamMgr class.
	 */
	DECLARE_COMPLEX_RESOURCE_LOAD_FUNCTION(ForceNoResourceSharing, not_needed);

	/*
	 * Should only be called by the FontMgr class.
	 */
	virtual void Destroy();

	/*!
	 * \brief Estimates the number of bytes used by the loaded resource.
	 */
	virtual uint GetDataUsageInBytes() const;

	/*!
	 * \brief Sets the volume of the Stream.
	 *
	 * \param[in] newVol the new volume between 0 (silent) and 255 (unneccessarily load).
	 */
	void SetVolume(byte newVol);

	/*!
	 * \brief Gets the volume of the Stream.
	 *
	 * \return The new sound value between 0 (silent) and 255 (unneccecarily load).
	 */
	byte GetVolume() const { return volume; }

	/*!
	 * \brief Gets the current levels of the stream.
	 *
	 * \param[out] left the levels for the left channel.
	 * \param[out] right the levels for the right channel.
	 */
	void GetLevels(float &left, float &right);
	
	/*!
	 * \brief Determines if the sound is currently playing.
	 *
	 * \return True if the sound is playing, false otherwise.
	 */
	bool IsPlaying();

	/*!
	 * \brief Plays the currently loaded sound.
	 */
	void Play();
	
	/*!
	 * \brief Stops playing the sound.
	 */
	void Stop();

	/*!
	 * \brief Pauses playing the sound.  The next call to Play will begin at the point Pause was called.
	 */
	void Pause();

private:
	FSOUND_STREAM *stream; // fmod sound handle

	byte volume;
	uint pauseTime;
	int channel;
};

DECLARE_RESOURCE_TYPE(SoundSample);
DECLARE_COMPLEX_RESOURCE_TYPE(SoundStream, ForceNoResourceSharing);

/*!
 * \brief Loads an SoundSample using the global SoundSampleManager.
 * 
 * \param[in] file the filename of the SoundSample to load assumed to be relative to the 'Sounds' data path.
 * \param[in] useResourcePath set to true to use the currently set resource path.  Defaults to true.
 *
 * \return The SoundSampleRef of the loaded SoundSample.
 *
 * NOTE: This should only be called after the global environment has been initialized.
 */
SoundSampleRef LoadSoundSampleResource(const char *file, bool useResourcePath = true);

/*!
 * \brief Loads an SoundStream using the global SoundStreamManager.
 * 
 * \param[in] file the filename of the SoundStream to load assumed to be relative to the 'Sounds' data path.
 * \param[in] useResourcePath set to true to use the currently set resource path.  Defaults to true.
 *
 * \return The SoundStreamRef of the loaded SoundStream.
 *
 * NOTE: This should only be called after the global environment has been initialized.
 */
SoundStreamRef LoadSoundStreamResource(const char *file, bool useResourcePath = true);

/*!
 * \class Sound
 * \headerfile Sound.h "Sound/Font.h"
 * \brief The Sound class.
 *
 * A class loads and plays Sound effects.  The following sound formats are 
 * supported:
 * -# WAV
 * -# MP2
 * -# MP3
 * -# OGG
 * -# RAW 
 */
class Sound
{
public:
	Sound();
	~Sound();

	/*!
	 * \brief Loads the specified sound.
	 *
	 * \param[in] file the filename of the sound to load.
	 * \param[in] useResourcePath set to true to use the currently set resource path.  Defaults to true.
	 */
	void Load(const char *file, bool useResourcePath = true);
	
	/*!
	 * \brief Determines if the sound is currently playing.
	 *
	 * \return True if the sound is playing, false otherwise.
	 */
	bool IsPlaying();

	/*!
	 * \brief Plays the currently loaded sound.
	 */
	void Play();
	
	/*!
	 * \brief Stops playing the sound.
	 */
	void Stop();
	
	/*!
	 * \brief Sets the volume of the music.
	 *
	 * \param[in] newVol the new volume between 0 (silent) and 255 (unnecessarily load).
	 */
	void SetVolume(byte newVol);
	
	/*!
	 * \brief Gets the volume of the music.
	 *
	 * \return The new sound value between 0 (silent) and 255 (unnecessarily loud).
	 */
	byte GetVolume() const;

private:
	SoundSampleRef sample; // sound resource
	
	int channel; // the channel that the sound is playing on
	byte volume;
};

/*!
 * \class Music
 * \headerfile Sound.h "Sound/Font.h"
 * \brief The Music class.
 *
 * A class that loads and plays Music.  The following sound formats are 
 * supported:
 * -# WAV
 * -# MP2
 * -# MP3
 * -# OGG
 * -# RAW
 *
 * NOTE: Music is not shared between duplicate Music classes.
 */
class Music
{
public:
	
	/*!
	 * \brief Default constructor.
	 */
	Music() {}
	
	/*!
	 * \brief Constructor that loads the specified music file.
	 *
	 * \param[in] file the filename of the music file to load.
	 */
	Music(const char *file) { Load(file); }
	
	/*!
	 * \brief Returns true is the Music stream is valid, false otherwise.
	 */
	bool IsValid() const { return stream.IsValid(); }
	
	/*!
	 * \brief Loads the specified music file.
	 *
	 * \param[in] file the filename of the music file to load.
	 * \param[in] useResourcePath set to true to use the currently set resource path.  Defaults to true.
	 */
	void Load(const char *file, bool useResourcePath = true);
	
	/*!
	 * \brief Sets the volume of the music.
	 *
	 * \param[in] newVol the new volume between 0 (silent) and 255 (unneccessarily load).
	 */
	void SetVolume(byte newVol);
	
	/*!
	 * \brief Gets the volume of the music.
	 *
	 * \return The new sound value between 0 (silent) and 255 (unneccessarily loud).
	 */
	byte GetVolume() const;

	/*!
	 * \brief Gets the current levels of the stream.
	 *
	 * \param[out] left the levels for the left channel.
	 * \param[out] right the levels for the right channel.
	 */
	void GetLevels(float &left, float &right);
	
	/*!
	 * \brief Determines if the music file is currently playing.
	 *
	 * \return True if the sound is playing, false otherwise.
	 */
	bool IsPlaying();
		
	/*!
	 * \brief Plays the currently loaded music file.
	 */
	void Play();
		
	/*!
	 * \brief Stops playing the music file.
	 */
	void Stop();
	
	/*!
	 * \brief Pauses playing the music file.  The next call to Play will begin at the point Pause was called.
	 */
	void Pause();
		
private:
	SoundStreamRef stream; // stream resource
};

#endif
