/*! 
 * \file FormattedText.h
 * \brief Defines functions for formatting text.
 * \author Joe Forte
 * \date 6/7/2008
 *	
 * Defines functions that format strings given a Font and a description of the
 * bounds that the formatted text should fit into.
 */

#ifndef FORMATTEDTEXT_H
#define FORMATTEDTEXT_H

#include <vector>
#include <string>
#include "Math/MathCommon.h"
#include "Graphics/Font.h"

/*!
 * \brief Formats the given string with respect to a given area and font.
 *
 * Takes a string and splits it into a number of lines based on the given font and desired area.
 * After formatting, the \a text vector can be drawn from top to bottom with the font \a f to 
 * to produce the correctly formatted block of text.
 *
 * \param[in] str the string to be formatted.
 * \param[in] f the font used for formatting.
 * \param[in] area the area the str will be formatted to fit into.  NOTE: \a x and \a y are ignored.
 * \param[out] text the lines of text split from \a str.
 *
 * \return the number of lines added to text.
 *
 * \todo take area into consideration.
 */
int FormatText(const char *str, FontRef f, const Rectf &area, std::vector<std::string> &text);

/*!
 * \brief Formats the given string not taking area into consideration.
 *
 * Takes a string and splits it into a number of lines based on the given font. After formatting, 
 * the \a text vector can be drawn from top to bottom with the font \a f to to produce the correctly 
 * formatted block of text.
 *
 * \param[in] str the string to be formatted.
 * \param[in] f the font used for formatting.
 * \param[out] text the lines of text split from \a str.
 *
 * \return the number of lines added to text.
 */
int FormatText(const char *str, FontRef f, std::vector<std::string> &text);

/*!
 * \brief Draws the given formatted text.
 *
 * \param[in] f the font to draw the text with.
 * \param[in] p the top left position in normalized devices coordintes where the text will be draw.
 * \param[in] c the color to draw the text.
 * \param[in] text the text to draw.  Assumed to be generated with respect to the font \a f.
 * \see FormatText
 */
void DrawFormattedText(FontRef f, const Point2f &p, const Color4f &c, const std::vector<std::string> &text);

/*!
 * \brief Draws a sub-sequance of the given formatted text.
 *
 * \param[in] f the font to draw the text with.
 * \param[in] p the top left position in normalized devices coordintes where the text will be draw.
 * \param[in] c the color to draw the text.
 * \param[in] text the text to draw.  Assumed to be generated with respect to the font \a f.
 * \param[in] offset the offset from the start of the \a text vector to draw from.
 * \param[in] count the number of lines to draw from the \a text vector.
 * \see FormatText
 */
void DrawFormattedText(FontRef f, const Point2f &p, const Color4f &c, const std::vector<std::string> &text, uint offset, uint count);

/*!
 * \brief Wraps text to be within a given max width.
 *
 * \param[in] f the font used to wrap the text.
 * \param[in] width the max width to wrap text against.
 * \param[in] str the string to wrap.
 * \param[out] txt vector that holds the wrapped text.  Where each line in the vector fits into the given width.
 *
 * \return The number of lines added to the txt vector.
 *
 * NOTE: Any special characters (newlines, tab etc.) are ignored.
 */
int WrapText(FontRef f, float width, const std::string &str, std::vector<std::string> &txt);

/*!
 * \brief Wraps a vector of words to be within a given max width.
 *
 * \param[in] f the font used to wrap the text.
 * \param[in] width the max width to wrap text against.
 * \param[in] words the vector of words to wrap.  Assumes spaces are present in the words themselves.
 * \param[out] txt vector that holds the wrapped text.  Where each line in the vector fits into the given width.
 *
 * \return The number of lines added to the txt vector.
 *
 * \see Split
 *
 * NOTE: Any special characters (newlines, tab etc.) are ignored.
 */
int WrapText(FontRef f, float width, const std::vector<std::string> &words, std::vector<std::string> &txt);

#endif
