/*!
 * \file StringUtility.cpp
 * \brief Implements functions for manipulating strings.
 * \author Joe Forte
 * \date 9/13/2008
 */

#include <sstream>
#include "Utility/StringUtility.h"

//--------------------------------------------------------------------
bool IsInteger(const std::string &str, int *val)
{
	std::string s = Trim(str);
	std::istringstream iss(s);
	
	// Checks if the value is in hex format (proceeded by 0x)
	if( s.length() > 1 && Match(s[0], "0") && Match(s[1], "xX") )
	{
		iss >> std::hex;
	}
	
	int i;
	if( (iss >> i).fail() )
		return false;
	
	if(val)
		*val = i;
	return true;
}

//--------------------------------------------------------------------
bool IsWholeInteger(const std::string &str, int *val)
{
	int i;
	if(IsInteger(str, &i) && str.find_first_of(".") == std::string::npos)
	{
		if(val) 
			*val = i;
		return true;
	}
	
	return false;
}

//--------------------------------------------------------------------
bool IsFloat(const std::string &str, float *val)
{
	std::istringstream iss(str);
	
	float f;
	if( (iss >> f).fail() )
		return false;
	
	if(val)
		*val = f;
	return true;
}

//--------------------------------------------------------------------
bool IsBoolean(const std::string &str, bool *val)
{
	std::string s = Trim(str);
	ToLower(s);
	
	if(s.length() == 1)
	{
		if(Match(s[0], "0f"))
		{
			if(val) *val = false;
			return true;
		}
		else if(Match(s[0], "1t"))
		{
			if(val) *val = true;
			return true;
		}
		else
		{
			return false;
		}
	}
	else if(s == "true")
	{
		if(val) *val = true;
		return true;
	}
	else if(s == "false")
	{
		if(val) *val = false;
		return true;
	}
	
	return false;
}

//--------------------------------------------------------------------
void Replace(std::string &str, const char *match, char replace)
{
	size_t len = str.length();
	for(size_t i = 0; i < len; i++)
	{
		if(Match(str[i], match))
			str[i] = replace;
	}
}

//--------------------------------------------------------------------
void Replace(std::string &str, const char *match, const char *replace)
{
	const size_t replaceLen = strlen(replace);
	size_t len = str.length();
	size_t i = 0;
	
	while(i < len)
	{
		if(Match(str[i], match))
		{
			str.replace(i, 1, replace);
			len = len + replaceLen;
			i = i + replaceLen;
		}
		else
		{
			i++;
		}
	}
}

//--------------------------------------------------------------------
std::string Trim(const std::string &str, const char *what)
{
	size_t start = str.find_first_not_of(what);
	size_t end = str.find_last_not_of(what);
	
	if(start == std::string::npos)
		return str.substr(0, end + 1);
	
	return str.substr(start, end - start + 1);
}

//--------------------------------------------------------------------
std::vector<std::string> Split(const std::string &str, const char *delim, bool skipDeliminator)
{
	std::vector<std::string> result;
	Split(str, delim, result, skipDeliminator);
	return result;
}

//--------------------------------------------------------------------
size_t Split(const std::string &str, const char *delim, std::vector<std::string> &result, bool skipDeliminator)
{
	size_t count = 0;
	
	size_t tokenStart = 0;
	size_t tokenLen = 0;
	
	const size_t len = str.size();
	
	size_t i = 0;
	while(i < len)
	{
		if(Match(str[i], delim))
		{
			if(skipDeliminator)
			{
				if(tokenLen > 0)
				{
					result.push_back(str.substr(tokenStart, tokenLen));
					count++;
				}
				
				tokenStart = i + 1;
				tokenLen = 0;
			}
			else // deliminators are stored in the substrings
			{
				if(tokenLen > 0)
				{
					if(Match(str[i-1], delim)) // groups sequential deliminators into one substring
					{
						tokenLen++;
					}
					else
					{
						result.push_back(str.substr(tokenStart, tokenLen));
						tokenStart = i;
						tokenLen = 1;
						
						count++;
					}
				}
				else
				{
					tokenLen++;
				}
			}
		}
		else
		{
			tokenLen++;
		}
		
		i++;
	}
	
	if(tokenLen > 0)
	{
		result.push_back(str.substr(tokenStart, tokenLen));
		count++;
	}
	
	return count;	
}
