/*!
 * \file StringUtility.h
 * \brief Defines functions for manipulating strings.
 * \author Joe Forte
 * \date 8/4/2008
 */

#ifndef STRINGUTILITY_H
#define STRINGUTILITY_H

#include <assert.h>
#include <string>
#include <vector>
#include <algorithm>

/*!
 * \brief Transforms a string to lower case.
 *
 * \param[in,out] str the string to make lower case.
 */
inline void ToLower(std::string &str)
{
	std::transform(str.begin(), str.end(), str.begin(), ::tolower);
}

/*!
 * \brief Transforms a string to upper case.
 *
 * \param[in,out] str the string to make upper case.
 */
inline void ToUpper(std::string &str)
{
	std::transform(str.begin(), str.end(), str.begin(), ::toupper);
}

/*!
 * \brief Matches a single character to a string of characters. 
 *
 * \param[in] c the character to find a match for.
 * \param[in] match a string of characters to match \a c against.
 *
 * \return True if \a c matches a character in \a match, false otherwise.
 */
inline bool Match(char c, const char *match)
{	
	while(*match)
	{
		if(*match == c)
			return true;
		match++;
	}
	
	return false;
}

/*!
 * \brief Determines if the given string is an integer.
 *
 * \param[in] str the string to check.
 * \param[out] val the integer that the string represents. Defaults to NULL.
 *
 * \return true if \a str represents an integer,  false otherwise.
 */
bool IsInteger(const std::string &str, int *val = NULL);

/*!
 * \brief Determines if the given string is strictly an integer.
 *
 * \param[in] str the string to check.
 * \param[out] val the integer that the string represents. Defaults to NULL.
 *
 * \return true if \a str represents an integer,  false otherwise.
 */
bool IsWholeInteger(const std::string &str, int *val = NULL);

/*!
 * \brief Determines if the given string is a float.
 *
 * \param[in] str the string to check.
 * \param[out] val the float that the string represents. Defaults to NULL.
 *
 * \return true if \a str represents an float, false otherwise.
 */
bool IsFloat(const std::string &str, float *val = NULL);

/*!
 * \brief Determines if the given string is a boolean (0, 1, T, F, True, False, etc.)
 *
 * \param[in] str the string to check.
 * \param[out] val the bool that the string represents. Defaults to NULL.
 *
 * \return true if \a str represents a boolean, false otherwise.
 */
bool IsBoolean(const std::string &str, bool *val = NULL);

/*!
 * \brief Replaces matching characters with a spacific single char.
 *
 * \param[in] str the string to alter.
 * \param[in] match the characters to match.
 * \param[in] replace the character to replace matches with.
 */
void Replace(std::string &str, const char *match, char replace);

/*!
 * \brief Replaces matching characters with a spacific string.
 *
 * \param[in] str the string to alter.
 * \param[in] match the characters to match.
 * \param[in] replace the character string to replace matches with.
 */
void Replace(std::string &str, const char *match, const char *replace);

/*!
 * \brief Trims leading and ending characters from a string.
 *
 * \param[in] str the string to trim.
 * \param[in] what the characters to be trimed. Defaults to space and newline characters.
 *
 * \return the trimmed string.
 */
std::string Trim(const std::string &str, const char *what = " \t\n\r");

/*!
 * \brief Splits a string into multiple substrings.
 *
 * \param[in] str the string to split.
 * \param[in] delim a string of deliminating characters.  Each individual character in the string is considered a deliminator.  Defaults to " ".
 * \param[in] skipDeliminator set to false to not have deliminators present in the output.  Defaults to true.
 *
 * \return The vector of substrings split from \a str.
 */
std::vector<std::string> Split(const std::string &str, const char *delim = " ", bool skipDeliminator = true);

/*!
 * \brief Splits a string into multiple substrings.
 *
 * \param[in] str the string to split.
 * \param[in] delim a string of deliminating characters.  Each individual character in the string is considered a deliminator.
 * \param[out] result stores the split substrings of \a str.
 * \param[in] skipDeliminator set to false to not have deliminators present in the output.  Defaults to true.
 *
 * \return The number of substrings added to \a result.
 */
size_t Split(const std::string &str, const char *delim, std::vector<std::string> &result, bool skipDeliminator = true);

#endif
